/************************************************************************/
/* sofi.c                                                               */
/*                                                                      */
/* Functions to calculate radiation in the umbral shadow during the     */
/* total eclipse at the 29.03.2006 using MYSTIC.                        */
/*                                                                      */   
/* Author: Claudia Emde                                                 */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <errno.h>

#include "sofi.h"
#include "r250.h"


#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


/* prototypes */
/*
static photon_struct *generate_photon_sofi(int source,
					   atmosphere_struct *atmos, 
					   albedo_struct *albedo,
					   sample_struct *sample, 
					   double *pd,
					   double sza, double phi,
					   int ipa);
*/



/***********************************************************************************/
/* Function: sample_photons_sofi_060329                                   @62_30i@ */
/* Description:                                                                    */
/*  Generate a probability distribution functions which corresponds to the         */
/*  irradiance distribution at the top pf the atmosphere during the total          */
/*  eclipse of the sun at 29-03-2006 in Turkey.                                    */
/*  The integrated probability is returned.                                        */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/
void sample_photons_sofi(float wvnmlo, float wvnmhi, double *pd)
{
  double *x,*r;
  double *Ic, *Sl, *Ic_int, *alpha; 
  int Nr, i, j;
  /* Definition of constants:*/ 
  double planck=6.63e-34;
  double speed_of_light=2.998e8;
  double boltzmann=1.38e-23;
  double T_S=5740.;
  
  /* Wavelength */
  double lambda=2./((wvnmlo+wvnmhi)*100.);
  /* Sampling will be done in km steps, counted from X=0, corresponding to the 
     centre of the umbral shadow. */
  int Nd=5000;
  /*extraterrestrial solar irradiance */
  double E_0=1368.;
  /* Intensity in centre of solar disk */
  double I_0;
  /* radtio between moon diameter and sun diameter */
  double ratio=1.0494;
  /* Sun disk radius normalized to x */
  double r_M=ratio;
  /* calculated from data from Nasa report, corresponds to approx 1km */
  double dx=0.0006; 
  /* Limb darkening coefficient*/
  double beta=3*planck*speed_of_light*sqrt(sqrt(2.))/(8*boltzmann* lambda * T_S); 
  double I_int=0.;
  
  /* Distance between centre of moon and centre of sun*/ 
  x = (double*) calloc(Nd,sizeof(double));
  /* Irradiance as a function of x*/
  Ic = (double*) calloc(Nd,sizeof(double));
  Ic_int = (double*) calloc(Nd,sizeof(double));
  
  /* Initialize x. This is *not* X in Koepke! Normalisation not included. 
     X=x/(r_M+r_S) */
  x[0]=0.;
  for (j=1; j<Nd; j++)
    x[j] = x[j-1]+dx;
  
  /* Grid for radius on solar disk, needed for integration of apparant irradiance. 
     Also in units of apparent moon radius.*/
  Nr=1001; 
  r = (double*) calloc(Nr,sizeof(double));
  /* Angle for integration of solar disk.*/
  alpha = (double*) calloc(Nr,sizeof(double));
  /* Solar limb darkening */
  Sl = (double*) calloc(Nr,sizeof(double)); 

  r[0]=0.;
  Sl[0]=1.;
  for (i=1; i<Nr; i++)
    {
      r[i] = r[i-1]+1./(Nr-1);
      Sl[i] = (1+beta*sqrt(fabs(1.-r[i]*r[i])))/(1+beta);
    }     
  /* Integrate over solar disk to get I_0, take numerical integration
     for consistency */
  for (i=1; i<Nr-1; i++)
    I_int+=(Sl[i]*r[i]+Sl[i+1]*r[i+1])*0.5*(r[i+1]-r[i]);
  
  /* Intensity in centre of solar disk */
  I_0=E_0/(2.*PI*I_int);
  
  /* Calculate irradiance as a function of X*/
  for (j=0; j<Nd; j++)
    {
      for (i=0; i<Nr; i++)
        {
          if(x[j]-r_M >=r[i] && x[j] > r_M)
            alpha[i]=0.;
          else if(fabs(x[j]-r_M)<r[i] && x[j]!=0.)
            alpha[i]=acos((r[i]*r[i]+x[j]*x[j]-r_M*r_M)/(2.*r[i]*x[j]));
          else if( r_M-x[j] >= r[i] )
            alpha[i] = PI; 
          else
            fprintf (stderr, "Error in calculation of alpha. x[j] %f, r[i] %f,i %d \n",
                     x[j], r[i], i);
        }
      
      for (i=0; i<Nr-1; i++) 
        Ic[j]+=2.*PI*I_0*(((1.-alpha[i]/PI)*Sl[i]*r[i]+(1.-alpha[i+1]/PI)
                           *Sl[i+1]*r[i+1])*(r[i+1]-r[i])/2.);
    }
  
  /* Photon density function */
  for (i=0; i<Nd; i++)
    pd[i] = Ic[i]/E_0;

  /*  pd[i] = (Ic[i]+1.1e-7*I_0)/E_0;
      1e-7: Upper limit of corona intensity.*/
  
  /* for (i=0;i<Nd; i++)
     fprintf(stderr, "pd[i] %f Ic[i] %f  i:%d \n", pd[i], Ic[i], i);*/
  
  
  /* Free variables.*/
  free(r);
  free(alpha);
  free(Sl);
}

/***********************************************************************************/
/* Function: generate_photon_sofi_060329                                  @62_30i@ */
/* Description:                                                                    */
/*  Generate a photon and inititalize according to a source corresponding to a     */
/*  partly covered sun disk.                                                       */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/


/* static photon_struct *generate_photon_sofi(int source, */
/* 					   atmosphere_struct *atmos,  */
/* 					   albedo_struct *albedo, */
/* 					   sample_struct *sample,  */
/* 					   double *pd, */
/* 					   double sza, double phi,  */
/* 					   int ipa) */

/* { */
/*   int i_re=0; */
/*   double sinphi=0, cosphi=0, cossza=0, sinsza=0; */
/*   double alpha=0, x=0, y=0, r=0, dx=0, dy=0; */
  
/*   photon_struct *p = calloc_photon (sample); */
/*   /\* work on a copy of the photon, not the original one *\/ */
  
  

/*   /\* defaults *\/ */
/*   p->scattercounter=0;  /\* number of scattering events *\/ */
/*   p->direct=1;          /\* indicates direct of diffuse *\/ */
/*   p->pathlength=0;      /\* photon pathlength           *\/ */
/*   p->boundary=NONE; */
/*   p->boundary2D=0; */
/*   p->crossingboundary=0; */

/*   /\* reset pathlengths *\/ */
/*   p->tauabs.air = 0; */
/*   p->tauabs.aer = 0; */
/*   p->tauabs.ozo = 0; */
/*   p->tauabs.cld = 0; */
/*   p->tauabs.ice = 0; */
/*   p->tauabs.tot = 0; */

/*   /\* reset reallyhit *\/ */
/*   p->reallyhit[0]=0; */
/*   p->reallyhit[1]=0; */
/*   p->reallyhit[2]=0; */

/*   /\* independent pixel *\/ */
/*   p->ipa = ipa; */
    
/*   p->direct=1;   /\* direct beam source *\/ */
  
/*   /\* sine and cosine of solar zenith *\/ */
/*   cossza = cos(sza/180.0*PI); */
/*   sinsza = sin(sza/180.0*PI); */
  
/*   /\* Different definition of azimuth angle *\/ */
/*   phi = phi-180.; */

/*   /\* sine and cosine of solar azimuth *\/ */
/*   cosphi = cos(phi/180.0*PI); */
/*   sinphi = sin(phi/180.0*PI);   */

/*   /\* Start position of the photon*\/ */
/*   p->x[0] = atmos->xmax * dr250(); */
/*   p->x[1] = atmos->ymax * dr250(); */
  
/*   p->ic = (int) p->x[0]/atmos->delX;  */
/*   p->jc = (int) p->x[1]/atmos->delY;   */

/*   /\* Centre of umbral shadow at TOA *\/ */
/*   r=tan(sza/180.0*PI) * atmos->Z[atmos->Nz]; */
/*   dx=r*sinphi;  */
/*   dy=r*cosphi; */

/*   /\* x and y coordinates, (x,y)=0 corresponds to centre of ellipse *\/  */
/*   x=p->x[0]-200e3+dx;  */
/*   y=p->x[1]-200e3+dy;  */

/*   /\* Starting point in polar coordinates *\/ */
/*   r=sqrt(pow(x,2)+pow(y,2)); */
/*   alpha = atan(x/y); */
  
/*   /\* Equivalent radius, if the ellipsoidal shadow is transformed back to a circle,  */
/*      for which the probability density function is given.*\/ */
/*   i_re=(int)fabs(sqrt(pow(r,2)/ */
/*                       (pow(cos(alpha-phi/180.0*PI),2)/pow(cossza,2)+ */
/*                        pow(sin(alpha-phi/180.0*PI),2))))/1000.; */
  
/*   /\* We can use tauabs as photon weight: */
/*      exp(-tauabs) is the weight of the photon, which is finally used in the  */
/*      averaging. Therefore tauabs=-log(weight).*\/ */
/*   /\* */
/*   fprintf(stderr, "p->dir.cotheta: %d   p->pdir[0]: %d \n", p->dir.cotheta,  */
/*   p->pdir[0]);*\/ */
/*   p->tauabs.tot = -log(pd[i_re]); */
  
/*   /\* vertical start position: TOA *\/ */
/*   p->x[2] = atmos->Z[atmos->Nz]; */
/*   p->kc   = atmos->Nz-1; */
  
/*   /\* initialize direction using the specified solar zenith and azimuth angles *\/ */
/*   init_direction (sinsza, cossza, sinphi, cosphi, &(p->dir)); */
  
/*   return p; */
/* } */

