/*--------------------------------------------------------------------
 * $Id: albedo.c,v 1.12 2006/12/17 23:45:52 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include "uvspec.h"
#include "fortran_and_c.h"
#include "f77-uscore.h"
#include "ascii.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif

#define nigbp 20
#define nband 15
#define PI 3.14159265358979323846264338327


/************************************/
/* prototypes of internal functions */
/************************************/

static int read_albedo (char *filename, alb_out_struct *alb_out);
static int read_rpv    (char *filename, rpv_out_struct *rpv, float scale);
static int interpolate_rpv (rpv_out_struct *rpv, float *lambda, 
			    int nlambda_lower, int nlambda_upper);
static int read_surfaces (char *filename, surfaces_out_struct *surf,
			  float *lambda, int nlambda,
			  int nlambda_lower, int nlambda_upper, float scale);

void F77_FCN (landspec) (int *igbp, float *u0, float *wv, float *salb15, float *bbalb);
void F77_FCN (getemiss) (int *igbp, float *ems12);
void F77_FCN (swdnwgts) (float *CSZ, float *PW, float *WGT);

static int get_surface_type_from_map (float latitude, float longitude, alb_out_struct *alb, 
				      char *filename, int verbose, int quiet);
static void number2surfacestring (int surface_nr, char *surface_string);


/**************************************************************/
/* Setup surface albedo.                                      */
/**************************************************************/

int setup_albedo (input_struct input, output_struct *output)
{
  int status=0, iv=0, linear=0;

  char *filename       = calloc (FILENAME_MAX+1, sizeof(char));
  char *surface_string = calloc (30,             sizeof(char));

  /* Input Parameters (solar and thermal) */
  /* igbp - IGBP scene identification index [1 - 20] */
  int band;
  float *bands;
  float *subbands;

  /* Input Parameters (solar): */
  /*  u0   - Cosine of the solar zenith angle [0.0 - 1.0] */
  /*  wv   - PRECIPITABLE WATER (cm) Range (0 to~7) */
  float u0, wv;

  /* Output Parameters: */
  /* salb15 - spectral albedo in 15 shortwave bands */
  /* wgt15  - estimated weights for salb15 */
  /* bbalb  - solar broad band albedo */
  float *salb15   = calloc (15, sizeof(float));
  float *wgt15    = calloc (15, sizeof(float));
  float bbalb;

  /* Output Parameters: */
  /* ems12 - spectral emissivity in 12 longwave bands */
  float *ems12   = calloc (12, sizeof(float));

  float sum_wgt=0;
  float sum_alb=0;
  int iq=0;

  float *file_lower_lambda_albedo=NULL;
  float *file_upper_lambda_albedo=NULL;

  char tmp_string[2];

  output->alb.albedo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

  output->rpv.rho0_t   = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.k_t      = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.theta_t  = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.scale_t  = (float *) calloc (output->wl.nlambda_t, sizeof(float));


  /* search surface type in the map according to lat, lon */
  if (input.alb.map == TRUE) {
    status = get_surface_type_from_map (input.latitude,input.longitude, &(output->alb), 
                                        input.filename[FN_ALBEDO_MAP], input.verbose, input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d reading surface type map in setup_albedo (albedo.c) \n", status);
      return status;
    }
  }


  switch(input.alb.source) {
  case ALBEDO_CONSTANT:
    for (iv=0;iv<output->wl.nlambda_t;iv++)
      output->alb.albedo_t[iv] = input.alb.albedo; 
    break;
  case ALBEDO_FROM_FILE:
  case ALBEDO_USER_LIBRARY:

    if (input.alb.source == ALBEDO_FROM_FILE)
      strcpy (filename, input.filename[FN_ALBEDO]);
    else if (input.alb.source == ALBEDO_USER_LIBRARY) {      
      /* create albedo file name */
      if (input.alb.surface < 1 || 99 < input.alb.surface) {
        fprintf (stderr, "Error, albedo_surface %5d out of range \n", input.alb.surface);
        fprintf (stderr, "       only range [ 1, 99] is valid\n");
        return -1;
      }
      strcpy (filename, input.filename[FN_ALBEDO_LIB_PATH]);
      strcat (filename, "albedo_");
      if (input.alb.surface < 10)
        sprintf(tmp_string, "0%1d",input.alb.surface);
      else
        sprintf(tmp_string, "%2d",input.alb.surface);
      strcat (filename, tmp_string);
      strcat (filename, ".dat");
    }
    else {
      fprintf (stderr, "Program bug in albedo.c\n");
      return -1;
    }

    /* black underground, for safety set albedo_t = 0 (should already be 0) */
    if (input.alb.source == ALBEDO_USER_LIBRARY && input.alb.surface == 0) {
      for (iv=0; iv<output->wl.nlambda_t; iv++)
        output->alb.albedo_t[iv]=0;
    }
    else {
      status = read_albedo (filename, &output->alb);
      if (status != 0) {
        fprintf (stderr, "Error reading albedo file: %s\n", input.filename[FN_ALBEDO]);
        return status;
      }

      free(filename); 
    
      /**** Interpolate albedo to transmission grid ****/
    
      linear = 1;  /**** Use linear interpolation for the albedo, otherwise 
                         negative albedos or albedos larger than 1 may occur, 
                         I should know :-) AKy, ****/ 

      status = arb_wvn2 (output->alb.file_nlambda_albedo, output->alb.file_lambda_albedo,
                         output->alb.file_albedo, 
                         output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, 
                         output->wl.lambda_t, output->alb.albedo_t, 
                         linear);

      if (status != 0) {
        fprintf (stderr, "Error %d returned by arb_wvn2()\n", status);
        return status;
      }
    }
    
    break;
  case ALBEDO_IGBP_LIBRARY:

    /* get input values for albedo functions */

    /* Cosine of the solar zenith angle [0.0 - 1.0] */
    /* if there are clouds, then albedo is calculated as if the SZA is 53 degrees */
    if (output->wc.occurrence == TRUE || output->ic.occurrence == TRUE)
      u0 = cos(53./360.*2.*PI);
    else
      u0 = cos(input.atm.sza/360*2*PI); 

    /* Precipitable water in cm (== 10 kg/m2), range = [ 0.0, 7.0]  */
    wv = 0.1 * output->precip_water;  /* 1/10. == kg/m2 -> cm */ 
    if (wv > 7.0) { 
      wv = 7.0;  /* maximum range given by parameterization */
                 /* should libRadtran give a warning here?  */
    }

    if (output->alb.surface == 0) /* black surface */
      return 0;
    else {
      number2surfacestring(output->alb.surface, surface_string);
      if (input.verbose)
        fprintf (stderr, "    IGBP surface type nr %d, %s\n", output->alb.surface, surface_string);

      strcpy (input.filename[FN_ALBEDO_LIB_WVL], input.filename[FN_PATH]);
      strcat (input.filename[FN_ALBEDO_LIB_WVL], "/albedo/IGBP/wvl_albedo.dat");

      /* read center wavelength and band limits [wavenumbers] from file */
      status = read_5c_file_float (input.filename[FN_ALBEDO_LIB_WVL],
				   &(bands),
				   &(subbands),
				   &(output->alb.file_lambda_albedo),
                                   &(file_lower_lambda_albedo),
                                   &(file_upper_lambda_albedo),
				   &(output->alb.file_nlambda_albedo));

      if (output->alb.surface < 1 || 20 < output->alb.surface) {
        fprintf (stderr, "Error: surface type nr %d for albedo_surface not known!\n", output->alb.surface);
        return -1;
      }

      for (band=0; band<output->alb.file_nlambda_albedo; band++) {
        file_lower_lambda_albedo[band] *= 1000.0; /* 1000 == micro m -> nm */
        file_upper_lambda_albedo[band] *= 1000.0; /* 1000 == micro m -> nm */
        output->alb.file_lambda_albedo[band] *= 1000.0; /* 1000 == micro m -> nm */
      }

      /* solar albedo */
      F77_FCN (landspec) (&output->alb.surface, &u0, &wv, salb15, &bbalb);
      /* estimated weights for solar albedo */
      F77_FCN (swdnwgts) (&u0, &wv, wgt15);
      /* thermal albedo */
      F77_FCN (getemiss) (&output->alb.surface, ems12);

      output->alb.file_albedo = calloc (15+12, sizeof(float));

      for (band=0;band<(15+12);band++) {
        if (band<15)
          output->alb.file_albedo[band] = salb15[band];
	else
	  output->alb.file_albedo[band] = 1 - ems12[band-15];
      }

      if (input.verbose) { 
        fprintf(stderr,"\n    band    lambda     lambda      lambda    albedo    weight\n");
        fprintf(stderr,  "            lower      center      upper                     \n");
        fprintf(stderr,  "   -----------------------------------------------------------");
        for (band=0;band<(15+12);band++) {
          fprintf(stderr,"\n    %3d %10.2f %10.2f %11.2f   %8.6f ",
                  band+1, 
                  file_lower_lambda_albedo[band], 
                  output->alb.file_lambda_albedo[band], 
                  file_upper_lambda_albedo[band], 
		  output->alb.file_albedo[band]);
          if (band<15)
            fprintf(stderr," %8.6f", wgt15[band]);
        }
        fprintf(stderr,"\n\n");
      }

    }

    if (input.ck_scheme == CK_FU) {

      /* set wavelength grid */
      output->alb.file_nlambda_albedo = output->wl.nlambda_t;

      for (band = output->wl.nlambda_rte_lower; band <= output->wl.nlambda_rte_upper; band++) {
	if (band == 0) {
          sum_wgt=0;
          sum_alb=0;
          for (iq=0;iq<10;iq++) {
            sum_wgt += wgt15[iq];
            sum_alb += wgt15[iq]*output->alb.file_albedo[iq];
          }
          output->alb.albedo_t[band] = sum_alb / sum_wgt; 
	}
	else
          output->alb.albedo_t[band] = output->alb.file_albedo[band+9];
      }
    }
    else {

      /* spectral albedo constant inside one band */

      for (iv=0; iv<output->wl.nlambda_t; iv++) {
        output->alb.albedo_t[iv] = -999;
        for (band = 0; band < output->alb.file_nlambda_albedo; band ++) {
          if ((file_lower_lambda_albedo[band] <= output->wl.lambda_t[iv]) && 
              (output->wl.lambda_t[iv] <= file_upper_lambda_albedo[band])) {
	    /* fprintf (stderr, "band[%2d]=%f,lambda_t[%2d]=%f, band[%d]=%f \n",             */
            /*   band,  output->alb.file_lambda_albedo[band]  , iv, output->wl.lambda_t[iv], */
            /*   band+1,output->alb.file_lambda_albedo[band+1]);                            */
            output->alb.albedo_t[iv] = output->alb.file_albedo[band];
	    break;
	  }
	}
        if (output->alb.albedo_t[iv] == -999) {
          if (4000.0 <= output->wl.lambda_t[iv] && output->wl.lambda_t[iv] <= 4545.45) { 
            /* not defined in this region: make linear interpolation (is this OK????) */
            output->alb.albedo_t[iv] = output->alb.file_albedo[14] + 
	          (output->wl.lambda_t[iv]-4000.0) / (4545.45 - 4000.0) * 
                  (output->alb.file_albedo[15]-output->alb.file_albedo[14]);
	  }
          else {
            fprintf (stderr, "Error %d at setup albedo_IGBP! \n", -999);
            return -999;
	  }
	}
      }
    }
    break;
  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme\n");
    return -1;
  }
 

  /**** Check interpolated albedo values ****/

  for (iv=0; iv<output->wl.nlambda_t; iv++) {
/*      if (input.verbose) { */
/*        fprintf (stderr, "lambda   albedo"); */
/*        fprintf (stderr, "%10.3f %9.5f\n",output->wl.lambda_t[iv], output->alb.albedo_t[iv]); */
/*      } */
    if (output->alb.albedo_t[iv] > 1.0 || output->alb.albedo_t[iv] < 0.0)  {
      fprintf (stderr, "Albedo interpolated from %s out of range:\n", input.filename[FN_ALBEDO]);
      fprintf (stderr, "wavelength: %f, albedo %f\n", 
                 output->wl.lambda_t[iv], output->alb.albedo_t[iv]);
      fprintf (stderr, "Modify albedo file %s\n", input.filename[FN_ALBEDO]);
      return -1;
    } 
  }


  /* read BRDF */
  if (strlen(input.filename[FN_RPV]) > 0) {
    
    status = read_rpv (input.filename[FN_RPV], &(output->rpv), input.rpv.scale);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_rpv (%s)\n", status, input.filename[FN_RPV]);
      return status;
    }
    

    status = interpolate_rpv (&(output->rpv), output->wl.lambda_t, 
			      output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by interpolate_rpv()\n", status);
      return status;
    }
  }
  else {
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      output->rpv.rho0_t  [iv] = input.rpv.rho0;    
      output->rpv.k_t     [iv] = input.rpv.k;    
      output->rpv.theta_t [iv] = input.rpv.theta;    
      output->rpv.scale_t [iv] = input.rpv.scale;    
    }
  }


  /* read surface information for MYSTIC */
  if (strlen(input.rte.mc.filename[FN_MC_RPV]) > 0) {

    status = read_surfaces(input.rte.mc.filename[FN_MC_RPVTYPE], &(output->surf),
			   output->wl.lambda_t, output->wl.nlambda_t,
			   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper, input.rpv.scale);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_surfaces(%s)\n", status, input.rte.mc.filename[FN_MC_RPV]);
      return status;
    }

  }    

  return 0;
}





/**************************************************************/
/* Read 2-column albedo file.                                 */
/**************************************************************/

static int read_albedo (char *filename, alb_out_struct *alb_out) 
{
  int status=0;
  
  status = read_2c_file_float (filename, 
			       &(alb_out->file_lambda_albedo),
			       &(alb_out->file_albedo),
			       &(alb_out->file_nlambda_albedo));

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  return 0;
}


/**************************************************************/
/* Read 4-column RPV file.                                    */
/**************************************************************/

static int read_rpv (char *filename, rpv_out_struct *rpv, float scale)
{
  int iv=0, status=0;

  status = read_5c_file_float (filename,
			       &(rpv->lambda),
			       &(rpv->rho0),
			       &(rpv->k),
			       &(rpv->theta),
			       &(rpv->scale),
			       &(rpv->n));

  if (status!=0) {

    /* try 4 columns instead */
    status = read_4c_file_float (filename,
				 &(rpv->lambda),
				 &(rpv->rho0),
				 &(rpv->k),
				 &(rpv->theta),
				 &(rpv->n));
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* allocate memory for scaling factor and set to constant */
    rpv->scale = calloc (rpv->n, sizeof(float));
    for (iv=0; iv<rpv->n; iv++)
      rpv->scale[iv] = scale;
  }

  return 0; /* if o.k. */
}  



/**************************************************************/
/* Interpolate RPV to internal wavelength grid.               */
/**************************************************************/

static int interpolate_rpv (rpv_out_struct *rpv, float *lambda, 
			    int nlambda_lower, int nlambda_upper)
{
  int status=0;

  int linear=1;   /* linear interpolation */

  /* interpolate to internal RTE grid */
  status = arb_wvn2 (rpv->n, rpv->lambda, rpv->rho0, 
		     nlambda_lower, nlambda_upper, lambda, rpv->rho0_t, 
		     linear);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn2 (rho0)\n", status);
    return status;
  }
  
  status = arb_wvn2 (rpv->n, rpv->lambda, rpv->k, 
		     nlambda_lower, nlambda_upper, lambda, rpv->k_t, 
		     linear);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn2 (k)\n", status);
    return status;
  }
  
  status = arb_wvn2 (rpv->n, rpv->lambda, rpv->theta, 
		     nlambda_lower, nlambda_upper, lambda, rpv->theta_t, 
		     linear);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn2 (theta)\n", status);
    return status;
  }
  
  status = arb_wvn2 (rpv->n, rpv->lambda, rpv->scale, 
		     nlambda_lower, nlambda_upper, lambda, rpv->scale_t, 
		     linear);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn2 (scale)\n", status);
    return status;
  }
  
  return 0;  /* if o.k. */
}
  
  

/**************************************************************/
/* Read surfaces from file.                                   */
/**************************************************************/

static int read_surfaces (char *filename, surfaces_out_struct *surf, float *lambda, int nlambda,
			  int nlambda_lower, int nlambda_upper, float scale) 
{
  int il=0, status=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;
  char ***string=NULL;

  status = ASCII_checkfile (filename, 
			    &rows,
			    &min_columns,
			    &max_columns,
			    &max_length);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  if (min_columns<2)
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
  
  if (min_columns != max_columns)
    fprintf (stderr, "*** Warning, inconsistent number of columns in %s\n", filename);

  status = ASCII_calloc_string (&string,
				rows,
				max_columns,
				max_length);

  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory\n", status);
    return status;
  }

  status = ASCII_readfile (filename, string);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  surf->n = rows;
  surf->label    = calloc (surf->n, sizeof (char *));
  surf->filename = calloc (surf->n, sizeof (char *));
  surf->rpv      = calloc (surf->n, sizeof (rpv_out_struct));

  for (il=0; il<surf->n; il++) {
    surf->label[il]    = calloc (max_length+1, sizeof(char));
    surf->filename[il] = calloc (max_length+1, sizeof(char));

    strcpy (surf->label[il], string[il][0]);
    strcpy (surf->filename[il], string[il][1]);
  }
  
  ASCII_free_string (string, rows, max_columns);

  /* now read filenames and interpolate data to internal resolution */
  for (il=0; il<surf->n; il++) {
    fprintf (stderr, " ... reading type %s from %s\n", surf->label[il], surf->filename[il]);

    /* read wavelength-dependent RPV data */
    status = read_rpv (surf->filename[il], &(surf->rpv[il]), scale);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_rpv(%s)\n", status, surf->filename[il]);
      return status;
    }

    /* allocate memory */
    surf->rpv[il].rho0_t   = (float *) calloc (nlambda, sizeof(float));
    surf->rpv[il].k_t      = (float *) calloc (nlambda, sizeof(float));
    surf->rpv[il].theta_t  = (float *) calloc (nlambda, sizeof(float));
    surf->rpv[il].scale_t  = (float *) calloc (nlambda, sizeof(float));

    /* interpolate to internal wavelength grid */
    status = interpolate_rpv (&(surf->rpv[il]), lambda, nlambda_lower, nlambda_upper);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by interpolate_rpv()\n", status);
      return status;
    }
  }

  return 0;
}

static int get_surface_type_from_map (float lat, float lon, alb_out_struct *alb, char *filename, int verbose, int quiet)
{

  int status = 0;

#if HAVE_LIBNETCDF

  int ncid=0;
  int    idd_nlat, idd_nlon=0, id_data=0;
  size_t     nlat=0,   nlon=0;

  float i, j;
  size_t start[2] = {0,0};
  unsigned char igbp=0;


  if (verbose) {
    fprintf (stderr, " ... search surface type in map: \n");
    fprintf (stderr, "     %s \n", filename);
    fprintf (stderr, "     map size = [%8.3f,%8.3f] x [%8.3f,%8.3f]\n", 
                           alb->map_lat_min/3600, alb->map_lat_max/3600, 
                           alb->map_lon_min/3600, alb->map_lon_max/3600);
  }

  /* map reference system:                      */
  /* southpole = 0 -> northpole = 180,          */
  /*   shift of all lat not nessesary, as we need differences 
       (that would be [-90,90]-> lat+90-> [0,180]) */
  /* Greenwich = 0->(eastward)->360 Greenwich */
  /*   it is taken account for this in uvspec.c (albedo checks) */


  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
    return status;
  }

  /* get dimension id for "nlat" */
  status = nc_inq_dimid (ncid, "nlat", &idd_nlat);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nlat" */
  status = nc_inq_dimlen (ncid, idd_nlat, &nlat);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* get dimension id for "nlon" */
  status = nc_inq_dimid (ncid, "nlon", &idd_nlon);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nlon" */
  status = nc_inq_dimlen (ncid, idd_nlon, &nlon);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* get variable id for "elev" (elevation) */
  status = nc_inq_varid (ncid, "igbp", &id_data);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d, while getting ncid from %s\n", status, filename);
    return status;
  }


  if (verbose)
    fprintf (stderr, "     found %d x %d data points\n", nlat, nlon);

  i = (lat - alb->map_lat_min) / (alb->map_lat_max - alb->map_lat_min) * nlat;
  j = (lon - alb->map_lon_min) / (alb->map_lon_max - alb->map_lon_min) * nlon;

  /* rounding (float -> int) */
  /* latitude, index in y-direction, from south to north */ 
  start[0] =  i;
  /* correction integer rounding at the northern boundary of the map */
  if (start[0]==nlat) start[0] = nlat-1;

  /* locations on the boundaries are treated like the northward pixel */
  /* index in x-direction, from Greenwich to east */
  start[1] =  j;
  /* locations on the boundaries are treated like the eastward pixel */
  /* correction integer rounding at the eastern boundary of the map */
  if (start[1]==nlon) start[1] = nlon-1;  

  if (verbose) {
    fprintf (stderr, "     look up surface type at: lat = %f, lon = %f, ", lat/3600, lon/3600);
    fprintf (stderr, " element: %6d x %6d \n", start[0]+1,start[1]+1);
  }

  status = nc_get_var1_uchar (ncid, id_data, start, &igbp);

  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  alb->surface = igbp;
  if (!quiet)
    fprintf (stderr, " ... surface type is surface type nr: %d\n", igbp);

#else
  fprintf (stderr, " ******************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot       *\n");
  fprintf (stderr, " * use the albedo_map option. Please get netcdf and rebuild.      *\n");
  fprintf (stderr, " ******************************************************************\n");
  return -1;
#endif

  return status;
}

static void number2surfacestring (int surface_nr, char *surface_string) {

  if      (surface_nr ==  0) strcpy (surface_string, "BLACK"); 
  else if (surface_nr ==  1) strcpy (surface_string, "EVERGREEN_NEEDLE_FOREST");  
  else if (surface_nr ==  2) strcpy (surface_string, "EVERGREEN_BROAD_FOREST"); 
  else if (surface_nr ==  3) strcpy (surface_string, "DECIDUOUS_NEEDLE_FOREST");
  else if (surface_nr ==  4) strcpy (surface_string, "DECIDUOUS_BROAD_FOREST");
  else if (surface_nr ==  5) strcpy (surface_string, "MIXED_FOREST");
  else if (surface_nr ==  6) strcpy (surface_string, "CLOSED_SHRUBS");
  else if (surface_nr ==  7) strcpy (surface_string, "OPEN_SHRUBS");
  else if (surface_nr ==  8) strcpy (surface_string, "WOODY_SAVANNA"); 
  else if (surface_nr ==  9) strcpy (surface_string, "SAVANNA"); 
  else if (surface_nr == 10) strcpy (surface_string, "GRASSLAND");
  else if (surface_nr == 11) strcpy (surface_string, "WETLAND"); 
  else if (surface_nr == 12) strcpy (surface_string, "CROPLAND");
  else if (surface_nr == 13) strcpy (surface_string, "URBAN");
  else if (surface_nr == 14) strcpy (surface_string, "CROP_MOSAIC"); 
  else if (surface_nr == 15) strcpy (surface_string, "ANTARCTIC_SNOW");
  else if (surface_nr == 16) strcpy (surface_string, "DESERT");
  else if (surface_nr == 17) strcpy (surface_string, "OCEAN_WATER"); 
  else if (surface_nr == 18) strcpy (surface_string, "TUNDRA"); 
  else if (surface_nr == 19) strcpy (surface_string, "FRESH_SNOW"); 
  else if (surface_nr == 20) strcpy (surface_string, "SEA_ICE");
  else {
    fprintf(stderr,"Unknown surface number %d\n", surface_nr); 
  }

}
