/*--------------------------------------------------------------------
 * $Id: mie.c,v 1.12 2006/12/17 23:45:52 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ascii.h"
#include "mie.h"
#include "numeric.h"
#include "getopt.h"

#define PI 3.1415926535897932
#define MAXLINE 255


int    setup_wlgrid_and_read_ref_index(input_struct input, output_struct *output, 
                                       double **ri_real, double **ri_cplx, 
                                       int *lambda_i_start, int *lambda_i_end);
int    calc_distribution_radius_grid  (input_struct input, output_struct *output, int iv);
int    calc_distribution_number_dens  (input_struct input, output_struct *output, int iv, int ir);
double GAMMA_LN(float xx);
double double_min(double x, double y);
double double_max(double x, double y);
int    write_output (input_struct input, output_struct *output, int ir, int iv);
char   *replace(char *string, char *oldpiece, char *newpiece);
int    make_reff (float *LWC, int n, float dropconc, int distribution, float alpha, float **r_eff);
int mie_calc_sizedist_optimised (mie_inp_struct input, mie_out_struct *output, 
		       int program, int medium, mie_complex crefin, 
		       float wavelength, float temperature, 
		       double *x_size, double *y_size, int n_size,
		       double *beta, double *omega, double *g,
		       mie_complex *ref, 
                       int ir, float **qext_array, float **qsca_array, float **gsca_array, float ****pmom_array);


int mie (input_struct input, output_struct *output)
{
  int ip=0, im=0, ir=0, iv=0, status=0;
  int first=1;

  double wavelength=0;

  float *qext_array;
  float *qsca_array;
  float *gsca_array;
  float ***pmom_array;  /* Phase function moments */

  double *ri_real;
  double *ri_cplx;

  double x, ext, sca, vol;
  int lambda_i_start=0;
  int lambda_i_end;

  output->ref.re = 0.0;
  output->ref.im = 0.0; 

  output->mie.pmom = calloc (4, sizeof(float *));
  for (ip=0; ip<4; ip++)
    output->mie.pmom[ip] = calloc (input.mie.nmom+1, sizeof(float));

  /* setup wavelength grid and            */
  /* read refraction index file, if given */
  status = setup_wlgrid_and_read_ref_index(input, output, &ri_real, &ri_cplx, &lambda_i_start, &lambda_i_end);
  if (status != 0) {
    fprintf (stderr, "Error %d returned by setup_wlgrid_and_read_ref_index()\n", status);
    return status;
  }
  

  /* determine the number of r_eff to be simulated */
  switch(input.distribution) {
  case DIST_NONE:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    /* determine radius r_eff */
    if (input.r_eff_step != 0.0)
      /* 1.E-7 rouding error */
      output->n_r_eff = (int) floor((input.r_eff_max - input.r_eff_min+1.E-7) / input.r_eff_step); 
    else
      output->n_r_eff = 0;
  
    output->r_eff = (double *) calloc (output->n_r_eff+1, sizeof(double));
    for (ir=0; ir<=output->n_r_eff; ir++)
      output->r_eff[ir] = input.r_eff_min + ir*input.r_eff_step;
    break;
  case DIST_FILE:
    /* size distribution file, can only be valid for one r_eff */
    output->n_r_eff = 0;
    output->r_eff = (double *) calloc (output->n_r_eff+1, sizeof(double));
    output->r_eff[0] = -999.0;
    break;
  default:
    fprintf (stderr, "Error, distribution number in mie.c %d\n", input.distribution);
    return -3;
    break;
  }

  output->warnings = calloc (output->wl.nlambda+1, sizeof(int *));
  for (iv=0; iv<=output->wl.nlambda; iv++)
    output->warnings[iv] = calloc (output->n_r_eff+1, sizeof(int)); 

  /* call the mie programms, who will do the main work */
  /* staring r_eff-loop and lambda-loop */


  for (iv=lambda_i_start; iv<=lambda_i_end; iv++) {

    /* change wavelength dependent properties */
    wavelength = (output->wl.lambda[iv]) * 1.e-03;  /* Convert from nm to micro meter */
    if (strlen (input.ri_filename)>0) {
      input.crefin.re = ri_real[iv];
      input.crefin.im = ri_cplx[iv];
    }

    /* generate the grid of the droplet distribution */
    status = calc_distribution_radius_grid (input, output, iv); 
    if (status!=0) {
      fprintf (stderr, "Error %d returned by make_distribution()\n", status);
      return status;
    }

    qext_array = (float *) calloc (output->dist.n_dens, sizeof(float));
    qsca_array = (float *) calloc (output->dist.n_dens, sizeof(float));
    gsca_array = (float *) calloc (output->dist.n_dens, sizeof(float));

    pmom_array = calloc (4, sizeof(float **));
    for (ip=0; ip<4; ip++) {
      pmom_array[ip] = calloc (input.mie.nmom+1, sizeof(float *));
      for (im=0; im<input.mie.nmom+1; im++)
	pmom_array[ip][im] = calloc (output->dist.n_dens, sizeof(float));
    }
    
    for (ir=0; ir<=output->n_r_eff; ir++) {

      if (ir == 0)
        fprintf (stderr, " ... start Mie-calculation for lambda = %10.3f nm  ", output->wl.lambda[iv]);

      /* number density distribution per radius intervall,*/ 
      /* unfortunately also wavelength dependent          */
      status = calc_distribution_number_dens (input, output, iv, ir); 
      if (status!=0) {
	fprintf (stderr, "Error %d returned by make_distribution()\n", status);
	return status;
      }

      switch(input.distribution) {
      case DIST_NONE:
        /* one single radius */
        status = mie_calc (input.mie, &(output->mie), 
			   input.program, input.medium, input.crefin,
			   wavelength, output->r_eff[ir], input.temperature, &(output->ref));
        if (status!=0) {
	  fprintf (stderr, "Error %d returned by mie_calc()\n", status);
	  return status;
        }
        /* calculate beta and omega */
        x = (2.0 * PI * output->r_eff[ir]) / wavelength; /* (micro m / micro m) */
        ext = PI * x*x * output->mie.qext;
	sca = PI * x*x * output->mie.qsca;
	vol = (4.0/3.0) * PI * pow((output->r_eff[ir]*1E-6),3.0); /* micro m -> m */
        output->beta  = ext / vol;
        output->omega = sca / ext;
	break;
      case DIST_FILE:
      case DIST_GAMMA:
      case DIST_LOGNORMAL:
        /* droplet radius distribution */
        status = mie_calc_sizedist_optimised (input.mie, &(output->mie), 
    				    input.program, input.medium, input.crefin,
				    wavelength, input.temperature, output->dist.radius, 
                                    output->dist.number_dens, output->dist.n_dens,
				    &output->beta, &output->omega, &output->g, &(output->ref), 
                                    ir, &qext_array, &qsca_array, &gsca_array, &pmom_array);
        if (status!=0) {
	  fprintf (stderr, "Error %d returned by mie_calc_sizedist_optimised()\n", status);
	  return status;
        }
	break;
      default:
        fprintf (stderr, "Error, distribution number in mie.c %d\n", input.distribution);
        return -4;
        break;
      }

      /* write output */
      status = write_output (input, output, ir, iv);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by write_output()\n", status);
	return status;
      }
 
      /* for debugging: print cumulative probability distribution      */
      /* to 'cumulative.dat':                                          */
      /* cumulative_probability (output->mie.pmom[0], input.mie.nmom); */

      /* for debugging: print phase function to 'phase.dat':           */
      /* phase_function (output->mie.pmom[0], input.mie.nmom);         */

    } /* for (ir=0; ir<=output->n_r_eff; ir++) { */ 

    /* free distribution dependent properties */
    free(qext_array);
    free(qsca_array);
    free(gsca_array);

    for (ip=0; ip<4; ip++) {
      for (im=0; im<input.mie.nmom+1; im++)
        free(pmom_array[ip][im]);
      free(pmom_array[ip]);
    }
    free(pmom_array);

  } /* for (iv=0; iv<=output->wl.nlambda; iv++) { ...*/
  

  /* write warnings */
  if (input.output_format == OUTPUT_FORMAT_CLOUDPROP) {
    for (iv=0; iv<=output->wl.nlambda; iv++) {
      for (ir=0; ir<=output->n_r_eff; ir++) {
        if (output->warnings[iv][ir]==1) {
          if (first == 1) { /* true */
            fprintf (stderr, "*** Warning: maxmimum accuracy of the phase function not reached\n");
            fprintf (stderr, "*** with %4d Legendre coefficients.\n",input.mie.nmom);
            fprintf (stderr, "*** In order to get more accurate results\n"); 
            fprintf (stderr, "*** please increase 'nmom' in the mie input file!\n"); 
            first = 0;
	  }
          fprintf (stderr, "          lambda = %10.3f nm, r_eff = %8.3f micro meter\n", output->wl.lambda[iv], output->r_eff[ir]);
	}
      } /* for (ir=0; ir<=output->n_r_eff; ir++) { */ 
    } /* for (iv=0; iv<=output->wl.nlambda; iv++) { ...*/
  }

  /*free memory */
  for (iv=0; iv<=output->wl.nlambda; iv++)
    free(output->warnings[iv]);  
  free(output->warnings);

  return 0;
}


int setup_wlgrid_and_read_ref_index(input_struct input, output_struct *output, 
                                    double **ri_real, double **ri_cplx, 
                                    int *lambda_i_start, int *lambda_i_end)
{  
  int status = 0;
  int ir=0, iv=0;

  /* read refractive index */
  if (strlen (input.ri_filename)>0) {
    status = read_3c_file (input.ri_filename, 
			   &output->wl.lambda, 
                           ri_real, ri_cplx, 
                           &output->wl.nlambda);
    
    if (status != 0) {
      fprintf (stderr, "error %d reading %s\n", status, input.ri_filename);
      return status;
    }

    fprintf (stderr, " ... read %d data points from %s\n", 
	     output->wl.nlambda, input.ri_filename);


    /* check if data points are sorted in ascending order */
    for (ir=0; ir<output->wl.nlambda-1; ir++)
      if (output->wl.lambda[ir+1] < output->wl.lambda[ir]) {
	fprintf (stderr, "error: %s not sorted in ascending order\n", 
		 input.ri_filename);
	return -1;
      }

    /* special convention here */
    output->wl.nlambda--;
  }
  else {
    if (input.ck_scheme == CK_NONE) {
      if (input.wl.wl_step != 0.0)
        output->wl.nlambda = (int) (input.wl.wl_end-input.wl.wl_start) / input.wl.wl_step;
      else
        output->wl.nlambda = 0;

      output->wl.lambda = (double *) calloc (output->wl.nlambda+1, sizeof(double));
      for (iv=0; iv<=output->wl.nlambda; iv++)
        output->wl.lambda[iv] = input.wl.wl_start + input.wl.wl_step*iv;
    }
    else {

      switch(input.ck_scheme) {
      case CK_KATO:
      case CK_KATO2:
      case CK_KATO2_96:
        fprintf (stderr, " ... Kato wavelength bands (Planck weighted center wavelength)\n");
        output->wl.nlambda = 32;
        output->wl.lambda  = (double *) calloc (output->wl.nlambda, sizeof(double));
        output->wl.lambda[ 0] =  257.176; /* in nm */
        output->wl.lambda[ 1] =  277.541;
        output->wl.lambda[ 2] =  294.485;
        output->wl.lambda[ 3] =  316.585;
        output->wl.lambda[ 4] =  345.787;
        output->wl.lambda[ 5] =  384.357;
        output->wl.lambda[ 6] =  429.336;
        output->wl.lambda[ 7] =  483.697;
        output->wl.lambda[ 8] =  527.705;
        output->wl.lambda[ 9] =  544.270;
        output->wl.lambda[10] =  557.174;
        output->wl.lambda[11] =  583.737;
        output->wl.lambda[12] =  613.952;
        output->wl.lambda[13] =  643.529;
        output->wl.lambda[14] =  674.519;
        output->wl.lambda[15] =  693.235;
        output->wl.lambda[16] =  721.382;
        output->wl.lambda[17] =  764.185;
        output->wl.lambda[18] =  814.807;
        output->wl.lambda[19] =  864.118;
        output->wl.lambda[20] =  926.245;
        output->wl.lambda[21] = 1005.830;
        output->wl.lambda[22] = 1113.970;
        output->wl.lambda[23] = 1330.000;
        output->wl.lambda[24] = 1558.310;
        output->wl.lambda[25] = 1764.690;
        output->wl.lambda[26] = 2045.010;
        output->wl.lambda[27] = 2206.390;
        output->wl.lambda[28] = 2570.000;
        output->wl.lambda[29] = 3277.760;
        output->wl.lambda[30] = 3785.440;
        output->wl.lambda[31] = 4349.730;
	break;
      case CK_FU:
        fprintf (stderr, " ... Fu wavelength bands (Planck weighted center wavelength)\n");
        output->wl.nlambda = 18;
        output->wl.lambda  = (double *) calloc (output->wl.nlambda, sizeof(double));
        output->wl.lambda[ 0] =   490.000; /* in nm */
        output->wl.lambda[ 1] =   920.000;
        output->wl.lambda[ 2] =  1540.000;
        output->wl.lambda[ 3] =  2150.000;
        output->wl.lambda[ 4] =  2890.000;
        output->wl.lambda[ 5] =  3730.660;
        output->wl.lambda[ 6] =  4941.870;
        output->wl.lambda[ 7] =  5599.780;
        output->wl.lambda[ 8] =  6569.790;
        output->wl.lambda[ 9] =  7577.880;
        output->wl.lambda[10] =  8528.080;
        output->wl.lambda[11] =  9628.070;
        output->wl.lambda[12] = 11300.000;
        output->wl.lambda[13] = 13628.400;
        output->wl.lambda[14] = 16600.000;
        output->wl.lambda[15] = 21300.000;
        output->wl.lambda[16] = 29400.000;
        output->wl.lambda[17] = 50600.000;
	break;
      case CK_AVHRR_KRATZ:
        fprintf (stderr, " ... AVHRR-Kratz wavelength bands (Planck weighted center wavelength)\n");
        output->wl.nlambda = 16;
        output->wl.lambda  = (double *) calloc (output->wl.nlambda, sizeof(double));
        output->wl.lambda[ 0] =   569.0820; /* in nm */
        output->wl.lambda[ 1] =   588.9440;
        output->wl.lambda[ 2] =   612.5150;
        output->wl.lambda[ 3] =   653.5070;
        output->wl.lambda[ 4] =   717.0210;
        output->wl.lambda[ 5] =   762.2760;
        output->wl.lambda[ 6] =   834.9930;
        output->wl.lambda[ 7] =   940.2140;
        output->wl.lambda[ 8] =  1007.2600;
        output->wl.lambda[ 9] =  3543.7400;
        output->wl.lambda[10] =  3644.8600;
        output->wl.lambda[11] =  3743.4100;
        output->wl.lambda[12] =  3848.2200;
        output->wl.lambda[13] =  3954.7500;
        output->wl.lambda[14] = 10783.6000;
        output->wl.lambda[15] = 11870.9000;
	break;
      case CK_GENERIC:
        fprintf (stderr, "Error: not jet implemented!\n");
        return -2;
	break;
      default:
        fprintf (stderr, "Error: unsupported correlated-k scheme, nr %2d\n", input.ck_scheme);
        return -1;
      }

      input.wl.wl_start = output->wl.lambda[0];
      input.wl.wl_end   = output->wl.lambda[output->wl.nlambda-1];

      /* special convention here */
      output->wl.nlambda--;

      for (iv=0; iv<=output->wl.nlambda; iv++)
        fprintf (stderr, "   lambda[%3d] = %10.3f nm\n", iv+1, output->wl.lambda[iv]);
    }
  }

  /* setup of wavelength range */
  if (input.lambda_i_start == -999)  
    (*lambda_i_start) = 0;
  else {
    if (input.lambda_i_start < 1) {
      fprintf(stderr,"  wavelength_index_start = %d has to be 1 or larger \n",input.lambda_i_start);
      status=-1;
    }
    if (output->wl.nlambda < input.lambda_i_start) {
      fprintf(stderr,"  wavelength_index_start = %d, but there are only %d wavelength bands \n",
                        input.lambda_i_start, output->wl.nlambda);
      status=-2;
    }
    (*lambda_i_start) = input.lambda_i_start-1;
  }

  if (input.lambda_i_end == -999)  
    (*lambda_i_end) = output->wl.nlambda;
  else {
    if (input.lambda_i_end < (*lambda_i_start)) {
      fprintf(stderr,"  wavelength_index_end = %d has to be larger than wavelength_index_start = %d\n",
                        input.lambda_i_end, (*lambda_i_start));
      status=-3;
    }
    if (output->wl.nlambda < input.lambda_i_end-1) {
      fprintf(stderr,"  wavelength_index_end = %d, but there are only %d wavelength bands \n",
                        input.lambda_i_end+1, output->wl.nlambda+1);
      status=-4;
    }
    (*lambda_i_end) = input.lambda_i_end-1;
  }


  /* check wavelength rage of water refractive index parameterisation */
  if (input.medium == ICE) {
    
    if (input.wl.wl_start < 0.0443*1000.) {
      fprintf(stderr,"******* ERROR >>>>>>  REFICE--wavelength outside table boundaries\n");
      fprintf(stderr,"******* ERROR >>>>>>  wavelength_start=%14.7g nm, lambda_min = %7.3f nm\n",input.wl.wl_start,0.0443*1000.);
      status=-1;
    }
    if (8.600E+06*1000. < input.wl.wl_end) {
      fprintf(stderr,"******* ERROR >>>>>>  REFICE--wavelength outside table boundaries\n");
      fprintf(stderr,"******* ERROR >>>>>>  lambda_max = %7.3f nm, wavelength_end=%14.7g nm\n",8.600E+06*1000.,input.wl.wl_end);
      status=-2;
    }
    if (status != 0)
      return status;
  }

  return status;
}

int calc_distribution_radius_grid(input_struct input, output_struct *output, int iv)
{
  float radmin, radmax, rad, delta_rad;
  float x, delta_x;
  int is;
  int status=0;

  /* fprintf(stderr,"calculate droplet radius distribution\n"); */

  switch(input.distribution) {
  case DIST_NONE:
    /* single radius calculation, nothing to do */
    output->dist.n_dens = 1;
    break;
  case DIST_FILE:
    /* read size distribution from file */
    status = read_2c_file (input.sd_filename, &output->dist.radius, &output->dist.number_dens, &output->dist.n_dens);
    
    if (status != 0) {
      fprintf (stderr, "error %d reading %s\n", status, input.sd_filename);
      return status;
    }

    fprintf (stderr, " ... read %d data points from %s\n", 
	     output->dist.n_dens, input.sd_filename);
    break;
  case DIST_GAMMA:
  case DIST_LOGNORMAL:

    radmin = 0.02 * input.r_eff_min; /* cover the whole droplet spectrum   */
    radmax = 8.00 * input.r_eff_max; /* in order to save mie-calculations  */

    /* count number of radii */
    /* the grid of dist.radius is independent of ir, that means output->r_eff[ir] */
    rad = radmin;
    is = 0;
    while (rad <= radmax) {
      x = 2*PI*rad / (output->wl.lambda[iv]*1.e-03);  /* 1E-3 = nm -> micro m */
      delta_x = double_max(0.03,0.03*sqrt(x));
      delta_rad = delta_x * (output->wl.lambda[iv]*1.e-03) / (2*PI);
      rad = rad + delta_rad;
      is = is + 1;
    }
    output->dist.n_dens = is;

    /* callocation */
    if ((output->dist.radius = calloc(output->dist.n_dens, sizeof(double)))==NULL) {
      fprintf (stderr, "Error: Allocation of output->dist.radius (make_distribution)\n");
      return -2;
    }
    if ((output->dist.number_dens = calloc(output->dist.n_dens, sizeof(double)))==NULL) {
      fprintf (stderr, "Error: Allocation of output->dist.radius (make_distribution)\n");
      return -3;
    } 

    /* calculate number density */
    rad = radmin;
    is = 0;
    while (rad <= radmax) {
      x = 2*PI*rad / (output->wl.lambda[iv]*1.e-03);  /* 1E-3 = nm -> micro m */
      delta_x = double_max(0.03,0.03*sqrt(x));
      delta_rad = delta_x * (output->wl.lambda[iv]*1.e-03) / (2*PI);
      rad = rad + delta_rad;
      output->dist.radius[is]      = rad;
      is = is + 1;
    }
    break;
  default:
    fprintf (stderr, "Error in make_distribution : unrecognized distribution %d\n", input.distribution);
    break;
  } /* switch(input.distribution) {... */


  /* check if data points are sorted in ascending order */
  switch(input.distribution) {
  case DIST_NONE:
    break;
  case DIST_FILE:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    /* for gamma and lognormal this is not nessesary, but it costs almost no time */

    for (is=0; is<output->dist.n_dens-1; is++)
      if (output->dist.radius[is+1]<output->dist.radius[is]) {
        if (input.distribution == DIST_FILE)
          fprintf (stderr, "error: %s not sorted in ascending order\n", 
	                    input.sd_filename);
        else
          fprintf (stderr, "error: radius in distribution not sorted in ascending order, program bug!!!\n");
        return -5;
      }
    
    if (output->dist.radius[0]==0) {
      output->dist.radius[0] = 0.1 * output->dist.radius[1];
      fprintf (stderr, " ... cannot handle r=0; setting first data point to r=%g\n",
	     output->dist.radius[0]);
    }
    break;
  default:
    fprintf (stderr, "Error in make_distribution : unrecognized distribution %d\n", input.distribution);
    break;
  }

  return status;
}

int calc_distribution_number_dens (input_struct input, output_struct *output, int iv, int ir)
{
  float rad;
  float A, B;
  float LWC=1.0;
  int is;
  int status=0;

  /* fprintf(stderr,"calculate droplet radius distribution\n"); */

  switch(input.distribution) {
  case DIST_NONE:
    /* single radius calculation, nothing to do */
    break;
  case DIST_FILE:
    /* only one effective radius */
    if (ir > 0) {
      fprintf(stderr,"Error in make_distribution: more than one loop in ir for distribution \n");
      fprintf(stderr,"    for distribution read from file (obviously only one reff possible)\n");
      return -1;
    }
    break;
  case DIST_GAMMA:
  case DIST_LOGNORMAL:

    if (input.distribution == DIST_GAMMA) {
      B = (input.alpha+3)/output->r_eff[ir];
      A = (0.75/PI)*pow(B,(input.alpha+4))*1000./exp(GAMMA_LN(input.alpha+4.));
    }
    else if (input.distribution == DIST_LOGNORMAL) {
      B = output->r_eff[ir]*exp(-2.5*input.alpha*input.alpha);
      A = 1000./((4./3.*PI)* sqrt(2*PI)*input.alpha * B*B*B *exp(4.5*input.alpha*input.alpha));
    }
    else{
      fprintf(stderr,"Unknown 'distribution %d' in make_distribution\n",input.distribution);
      status--;
      return status;
    }

    if (output->rho_medium <= 0.0) {
      fprintf (stderr, "Error, rho_medium = %f is 0 or negative!\n",output->rho_medium);
      return -1;
    }
    A = A*LWC/output->rho_medium;

    /* calculate number density */
    for (is=0;is<output->dist.n_dens;is++) {

      rad = output->dist.radius[is];

      /* calculate particle density for each radius */ 
      if (input.distribution == DIST_GAMMA)
        output->dist.number_dens[is] = A * pow(rad,input.alpha) * exp(-B*rad);
      else if (input.distribution == DIST_LOGNORMAL)
        output->dist.number_dens[is] = A/rad * exp(-0.5* pow(log(rad/B),2) / pow(input.alpha,2));
      else {
        fprintf(stderr,"Error in make_distribution : unrecognized distribution %d\n",input.distribution);
        return -4;
      }

      /* fprintf(stderr,"is = %3d, radius = %8.4f, number_dens = %13.6f, is=%3d \n", */
      /*           is, output->dist.radius[is], output->dist.number_dens[is], is);   */
    }
    break;
  default:
    fprintf (stderr, "Error in make_distribution : unrecognized distribution %d\n", input.distribution);
    break;
  } /* switch(input.distribution) {... */

  return status;
}


/* get the effective radius from LWC and number density, assuming a distribution */
int make_reff (float *LWC, int n, float dropconc, int distribution, float alpha, float **r_eff)
{
  int ir=-999;
  int status=0;

  for (ir=0; ir<=n; ir++) {
    if (distribution == DIST_GAMMA) {
      (*r_eff)[ir] = 100.0 * (LWC[ir] *pow((0.75*(alpha+3)*(alpha+3))/(PI*(alpha+1)*(alpha+2)*dropconc),(1.0/3.0)));
    }
    else if (distribution == DIST_LOGNORMAL) {
      (*r_eff)[ir] = 100.0 * pow((3.*LWC[ir]*exp(3.*alpha*alpha)/(4.*PI*dropconc)), (1.0/3.0));
    }
    else {
      fprintf(stderr,"Unknown 'distribution %d' in make_reff\n",distribution);
      status--;
    }
  }
  return status;
}



double GAMMA_LN(float xx)
{
  int j=0;
  double *cof;
  double stp = 2.50662827465;
  double half = 0.5, one=1.0, fpf = 5.5, x, tmp, ser;
  double gammaln = 0.0;

  cof  = (double *) calloc (6, sizeof(double));
  cof[0] =  76.18009173;
  cof[1] = -86.50532033;
  cof[2] =  24.01409822;
  cof[3] =  -1.231739516;
  cof[4] =   0.120858003;
  cof[5] =  -0.53638200;

  x = xx - one;
  tmp = x + fpf;
  tmp = (x+half)*log(tmp)-tmp;
  ser=one;
  for (j=0; j<=5; j++) {
    x = x + one;
    ser = ser + cof[j]/x; 
  }
  gammaln=tmp+log(stp*ser);
  return gammaln;
}


double double_min(double x, double y)
{
  if (x <= y)
    return x;
  else
    return y;
}

double double_max(double x, double y)
{
  if (x >= y)
    return x;
  else
    return y;
}



/***********************************************************************************/
/* Function: mie_calc_sizedist_optimised                                           */
/* Description:                                                                    */
/*  Mie calculations, using the MIEV0 and BHMIE codes by Warren Wiscombe           */
/*  (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman              */
/*  (ftp://astro.princeton.edu/draine/scat/bhmie/).                                */
/*                                                                                 */
/* Parameters:                                                                     */
/*  mie_inp_struct input:    mie input structure (see src_c/miecalc.h)             */
/*  mie_out_struct *output:  mie output structure (see src_c/miecalc.h)            */
/*  int program:             MIEV0 or BHMIE                                        */
/*  int medium:              WATER, ICE, or USER; if USER, the refractive index    */
/*                           is read from crefin                                   */
/*  mie_complex crefin:      Complex refractive index (both numbers positive)      */
/*  float wavelength:        wavelength [micron]                                   */
/*  float temperature:       temperature                                           */
/*  double *x_size:          size distribution, radius [um]                        */
/*  double *y_size:          size distribution, n(r)                               */
/*  int n_size:              size distribution, number of radii                    */
/*  double *beta:            extinction coefficient [km-1] per unit                */
/*                           liquid water content (returned)                       */
/*  double *omega:           Single scattering albedo (returned)                   */
/*  double *g:               Asymmetry parameter (returned)                        */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                                 */
/* Remark: this is almost the same function than mie_calc_sizedist                 */
/*         but the call of the mie_calc() function is only executed,               */
/*         if this is the first call in the radius-loop                            */
/*         as the dist_radius grid and therefor the MIE-Properties does not change */
/*         but the concentration does change with r_eff and that means             */
/*         that the integration is executed every turn                             */
/***********************************************************************************/

int mie_calc_sizedist_optimised (mie_inp_struct input, mie_out_struct *output, 
		       int program, int medium, mie_complex crefin, 
		       float wavelength, float temperature, 
		       double *x_size, double *y_size, int n_size,
		       double *beta, double *omega, double *g, 
		       mie_complex *ref, 
                       int ir, float **qext_array, float **qsca_array, float **gsca_array, float ****pmom_array)
{
  int status=0, is=0, ip=0, im=0;

  double xsquared=0, xcubed=0, norm=0;
  
  double ext=0, sca=0, asy=0, vol=0;


  /* allocate memory for fields */
  double *extinction = calloc (n_size, sizeof (double));
  double *scattering = calloc (n_size, sizeof (double));
  double *asymmetry  = calloc (n_size, sizeof (double));
  double *volume     = calloc (n_size, sizeof (double));


  double ***pmom=NULL;

  if (input.nmom>0) {
    pmom     = calloc (4, sizeof(double **));
   
    for (ip=0; ip<4; ip++) {
      pmom    [ip] = calloc (input.nmom+1, sizeof(double *));
      
      for (im=0; im<=input.nmom; im++)
        pmom[ip][im] = calloc (n_size, sizeof(double));
    }
  }

  
  for (is=0; is<n_size; is++) {

    if (ir == 0) {
      if (is % 4 == 0)
        fprintf (stderr, "\b|");
      else if (is % 4 == 1)
        fprintf (stderr, "\b/");
      else if (is % 4 == 2)
        fprintf (stderr, "\b-");
      else if (is % 4 == 3)
        fprintf (stderr, "\b\\");

      status = mie_calc (input, output, program, medium, crefin, 
	  	         wavelength, x_size[is], temperature, ref);
      if (status!=0) {
        fprintf (stderr, "Error %d returned by mie_calc()\n", status);
        return status;
      }

      /* save results for further calculations */
      (*qext_array)[is] = output->qext;
      (*qsca_array)[is] = output->qsca;
      (*gsca_array)[is] = output->gsca;

      if (input.nmom>0)
        for (ip=0; ip<4; ip++)
          for (im=0; im<=input.nmom; im++)
	    (*pmom_array)[ip][im][is] = output->pmom[ip][im];
    }
    else {
      /* use the saved results */
      output->qext = (*qext_array)[is];
      output->qsca = (*qsca_array)[is];
      output->gsca = (*gsca_array)[is];

      if (input.nmom>0)
        for (ip=0; ip<4; ip++)
          for (im=0; im<=input.nmom; im++)
	    output->pmom[ip][im] = (*pmom_array)[ip][im][is];
    }
    
    /* start integration */
    xsquared = x_size[is] * x_size[is] * 1e-12;
    xcubed   = xsquared   * x_size[is] * 1e-6;
    
    extinction[is] = output->qext * xsquared * y_size[is];
    scattering[is] = output->qsca * xsquared * y_size[is];
    asymmetry [is] = output->gsca * scattering[is]; 
    volume    [is] = xcubed * y_size[is];

    /* quote from MIEV.doc, concerning the normalization of pmom[]:   */
    /*	                                                              */
    /*	The normalized moments are                                    */
    /*      4 / ( XX**2 * QSCA ) * PMOM,  but it is PMOM itself, not  */
    /*      these normalized moments, which should be integrated over */
    /*      a size distribution.                                      */

    if (input.nmom>0)
      for (ip=0; ip<4; ip++)
        for (im=0; im<=input.nmom; im++)
	  pmom[ip][im][is] = output->pmom[ip][im] * y_size[is];

  } /* for (is=0; is<n_size; is++) { */

  if (ir == 0)
    fprintf (stderr, "\n\a");
  
  norm = integrate (x_size, y_size,     n_size);
  
  ext  = integrate (x_size, extinction, n_size) / norm;
  sca  = integrate (x_size, scattering, n_size) / norm;
  asy  = integrate (x_size, asymmetry,  n_size) / norm;
  vol  = integrate (x_size, volume,     n_size) / norm;
  
  if (input.nmom>0)
    for (ip=0; ip<4; ip++)
      for (im=0; im<=input.nmom; im++)
	output->pmom[ip][im] = integrate (x_size, pmom[ip][im], n_size) / norm;
  

  *beta  = 3.0*ext/(4.0*vol)/1000.0;
  *omega = sca/ext;
  *g     = asy/sca;

  /* free memory */
  free (extinction);
  free (scattering);
  free (asymmetry);
  free (volume);

  if (input.nmom>0) {
    for (ip=0; ip<4; ip++) {
      for (im=0; im<=input.nmom; im++)
	free(pmom[ip][im]);
      
      free(pmom[ip]);
    }    
    free(pmom);
  }

  return 0;
}



/* write all the output to the screen or into files */
int write_output (input_struct input, output_struct *output, int ir, int iv)
{

  int  io=0;
  int  nmom=-999;
  int  im=0;
  char outputfile[255];
  char ck_band_str[3];
  char tmp_str[50];
  int  status=0;

  if (input.output_format == OUTPUT_FORMAT_NORMAL) {

    if (input.n_output_user == 0) {  
      /* default output */
      input.n_output_user = 8;
      input.output_user = calloc (input.n_output_user, sizeof (int));
      input.output_user[0]=OUTPUT_USER_WAVE;
      input.output_user[1]=OUTPUT_USER_REFRAC_REAL;
      input.output_user[2]=OUTPUT_USER_REFRAC_IMAG;
      input.output_user[3]=OUTPUT_USER_QEXT;
      input.output_user[4]=OUTPUT_USER_OMEGA;
      input.output_user[5]=OUTPUT_USER_GG;
      input.output_user[6]=OUTPUT_USER_SPIKE;
      input.output_user[7]=OUTPUT_USER_PMOM;
    }

    for (io=0; io<input.n_output_user; io++) {
      switch (input.output_user[io]) {
	case OUTPUT_USER_NONE:
	  fprintf (stdout, "%9s ", " ");
	  break;
	case OUTPUT_USER_WAVE:
	  fprintf (stdout, "%9.3f ", output->wl.lambda[iv]);
	  break;
	case OUTPUT_USER_WAVENUMBER:
	  fprintf (stdout, "%9.3f ", 1E7/output->wl.lambda[iv]);
	  break;
	case OUTPUT_USER_R_EFF:
	  fprintf (stdout, "%9.3f ", output->r_eff[ir]);
	  break;
	case OUTPUT_USER_REFRAC_REAL:
	  fprintf (stdout, "%12.6e ", output->ref.re);
	  break;
	case OUTPUT_USER_REFRAC_IMAG:
	  fprintf (stdout, "%12.6e ", output->ref.im);
	  break;
	case OUTPUT_USER_QEXT:
          if (input.distribution == DIST_FILE || 
              input.distribution == DIST_GAMMA || 
              input.distribution == DIST_LOGNORMAL)
	    fprintf (stdout, "%12.6e ", output->beta);
          else
            fprintf (stdout, "%12.6e ", output->mie.qext);
	  break;
        case OUTPUT_USER_QSCA:
          if (input.distribution == DIST_FILE || 
              input.distribution == DIST_GAMMA || 
              input.distribution == DIST_LOGNORMAL)
            fprintf (stdout, "%12.6e ", output->omega*output->beta);
          else
	    fprintf (stdout, "%12.6e ", output->mie.qsca);
	  break;
	case OUTPUT_USER_OMEGA:
          if (input.distribution == DIST_FILE || 
              input.distribution == DIST_GAMMA || 
              input.distribution == DIST_LOGNORMAL)
            fprintf (stdout, "%12.6e ", output->omega);
          else
	    fprintf (stdout, "%12.6e ", output->mie.qsca/output->mie.qext);
	  break;
	case OUTPUT_USER_GG:
          if (input.distribution == DIST_FILE || 
              input.distribution == DIST_GAMMA || 
              input.distribution == DIST_LOGNORMAL)
	    fprintf (stdout, "%12.6e ", output->g);
          else
	    fprintf (stdout, "%12.6e ", output->mie.gsca);
	  break;
	case OUTPUT_USER_SPIKE:
          if (input.distribution == DIST_FILE || 
              input.distribution == DIST_GAMMA || 
              input.distribution == DIST_LOGNORMAL)
	    fprintf (stdout, "%12.6e ", -1.0);
          else
	    fprintf (stdout, "%12.6e ", output->mie.spike);
	  break;
	case OUTPUT_USER_PMOM:
          if (input.mie.nmom>0)
	    for (im=0; im<=input.mie.nmom; im++)
	      fprintf (stdout, " %12.6e", output->mie.pmom[0][im]/output->mie.pmom[0][0]);	        
	  break;
	default:
	  fprintf (stderr, "Error, unknown user output %d\n", input.output_user[io]);
	  break;
      } /* switch (input.output_user[io]) { */
    } /* for (io=0; io<input.n_output_user; io++) { */
    fprintf (stdout, "\n");
    fflush (stdout);
  } /* if (input.output_user[0] != OUTPUT_USER_CLOUDPROP) { */

  else if (input.output_format == OUTPUT_FORMAT_CLOUDPROP) {

    if (ir == 0) {
      /* generate file name */
      switch (input.medium) {
      case WATER:
        strcpy (outputfile, "wc.");          
	break;
      case ICE:
        strcpy (outputfile, "ic.");   
        break;
      case USER:
        strcpy (outputfile, "uc.");   
	break;
      }

      sprintf(tmp_str, "%3d",iv+1);
      strcpy (ck_band_str, replace(tmp_str," ","0"));

      strcat (outputfile, ck_band_str);
      strcat (outputfile, ".mie");
      fprintf (stderr, " ... generating file: %s\n", outputfile);
          
      if ((output->ptr_output = fopen (outputfile, "w")) == NULL) {
        fprintf (stderr, "Error, cannot open file: %s in write_output! \n", outputfile);   
        return -1;         
      }
    }

    fprintf(stderr,"   lambda = %10.3f nm (%3d of %3d), r_eff = %8.3f micro meter (%3d of %3d) \n", 
                   output->wl.lambda[iv], iv+1,output->wl.nlambda+1, output->r_eff[ir], ir+1,output->n_r_eff+1);

    if (ir == 0) {
      switch (input.medium) {
      case WATER:
        fprintf (output->ptr_output, "! Mie table vs. effective radius (LWC=1 g/m^3)\n");
	break;
      case ICE:
        fprintf (output->ptr_output, "! Mie table vs. effective radius (IWC=1 g/m^3)\n");
	break;
      case USER:
        fprintf (output->ptr_output, "! Mie table vs. effective radius (ATTENTION!!! assume rho_medium = %7.4f g/cm3)\n", 
                                      output->rho_medium);
	break;
      }
      fprintf (output->ptr_output," %12.6e  wavelength (micron) \n",output->wl.lambda[iv]/1000.);
      fprintf (output->ptr_output," %12.6e  %12.6e  index of refraction \n", output->ref.re, output->ref.im);
      if (input.distribution == DIST_NONE) {
        fprintf (output->ptr_output,"S  distribution type ! one Single radius  \n");
        fprintf   (output->ptr_output,"-999.99  distribution shape parameter !(dummy to fit the format)\n");
      }
      else if (input.distribution == DIST_GAMMA) {
        fprintf (output->ptr_output,"G  distribution type\n");
        fprintf   (output->ptr_output,"%7.4f  distribution shape parameter\n", input.alpha);
      }
      else if (input.distribution == DIST_LOGNORMAL) {
        fprintf (output->ptr_output,"L  distribution type\n");
        fprintf   (output->ptr_output,"%7.4f  distribution shape parameter\n", input.alpha);
      }
      else if (input.distribution == DIST_FILE) {
        fprintf (output->ptr_output,"U  distribution type ! User defined, from file \n");
        fprintf   (output->ptr_output,"-999.99  distribution shape parameter !(dummy to fit the format)\n");
      }


      fprintf (output->ptr_output,"%4d %8.4f %8.4f  number, starting, ending effective radius \n", 
	                output->n_r_eff+1, output->r_eff[0], output->r_eff[output->n_r_eff]); 
    } /* if (ir == 0) { */

    for (im=0;im<input.mie.nmom;im++)
      if (fabs((2*im+1)*output->mie.pmom[0][im]/output->mie.pmom[0][0]) < 0.000005) break;
    nmom=im-1;
    /* remember to write a warning, if there could be more Legendre coefficients */
    if (im == input.mie.nmom) {
      fprintf (stderr, "*** Warning: maxmimum possible accuracy of the phase function not reached.\n");
      fprintf (stderr, "*** You may want to calculate more moments.\n");
      output->warnings[iv][ir]=1;
    }

/*     if (input.distribution == DIST_GAMMA ||  */
/*         input.distribution == DIST_LOGNORMAL || */
/*         input.distribution == DIST_FILE) */
      fprintf (output->ptr_output,"%9.4f  %12.5e  %12.5e %6d  Reff  Ext  Alb  Nleg\n", output->r_eff[ir], output->beta/output->rho_medium, 
                                                                               output->omega, nmom);
/*     else */
/*       fprintf (output->ptr_output,"%9.4f  %12.5e  %12.5e %6d  Reff  Ext  Alb  Nleg\n", output->r_eff[ir], output->mie.qext*1E+6*1E-3, /\* 1E+6 = cm-3/m3 -> m3/m3 *\/ */
/*                                                                                       output->mie.qsca/output->mie.qext, nmom);       /\* 1E-3 = km -> m *\/ */

    if (nmom>0) {
      fprintf (output->ptr_output, "  ");
      for (im=0; im<=nmom; im++)
	fprintf (output->ptr_output, " %9.5f", (2*im+1) * output->mie.pmom[0][im]/output->mie.pmom[0][0]);
    }

    fprintf (output->ptr_output, "\n");
    fflush (output->ptr_output);  

    if (ir == output->n_r_eff)
      fclose(output->ptr_output);

  } /* else if (input.output_format == OUTPUT_FORMAT_CLOUDPROP) ... */ 
  return status;
}


char *replace(char *string, char *oldpiece, char *newpiece) {

   int str_index, newstr_index, oldpiece_index, end,

      new_len, old_len, cpy_len;
   char *c;
   static char newstring[MAXLINE];

   if ((c = (char *) strstr(string, oldpiece)) == NULL)

      return string;

   new_len        = strlen(newpiece);
   old_len        = strlen(oldpiece);
   end            = strlen(string)   - old_len;
   oldpiece_index = c - string;


   newstr_index = 0;
   str_index = 0;
   while(str_index <= end && c != NULL)
   {

      /* Copy characters from the left of matched pattern occurrence */
      cpy_len = oldpiece_index-str_index;
      strncpy(newstring+newstr_index, string+str_index, cpy_len);
      newstr_index += cpy_len;
      str_index    += cpy_len;

      /* Copy replacement characters instead of matched pattern */
      strcpy(newstring+newstr_index, newpiece);
      newstr_index += new_len;
      str_index    += old_len;

      /* Check for another pattern match */
      if((c = (char *) strstr(string+str_index, oldpiece)) != NULL)
         oldpiece_index = c - string;

   }
   /* Copy remaining characters from the right of last matched pattern */    strcpy(newstring+newstr_index, string+str_index);

   return newstring;
} 
