/*--------------------------------------------------------------------
 * $Id: molecular.c,v 1.36 2006/12/17 23:40:31 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include "uvspec.h"
#include "ascii.h"
#include "numeric.h"

/************************************/
/* prototypes of internal functions */
/************************************/

static int crs_o3  (crs_out_struct *crs_out, int first, 
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type);

static int crs_co2 (crs_out_struct *crs_out, int first, 
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type);

static int crs_o2 (crs_out_struct *crs_out, int first, 
		   float *lambda, int n_lambda, 
		   float *temper, int ntemp, char *path,
		   int type, int quiet);

static int crs_no2 (crs_out_struct *crs_out,
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type);

static int crs_bro (crs_out_struct *crs_out,
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type);

static int crs_oclo (crs_out_struct *crs_out, int first, 
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type);

static int crs_hcho (crs_out_struct *crs_out, int first, 
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type);

static int read_crs (crs_out_struct *crs_out,
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path, int mol_id);

static int closest_above(float lambda, float* lambda_raw, int n_crs); 
static int closest_below(float lambda, float* lambda_raw, int n_crs);
static double quadratic(double x, double x0, double x1, double x2, 
		       double y0, double y1, double y2);
static double levelsum (atm_out_struct *atm, int mol);


/**************************************************************/
/* Setup absorption by molecules                              */
/**************************************************************/

int setup_crs (input_struct input, output_struct *output)
{
  int status = 0, i=0, lc=0, iv=0;
  static int first = 1;
  
  /* Allocate memory for crs; currently Rayleigh scattering   */
  /* and absorption by O3, O2, H2O, CO2, NO2, BRO, OCLO, HCHO */

  if (first) {
    status  = ASCII_calloc_float_3D(&output->crs.crs, output->atm.nlev, 
				    MOL_NN, output->wl.nlambda_t);
    if (status!=0) { 
      fprintf (stderr, "Error %d allocating memory for output->crs.crs\n", status); 
      return status; 
    } 
    status  = ASCII_calloc_float(&output->crs.crs_amf, output->wl.nlambda_t, output->atm.nlev);
    if (status!=0) { 
      fprintf (stderr, "Error %d allocating memory for output->crs.crs_amf\n", status); 
      return status; 
    } 
  }

  switch(input.ck_scheme) {
  case CK_NONE:

    /* First read in all default cross sections;         */
    /* cross sections are only read if the concentration */
    /* is larger than 0 in at least one of the levels    */

    switch (output->atm.molabs) {
    case MOLABS_CALC:

      if (input.verbose)
	fprintf (stderr, " ... reading cross sections\n");

      /* O3 absorption */
      if (levelsum (&(output->atm), MOL_O3) > 0) {
	status = crs_o3 (&output->crs,
			 first, 
			 output->wl.lambda_t, output->wl.nlambda_t, 
			 output->atm.microphys.temper, output->atm.nlev, 
			 input.filename[FN_PATH], input.ozone_spectrum);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up O3 cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading O3 cross section because concentration is zero\n");
      }      

      /* CO2 absorption */
      if (levelsum (&(output->atm), MOL_CO2) > 0) {
	status = crs_co2 (&output->crs,
			  first, 
			  output->wl.lambda_t, output->wl.nlambda_t, 
			  output->atm.microphys.temper, output->atm.nlev, 
			  input.filename[FN_PATH], input.co2_spectrum);
      
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up CO2 cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading CO2 cross section because concentration is zero\n");
      }      

      /* O2 absorption */
      if (levelsum (&(output->atm), MOL_O2) > 0) {
	status = crs_o2 (&output->crs,
			 first,
			 output->wl.lambda_t, output->wl.nlambda_t, 
			 output->atm.microphys.temper, output->atm.nlev, 
			 input.filename[FN_PATH], input.o2_spectrum, input.quiet);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up O2 cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading O2 cross section because concentration is zero\n");
      }      

      /* NO2 absorption */
      if (levelsum (&(output->atm), MOL_NO2) > 0) {
	status = crs_no2 (&output->crs,
			  output->wl.lambda_t, output->wl.nlambda_t, 
			  output->atm.microphys.temper, output->atm.nlev, 
			  input.filename[FN_PATH], input.no2_spectrum);
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up NO2 cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading NO2 cross section because concentration is zero\n");
      }      

      /* BRO absorption */
      if (levelsum (&(output->atm), MOL_BRO) > 0) {
	status = crs_bro (&output->crs,
			  output->wl.lambda_t, output->wl.nlambda_t, 
			  output->atm.microphys.temper, output->atm.nlev, 
			  input.filename[FN_PATH], input.bro_spectrum);
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up BrO cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading BRO cross section because concentration is zero\n");
      }      

      /* OCLO absorption */
      if (levelsum (&(output->atm), MOL_OCLO) > 0) {
	status = crs_oclo (&output->crs, first, 
			   output->wl.lambda_t, output->wl.nlambda_t, 
			   output->atm.microphys.temper, output->atm.nlev, 
			   input.filename[FN_PATH], input.oclo_spectrum);
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up OClO cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading OCLO cross section because concentration is zero\n");
      }      
	
      /* HCHO absorption */
      if (levelsum (&(output->atm), MOL_HCHO) > 0) {
	status = crs_hcho (&output->crs, first, 
			   output->wl.lambda_t, output->wl.nlambda_t, 
			   output->atm.microphys.temper, output->atm.nlev, 
			   input.filename[FN_PATH], input.hcho_spectrum);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d setting up HCHO cross section.\n", status);
	  return status;
	}
      }
      else {
	if (input.verbose)
	  fprintf (stderr, " ... not reading HCHO cross section because concentration is zero\n");
      }      

      /* Next see if any cross section are given with crs_file, if so, overwrite */
      
      for (i=0; i<MOL_NN; i++) {
	if (strlen(output->crs.filename[i]) > 0) {
	  status = read_crs (&output->crs, 
			     output->wl.lambda_t, output->wl.nlambda_t, 
			     output->atm.microphys.temper, output->atm.nlev, 
			     input.filename[FN_PATH], i);
	}
	if (status!=0) {
	  fprintf (stderr, "Error %d reading cross section file %s.\n", 
		   status, output->crs.filename[i]);
	  return status;
	}
      }

      /* Set crs for amf calculations */
      if (output->atm.microphys.denstab_id > 0) {
	for (lc=0; lc<output->atm.nlev; lc++) 
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->crs.crs_amf[iv][lc] = output->crs.crs[lc][output->atm.microphys.denstab_id][iv];
      }

      break;

    case MOLABS_FILE_MONO:
    case MOLABS_FILE_SPEC:
    case MOLABS_NONE:
      if (input.verbose)
	fprintf (stderr, " ... not reading cross sections\n");

      /* no need to read anything */
      break;

    default:
      fprintf (stderr, "Error, unknown molecular absorption option %d\n", output->atm.molabs);
      return -1;
    }      

    break;

  case CK_FU:
  case CK_GENERIC:       /* ??? need to check if that is appropriate here ??? */
  case CK_AVHRR_KRATZ:   /* ??? need to check if this is appropriate here ??? */
  case CK_LOWTRAN:       /* ??? need to check if this is appropriate here ??? */
  case CK_KATO:
  case CK_KATO2:
  case CK_KATO2_96:

    if (input.atm.interpol_method_press != INTERP_METHOD_LOG) {
      fprintf (stderr, "pressure interpolation is assumed to be logarithmic\n");
      fprintf (stderr, "please change programm code for kato in crs_ck (in ck.c) \n");
      return -1;
    }

    /* setup correlated k-tables */
    status = crs_ck (&(output->ck),     
		     output->wl.lambda_t, output->wl.nlambda_t, 
		     output->atm.microphys.temper, output->atm.microphys.temper_avg, 
                     output->atm.microphys.press, output->atm.zd,
		     output->atm.microphys.dens, output->atm.microphys.dens_avg,
                     output->atm.nlev,
		     output->mixing_ratio, input.ck_h2ocont, input.ck_o4abs, output->atm.sza_t,
		     input.filename[FN_PATH], input.filename[FN_CK_GENERIC], first, input.verbose,
                     input.wcipa, input.icipa, output->nipa,
                     input.wc.wv_saturation, input.ic.wv_saturation,
		     &output->crs_ck,
                     input, output);   /* in ck.c */
    
    if (status!=0) {
      fprintf (stderr, "Error %d setting up cross sections for correlated k\n", status);
      return status;
    }

    break;

  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme\n");
    return -1;
    break;
  }
  
  first=0;
  return 0;
}

/**************************************************************/
/* Setup molecular (Rayleigh) scattering                      */
/**************************************************************/

int setup_rayleigh (input_struct input, output_struct *output)
{
  int status=0, iv=0;

  /* allocate memory for depolarization and initialize */
  output->rayleigh_depol = calloc (output->wl.nlambda_t, sizeof(float));

  for (iv=0; iv<output->wl.nlambda_t; iv++)
    if (input.rayleigh_crs == RAYLEIGH_NICOLET && input.rayleigh_depol == (float) -999.0)
      output->rayleigh_depol[iv] = 0;
    else
      output->rayleigh_depol[iv] = input.rayleigh_depol;

  switch (input.rayleigh_crs) {
  case RAYLEIGH_BODHAINE:
    if (input.verbose)
      fprintf (stderr, " ... Using Bodhaine et al. Rayleigh scattering cross section\n");
    status = crs_rayleigh_bodhaine (output->wl.lambda_t, output->wl.nlambda_t, 
				    output->mixing_ratio[MX_CO2], 
				    output->rayleigh_depol, output->crs.crs);

    if (status!=0) {
      fprintf (stderr, "Error %d setting up Rayleigh cross section from Bodhaine\n", status);
    }
    break;
  case RAYLEIGH_NICOLET:
    if (input.verbose)
      fprintf (stderr, " ... Using Nicolet Rayleigh scattering cross section\n");
    status = crs_rayleigh_nicolet (output->wl.lambda_t, output->wl.nlambda_t, output->crs.crs);
    if (status!=0) {
      fprintf (stderr, "Error %d setting up Rayleigh cross section from Nicolet\n", status);
    }
    break;
  default:
    fprintf (stderr, "Error, unknown ozone absorption cross section %d\n", status);
  }

  return status;
}


/***********************************************************************************/
/* Function: crs_o3                                                                */
/* Description:                                                                    */
/*  Calculates ozone cross sections as a function of wavelength and temperature.   */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  int first: If 1, temperature coefficients for ozone will be calculated at      */
/*             wavelengths lambda; if 0, the temperature coefficients from an      */
/*             earlier call will be used. first should obviously be 1 the first    */
/*             time crs_o3 is called. It should also be 1 if lambda is changed.    */
/*             It does not need to be 1 if temper is changed.                      */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The ozone cross section data to be used,                        */
/*                 1 (Molina and Molina, 1986);                                    */
/*                 2 (Daumont et al., 1992; Malicet et al., 1995)                  */
/*                 3 (Paur and Bass, 1985)                                         */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_o3 (crs_out_struct *crs_out, int first, 
		   float *lambda, int n_lambda, 
		   float *temper, int ntemp, char *path,
		   int type)
{
  int ia=0, ib=0, it=0, iv=0;
  int n_crs=0, status=0;
  static float fact=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  static float *lambda_raw=NULL, *coeff_crs[3]={NULL,NULL,NULL};
  float a=0.0, b=0.0, gradient=0.0;

  if (first) {
    if (n_lambda <= 0 && ntemp != 0)
      return 2;
    else { 
      if (n_lambda != 0 &&  ntemp <= 0)
	return 3;
    }

    switch (type) {
    case O3_MOLINA:
      strcpy(filenm,"crs/crs_o3_mol_cf.dat");
      fact     = 1.0E-20;
      break;
    case O3_DAUMONT:
      strcpy(filenm,"crs/crs_o3_dau_cf.dat");
      fact     = 1.0E-20;
      break;
    case O3_BASSPAUR:
      strcpy(filenm,"crs/crs_o3_pab_cf.dat");
      fact     = 1.0E-20;
      break;
    default:
      fprintf (stderr, "Error, unknown ozone absorption cross section %d\n", status);
      return status;
    }
    
    strcpy(fullnm,path);
    strcat(fullnm,filenm);

    status = read_4c_file_float (fullnm, 
				 &lambda_raw, 
				 &(coeff_crs[0]), &(coeff_crs[1]), &(coeff_crs[2]), 
				 &n_crs);

    if (status!=0) { 
      fprintf (stderr, "Error %d opening file %s\n", status, fullnm); 
      return status; 
    } 

  } /* if (first) */
  
  for (iv=0; iv<n_lambda; iv++) {
    ia=closest_above(lambda[iv], lambda_raw, n_crs);
    ib=closest_below(lambda[iv], lambda_raw, n_crs);

    if (ia > (n_crs-1) || ib<0) {
      fprintf (stderr, "Error (crs_o3), wavelength %.1f nm out of range!\n", lambda[iv]);
      fprintf (stderr, "Cannot do a monochromatic calculation at this wavelength;\n");
      fprintf (stderr, "for infrared, have a look at the 'correlated_k' option!\n");
      fprintf (stderr, "lambda: %f n_crs: %d ia: %d ib: %d\n",lambda[iv], n_crs, ia, ib);
      return 9;
    }
    
    for (it=0; it<ntemp; it++) {
      if (ia > (n_crs-1) || ib<0)
	crs_out->crs[it][MOL_O3][iv] = 0;
      else {
	a = coeff_crs[0][ia] + 
	  coeff_crs[1][ia] * (temper[it]-273.15) + 
	  coeff_crs[2][ia]*(temper[it]-273.15)*(temper[it]-273.15);
	if (ia == ib) {
	  crs_out->crs[it][MOL_O3][iv] = a*fact;
	}
	else {
	  b=coeff_crs[0][ib] + 
	    coeff_crs[1][ib] * (temper[it]-273.15) + 
	    coeff_crs[2][ib]*(temper[it]-273.15)*(temper[it]-273.15);
	  gradient=(a-b)/(lambda_raw[ia]-lambda_raw[ib]);
	  crs_out->crs[it][MOL_O3][iv] = (gradient*(lambda[iv]-lambda_raw[ib])+b) * fact;
	  /*printf("%d %d %f %e\n",it,iv,lambda[iv],crs_out->crs[it][MOL_O3][iv]);*/
	}
      }
    }
  }

  return 0;
}



/***********************************************************************************/
/* Function: crs_co2                                                               */
/* Description:                                                                    */
/*  Calculates CO2 cross sections as a function of wavelength and temperature.     */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  int first: If 1, temperature coefficients for ozone will be calculated at      */
/*             wavelengths lambda; if 0, the temperature coefficients from an      */
/*             earlier call will be used. first should obviously be 1 the first    */
/*             time crs_co2 is called. It should also be 1 if lambda is changed.   */
/*             It does not need to be 1 if temper is changed.                      */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The CO2 cross section data to be used,                          */
/*                 1 (Lewis and Carver, 1983);                                     */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_co2 (crs_out_struct *crs_out, int first, 
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type)
{
  int ia=0, ib=0, it=0, iv=0;
  int n_crs=0, status=0;
  static float fact=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  static float *lambda_raw=NULL, *coeff_crs[3]={NULL,NULL,NULL};
  float a=0, b=0, gradient=0;

  if (first) {
    if (n_lambda <= 0 && ntemp != 0)
      return 2;
    else { 
      if (n_lambda != 0 &&  ntemp <= 0)
	return 3;
    }
    
    switch (type) {
    case CO2_LEWIS:
      strcpy(filenm,"crs/crs_co2_lew_cf.dat");
      break;
    default:
      fprintf (stderr, "Error, unknown CO2 absorption cross section %d\n", type);
      return status;
    }

    
    strcpy(fullnm,path);
    strcat(fullnm,filenm);
    
    status = read_4c_file_float (fullnm, 
				 &lambda_raw, 
				 &(coeff_crs[0]), &(coeff_crs[1]), &(coeff_crs[2]), 
				 &n_crs);
    
    if (status!=0) { 
      fprintf (stderr, "Error %d reading file %s\n", status, fullnm); 
      return status; 
    } 
    
  } /* if (first) */
  
  for (iv=0; iv<n_lambda; iv++) {
    ia=closest_above(lambda[iv], lambda_raw, n_crs);
    ib=closest_below(lambda[iv], lambda_raw, n_crs);

    if (ia > (n_crs-1) || ib<0) {
      fprintf (stderr, "Error (crs_co2), wavelength %.1f nm out of range!\n", lambda[iv]);
      fprintf (stderr, "Cannot do a monochromatic calculation at this wavelength;\n");
      fprintf (stderr, "for infrared, have a look at the 'correlated_k' option!\n");
      fprintf (stderr, "lambda: %f n_crs: %d ia: %d ib: %d\n",lambda[iv], n_crs, ia, ib);
      return 9;
    }

    for (it=0; it<ntemp; it++) {
      a = coeff_crs[0][ia] + 
	coeff_crs[1][ia] * (temper[it]-273.15) + 
	coeff_crs[2][ia]*(temper[it]-273.15)*(temper[it]-273.15);
      if (ia == ib) {
	crs_out->crs[it][MOL_CO2][iv] = a*fact;
      }
      else {
	b = coeff_crs[0][ib] + 
	  coeff_crs[1][ib] * (temper[it]-273.15) + 
	  coeff_crs[2][ib] * (temper[it]-273.15)*(temper[it]-273.15);
	gradient = (a-b) / (lambda_raw[ia]-lambda_raw[ib]);
	crs_out->crs[it][MOL_CO2][iv] = (gradient*(lambda[iv]-lambda_raw[ib])+b) * fact;
      }
    }
  }

  return 0;
}



/***********************************************************************************/
/* Function: crs_o2                                                                */
/* Description:                                                                    */
/*  Calculates O2 cross sections as a function of wavelength and temperature.      */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  int first: If 1, temperature coefficients for O2 will be calculated at         */
/*             wavelengths lambda; if 0, the temperature coefficients from an      */
/*             earlier call will be used. first should obviously be 1 the first    */
/*             time crs_o2 is called. It should also be 1 if lambda is changed.    */
/*             It does not need to be 1 if temper is changed.                      */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The O2 cross section data to be used,                           */
/*                 1 (default);                                                    */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_o2 (crs_out_struct *crs_out, int first, 
		   float *lambda, int n_lambda, 
		   float *temper, int ntemp, char *path,
		   int type, int quiet)
{
  int ia=0, ib=0, it=0, iv=0, i=0, j=0;
  int n_crs_SR=0, n_crs_Yo88=0, n_crs_Yo2005=0, n_crs_Og75=0, n_crs_Le83=0, status=0, n_tmp=0;
  static float fact_SR=0;
  char filenm[3][FILENAME_MAX+200]={"","",""}, fullnm[FILENAME_MAX+200]="";
  static float *lambda_SR=NULL, *lambda_Yo88=NULL, *lambda_Yo2005=NULL, *lambda_Og75=NULL, *lambda_Le83=NULL;
  static float *crs_SR[3][3], *crs_Yo88=NULL, *crs_Yo2005=NULL, *crs_Og75=NULL, *crs_Le83_288=NULL, *crs_Le83_alpha=NULL;
  float *tmp=NULL;
  float *tmpw=NULL, *tmp0=NULL, *tmp1=NULL, *tmp2=NULL;
  float a=0.0, b=0.0, d=0.0, gradient=0.0;
  
  if (first) {
    if (n_lambda <= 0 && ntemp != 0)
      return 2;
    else { 
      if (n_lambda != 0 &&  ntemp <= 0)
	return 3;
    }

    /* read Schumann-Runge from Minschwaner et al [1992] */
    if (lambda[0] <= (float) (1e7/49000.5) && lambda[n_lambda-1] >= (float) (1e7/57000.0)) {

      if (!quiet) {
	fprintf (stderr, " ... reading temperature-dependent O2 Schumann-Runge cross sections\n");
	fprintf (stderr, " ... by Minschwaner et al. [1992]\n");
      }

      switch (type) {
      case O2_DEFAULT:
	strcpy(filenm[0],"crs/Minschwaner_O2_fitcoef_cold.txt");
	strcpy(filenm[1],"crs/Minschwaner_O2_fitcoef_mid.txt");
	strcpy(filenm[2],"crs/Minschwaner_O2_fitcoef_hot.txt");
	fact_SR = 1.0E-20;
	break;
      default:
	fprintf (stderr, "Error, unknown oxygen absorption cross section %d\n", status);
	return status;
      }
      
      for (i=0;i<3;i++) {
	strcpy(fullnm,path);
	strcat(fullnm,filenm[i]);
	status = read_4c_file_float (fullnm, 
				     &tmp, 
				     &(crs_SR[i][0]), &(crs_SR[i][1]), &(crs_SR[i][2]), 
				     &n_tmp);
	
	if (status!=0) { 
	  fprintf (stderr, "Error %d reading file %s\n", status, fullnm); 
	  return status; 
	} 

	/* sort by wavelength, assuming that Minschwaner was sorted by wavenumber */
	tmpw = calloc (n_tmp, sizeof(float));
	tmp0 = calloc (n_tmp, sizeof(float));
	tmp1 = calloc (n_tmp, sizeof(float));
	tmp2 = calloc (n_tmp, sizeof(float));
	
	for (j=0; j<n_tmp; j++) {
	  tmpw[j] = tmp[n_tmp-1-j];
	  tmp0[j] = crs_SR[i][0][n_tmp-1-j];
	  tmp1[j] = crs_SR[i][1][n_tmp-1-j];
	  tmp2[j] = crs_SR[i][2][n_tmp-1-j];
	}
	
	for (j=0; j<n_tmp; j++) {
	  tmp[n_tmp-1-j]       = tmpw[j];
	  crs_SR[i][0][j] = tmp0[j];
	  crs_SR[i][1][j] = tmp1[j];
	  crs_SR[i][2][j] = tmp2[j];
	}
	
	free(tmpw); free(tmp0); free(tmp1); free(tmp2); 
	

	if (i==0) {
	  n_crs_SR = n_tmp;
	  lambda_SR = calloc (n_crs_SR, sizeof(float));
	  for (iv=0; iv<n_crs_SR; iv++)
	    lambda_SR[iv] = 1.0E7/tmp[n_crs_SR-1-iv];
	  
	  /* check if the data is sorted by wavelength */
	  for (iv=1;iv<n_crs_SR;iv++)
	    if (lambda_SR[iv] <= lambda_SR[iv-1]) {
	      fprintf (stderr, "Error, %s not sorted by wavenumber\n", fullnm);
	      return -1;
	    }
	}
	else {
	  if (n_crs_SR!=n_tmp) {
	    fprintf (stderr, "Error, Minschwaner et al. [1992] wavelength grids\n");
	    fprintf (stderr, "not equal: different number of grid points.\n");
	    return -1;
	  }
	  for (iv=0;iv<n_crs_SR;iv++) {
	    if (fabs((lambda_SR[iv]-1.0E7/tmp[n_crs_SR-1-iv])/lambda_SR[iv]) > 1e-6)  {
	      fprintf (stderr, "Error, Minschwaner et al. [1992] wavelength grids\n");
	      fprintf (stderr, "not equal: %f vs. %f\n", 
		       lambda_SR[iv], (float) (1.0E7/tmp[n_crs_SR-1-iv]));
	      return -1;
	    }
	  }
	}
	free (tmp);
      }
    }

    /* Yoshino et al. [1988] */
    if (lambda[0] <= 240.0 && lambda[n_lambda-1] >= 205.0) { 

      if (!quiet)
	fprintf (stderr, " ... reading O2 Herzberg continuum, 205-240nm, from Yoshino et al. [1988]\n");

      switch (type) {
      case O2_DEFAULT:
	strcpy(filenm[0],"crs/crs_o2_yoshino1988.dat");
	break;
      default:
	fprintf (stderr, "Error, unknown oxygen absorption cross section %d\n", status);
	return status;
      }

      strcpy(fullnm,path);
      strcat(fullnm,filenm[0]);

      status = read_2c_file_float (fullnm, 
				   &lambda_Yo88, 
				   &crs_Yo88, &n_crs_Yo88);

      if (status!=0) { 
	fprintf (stderr, "Error %d reading file %s\n", status, fullnm); 
	return status; 
      } 
    }

    /* Yoshino et al. [2005], 160nm - 175.5 nm */
    if (lambda[0] <= (float) (1e7/57000.0) && lambda[n_lambda-1] >= 160.0) { 

      if (!quiet)
	fprintf (stderr, " ... reading O2 Schumann-Runge continuum, 160-175nm, from Yoshino et al. [2005]\n");

      switch (type) {
      case O2_DEFAULT:
	strcpy(filenm[0],"crs/crs_o2_yoshino2005.dat");
	break;
      default:
	fprintf (stderr, "Error, unknown oxygen absorption cross section %d\n", status);
	return status;
      }

      strcpy(fullnm,path);
      strcat(fullnm,filenm[0]);

      status = read_2c_file_float (fullnm, 
				   &lambda_Yo2005, 
				   &crs_Yo2005, &n_crs_Yo2005);

      if (status!=0) { 
	fprintf (stderr, "Error %d reading file %s\n", status, fullnm); 
	return status; 
      } 
    }      

    /* Ogawa and Ogawa [1975], 108nm - 160 nm */
    if (lambda[0] <= (160.0) && lambda[n_lambda-1] >= 108.0) { 

      if (!quiet)
	fprintf (stderr, " ... reading Schumann-Runge continuum, 108-160nm, from Ogawa and Ogawa [1975]\n");

      switch (type) {
      case O2_DEFAULT:
	strcpy(filenm[0],"crs/crs_o2_ogawa1975.dat");
	break;
      default:
	fprintf (stderr, "Error, unknown oxygen absorption cross section %d\n", status);
	return status;
      }

      strcpy(fullnm,path);
      strcat(fullnm,filenm[0]);

      status = read_2c_file_float (fullnm, 
				   &lambda_Og75, 
				   &crs_Og75, &n_crs_Og75);

      if (status!=0) { 
	fprintf (stderr, "Error %d reading file %s\n", status, fullnm); 
	return status; 
      } 
    }      

    /* Lewis [1983], 121.4nm - 121.9 nm (Lyman alpha region) */
    if (lambda[0] <= (121.9) && lambda[n_lambda-1] >= 121.4) { 

      if (!quiet)
	fprintf (stderr, " ... reading Schumann-Runge continuum, 121.4-121.9nm, from Lewis [1983]\n");

      switch (type) {
      case O2_DEFAULT:
	strcpy(filenm[0],"crs/crs_o2_lewis1983.dat");
	break;
      default:
	fprintf (stderr, "Error, unknown oxygen absorption cross section %d\n", status);
	return status;
      }

      strcpy(fullnm,path);
      strcat(fullnm,filenm[0]);

      status = read_3c_file_float (fullnm, 
				   &lambda_Le83, 
				   &crs_Le83_288, &crs_Le83_alpha, &n_crs_Le83);

      if (status!=0) { 
	fprintf (stderr, "Error %d reading file %s\n", status, fullnm); 
	return status; 
      } 
    }
  } /* if (first) */
  

  /* calculate absorption cross section */
  for (iv=0; iv<n_lambda; iv++) {
    
    /* Schumann-Runge range */
    if (lambda[iv] <= (float) (1e7/49000.5) && lambda[iv] >= (float) (1e7/57000.0)) {

      ia=closest_above(lambda[iv], lambda_SR, n_crs_SR);
      ib=closest_below(lambda[iv], lambda_SR, n_crs_SR);
      
      for (it=0; it<ntemp; it++) {
	if (temper[it]>=130&&temper[it]<=190)
	  i=0;
	
	if (temper[it]>190&&temper[it]<=280)
	  i=1;
	
	if (temper[it]>280&&temper[it]<=500)
	  i=2;
	
	if (temper[it]<130||temper[it]>500) {
	  fprintf (stderr, "Error, temperature %f out of range of the\n", temper[it]);
	  fprintf (stderr, "Minschaner et al. [1997] Schumann-Runge parameterization\n");
	  
	  return -1;
	}
	
	/* d is the transformed temperature variable defined in eq. (6): */
	/* d = [(T-100)/10]**2.                                          */
	
	d = ((temper[it]-100.0)/10.0)*((temper[it]-100.0)/10.0);
	
	/* assume that cross section outside the wavelength region is zero */
	if (ia > (n_crs_SR-1) || ib<0)
	  crs_out->crs[it][MOL_O2][iv]=0;
	else {
	  a = crs_SR[i][0][ia] * d*d + crs_SR[i][1][ia] * d + crs_SR[i][2][ia];
	  if (ia == ib) {
	    crs_out->crs[it][MOL_O2][iv] = a * fact_SR;
	  }
	  else {
	    b=crs_SR[i][0][ib] * d*d + crs_SR[i][1][ib] * d + crs_SR[i][2][ib];
	    gradient=(a-b)/(lambda_SR[ia]-lambda_SR[ib]);
	    crs_out->crs[it][MOL_O2][iv] = (gradient*(lambda[iv]-lambda_SR[ib])+b) * fact_SR;
	  }
	}
      }
    }

    /* add Herzberg continuum to Schumann-Runge range; eq (3) from Yoshino et al [1988] */
    if (lambda[iv] >= 194.0 && lambda[iv] < 205.0)
      for (it=0; it<ntemp; it++)
	crs_out->crs[it][MOL_O2][iv] += 6.884*(204.87/lambda[iv]) * exp(-69.7374*pow(log(204.87/lambda[iv]),2))*1e-24;
    
    
    /* Yoshino et al. [1988] */
    if (lambda[iv] >= 205 && lambda[iv] <= 240) { 

      ia = closest_above(lambda[iv], lambda_Yo88, n_crs_Yo88);
      ib = closest_below(lambda[iv], lambda_Yo88, n_crs_Yo88);

      for (it=0; it<ntemp; it++) {
	if (ia > (n_crs_Yo88-1) || ib<0)
	  crs_out->crs[it][MOL_O2][iv]=0;
	else {
	  if (ia == ib)
	    crs_out->crs[it][MOL_O2][iv] = crs_Yo88[ia];
	  else {
	    gradient=(crs_Yo88[ia] - crs_Yo88[ib]) / (lambda_Yo88[ia] - lambda_Yo88[ib]);
	    crs_out->crs[it][MOL_O2][iv] = (gradient * (lambda[iv]-lambda_Yo88[ib]) + crs_Yo88[ib]);
	  }
	}
      }
    }

    /* Yoshino et al. [2005] */
    if (lambda[iv] <= (float) (1e7/57000.0) && lambda[iv] >= 160.0) { 

      ia = closest_above(lambda[iv], lambda_Yo2005, n_crs_Yo2005);
      ib = closest_below(lambda[iv], lambda_Yo2005, n_crs_Yo2005);

      for (it=0; it<ntemp; it++) {
	if (ia > (n_crs_Yo2005-1) || ib<0)
	  crs_out->crs[it][MOL_O2][iv]=0;
	else {
	  if (ia == ib)
	    crs_out->crs[it][MOL_O2][iv] = crs_Yo2005[ia];
	  else {
	    gradient=(crs_Yo2005[ia] - crs_Yo2005[ib]) / (lambda_Yo2005[ia] - lambda_Yo2005[ib]);
	    crs_out->crs[it][MOL_O2][iv] = (gradient * (lambda[iv]-lambda_Yo2005[ib]) + crs_Yo2005[ib]);
	  }
	}
      }
    }

    /* Ogawa and Ogawa [1975], 108nm - 160 nm */
    if (lambda[iv] <= 160.0 && lambda[iv] >= 108.0) { 

      ia = closest_above(lambda[iv], lambda_Og75, n_crs_Og75);
      ib = closest_below(lambda[iv], lambda_Og75, n_crs_Og75);

      for (it=0; it<ntemp; it++) {
	if (ia > (n_crs_Og75-1) || ib<0)
	  crs_out->crs[it][MOL_O2][iv]=0;
	else {
	  if (ia == ib)
	    crs_out->crs[it][MOL_O2][iv] = crs_Og75[ia];
	  else {
	    gradient=(crs_Og75[ia] - crs_Og75[ib]) / (lambda_Og75[ia] - lambda_Og75[ib]);
	    crs_out->crs[it][MOL_O2][iv] = (gradient * (lambda[iv]-lambda_Og75[ib]) + crs_Og75[ib]);
	  }
	}
      }
    }

    /* Lewis [1983], 121.4nm - 121.9 nm (Lyman alpha region) */
    if (lambda[iv] <= (121.9) && lambda[iv] >= 121.4) { 

      ia = closest_above(lambda[iv], lambda_Le83, n_crs_Le83);
      ib = closest_below(lambda[iv], lambda_Le83, n_crs_Le83);

      /* temperature dependence in the Lyman alpha region */
      for (it=0; it<ntemp; it++) {

	a = crs_Le83_288[ia] + crs_Le83_alpha[ia] * (temper[it]-288.0);
	if (ia > (n_crs_Le83-1) || ib<0)
	  crs_out->crs[it][MOL_O2][iv]=0;
	else {
	  if (ia == ib)
	    crs_out->crs[it][MOL_O2][iv] = a;
	  else {
	    b = crs_Le83_288[ib] + crs_Le83_alpha[ib] * (temper[it]-288.0);
	    gradient=(a-b)/(lambda_Le83[ia]-lambda_Le83[ib]);
	    crs_out->crs[it][MOL_O2][iv] = (gradient*(lambda[iv]-lambda_Le83[ib])+b);
	  }
	}
      }
    }

  }
  
  return 0;
}



/***********************************************************************************/
/* Function: crs_no2                                                               */
/* Description:                                                                    */
/*  Calculates NO2 cross sections as a function of wavelength and temperature.     */ 
/*  ??? Temperature dependence is not yet considered ???                           */
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The  cross section data to be used,                             */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_no2 (crs_out_struct *crs_out,
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type)
{
  int it=0, iv=0;
  int n_raw=0, status=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  
  double crs=0;

  double *lambda_raw=NULL, *coeffc_raw=NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;

  if (n_lambda <= 0 && ntemp != 0)
    return 2;
  else { 
    if (n_lambda != 0 &&  ntemp <= 0)
      return 3;
  }

  switch (type) {
  case NO2_GOME:
    strcpy(filenm,"crs/crs_no2_gom.dat");
    break;
  default:
    fprintf (stderr, "Error, unknown NO2 absorption cross section %d\n", type);
    return status;
  }
    
  strcpy(fullnm,path);
  strcat(fullnm,filenm);

  /* read cross section data */
  status = read_2c_file (fullnm,
			 &lambda_raw, &coeffc_raw, &n_raw);
  
  if (status!=0) { 
    fprintf (stderr, "Error %d reading %s\n", status, fullnm); 
    return status; 
  } 
  
  /* linear interpolation to user-defined wavelength grid */
  status = linear_coeffc (lambda_raw, coeffc_raw, n_raw, 
			  &a0, &a1, &a2, &a3);
  if (status!=0) { 
    fprintf (stderr, "Error %d calculating interpolation coefficients for NO2 cross section\n", 
	     status); 
    return status; 
  } 

  for (iv=0; iv<n_lambda; iv++) {

    if (lambda[iv] < lambda_raw[0] || lambda[iv] > lambda_raw[n_raw-1])
      crs=0;
    else
      status = calc_splined_value (lambda[iv], &crs, lambda_raw, n_raw,
				   a0, a1, a2, a3);
    

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating NO2 cross section to %f nm\n", 
	       status, lambda[iv]);
      return status;
    }

    /* ??? temperature dependence not yet considered; ??? */
    /* ??? use constant cross section for all levels  ??? */
    for (it=0; it<ntemp; it++) 
      crs_out->crs[it][MOL_NO2][iv] = crs;
  }


  return 0;
}

/***********************************************************************************/
/* Function: crs_bro                                                               */
/* Description:                                                                    */
/*  Calculates BrO cross sections as a function of wavelength and temperature.     */ 
/*  ??? Temperature dependence is not yet considered ???                           */
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The  cross section data to be used,                             */
/*                 no choices yet                                                  */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_bro (crs_out_struct *crs_out,
		    float *lambda, int n_lambda, 
		    float *temper, int ntemp, char *path,
		    int type)
{
  int it=0, iv=0;
  int n_raw=0, status=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  
  double crs=0;

  double *lambda_raw=NULL, *coeffc_raw=NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;

  if (n_lambda <= 0 && ntemp != 0)
    return 2;
  else { 
    if (n_lambda != 0 &&  ntemp <= 0)
      return 3;
  }

  switch (type) {
  case BRO_IASB:
    strcpy(filenm,"crs/crs_bro_wahner_228.dat");
    break;
  default:
    fprintf (stderr, "Error, unknown BrO absorption cross section %d\n", type);
    return status;
  }
    
  strcpy(fullnm,path);
  strcat(fullnm,filenm);

  /* read cross section data */
  status = read_2c_file (fullnm,
			 &lambda_raw, &coeffc_raw, &n_raw);
  
  if (status!=0) { 
    fprintf (stderr, "Error %d reading %s\n", status, fullnm); 
    return status; 
  } 
  
  /* linear interpolation to user-defined wavelength grid */
  status = linear_coeffc (lambda_raw, coeffc_raw, n_raw, 
			  &a0, &a1, &a2, &a3);
  if (status!=0) { 
    fprintf (stderr, "Error %d calculating interpolation coefficients for BrO cross section\n", 
	     status); 
    return status; 
  } 

  for (iv=0; iv<n_lambda; iv++) {

    if (lambda[iv] < lambda_raw[0] || lambda[iv] > lambda_raw[n_raw-1])
      crs=0;
    else
      status = calc_splined_value (lambda[iv], &crs, lambda_raw, n_raw,
				   a0, a1, a2, a3);
    

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating BrO cross section to %f nm\n", 
	       status, lambda[iv]);
      return status;
    }

    /* ??? temperature dependence not yet considered; ??? */
    /* ??? use constant cross section for all levels  ??? */
    for (it=0; it<ntemp; it++) 
      crs_out->crs[it][MOL_BRO][iv] = crs;
  }


  return 0;
}



/***********************************************************************************/
/* Function: crs_oclo                                                              */
/* Description:                                                                    */
/*  Calculates oclo  cross sections as a function of wavelength and temperature.   */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  int first: If 1, temperature coefficients for ozone will be calculated at      */
/*             wavelengths lambda; if 0, the temperature coefficients from an      */
/*             earlier call will be used. first should obviously be 1 the first    */
/*             time crs_oclo is called. It should also be 1 if lambda is changed.  */
/*             It does not need to be 1 if temper is changed.                      */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The cross section data to be used,                              */
/*                 1 (Molina and Molina, 1986);                                    */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_oclo (crs_out_struct *crs_out, int first, 
		   float *lambda, int n_lambda, 
		   float *temper, int ntemp, char *path,
		   int type)
{
  int ia=0, ib=0, it=0, iv=0;
  int n_crs=0, status=0;
  static float fact=1.e-20;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  static float *lambda_raw=NULL, *crs_at_T[3]={NULL,NULL,NULL};
  float a=0.0, b=0.0, gradient=0.0;
  double T0=204, T1=296, T2=378;
  double s0=0, s1=0, s2=0;

  if (first) {
    if (n_lambda <= 0 && ntemp != 0)
      return 2;
    else { 
      if (n_lambda != 0 &&  ntemp <= 0)
	return 3;
    }

    switch (type) {
    case OCLO_WAHNER:
      strcpy(filenm,"crs/crs_oclo.dat");
      break;
    default:
      fprintf (stderr, "Error, unknown oclo absorption cross section %d\n", status);
      return status;
    }
    
    strcpy(fullnm,path);
    strcat(fullnm,filenm);

    status = read_4c_file_float (fullnm, 
				 &lambda_raw, 
				 &(crs_at_T[0]), &(crs_at_T[1]), &(crs_at_T[2]), 
				 &n_crs);

    if (status!=0) { 
      fprintf (stderr, "Error %d opening file %s\n", status, fullnm); 
      return status; 
    } 

  } /* if (first) */
  
  for (iv=0; iv<n_lambda; iv++) {
    ia=closest_above(lambda[iv], lambda_raw, n_crs);
    ib=closest_below(lambda[iv], lambda_raw, n_crs);

    if (ia > (n_crs-1) || ib<0) {
      fprintf (stderr, "Error (crs_oclo), wavelength %.1f nm out of range!\n", lambda[iv]);
      fprintf (stderr, "Cannot do a monochromatic calculation at this wavelength;\n");
      fprintf (stderr, "for infrared, have a look at the 'correlated_k' option!\n");
      fprintf (stderr, "lambda: %f n_crs: %d ia: %d ib: %d\n",lambda[iv], n_crs, ia, ib);
      return 9;
    }

    for (it=0; it<ntemp; it++) {
      if (fabs(temper[it]-T0) < 0.001) {
	a = crs_at_T[0][ia];
	b = crs_at_T[0][ib];
      }
      else if (fabs(temper[it]-T1) < 0.001) {
	a = crs_at_T[1][ia];
	b = crs_at_T[1][ib];
      }
      else if (fabs(temper[it]-T2) < 0.001) {
	a = crs_at_T[2][ia];
	b = crs_at_T[2][ib];
      }
      else {
	s0 = crs_at_T[0][ia]*1e+20;
	s1 = crs_at_T[1][ia]*1e+20;
	s2 = crs_at_T[2][ia]*1e+20;
	a = quadratic(temper[it], T0, T1, T2, s0, s1, s2);
	s0 = crs_at_T[0][ib]*1e+20;
	s1 = crs_at_T[1][ib]*1e+20;
	s2 = crs_at_T[2][ib]*1e+20;
	b = quadratic(temper[it], T0, T1, T2, s0, s1, s2);
	a *= fact;
	b *= fact;
      }
      
      if (ia == ib) {
	crs_out->crs[it][MOL_OCLO][iv] = a;
      }
      else {
	gradient=(a-b)/(lambda_raw[ia]-lambda_raw[ib]);
	crs_out->crs[it][MOL_OCLO][iv] = (gradient*(lambda[iv]-lambda_raw[ib])+b);
      }
    }
  }

  return 0;
}

/***********************************************************************************/
/* Function: crs_hcho                                                              */
/* Description:                                                                    */
/*  Calculates hcho  cross sections as a function of wavelength and temperature.   */ 
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  int first: If 1, temperature coefficients for ozone will be calculated at      */
/*             wavelengths lambda; if 0, the temperature coefficients from an      */
/*             earlier call will be used. first should obviously be 1 the first    */
/*             time crs_hcho is called. It should also be 1 if lambda is changed.  */
/*             It does not need to be 1 if temper is changed.                      */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int type:      The cross section data to be used,                              */
/*                 1 (Molina and Molina, 1986);                                    */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int crs_hcho (crs_out_struct *crs_out, int first, 
		     float *lambda, int n_lambda, 
		     float *temper, int ntemp, char *path,
		     int type)
{
  int it=0, iv=0;
  int n_raw=0, status=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  
  double crs=0;

  double *lambda_raw=NULL, *coeffc_raw=NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;

  if (n_lambda <= 0 && ntemp != 0)
    return 2;
  else { 
    if (n_lambda != 0 &&  ntemp <= 0)
      return 3;
  }

  switch (type) {
  case HCHO_IASB:
    strcpy(filenm,"crs/crs_hcho_cantrell.dat");
    break;
  default:
    fprintf (stderr, "Error, unknown HCHO absorption cross section %d\n", status);
    return status;
  }
    
  strcpy(fullnm,path);
  strcat(fullnm,filenm);
  
  /* read cross section data */
  status = read_2c_file (fullnm,
			 &lambda_raw, &coeffc_raw, &n_raw);
  
  if (status!=0) { 
    fprintf (stderr, "Error %d reading %s\n", status, fullnm); 
    return status; 
  } 
  
  /* linear interpolation to user-defined wavelength grid */
  status = linear_coeffc (lambda_raw, coeffc_raw, n_raw, 
			  &a0, &a1, &a2, &a3);
  if (status!=0) { 
    fprintf (stderr, "Error %d calculating interpolation coefficients for HCHO cross section\n", 
	     status); 
    return status; 
  } 
  
  for (iv=0; iv<n_lambda; iv++) {
    
    if (lambda[iv] < lambda_raw[0] || lambda[iv] > lambda_raw[n_raw-1])
      crs=0;
    else
      status = calc_splined_value (lambda[iv], &crs, lambda_raw, n_raw,
				   a0, a1, a2, a3);
    
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating HCHO cross section to %f nm\n", 
	       status, lambda[iv]);
      return status;
    }
    
    /* ??? temperature dependence not yet considered; ??? */
    /* ??? use constant cross section for all levels  ??? */
    for (it=0; it<ntemp; it++) 
      crs_out->crs[it][MOL_HCHO][iv] = crs;
  }
  
  return 0;
}

/***********************************************************************************/
/* Function: read_crs                                                              */
/* Description:                                                                    */
/*  Read a cross section versus wavelength file. No temperature dependence may     */
/*  be specified.                                                                  */
/*                                                                                 */
/* Parameters:                                                                     */
/*  crs_out_struct *crs_out: structure holding the cross sections                  */
/*  float *lambda: Array of wavelengths [nm].                                      */
/*  int n_lambda:  Number of wavelengths.                                          */
/*  float *temper: Temperatures [K] at the atmospheric levels.                     */
/*  int n_temp:    Number of temperatures.                                         */ 
/*  char *path:    Location of the libRadtran data files.                          */
/*  int mol_id:    Molecule identifier                                             */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

static int read_crs (crs_out_struct *crs_out,
		     float *lambda, int n_lambda, 
		     float *temper, int ntemp, char *path, int mol_id)
{
  int it=0, iv=0;
  int n_raw=0, status=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  
  double crs=0;

  double *lambda_raw=NULL, *coeffc_raw=NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;

  if (n_lambda <= 0 && ntemp != 0)
    return 2;
  else { 
    if (n_lambda != 0 &&  ntemp <= 0)
      return 3;
  }

  strcpy(filenm,crs_out->filename[mol_id]);
    
  strcat(fullnm,filenm);

  /* read cross section data */
  status = read_2c_file (fullnm, &lambda_raw, &coeffc_raw, &n_raw);
  
  if (status!=0) { 
    fprintf (stderr, "Error %d reading %s\n", status, fullnm); 
    return status; 
  } 
  
  /* linear interpolation to user-defined wavelength grid */
  status = linear_coeffc (lambda_raw, coeffc_raw, n_raw, 
			  &a0, &a1, &a2, &a3);
  if (status!=0) { 
    fprintf (stderr, "Error %d calculating interpolation coefficients for cross section file %s\n", 
	     status, crs_out->filename[mol_id]); 
    return status; 
  } 

  for (iv=0; iv<n_lambda; iv++) {

    if (lambda[iv] < lambda_raw[0] || lambda[iv] > lambda_raw[n_raw-1])
      crs=0;
    else
      status = calc_splined_value (lambda[iv], &crs, lambda_raw, n_raw,
				   a0, a1, a2, a3);
    
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating cross section file %s to %f nm\n", 
	       status, crs_out->filename[mol_id], lambda[iv]);
      return status;
    }

    /* ??? temperature dependence not yet considered; ??? */
    /* ??? use constant cross section for all levels  ??? */
    for (it=0; it<ntemp; it++) 
      crs_out->crs[it][mol_id][iv] = crs;
  }

  return 0;

}

static int closest_above (float lambda, float *lambda_raw, int n_crs) 
{
  int iv=0;
  
  for (iv=n_crs-1;iv>=0;iv--)  {
    if ((lambda_raw[iv]-lambda) == 0.0) 
      return iv;
    else {
      if ((lambda_raw[iv]-lambda) < 0.0) 
	return iv+1;
    }
  }
  return 0;
}


static int closest_below (float lambda, float *lambda_raw, int n_crs) 
{
  int iv=0;

  for (iv=0; iv<n_crs; iv++) {
    if ((lambda_raw[iv]-lambda) == 0.0) 
      return iv;
    else  {
      if ((lambda_raw[iv]-lambda) > 0.0) 
	return iv-1;
    }
  }
  return 0;
}

static double quadratic(double x, double x0, double x1, double x2, double y0, double y1, double y2)
{
  double y=0.0;
  double a=0, b=0, c=0;
  
  c = (1/(x2-x1))*(((y2-y0)/(x2-x0))-((y1-y0)/(x1-x0)));
  b = (y1-y0)/(x1-x0)-c*(x1+x0);
  a = y0 - b*x0 -c*x0*x0;

  y = a + b*x + c*x*x;

  return y;
}



static double levelsum (atm_out_struct *atm, int mol)
{
  int lc=0;
  double sum=0;

  for (lc=0;lc<atm->nlev;lc++)
    sum+=atm->microphys.dens[mol][lc];

  return sum;
}
