#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <getopt.h>

#include "ascii.h"
#include "numeric.h"
#include "sofi.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


#define PROGRAM "sofi"
#define VERS "0.1"

/*************************************************************/
/* print usage information                                   */
/*************************************************************/


static void print_usage (char *filename)
{
  fprintf (stderr, "%s %s - Calculate irradiance or diffus flux during a total eclipse\n\n",
           PROGRAM, VERS); 
  fprintf (stderr, "written by Claudia Emde,\n");
  fprintf (stderr, "           DLR, e-Mail claudia.emde@dlr.de\n");
  fprintf (stderr, "Version %s finished May 30, 2006\n\n", VERS);
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "-h            display help message\n");
  fprintf (stderr, "-z <degrees>  solar zenith angle\n");
  fprintf (stderr, "-a <degrees>  solar azimuth angle \n");
  fprintf (stderr, "-l <nm>       wavelength \n");
  fprintf (stderr, "-b <km>       height of atmosphere \n");
  fprintf (stderr, "-c            calculate clearsky \n");
  fprintf (stderr, "-s            calculate standard deviation \n");
  fprintf (stderr, "USAGE: %s [options] <filename>\n\n", filename);
  fprintf (stderr, "%s calculates the irradiance or the diffus flux \n", PROGRAM); 
  fprintf (stderr, "during the solar eclipse at 29-03-2006.\n");
  fprintf (stderr, "The input file is the photon distribution\n");
  fprintf (stderr, "at the top of the atmosphere calculated using  \n");
  fprintf (stderr, "mystic (backward), e.g. mc35.rad. \n");
  fprintf (stderr, "The tool can also be used to calculte clearsky \n");
  fprintf (stderr, "irradiance or flux provided the mystic file above. \n");
  fprintf (stderr, "\n");
}

static int get_options (int argc, char **argv,
                        char *programname,
                        char *infilename,
                        double *sza,
                        double *saa,
                        double *lambda, 
                        double *z,
                        double *dx,
                        double *dy, 
                        int *clear, 
                        int *std)
{
  int c=0;
  char *dummy=NULL;
  
  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);
  
  /* set defaults */
  strcpy (infilename, "");
  *sza = 0.0;
  *saa = 0.0;
  *lambda = 400.0;
  *z=50.0;
  *dx=0.0;
  *dy=0.0;
  /* By default it is assumed that sofi should be calculated*/
  *clear=0;
  *std=0;
  
  while ((c=getopt (argc, argv, "hz:a:l:b:x:y:cs")) != EOF)  {
    switch(c)  {
    case 'h':  /* help */
      print_usage (programname);
      return (-1);
      break;
    case 'z':  /* number of legendre polynomials */
      *sza = strtod(optarg, &dummy);
      break;
    case 'a':
      *saa = strtod(optarg, &dummy);
      break;
    case 'l':
      *lambda = strtod(optarg, &dummy);
      break;
    case 'b':
      *z = strtod(optarg, &dummy);
      break;
    case 'x':
      *dx = strtod(optarg, &dummy);
      break; 
    case 'y':
      *dy = strtod(optarg, &dummy);
      break;
    case 'c':
      *clear = 1;
      break;
    case 's':
      *std = 1;
      break;
    default:
      print_usage (programname);
      return (-1);
    }
  }


  /* check number of remaining command line arguments */
  if (argc - optind != 1)  {
    print_usage (programname);
    return -1;
  }
  
  strncpy (infilename, argv[optind+0], FILENAME_MAX);
  
  return 0;  /* if o.k. */
}


int main(int argc, char **argv)
{
  int i=0, status=0;

  char programname[FILENAME_MAX], infilename[FILENAME_MAX]="";
  
  double sza=0.0;
  double saa=0.0;
  double lambda=0.0;
  double z_toa=50.0;
  double dx_moon=0.0;
  double dy_moon=0.0;
  int clear=0;
  int std=0; 
  double r_earth=6371211.0;
  double irr_tot=0.0;
  double r, r2, alpha, x, y, z, dx, dy, dx_sph, dy_sph;
  int i_re=0;
  double wvnmlo, wvnmhi;
  double *pd, *irr; 
  double cossaa, sinsaa, cossza, tansza, cos_alphasaa, sin_alphasaa;

  status = get_options (argc, argv, programname, infilename, &sza,
                        &saa, &lambda, &z_toa, &dx_moon, &dy_moon,  &clear, &std);

  if (status!=0)
    return status;
  
  /* Read mystic output */

  double *fi=NULL;
  double *se=NULL;
  double *th=NULL;
  double *fo=NULL;
  double *fv=NULL;
  double *si=NULL;
  double *sv=NULL;
  double *ei=NULL;
  int L=0;
  
  read_8c_file (infilename, &fi, &se, &th, &fo, &fv, &si,&sv, &ei, &L); 
  fprintf (stderr, " ... read %d data points from %s\n", L, infilename);
  
  if (status!=0) {
    fprintf (stderr, "error %d reading %s\n", status, infilename);
    return status;
  }

  if (sza >= 90) {
    fprintf (stderr, "Solar eclipse calculations not implemented for \n");
    fprintf (stderr, "solar zenith angles greater or equal 90 degrees. \n");
    return -1;
  }
  
  if (!clear) {
    /* Calculate probability density function. */
   
    wvnmhi=1.0/(lambda*100.0)*1e9;
    wvnmlo=1.0/(lambda*100.0)*1e9;
  
    pd = (double*) calloc(5000,sizeof(double));
    
    sample_photons_sofi(wvnmlo,wvnmhi,pd);

    /*for (i=0; i<5000; i++)
      fprintf(stdout, "%d   %g \n", i, pd[i]);*/ 
    
    /*height of the model atmosphere, convert to meter*/
    z_toa*=1000;
    dx_moon*=1000;
    dy_moon*=1000; 
    
    /*often needed expressions*/ 
    cossaa=cos(saa*PI/180.0);
    sinsaa=sin(saa*PI/180.0);
    tansza=tan(sza*PI/180.0);
    cossza=cos(sza*PI/180.0);
   
    
    /* calculate centre of umbral shadow at TOA */
    r=z_toa*tansza; 
    dx=r*sinsaa; 
    dy=r*cossaa;
    
    /* Sampled irradiance */
    irr = (double*) calloc(L,sizeof(double));
    
    for(i=0; i<L; i++)
      { 
        /* Coordinates on spherical sampling area */
        x=fi[i]-fi[0]-(fi[L-1]+fi[0])/2;
        y=se[i]-se[0]-(se[L-1]+se[0])/2;
        z=sqrt((r_earth+z_toa)*(r_earth+z_toa)-x*x-y*y)-r_earth;

        /* Calculate shift due to spherical sampling domain */
        /* Projection on surface where the weighting function is calculated */
        dx_sph=(z_toa-z)*tansza*sinsaa;
        dy_sph=(z_toa-z)*tansza*cossaa;
                    
        /* x and y coordinates, (x,y)=0 corresponds to centre of ellipse */ 
        x+=(dx-dx_moon+dx_sph);
        y+=(dy-dy_moon+dy_sph);
        
        alpha = atan(x/y);
        r2=x*x+y*y;

        cos_alphasaa=cos(alpha-saa/180.0*PI);
        sin_alphasaa=sin(alpha-saa/180.0*PI);

        /* ellitical shadow needs to be transformed to spherical shadow 
           corresponding to sun in nadir.*/
        i_re=(int)(fabs(sqrt(r2/(cos_alphasaa*cos_alphasaa/(cossza*cossza)+
                                 sin_alphasaa*sin_alphasaa)))/1000.0);
        

        /*fprintf(stderr, "i_re: %d %g \n", ei[i], pd[i_re]);*/
        if(std)
          irr[i]=ei[i]*ei[i]*pd[i_re]*pd[i_re];
        else
          irr[i]=ei[i]*pd[i_re];
        
        irr_tot+=irr[i];
                      
        /* Use this if you want to plot the sofi weighted photon distribution*/
        /*fprintf(stdout,  "%f  %f %g \n " ,fi[i], se[i], irr[i]);*/
      
      }
  }
  else
    for(i=0; i<L; i++)
      irr_tot+=ei[i];
  
  if(std)
    irr_tot=sqrt(irr_tot);

  fprintf(stderr,  "Total irradiance/flux: %g \n", irr_tot);
  return 0;
}  


