/*--------------------------------------------------------------------
 * $Id: ancillary.c,v 1.111 2006/12/17 23:45:52 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "uvspec.h"
#include "ckdfu.h"
#include "ascii.h"
#include "numeric.h"
#include "solver.h"

#if HAVE_LIBGSL 
#include <gsl/gsl_math.h>
#include <gsl/gsl_diff.h>
#endif

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

#define EPSILON 1E-6
#define s2day   3600.0*24.0 /* seconds to day */
#define mW2W    1.E-3;       /* mW to W */

/* prototypes of internal functions */
static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit);


static float radiance2bt (float rad, float *wvnmlo, float *wvnmhi, float *filter, int n);

static int output2bt (input_struct input, output_struct *output);
static int output2bt3D (input_struct input, output_struct *output);

static int read_photon_file (char *filename, float *lambda_t, int nlamdba_t, float **fraction);

static int select_wavelength_indices (float *lambda, int nlambda,
				      float *lambda_lower, float *lambda_upper, 
				      int start_index, int end_index, 
				      int quiet,
				      int *lower, int *upper);

static int set_transmittance_wl_grid (wl_inp_struct *wl_inp, wl_out_struct *wl_out, int quiet);
static int set_transmittance_wl_grid_lowtran (wl_inp_struct *wl_inp, wl_out_struct *wl_out, int quiet);
static int write_spectral3D (input_struct input, output_struct *output);



/***************************************************************/
/* Setup the wavelength grid for the transmission calculation. */
/***************************************************************/

int setup_wlgrid (input_struct input, output_struct *output)
{     
  int iv=0, status=0, monochromatic=0, wavelength_read=0, read_thermal_bands_file=0;
  float *lambda_lower=NULL, *lambda_upper=NULL;

  
  /* in case of RGB conversion we need an internal wavelength grid */
  if (input.processing==PROCESS_RGB || input.processing==PROCESS_RGBNORM) {
    /* reset wavelength range */
    input.wl.start = -999;
    input.wl.end   = -999;
    
    /* and replace transmittance grid */
    strcpy (input.filename[FN_WLTRANS], input.filename[FN_PATH]);
    strcat (input.filename[FN_WLTRANS], "solar_flux/rgb");

    if (!input.quiet) {
      fprintf (stderr, " ... ignoring user-defined wavelength range and using\n");
      fprintf (stderr, " ...   internal wavelength grid from %s\n", input.filename[FN_WLTRANS]);
    }
  }    


  if (input.wl.start>0 && input.wl.end>0) {
    output->wl.start = input.wl.start;
    output->wl.end   = input.wl.end;
  }

  output->wl.type = WLGRID_NONE;

  if (input.ck_scheme == CK_NONE || input.ck_scheme == CK_LOWTRAN) {   /* no correlated-k */

    if (strlen(input.filename[FN_WLBANDS]) > 0 && input.source == SRC_THERMAL) {  /* thermal_bands_file */
      
      output->wl.type = WLGRID_BANDS;
      output->wl.ignore_solar_file = 1; /* ignore solar file if FN_WLBANDS is used */

      if (!input.quiet) 
	fprintf (stderr, " ... reading thermal_bands_file from %s\n", input.filename[FN_WLBANDS]);

      /* read center wavelength and band limits [wavenumbers] from file */
      status = read_3c_file_float (input.filename[FN_WLBANDS],
				   &(output->wl.lambda_t),
				   &(lambda_lower), 
				   &(lambda_upper),
				   &output->wl.nlambda_t);

      if (status!=0) {
	fprintf (stderr, "Error %d opening %s\n", status, input.filename[FN_WLBANDS]);
	return status;
      }

      /* convert lower and upper wavelength bounds to wavenumbers */
      output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	output->wl.wvnmlo_t[iv] = 1.0E7 / lambda_upper[iv];
	output->wl.wvnmhi_t[iv] = 1.0E7 / lambda_lower[iv];
      }

      /* free temporary wavelengths arrays */
      free (lambda_lower);
      free (lambda_upper);

      if (input.wl.start < 0 || input.wl.end < 0) {
	output->wl.start = output->wl.lambda_t[0];
	output->wl.end   = output->wl.lambda_t[output->wl.nlambda_t-1];
	if (!input.quiet) 
	    fprintf (stderr, " ... setting wavelength range to %f - %fnm\n",
		     output->wl.start, output->wl.end);
      }

      /* select wavelength range */
      status = select_wavelength_indices (output->wl.lambda_t, output->wl.nlambda_t, 
				 &(output->wl.start), &(output->wl.end), 
				 input.wl.start_index, input.wl.end_index, 
				 input.quiet,
				 &(output->wl.nlambda_rte_lower), &(output->wl.nlambda_rte_upper));
      if (status!=0) {
	fprintf (stderr, "Error %d at no correlated-k in setup_wlgrid (ancillary.c) \n", status);
	return status;
      }


      read_thermal_bands_file = 1;

    }  /* end thermal_bands_file */
    else {  

      if (strlen(input.filename[FN_WLTRANS]) > 0) {  /* transmittance_wl_file */
	
	output->wl.type = WLGRID_USER;

	/* read wavelengths from transmittance file */
	status = read_1c_file_float (input.filename[FN_WLTRANS],
				     &(output->wl.lambda_t),
				     &output->wl.nlambda_t);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d opening %s\n", status, input.filename[FN_WLTRANS]);
	  return status;
	}
	
	if (input.wl.start < 0 || input.wl.end < 0) {
	  output->wl.start = output->wl.lambda_t[0];
	  output->wl.end   = output->wl.lambda_t[output->wl.nlambda_t-1];
	  if (!input.quiet) 
	    fprintf (stderr, " ... setting wavelength range to %f - %fnm\n",
		     output->wl.start, output->wl.end);
	}
	
	
	/* select wavelength range */
	status = select_wavelength_indices (output->wl.lambda_t, output->wl.nlambda_t, 
				   &(output->wl.start), &(output->wl.end), 
				   input.wl.start_index, input.wl.end_index, 
				   input.quiet,
				   &(output->wl.nlambda_rte_lower), &(output->wl.nlambda_rte_upper));
        if (status!=0) {
	  fprintf (stderr, "Error %d at (transmittance_wl_file) in setup_wlgrid (ancillary.c)\n", status);
	  return status;
        }
      } /* end transmittance_wl_file */
      else {

	if (strlen(input.filename[FN_MOLABS]) > 0) {

	  output->wl.type = WLGRID_MOLABS;

	  if (!input.quiet) {
	    fprintf (stderr, " ... molecular_tau_file specified but computational wavelength grid\n");
	    fprintf (stderr, " ... not explicitely defined; reading the wavelength grid\n");
	    fprintf (stderr, " ... from molecular_tau_file %s\n", input.filename[FN_MOLABS]);
	  }

	  status = read_molecular_absorption_lambda (input.filename[FN_MOLABS], input.quiet,
						     &output->wl.lambda_t, &output->wl.nlambda_t, &monochromatic);
	  if (status!=0) {
	    fprintf (stderr, "Error %d reading wavelength grid from %s\n", 
		     status, input.filename[FN_MOLABS]);
	    return -1;
	  }

	  wavelength_read=0;
	  if (!monochromatic) {
	    status = select_wavelength_indices (output->wl.lambda_t, output->wl.nlambda_t, 
				       &(output->wl.start), &(output->wl.end), 
				       input.wl.start_index, input.wl.end_index, 
				       input.quiet,
				       &(output->wl.nlambda_rte_lower), &(output->wl.nlambda_rte_upper));
            if (status!=0) {
	      fprintf (stderr, "Error %d at (!monochromatic) in setup_wlgrid (ancillary.c)\n", status);
	      return status;
            }

	    wavelength_read=1;
	  }
	}

	if (!wavelength_read) {
	  output->wl.type = WLGRID_UVSPEC;
	  
	  /* set up a reasonable wavelength grid for the transmittance calculation */
	  if (input.ck_scheme == CK_NONE)
	    status = set_transmittance_wl_grid(&input.wl, &output->wl, input.quiet);
	  else
	    status = set_transmittance_wl_grid_lowtran(&input.wl, &output->wl, input.quiet);

	  if (status!=0) {
	    fprintf (stderr, "Error %d setting up wavelength grid\n", status);
	    return status;
	  }
	  
	  /* here we need to calculate the full wavelength range */
	  output->wl.nlambda_rte_lower = 0;
	  output->wl.nlambda_rte_upper = output->wl.nlambda_t-1;
	}
      }

      /* allocate memory for the band limit arrays */
      output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      
      /* the default bandwidth is 1cm-1 to get the emittance per cm-1; */
      /* input.bandwidth can be set with thermal_bandwidth             */       
      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	output->wl.wvnmlo_t[iv] = 1.0E7 / output->wl.lambda_t[iv] - input.bandwidth / 2.0;
	output->wl.wvnmhi_t[iv] = output->wl.wvnmlo_t[iv] + input.bandwidth;
      }
    }
      
    /* check if the transmittance grid covers the required wavelength range */
    if (output->wl.lambda_t[0] > output->wl.start || 
	output->wl.lambda_t[output->wl.nlambda_t-1] < output->wl.end) {
      fprintf (stderr, "Error, transmittance calculation grid (%f - %f nm)\n", 
	       output->wl.lambda_t[0], output->wl.lambda_t[output->wl.nlambda_t-1]);
      fprintf (stderr, "does not cover the user-defined wavelength range (%f - %f nm)\n",
	       output->wl.start, output->wl.end);
      return -1;
    }
    
    /* now setup the ck structure for the LOWTRAN/SBDART table */
    if (input.ck_scheme == CK_LOWTRAN) {
      
      /* read LOWTRAN/SBDART tables */
      status = sbdart_readtables (&(output->ck), 
				  output->wl.nlambda_t,
				  input.filename[FN_PATH],
				  input.rte.mc.filename[FN_MC_PHOTONS],
				  input.quiet);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by sbdart_readtables()\n", status); 
	return status;
      }
    }

    /* need to read the photons file */
    if (input.ck_scheme == CK_NONE) {


      if (strlen(input.rte.mc.filename[FN_MC_PHOTONS])>0) {
	status = read_photon_file (input.rte.mc.filename[FN_MC_PHOTONS], 
				   output->wl.lambda_t, output->wl.nlambda_t, &(output->wl.pfraction));
	if (status!=0) {
	  fprintf (stderr, "Error %d reading photon file name %s\n", status, input.rte.mc.filename[FN_MC_PHOTONS]);
	  return status;
	}
      }
      else {
	output->wl.pfraction = calloc (output->wl.nlambda_t, sizeof (float));
	for (iv=0; iv<output->wl.nlambda_t; iv++)
	  output->wl.pfraction[iv] = 1.0 / (float) output->wl.nlambda_t;
      }
    }
  }
  else { /* correlated-k */

    output->wl.type = WLGRID_CK;

    /* read information about wavelength grid and quadrature points */
    switch(input.ck_scheme) {
    case CK_KATO:
    case CK_KATO2:
    case CK_KATO2_96:
      
      /* read Kato et al. [1999] tables */
      status = kato_readtables (input.ck_scheme, &(output->ck), input.filename[FN_PATH],
				input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by kato_readtables()\n", status); 
	return status;
      }
      
      break;

    case CK_FU:

      /* read Fu and Liou [1992/93] tables */
      status = fu_readtables (&(output->ck), input.filename[FN_PATH],
			      input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by fu_readtables()\n", status); 
	return status;
      }

      break;

    case CK_AVHRR_KRATZ:

      /* read Kratz [1999] tables */
      status = avhrr_kratz_readtables (&(output->ck), input.filename[FN_PATH],
				       input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by avhrr_kratz_readtables()\n", status); 
	return status;
      }
      
      break;

    case CK_GENERIC:

      /* read generic tables in CDF format */
      status = ck_generic_readtables (&(output->ck), input.filename[FN_CK_GENERIC],
				      input.rte.mc.filename[FN_MC_PHOTONS]);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ck_generic_readtables()\n", status); 
	return status;
      }
      
      break;

    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme\n");
      return -1;
    }


    /* copy center wavelengths to transmittance grid  */
    /* and set wavenumber intervals                   */

    output->wl.nlambda_t = output->ck.n_wvl;
    output->wl.lambda_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    
    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      output->wl.lambda_t[iv] = output->ck.wvlc[iv+1];
      output->wl.wvnmlo_t[iv] = output->ck.wvnlo[iv+1];
      output->wl.wvnmhi_t[iv] = output->ck.wvnhi[iv+1];
    }

    if (input.wl.start < 0 || input.wl.end < 0) {
      output->wl.start = output->wl.lambda_t[0];
      output->wl.end   = output->wl.lambda_t[output->wl.nlambda_t-1];
      if (!input.quiet) 
	  fprintf (stderr, " ... setting wavelength range to %f - %fnm\n",
		   output->wl.start, output->wl.end);
    }

    /* select the required wavelength range */
    status = select_wavelength_indices (output->wl.lambda_t, output->wl.nlambda_t, 
			       &(output->wl.start), &(output->wl.end), 
			       input.wl.start_index, input.wl.end_index, 
			       input.quiet,
			       &(output->wl.nlambda_rte_lower), &(output->wl.nlambda_rte_upper));
    if (status!=0) {
      fprintf (stderr, "Error %d at (copy center wavelengths) in setup_wlgrid (ancillary.c)\n", status);
      return status;
    }
  }
    
  if (!input.quiet)
    if (strlen(input.filename[FN_WLBANDS]) > 0 && !read_thermal_bands_file)
      fprintf (stderr, " ... ignoring thermal_bands_file %s\n", input.filename[FN_WLBANDS]);


  /* inform the user about wavelength selection */
  if (!input.quiet) {
    if (output->wl.nlambda_rte_upper != output->wl.nlambda_t-1 ||
	output->wl.nlambda_rte_lower != 0) {
      fprintf (stderr, " ... user wavelength range: %f - %f nm\n", 
	       input.wl.start, input.wl.end);
      fprintf (stderr, " ... selected wavelength indices %d - %d\n", 
	       output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper);
      fprintf (stderr, " ... %f - %f nm\n", 
	       output->wl.lambda_t[output->wl.nlambda_rte_lower], 
	       output->wl.lambda_t[output->wl.nlambda_rte_upper]);
    }
  }

  return 0;
}


/***************************************************************/
/* Interpolate from one wavelength grid to another, either     */
/* with natural cubic splines (linear=0), linear (linear=1),   */ 
/* or logarithmic (linear=2) if the input is sorted in         */
/* descending order, descend needs to be set to 1.             */
/***************************************************************/

int arb_wvn (int n_old_x, float *old_x, float *old_y,
	     int n_new_x, float *new_x, float *new_y,
	     int linear, int descend)
{
  int i=0, j=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double *tmp_y=NULL;
  double ynew=0;
  double tst1=0, tst2=0;
  double sum=0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  if (!descend) {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[i];
      y[i] = (double) old_y[i];
    }
  }
  else {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[n_old_x-1-i];
      y[i] = (double) old_y[n_old_x-1-i];
    }
  }

  /* check if the array is now sorted in ascending order */
  for (i=0; i<n_old_x-1; i++)
    if (x[i]>=x[i+1]) {
      fprintf (stderr, "Error, x not sorted in ascending order, arb_wvn() (in ancillary.c) \n");
      fprintf (stderr, "x[%d] =  %f, x[%d] = %f\n", i, x[i], i+1, x[i+1]);
      return -1;
    }

  /* check that range of x_new is <= range of x_old (if not checked -> segfault) */
  /* check minimum */
  for (i=0;i<n_new_x;i++)
    if (new_x[i] < x[0]) {
      fprintf (stderr, "Error, x_new[%d] = %f < min(x_old) = %f, arb_wvn() (in ancillary.c) \n",
                        i, new_x[i], x[0]);
      return -2;
    }
  /* check maximum */
  for (i=0;i<n_new_x;i++)
    if (new_x[i] > x[n_old_x-1]) {
      fprintf (stderr, "Error, x_new[%d] = %f > max(x_old) = %f, arb_wvn() (in ancillary.c) \n",
                        i, new_x[i], x[n_old_x-1]);
      return -3;
    }


  /* special check for log_spline interpolation */
  if (linear == 4) {
    sum = 0.0;
    for (i=0; i<n_old_x; i++) {
      sum += y[i];
      if (y[i] < 0.0) {
        fprintf (stderr, "Error, cannot use log_spline interpolation for profile with negative values\n");
        return -4;
      }
    }
    if (sum == 0.0) { /* x is zero in each layer, but grid must be changed anyway */
      free(x);
      free(y);
      for (i=0; i<n_new_x; i++) 
        new_y[i] = 0.0;
      return 0;
    }  
    else{             
      for (i=0; i<n_old_x; i++) {
        if (y[i] > 0.0)
          y[i] = log(y[i]);
        else 
          y[i] = -10.E+99; /* this is a little bit cheating, but it works */
      }
    }
  }

  switch (linear) {
  case 0: /* spline */
  case 4: /* log spline */

    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 1: /* linear */
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 2: /* log */
    tmp_y = (double *) calloc (n_new_x, sizeof(double));
    for (i=0; i<n_new_x; i++)  {
      
      j=0;
      while (new_x[i] > x[j])	
	j++;
    
      if (j>0) 
	j--;
    
      /* logarithmic interpolation (if reasonable) */
      tst1 = fabs(y[j+1]-y[j]);
      tst2 = (y[j+1]<y[j] ? y[j+1] : y[j]);
      if (tst1<=0.001*y[j] || tst2<=0) /* linear */
	tmp_y[i] = y[j] + (new_x[i]-x[j]) / (x[j+1]-x[j]) * (y[j+1]-y[j]);
      else /* logarithmic */
	tmp_y[i] = exp (log(y[j]) + (new_x[i]-x[j])/(x[j+1]-x[j]) * 
			(log(y[j+1])-log(y[j])));
    }
			  
    /* first interpolate, then copy because then source */
    /* and target may be one and the same               */
    for (i=0; i<n_new_x; i++)
      new_y[i] = (float) tmp_y[i];

    free (tmp_y);
    break;

  case 3: /* linear mixing ratio */
    fprintf (stderr, "Error, linear mixing interpolation not possible with arb_wvn.\n");
    fprintf (stderr, "Please use interpolate_density instead. \n");
    return -5;
    break;

  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -6;
  }


  switch (linear) {
  case 0: /* spline */
  case 1: /* linear */
  case 4: /* log spline */
  
    for (i=0; i<n_new_x; i++)  {
      status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);
      
      /* this check is new in 0.99-alpha-6; so far, no error */
      /* was reported if a value could not be interpolated   */
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by calc_splined_value (%g)\n",
		 status, new_x[i]);
	fprintf (stderr, " %d data points, x[0] = %g, x[%d] = %g\n", n_old_x, x[0], n_old_x-1, x[n_old_x-1]);
	return status;
      }

      new_y[i] = (float) ynew;
    }
    
    free(a0); free(a1); free(a2); free(a3);
    break; 

  case 2:
    /* no need to do anything because interpolation has already been done above */
    break;

  case 3: /* linear mixing ratio */
    fprintf (stderr, "Error, linear mixing interpolation not possible with arb_wvn.\n");
    fprintf (stderr, "Please use interpolate_density instead. \n");
    return -1;
    break;

  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -7;
  }

  if (linear == 4)
    for (i=0; i<n_new_x; i++)
      new_y[i] = exp(new_y[i]);

  free(x); free(y);
  
  return 0;
}




/***************************************************************/
/* Interpolate from one wavelength grid to another, either     */
/* with natural cubic splines (linear=0) or linear (linear=1); */ 
/* if the input is sorted in descending order, descend needs   */
/* to be set to 1. In contrast to arb_wvn(), values that       */
/* cannot be interpolated will be set to 0.                    */
/***************************************************************/

int arb_wvn_zero (int n_old_x, float *old_x, float *old_y,  
		  int n_new_x, float *new_x, float *new_y, 
		  int linear, int descend)
{
  int i=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  if (!descend) {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[i];
      y[i] = (double) old_y[i];
    }
  }
  else {
    for (i=0; i<n_old_x; i++) {
      x[i] = (double) old_x[n_old_x-1-i];
      y[i] = (double) old_y[n_old_x-1-i];
    }
  }

  switch (linear) {
  case 0:
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 1:
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
    break;

  case 2:
  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -1;
  }

  
  switch (linear) {
  case 0:
  case 1:
    for (i=0; i<n_new_x; i++)  {
      status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);
      
      if (status==0)
	new_y[i] = (float) ynew;
      else 
	new_y[i] = 0;
    }

    free(a0); free(a1); free(a2); free(a3);

    break;

  case 2:
  default:
    fprintf (stderr, "Error, unknown interpolation type %d\n", linear);
    return -1;
  }

  free(x); free(y);
    
  return 0;
}






/***************************************************************/
/* Interpolate from one wavelength grid to another, either     */
/* with natural cubic splines (linear=0) or linear (linear=1); */ 
/* interpolate only to a selected range of the output grid     */
/* (including n_new_x_lower and n_new_x_upper)                 */
/***************************************************************/

int arb_wvn2 (int n_old_x, float *old_x, float *old_y,  
	      int n_new_x_lower, int n_new_x_upper, float *new_x, float *new_y, int linear)
{
  int i=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  for (i=0;i<n_old_x;i++) {
    x[i] = (double) old_x[i];
    y[i] = (double) old_y[i];
  }

  if (linear) {
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
  }
  else {
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
  }


  for (i=n_new_x_lower; i<=n_new_x_upper; i++)  {
    status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);

    /* this check is new in 0.99-alpha-6; so far, no error */
    /* was reported if a value could not be interpolated   */

    if (status!=0) {
      fprintf (stderr, "Error %d returned by calc_splined_value (%g)\n",
	       status, new_x[i]);
      fprintf (stderr, " %d data points, x[0] = %g, x[%d] = %g\n", n_old_x, x[0], n_old_x-1, x[n_old_x-1]);
      return status;
    }

    new_y[i] = (float) ynew;
  }

  free(a0); free(a1); free(a2); free(a3);
  free(x); free(y);

  return 0;
}



/**************************************************************/
/* Interpolate a given profile (x,y) to a new grid xnew;      */
/* data are written to original array y and memory is         */
/* automatically reallocated                                  */
/**************************************************************/

int interpolate_profile (float *x, float **y, int n,  
			 float *xnew, int nnew, int interpol_method, 
                         int quiet)
{
  int status=0;
  float *tmp = calloc (nnew, sizeof(float));
  int lc=0, lc2=0, first=0;

  switch (interpol_method) {
  case INTERP_METHOD_SPLINE:
  case INTERP_METHOD_LINEAR:
  case INTERP_METHOD_LOG:
  case INTERP_METHOD_LOG_SPLINE:
    status = arb_wvn (n, x, *y, nnew, xnew, tmp, interpol_method, 1);
    break;

  case INTERP_METHOD_LINMIX:
    fprintf (stderr, "Error, linear mixing ratio not possible with interpolate_profile\n");
    fprintf (stderr, "Please use interpolate_density instead!\n");    
    return -1;    
    break;

  default:
    fprintf (stderr, "Error, unknown interpolation method in interpolate_profile\n");
    return -1;
  }

  /* testing for negativ density values */

  switch (interpol_method) {
  case INTERP_METHOD_SPLINE:
    for (lc=0; lc<nnew; lc++) {
      if (tmp[lc] < 0.0) {
        if (!quiet) {
          if (first == 0)
            fprintf (stderr, "***  Warning: In interpolate_density, automatic correction of negative density to 0.0\n");
          fprintf (stderr, "***  Warning: z[%d]= %5.1f, dens[%d]= %12.7e -> dens[%d]= 0.0\n",lc, x[lc],lc,tmp[lc],lc);
          tmp[lc] = 0.0;
          first = 1;
        }
      }
    }
    break;

  case INTERP_METHOD_LINEAR:
  case INTERP_METHOD_LOG:
  case INTERP_METHOD_LINMIX:
  case INTERP_METHOD_LOG_SPLINE:
    for (lc=0; lc<nnew; lc++) {
      if (tmp[lc] < 0.0) {
        if (abs(tmp[lc]) < EPSILON) {
          if (!quiet) {
            fprintf (stderr, "*** Warning, small negative density detected during interpolate_profile!\n"); 
            fprintf (stderr, "*** in layer = %d, dens = %e.\n", lc, tmp[lc]); 
            fprintf (stderr, "*** This may happen, when dens profiles contain 0.0 values.\n");  
            fprintf (stderr, "*** Setting to 0.0 automatically.\n"); 
	  }
	  tmp[lc]=0.0;
	}
        else {
          fprintf (stderr, "Error, negative density detected in interpolate_profile! lc=%d \n",lc);
          for (lc2=0; lc2<nnew; lc2++)
            fprintf (stderr," ### dens[%d]=%e\n", lc2, tmp[lc2]);
          return -1;
	}
      }
    }
    break;
  default:
    fprintf (stderr, "Error, unknown interpolation method in interpolate_profile\n");
    return -1;
  }

  if (status != 0) {
    fprintf (stderr, "Error %d interpolating profile\n", status);
    return status;
  }

  free(*y);
  *y = tmp;

  return 0;
}


/**************************************************************/
/* Interpolate a given profile (x,y) to a new grid xnew;      */
/* data are written to original array y and memory is         */
/* automatically reallocated                                  */
/**************************************************************/

int interpolate_density (float *x, float **y, int n, 
			 float *xnew, int nnew, 
			 int interpol_method, 
			 float *dens_air_old, float *dens_air, 
			 int quiet)
{
  int status=0;
  float *tmp = calloc (nnew, sizeof(float));
  int lc=0, lc2=0;
  int first = 0;

  switch (interpol_method) {
  case INTERP_METHOD_SPLINE:
  case INTERP_METHOD_LINEAR:
  case INTERP_METHOD_LOG:
  case INTERP_METHOD_LOG_SPLINE:
    status = arb_wvn (n, x, *y, nnew, xnew, tmp, interpol_method, 1);
    break;

  case INTERP_METHOD_LINMIX:
    /* convert from density to mixing ratio */
    for (lc=0; lc<n; lc++) {
      if (dens_air_old[lc] <= 0.0) {
	fprintf (stderr, "Error, cannot use linmix interpolation, when air density is <= 0.0\n");
	return -1;
      }
      else
	(*y)[lc] = (*y)[lc] / dens_air_old[lc];
    }
    /*linear interpolation of the mixing ratio*/
    status = arb_wvn (n, x, *y, nnew, xnew, tmp, INTERP_METHOD_LINEAR, 1);
    
    /* convert back to number density */
    for (lc=0; lc<nnew; lc++)
      tmp[lc] = tmp[lc] * dens_air[lc];     
    break;
  default:
    fprintf (stderr, "Error, unknown interpolation method in interpolate_profile\n");
    return -1;
  }

  /* testing for negativ density values */

  switch (interpol_method) {
  case INTERP_METHOD_SPLINE:
    for (lc=0; lc<nnew; lc++) {
      if (tmp[lc] < 0.0) {
        if (!quiet) {
          if (first == 0)
            fprintf (stderr, "***  Warning: In interpolate_density, automatic correction of negative density to 0.0\n");
          fprintf (stderr, "***  Warning: z[%d]= %5.1f, dens[%d]= %12.7e -> dens[%d]= 0.0\n",lc, x[lc],lc,tmp[lc],lc);
          tmp[lc] = 0.0;
          first = 1;
        }
      }
    }
    break;

  case INTERP_METHOD_LINEAR:
  case INTERP_METHOD_LOG:
  case INTERP_METHOD_LINMIX:
  case INTERP_METHOD_LOG_SPLINE:
    for (lc=0; lc<nnew; lc++) {
      if (tmp[lc] < 0.0) {
        if (abs(tmp[lc]) < EPSILON) {
          if (!quiet) {
            fprintf (stderr, "*** Warning, small negative density detected during interpolate_density!\n"); 
            fprintf (stderr, "*** in layer = %d, dens = %e.\n", lc, tmp[lc]); 
            fprintf (stderr, "*** This may happen, when dens profiles contain 0.0 values.\n");  
            fprintf (stderr, "*** Setting to 0.0 automatically.\n"); 
	  }
	  tmp[lc]=0.0;
	}
        else {
          fprintf (stderr, "Error, negative density detected in interpolate_density!, lc=%d \n",lc);
          for (lc2=0; lc2<nnew; lc2++)
            fprintf (stderr," ### dens[%d]=%e\n", lc2, tmp[lc2]);
          return -1;
	}
      }
    }
    break;
  default:
    fprintf (stderr, "Error, unknown interpolation method in interpolate_profile\n");
    return -1;
  }

  if (status != 0) {
    fprintf (stderr, "Error %d interpolating density\n", status);
    return status;
  }

  free(*y);
  *y = tmp;

  return 0;
}


/*******************************************************************/
/* Interpolate all given atmospheric profiles to a new z-grid      */
/* profiles are written to original arrays and memory is           */
/* automatically reallocated                                       */
/*******************************************************************/

int interpolate_atmosphere (float *z, float **p_p, float **p_T, float ***p_dens, int n,  
			    float *znew, int nnew,
                            int interpol_method_press, int interpol_method_temper, int *interpol_method_gas, 
                            int quiet)

{
  /* n is old number of levels, nnew is new number of levels */

  int status=0;
  int lc=-999,i=-999;
  /*  float boltzmann = 1.38065e-23; */  
  float *dens_air_old=NULL;

  /* Pressure */
  status=0;
  status += interpolate_profile (z, p_p, n, znew, nnew, interpol_method_press, quiet);

  /* Temperature */
  status += interpolate_profile (z, p_T, n, znew, nnew, interpol_method_temper, quiet);

  /* copy old air number density */
  dens_air_old = calloc (n, sizeof(float));
  for (lc=0; lc<n; lc++)
    dens_air_old[lc] = (*p_dens)[MOL_AIR][lc];


  /* Air */
  /* this is not 100% consistent with interpolation of p and T */
  status += interpolate_profile (z, &((*p_dens)[MOL_AIR]), n, znew, nnew, interpol_method_gas[MOL_AIR], quiet);

/*   /\* determine air number density from pressure and temperature *\/ */
/*   free((*p_dens)[MOL_AIR]); */
/*   (*p_dens)[MOL_AIR] = calloc (nnew, sizeof(float)); */
/*   for (lc=0; lc<nnew; lc++) */
/*     (*p_dens)[MOL_AIR][lc] =  (*p_p)[lc] / (boltzmann * (*p_T)[lc]) * 100.0 / 1e6; */


  /* Ozone, O2, water vapour, CO2, NO2, BRO, OClO, HCHO */
  for (i=0; i<MOL_NN; i++) {
    if (i != MOL_AIR)
      status += interpolate_density (z, &((*p_dens)[i]),  n, znew, nnew, interpol_method_gas[i],
                                     dens_air_old, (*p_dens)[MOL_AIR], quiet);
  }


  if (status!=0) {
    fprintf (stderr, "Error %d interpolating atmosphere\n", status);
    return status;
  }

  return 0;
}


/******************************************************************/
/*  saturate watervapour inside waterclouds also IPA dependent    */
/******************************************************************/

int saturate_wv_inside_waterclouds (input_struct input, output_struct *output, int ipa, int iv)
{  
  int   status=0;
  int  *cloud_indices=NULL;
/*   int   lc1=-666, lc2=-666; */
  int   lc;
  int   nlev = output->atm.nlev;
  int   lc_wc=0, wc_nlev=0;
  int   first_time=1;
  char *gas;
  float rel_humidity=0.0;

  /******    Saturation of air inside water clouds    *******/

  /* for IPA: replace the cloud by IPA-cloud */
  if (input.wcipa && input.wc.wv_saturation != SATURATION_NONE) {
    wc_nlev = output->wcipa[ipa].nlev;   /* number of cloud levels */
    for (lc_wc=0; lc_wc<wc_nlev; lc_wc++) {
      output->wc.optprop.dtau [iv][lc_wc] = output->wcipa[ipa].optprop.dtau [iv][lc_wc];
      output->wc.zd [lc_wc] = output->wcipa[ipa].zd[lc_wc];
    }
  }
  else wc_nlev = output->wc.nlev;
    
  /* saturate wator vapour in water cloud (at the levels, not layers) */
  if (output->wc.occurrence && input.wc.wv_saturation != SATURATION_NONE) {
  
    if (!input.quiet) {
      if (input.wcipa) 
        fprintf(stderr," ... saturate water wapour inside water clouds (IPA number %d) \n", ipa+1);
      else
        fprintf(stderr," ... saturate water wapour inside water clouds \n");
    }

    /* allocate cloud_indices */
    if ((cloud_indices = calloc(nlev, sizeof(int)))==NULL) {
      fprintf (stderr, "Error: Allocation of cloud_indices in saturate_water_clouds (ancillary.c)\n");
      return -1;
    }  

    find_cloud_indices (output->wc.optprop.dtau [iv], output->wc.zd, wc_nlev, 
                        output->atm.zd, nlev,
                        &(cloud_indices));

    for (lc=0; lc<nlev; lc++) {

      if (cloud_indices[lc] == 1) {

        if (first_time && input.verbose)
          fprintf(stderr,"       z       org_h2o     org_rh     new_h2o    new_rh \n");
        first_time=0;
        rel_humidity = output->atm.microphys.dens[MOL_H2O][lc] / vapor_pressure (output->atm.microphys.temper[lc]) * 100.0;
        if (input.verbose) fprintf(stderr,"    %5.1f %14.7e %7.1f",output->atm.zd[lc],
                 output->atm.microphys.dens[MOL_H2O][lc],rel_humidity);

        output->atm.microphys.dens[MOL_H2O][lc] = vapor_pressure (output->atm.microphys.temper[lc]) * input.wc.wv_sat_ratio/100.0;
        rel_humidity = output->atm.microphys.dens[MOL_H2O][lc] / vapor_pressure (output->atm.microphys.temper[lc]) * 100.0;
/*         output->atm.microphys.rh[lc] = input.wc.wv_sat_ratio; */

        if (input.verbose) fprintf(stderr," %14.7e %6.1f\n",
                  output->atm.microphys.dens[MOL_H2O][lc],rel_humidity);
      }
    }
  
    free(cloud_indices);
 
    if (first_time) {
      if (input.verbose)
        fprintf(stderr," ...  no (additional) cloud levels found \n");
    }
    else {
      /* recalculate dens_avg and precip_water */
      /* average water vapour density is also higher in the ajacent layers, */
      /* as average is calculated by the level properties */
      gas = gas_number2string(MOL_H2O);

      status = column (output->atm.microphys.dens[MOL_H2O], output->atm.zd, output->atm.nlev, output->alt.altitude,
                       input.atm.interpol_method_gas[MOL_H2O], output->atm.microphys.dens[MOL_AIR],
                       &(output->atm.microphys.dens_avg[MOL_H2O]),NO, &(output->column[MOL_H2O]));

      if (status!=0) {
        fprintf (stderr, "\nError %d calculating %s column!\n", status, gas);
        return status;
      }

      output->precip_water = output->column[MOL_H2O] * DU2particle * (M_H2O / N_MOL_CKDFU * 10.);
                             /*           DU         *  DU->cm-2   *       cm-2->kg/m2         */
    }
  }

  /* some additional verbose output */
  /* ------------------------------ */
/*   if ((input.verbose) &&  */
/*       ((input.wcipa && input.wc.wv_saturation != SATURATION_NONE) || */
/*        (input.icipa && input.ic.wv_saturation != SATURATION_NONE)) && */
/*       !first_time) { */
    
/*     fprintf (stderr, "-------------------------------------------------------------------------------------------------------------------------------------\n"); */
/*     fprintf (stderr, "  lc |  z[km]  |  Pressure  | Temp.  |    Air      |   Ozone     |     O2      | Water vap.  |    CO2      |    NO2      | \n"); */
/*     fprintf (stderr, "     |         |   [hPa]    |  [K]   |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    | \n"); */
/*     fprintf (stderr, "-------------------------------------------------------------------------------------------------------------------------------------\n"); */
  
/*     for (lc=0; lc<output->atm.nlev; lc++) */
/*       fprintf (stderr, "%5d   %6.2f   %10.5f   %6.2f   %.5e   %.5e   %.5e   %.5e   %.5e   %.5e\n", */
/*                lc, output->atm.zd[lc], */
/*                output->atm.microphys.press[lc], output->atm.microphys.temper[lc],output->atm.microphys.dens[MOL_AIR][lc], */
/*                output->atm.microphys.dens[MOL_O3][lc],output->atm.microphys.dens[MOL_O2][lc], output->atm.microphys.dens[MOL_H2O][lc], */
/*                output->atm.microphys.dens[MOL_CO2][lc],output->atm.microphys.dens[MOL_NO2][lc]); */

/*     fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n"); */
/*     fprintf (stderr, "%5s | %6.2f | %10.5f | %6.2f | %7.1ecm-2 |  %7.3f DU | %7.1ecm-2 | %6.3f kg/m2 | %7.1ecm-2 | %7.5f DU |\n", */
/*                      "sum", 0.0/0.0, 0.0/0.0, 0.0/0.0, output->column[MOL_AIR]*DU2particle, output->column[MOL_O3], */
/*                       output->column[MOL_O2]*DU2particle, output->precip_water, output->column[MOL_CO2]*DU2particle, output->column[MOL_NO2]); */
/*     fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n"); */
  
/*   } */

  return status;
}

/****************************************************************/
/*  saturate watervapour inside iceclouds also IPA dependent    */
/****************************************************************/

int saturate_wv_inside_iceclouds (input_struct input, output_struct *output, int ipa, int iv)
{  
  int status=0;
  int  *cloud_indices=NULL;
/*   int lc1=-666, lc2=-666; */
  int lc;
  int nlev = output->atm.nlev;
  int lc_ic=0, ic_nlev=0;
  int first_time=1;
  char *gas;
  float rel_humidity=0.0;

  /******    Saturation of air inside ice clouds    *******/

  /* for IPA: replace the cloud by IPA-cloud */
  if (input.icipa && input.ic.wv_saturation != SATURATION_NONE) {
    ic_nlev = output->icipa[ipa].nlev;   /* number of cloud levels */
    for (lc_ic=0; lc_ic<ic_nlev; lc_ic++) {
      output->ic.optprop.dtau [iv][lc_ic] = output->icipa[ipa].optprop.dtau [iv][lc_ic];
      output->ic.zd [lc_ic] = output->icipa[ipa].zd[lc_ic];
    }
  }
  else ic_nlev = output->ic.nlev;
    
  /* saturate wator vapour in ice cloud (at the levels, not layers) */
  if (output->ic.occurrence && input.ic.wv_saturation != SATURATION_NONE) {
        
    if (!input.quiet) {
      if (input.icipa) 
        fprintf(stderr," ... saturate water wapour inside ice clouds (IPA number %d) \n", ipa+1);
      else
        fprintf(stderr," ... saturate water wapour inside ice clouds \n");
    }

    /* allocate cloud_indices */
    if ((cloud_indices = calloc(nlev, sizeof(int)))==NULL) {
      fprintf (stderr, "Error: Allocation of cloud_indices in saturate_ice_clouds (ancillary.c)\n");
      return -1;
    }

    find_cloud_indices (output->ic.optprop.dtau [iv], output->ic.zd, ic_nlev, 
                        output->atm.zd, nlev,
                        &(cloud_indices));

    for (lc=0; lc<nlev; lc++) {

      if (cloud_indices[lc] == 1) {

        if (first_time && input.verbose)
          fprintf(stderr,"       z       org_h2o     org_rh     new_h2o    new_rh \n");
        first_time=0;
        rel_humidity = output->atm.microphys.dens[MOL_H2O][lc] / vapor_pressure (output->atm.microphys.temper[lc]) * 100.0;

        if (input.verbose) fprintf(stderr,"    %5.1f %14.7e %7.1f",output->atm.zd[lc],
                 output->atm.microphys.dens[MOL_H2O][lc],rel_humidity);

        output->atm.microphys.dens[MOL_H2O][lc] = vapor_pressure (output->atm.microphys.temper[lc]) * input.ic.wv_sat_ratio/100.0;
        rel_humidity = output->atm.microphys.dens[MOL_H2O][lc] / vapor_pressure (output->atm.microphys.temper[lc]) * 100.0;
/*         output->atm.microphys.rh[lc] = input.ic.wv_sat_ratio; */

        if (input.verbose) fprintf(stderr," %14.7e %6.1f\n",
                  output->atm.microphys.dens[MOL_H2O][lc],rel_humidity);
      }
    }
  
    free(cloud_indices);

    if (first_time) {
      if (input.verbose)
        fprintf(stderr," ...  no (additional) cloud levels found \n");
    }
    else {
      /* recalculate dens_avg and precip_water */
      /* average water vapour density is also higher in the ajacent layers, */
      /* as average is calculated by the level properties */
      gas = gas_number2string(MOL_H2O);

      status = column (output->atm.microphys.dens[MOL_H2O], output->atm.zd, output->atm.nlev, output->alt.altitude,
                       input.atm.interpol_method_gas[MOL_H2O], output->atm.microphys.dens[MOL_AIR],
                       &(output->atm.microphys.dens_avg[MOL_H2O]),NO,&(output->column[MOL_H2O]));

      if (status!=0) {
        fprintf (stderr, "\nError %d calculating %s column!\n", status, gas);
        return status;
      }

      output->precip_water = output->column[MOL_H2O] * DU2particle * (M_H2O / N_MOL_CKDFU * 10.);
                             /*           DU         *  DU->cm-2   *       cm-2->kg/m2         */
    }
  }


  /* some additional verbose output */
  /* ------------------------------ */
/*   if ((input.verbose) &&  */
/*       ((input.wcipa && input.wc.wv_saturation != SATURATION_NONE) || */
/*        (input.icipa && input.ic.wv_saturation != SATURATION_NONE)) && */
/*       !first_time) { */
    
/*     fprintf (stderr, "-------------------------------------------------------------------------------------------------------------------------------------\n"); */
/*     fprintf (stderr, "  lc |  z[km]  |  Pressure  | Temp.  |    Air      |   Ozone     |     O2      | Water vap.  |    CO2      |    NO2      | \n"); */
/*     fprintf (stderr, "     |         |   [hPa]    |  [K]   |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    | \n"); */
/*     fprintf (stderr, "-------------------------------------------------------------------------------------------------------------------------------------\n"); */
  
/*     for (lc=0; lc<output->atm.nlev; lc++) */
/*       fprintf (stderr, "%5d   %6.2f   %10.5f   %6.2f   %.5e   %.5e   %.5e   %.5e   %.5e   %.5e\n", */
/*                lc, output->atm.zd[lc], */
/*                output->atm.microphys.press[lc], output->atm.microphys.temper[lc],output->atm.microphys.dens[MOL_AIR][lc], */
/*                output->atm.microphys.dens[MOL_O3][lc],output->atm.microphys.dens[MOL_O2][lc], output->atm.microphys.dens[MOL_H2O][lc], */
/*                output->atm.microphys.dens[MOL_CO2][lc],output->atm.microphys.dens[MOL_NO2][lc]); */

/*     fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n"); */
/*     fprintf (stderr, "%5s | %6.2f | %10.5f | %6.2f | %7.1ecm-2 |  %7.3f DU | %7.1ecm-2 | %6.3f kg/m2 | %7.1ecm-2 | %7.5f DU |\n", */
/*                      "sum", 0.0/0.0, 0.0/0.0, 0.0/0.0, output->column[MOL_AIR]*DU2particle, output->column[MOL_O3], */
/*                       output->column[MOL_O2]*DU2particle, output->precip_water, output->column[MOL_CO2]*DU2particle, output->column[MOL_NO2]); */
/*     fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n"); */
  
/*   } */

  return status;
}


/***********************************************************/
/* find_clouds_indices (defined on clouds z-grid, n_cloud) */
/* on another z-grid (atmosphere, n_atm)                   */
/* return an array of size n_atm with value 1,             */
/* where clouds are                                        */
/***********************************************************/

int find_cloud_indices (float *dtau_cloud, float *z_cloud, int n_cloud, 
                        float *z_atm, int n_atm,
                        int **p_cloud_indices)
{
  int lc_c = -666; /* loop integer cloud */
  int lc   = -666; /* loop integer atmosphere */
  int lc1  = -666; /* index upper cloud level, cloud grid */
  int lc2  = -666; /* index lower cloud level, cloud grid */  
  
  /* loop through cloud profile */
  for (lc_c=0; lc_c < n_cloud; lc_c++) {

    /* we found the a cloud layer */
    if (dtau_cloud[lc_c] != 0.0) {

      /* new boundary indieces */ 
      lc1=-666; lc2=-666;

      /* search indeces z-atm-grid corresponding to this layer */
      for (lc=n_atm-1; lc>=0; lc--) {
        if (z_cloud[lc_c  ] == z_atm[lc]) lc1 = lc;
        if (z_cloud[lc_c+1] == z_atm[lc]) lc2 = lc;
        if (lc1 >= 0 && lc2 >= 0) {
          break;
        }
      }
      if (lc1 < 0 || lc2 < 0) {
        fprintf (stderr, "Error, cloud boundaries not found in the atmospheric profiles in setup_gases()\n");
        fprintf (stderr, "  But they have to be there in order to saturate water vapour inside clouds \n");
        fprintf (stderr, "  This is a program bug. Please contact the programmers and include the input file \n");
        return -2;
      }
      /* swap the numbers, if not in right order (for atmosphere file in reverse order) */
      if (lc2 < lc1) {
        lc  = lc1;
        lc1 = lc2;
        lc2 = lc ;
      }
      /* flag the indeces-array */
      for (lc = lc1; lc <= lc2; lc++)
        (*p_cloud_indices)[lc] = 1; 
    }
  }

  return 0;
}



/**************************************************************/
/* Define the internal wavelength grid for the transmittance  */
/* calculation.                                               */
/**************************************************************/

static int set_transmittance_wl_grid(wl_inp_struct *wl_inp, wl_out_struct *wl_out, int quiet)
{
  int iv=0;
  float lambda=0.0, lambdanew=0.0;
  static int firstsr=1, firsthz=1;
  
  /* Determine number of wavelengths needed for transmittance calculation */
  lambda = wl_inp->start;
  iv=0;

  if (wl_inp->start != wl_inp->end) { /* non-monochromatic calculation */
    while (lambda < wl_inp->end) {
      
      lambdanew = lambda;
      
      if (lambda < 121.0) {
	lambdanew += 1.0;
	if (lambdanew > 121.0) {
	  lambda = 121.0;
	  lambdanew = 121.0;
	}
      }
      
      /* Lyman alpha */
      if (lambda >= 121.0 && lambda < 122.0)
	lambdanew += 0.01;
      
      /* Schumann-Runge continuum */
      if (lambda >= 122.0 && lambda < 130) {
	lambdanew += 0.1;
	if (lambdanew >= 1.0E7/57000.0)
	  lambda += 0.1;
      }

      if (lambda >= 130.0 && lambda < 1.0E7/57000.0) {
	lambdanew += 0.5;
	if (lambdanew >= 1.0E7/57000.0)
	  lambda += 0.5;
      }
      
      /* Schumann-Runge bands */
      if (lambda <= 1.0E7/49000.5 && lambda >= 1.0E7/57000.0) {
	if (firstsr) {
	  if (iv>0)
	    lambdanew=1.0E7 / 57000.0;
	  else
	    lambdanew=1.0E7 / (floor(2.0*1.0E7/lambda)/2.0);
	  
	  firstsr=0;
	}
	else 
	  lambdanew = 1.0E7/(1.0E7/lambda - 0.5);
      }
      
      
      /* Herzberg continuum, Hartley-Huggins bands */
      if (lambda > 1.0E7/49000.5 && lambda < 350.0) {
	if (firsthz) {
	  lambdanew = ceil(2.0*lambda)/2.0;
	  if (lambdanew == lambda)
	    lambdanew += 0.5;
	  firsthz=0;
	}
	else
	  lambdanew += 0.5;
      }
      
      if (lambda >= 350.0)
	lambdanew += 1.0;
      
      lambda = lambdanew;

      iv++;
    }
  }

  firstsr=1;
  firsthz=1;
  wl_out->nlambda_t = iv+1;
  wl_out->lambda_t = (float *) calloc (wl_out->nlambda_t, sizeof(float));
  

  /* Set wavelengths for transmittance calculation */
  lambda = wl_inp->start;
  
  for (iv=0; iv<wl_out->nlambda_t; iv++) {
    wl_out->lambda_t[iv] = lambda;
    
    lambdanew = lambda;
    
    if (lambda < 121.0) {
      lambdanew += 1.0;
      if (lambdanew > 121.0) {
	lambda = 121.0;
	lambdanew = 121.0;
      }
    }
    
    /* Lyman alpha */
    if (lambda >= 121.0 && lambda < 122.0)
      lambdanew += 0.01;
    
    /* Schumann-Runge continuum */
    if (lambda >= 122.0 && lambda < 130) {
      lambdanew += 0.1;
      if (lambdanew >= 1.0E7/57000.0)
	lambda += 0.1;
    }

    if (lambda >= 130.0 && lambda < 1.0E7/57000.0) {
      lambdanew += 0.5;
      if (lambdanew >= 1.0E7/57000.0)
	lambda += 0.5;
    }
    
    /* Schumann-Runge bands */
    if (lambda <= 1.0E7/49000.5 && lambda >= 1.0E7/57000.0) {
      if (firstsr) {
	if (iv>0)
	  lambdanew=1.0E7 / 57000.0;
	else
	  lambdanew=1.0E7 / (floor(2.0*1.0E7/lambda)/2.0);
	
	firstsr=0;
      }
      else 
	lambdanew = 1.0E7/(1.0E7/lambda - 0.5);
    }
    
    
    /* Herzberg continuum, Hartley-Huggins bands */
    if (lambda > 1.0E7/49000.5 && lambda < 350.0) {
      if (firsthz) {
	lambdanew = ceil(2.0*lambda)/2.0;
	if (lambdanew == lambda)
	  lambdanew += 0.5;
	firsthz=0;
      }
      else
	lambdanew += 0.5;
    }
    
    if (lambda >= 350.0)
      lambdanew += 1.0;
    
    lambda = lambdanew;
  }

  return 0;
}



/**************************************************************/
/* Define the internal wavelength grid for the transmittance  */
/* calculation. LOWTRAN requires much less detail, in         */
/* particular below 300nm                                     */
/**************************************************************/

static int set_transmittance_wl_grid_lowtran (wl_inp_struct *wl_inp, wl_out_struct *wl_out, int quiet)
{
  int iv=0;
  float lambda=0.0, wvn_step_t=0.0;
  
  /* Determine number of wavelengths needed for transmittance calculation */
  lambda = wl_inp->start;
  iv = 1;
  while (lambda < wl_inp->end) {

    if (lambda < 350.0)
      wvn_step_t = 0.5;
    else
      wvn_step_t = 1.0;

    lambda += wvn_step_t;
    iv++;
  }

  wl_out->nlambda_t = iv;
  wl_out->lambda_t = (float *) calloc (wl_out->nlambda_t, sizeof(float));


  /* Set wavelengths for transmittance calculation */
  lambda = wl_inp->start;

  for (iv=0; iv<wl_out->nlambda_t; iv++) {
    wl_out->lambda_t[iv] = lambda;

    if (lambda < 350.0)
      wvn_step_t = 0.5;
    else
      wvn_step_t = 1.0;

    lambda += wvn_step_t;
  }

  return 0;
}



/**************************************************************/
/* Convolve the spectrum with a user-defined slit function    */
/**************************************************************/

int convolve (input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iv=0, iu=0, j=0, lev=0;
  int status=0;
  double *x_data=NULL;

  int n_slit=0;
  double *x_slit=NULL, *y_slit=NULL;


  /* read slit function from file */
  status = read_2c_file (input.filename[FN_SLITFUNCTION], &x_slit, &y_slit, &n_slit); 

  if (status!=0)  { 
    fprintf (stderr, "Error %d reading file %s\n", status, input.filename[FN_SLITFUNCTION]); 
    return status; 
  } 

  x_data = calloc (output->wl.nlambda_h, sizeof(double)); 

  for (iv=0; iv<output->wl.nlambda_h; iv++)
    x_data[iv] = (double) output->wl.lambda_h[iv];
  
  if (input.rte.solver == POLRADTRAN) {
    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Convolve up_flux and down_flux*/
      for (is=0; is<input.rte.pol_nstokes; is++) {

	status = cnvlv (x_data, output->up_flux[lev][is], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);

	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting up_flux\n", status); 
	  return status; 
	} 

	status = cnvlv (x_data, output->down_flux[lev][is], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);

	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting down_flux\n", status); 
	  return status; 
	} 

	/* Convolve up_rad and down_rad*/
	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.nstr/2; iu++) {
	    status = cnvlv (x_data, output->down_rad[lev][j][iu][is], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);

	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting down_rad\n", status); 
	      return status; 
	    } 

	    status = cnvlv (x_data, output->up_rad[lev][j][iu][is], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting up_rad\n", status); 
	      return status; 
	    } 
	  }
	}
      }
    }
  }
  else {

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Convolve rfldir, rfldn, flup and uavg for all cases */

      /* rfldir */
      status = cnvlv (x_data, output->rfldir[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);

      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting rfldir\n", status); 
	return status; 
      } 
      
      /* rfldn */
      status = cnvlv (x_data, output->rfldn[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting rfldn\n", status); 
	return status; 
      } 
      
      /* flup */
      status = cnvlv (x_data, output->flup[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting flup\n", status); 
	return status; 
      } 
      
      /* uavg */
      status = cnvlv (x_data, output->uavg[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, "Error %d convoluting uavg\n", status); 
	return status; 
      } 
      
      
      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {

	    /* rfldir3d */
	    status = cnvlv (x_data, output->rfldir3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting rfldir3d\n", status); 
	      return status; 
	    } 

	    /* rfldn3d */
	    status = cnvlv (x_data, output->rfldn3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting rfldn3d\n", status); 
	      return status; 
	    } 

	    /* flup3d */
	    status = cnvlv (x_data, output->flup3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting flup3d\n", status); 
	      return status; 
	    } 

	    /* uavgso3d */
	    status = cnvlv (x_data, output->uavgso3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting uavgso3d\n", status); 
	      return status; 
	    } 

	    /* uavgdn3d */
	    status = cnvlv (x_data, output->uavgdn3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting uavgdn3d\n", status); 
	      return status; 
	    } 

	    /* uavgup3d */
	    status = cnvlv (x_data, output->uavgup3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting uavgup3d\n", status); 
	      return status; 
	    } 

	    /* radiance3d */
	    status = cnvlv (x_data, output->radiance3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, "Error %d convoluting radiance3d\n", status); 
	      return status; 
	    } 
	  }
	

      if (input.rte.solver != TWOSTR) {
	
	/* Convolve uavgso, uavgdn, uavgup */
	
	/* uavgso */
	status = cnvlv (x_data, output->uavgso[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting uavgso\n", status); 
	  return status; 
	} 
	
	/* uavgdn */
	status = cnvlv (x_data, output->uavgdn[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting uavgdn\n", status); 
	  return status; 
	} 
	
	/* uavgup */
	status = cnvlv (x_data, output->uavgup[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, "Error %d convoluting uavgup\n", status); 
	  return status; 
	} 

	
	/* Convolve u0u */
	for (iu=0; iu<input.rte.numu; iu++) {
	  status = cnvlv (x_data, output->u0u[lev][iu], output->wl.nlambda_h,
			  x_slit, y_slit, n_slit);
	  if (status!=0)  { 
	    fprintf (stderr, "Error %d convoluting u0u\n", status); 
	    return status; 
	  } 
	}
	
	if (output->print_phi > 0) {
	  
	  /* Convolve uu */
	  for (j=0; j<input.rte.nphi; j++) {
	    for (iu=0; iu<input.rte.numu; iu++) {
	      status = cnvlv (x_data, output->uu[lev][j][iu], output->wl.nlambda_h,
			      x_slit, y_slit, n_slit);
	      
	      if (status!=0)  { 
		fprintf (stderr, "Error %d convoluting u0u\n", status); 
		return status; 
	      } 
	    }
	  }
	}
      }
    }

    /* 3D absorption */
    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<output->atm.Nzcld; ks++)
	if (output->atm.threed[ks])
	  for (is=0; is<output->atm.Nxcld; is++)
	      for (js=0; js<output->atm.Nycld; js++) {
		
		status = cnvlv (x_data, output->abs3d[ks][is][js], output->wl.nlambda_h,
				x_slit, y_slit, n_slit);
		if (status!=0)  { 
		  fprintf (stderr, "Error %d convoluting abs3d\n", status); 
		  return status; 
		} 
	      }
  }


  free(x_data);  

  return 0;
}



/**************************************************************/
/* Convolve a spectrum with a slit function.                  */
/* Internal function, used by convolve().                     */
/**************************************************************/

static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit) 
{
  int iv=0, status=0;
  double *y_data = calloc (n_spec, sizeof(double));
  double *tmp=NULL;

  for (iv=0; iv<n_spec; iv++) 
    y_data[iv] = (double) y_spec[iv];
  
  status = int_convolute (x_spec, y_data, n_spec, x_slit, y_slit, n_slit, &tmp);
  if (status!=0) 
    return status; 
  
  for (iv=0; iv<n_spec; iv++) 
    y_spec[iv] = (float) tmp[iv];
  
  free(y_data);
  free(tmp);

  return 0;
}


/**************************************************************/
/* Interpolate the hires spectrum to a user-defined           */
/* wavelength grid.                                           */
/**************************************************************/

int spline_interpolate(input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iu=0, iv=0, j=0, lev=0;
  int status=0;
  double *x_user=NULL;
  int linear=0;

  if (strlen(input.filename[FN_SPLINE]) > 0) {

    /* read file with user x values */
    status = read_1c_file (input.filename[FN_SPLINE], &x_user, &output->wl.nlambda_s);
    if (status!=0)  {
      fprintf (stderr, "spline_interpolate: error %d reading file %s\n", 
	       status, input.filename[FN_SPLINE]);
      return (status);
    }

    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));

    for (iv=0; iv<output->wl.nlambda_s; iv++)
      output->wl.lambda_s[iv] = (float) x_user[iv];

    /* Check wavelength ranges */
    status=0;
    if (output->wl.start > output->wl.lambda_s[0])  {
      fprintf (stderr, "Error, spline wavelength %f is smaller than wvn %f\n",
	       output->wl.lambda_s[0], output->wl.start);
      status--;
    }

    if (output->wl.end < output->wl.lambda_s[output->wl.nlambda_s-1])  {
      fprintf (stderr, "Error, spline wavelength %f is greater than wvn %f\n",
	       output->wl.lambda_s[output->wl.nlambda_s-1], output->wl.end);
      status--;
    }

    if (status!=0) 
      return status;
  }
  else {
    output->wl.nlambda_s = 
      (int) ((input.spline_lambda_1-input.spline_lambda_0)/input.spline_lambda_step+1);
    
    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));
    
    for (iv=0; iv<output->wl.nlambda_s; iv++)
      output->wl.lambda_s[iv] = input.spline_lambda_0 + (float) iv * input.spline_lambda_step;
  }

  /* Spline interpolate to given wavelengths */
  if (input.rte.solver == POLRADTRAN) {

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Spline interpolate up_flux and down_flux*/
      for (is=0; is<input.rte.pol_nstokes; is++) {

	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->up_flux[lev][is],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->up_flux[lev][is], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating up_flux\n", status);
	  return status;
	}

	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->down_flux[lev][is],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->down_flux[lev][is], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating down_flux\n", status);
	  return status;
	}

	/* Spline interpolate up_rad and down_rad */
	for (j=0; j<input.rte.nphi; j++) {

	  for (iu=0; iu<input.rte.nstr/2; iu++) {

	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->down_rad[lev][j][iu][is],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->down_rad[lev][j][iu][is],
			     linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating down_rad\n", status);
	      return status;
	    }

	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->up_rad[lev][j][iu][is],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->up_rad[lev][j][iu][is],
			     linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating up_rad\n", status);
	      return status;
	    }
	  }
	}
      }

      /* Spline interpolate heating rate */
      if (input.heating != NO_HEAT) {
        status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->heat[lev],  
                         output->wl.nlambda_s, output->wl.lambda_s, output->heat[lev], 
                         linear, 0);
        status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->w_zout[lev],  
                         output->wl.nlambda_s, output->wl.lambda_s, output->w_zout[lev], 
                         linear, 0);
      }
    }
  }
  else {
    for (lev=0; lev<input.atm.nzout; lev++) {
      
      /* Spline interpolate rfldir, rfldn, flup, uavg and heat for all cases */
      
      /* rfldir */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->rfldir[lev], 
		       linear, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating rfldir\n", status);
	return status;
      }
      
      /* rfldn */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->rfldn[lev], 
		       linear, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating rfldn\n", status);
	return status;
      }
      
      /* flup */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->flup[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->flup[lev], 
		       linear, 0);
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating flup\n", status);
	return status;
      }
      
      /* uavg */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->uavg[lev], 
		       linear, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d interpolating uavg\n", status);
	return status;
      }

      /* heating rates */
      if (input.heating != NO_HEAT) {
        status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->heat[lev],  
                         output->wl.nlambda_s, output->wl.lambda_s, output->heat[lev], 
                         linear, 0);
        status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->w_zout[lev],  
                         output->wl.nlambda_s, output->wl.lambda_s, output->w_zout[lev], 
                         linear, 0);
      }
      
      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {

	    /* rfldir3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldir3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->rfldir3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating rfldir3d\n", status);
	      return status;
	    }
	    
	    /* rfldn3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldn3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->rfldn3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating rfldn3d\n", status);
	      return status;
	    }
	    
	    /* rflup3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->flup3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->flup3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating flup3d\n", status);
	      return status;
	    }

	    /* uavgso3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgso3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->uavgso3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating uavgso3d\n", status);
	      return status;
	    }
	    
	    /* uavgdn3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->uavgdn3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating uavgdn3d\n", status);
	      return status;
	    }
	    
	    /* uavgup3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgup3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->uavgup3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating uavgup3d\n", status);
	      return status;
	    }
	    
	    /* radiance3d */
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->radiance3d[lev][is][js], 
			     output->wl.nlambda_s, output->wl.lambda_s, output->radiance3d[lev][is][js], 
			     linear, 0);
	    if (status!=0) {
	      fprintf (stderr, "Error %d interpolating radiance3d\n", status);
	      return status;
	    }
	  }
      
      if (input.rte.solver != TWOSTR) {
	
	/* Spline interpolate uavgso, uavgdn, uavgup */
	
	/* uavgso */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgso[lev], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating uavgso\n", status);
	  return status;
	}
	
	/* uavgdn */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgdn[lev], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating uavgdn\n", status);
	  return status;
	}
	
	/* uavgup */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgup[lev], 
			 linear, 0);

	if (status!=0) {
	  fprintf (stderr, "Error %d interpolating uavgup\n", status);
	  return status;
	}
	
	/* Spline interpolate u0u */
	for (iu=0; iu<input.rte.numu; iu++) {
	  status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lev][iu],  
			   output->wl.nlambda_s, output->wl.lambda_s, output->u0u[lev][iu], 
			   linear, 0);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d interpolating u0u\n", status);
	    return status;
	  }
	}
	
	if (output->print_phi > 0) {
	  
	  /* Spline interpolate uu */
	  for (j=0; j<input.rte.nphi; j++) {
	    for (iu=0; iu<input.rte.numu; iu++) {
	      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uu[lev][j][iu],  
			       output->wl.nlambda_s, output->wl.lambda_s, output->uu[lev][j][iu], 
			       linear, 0);
	      
	      if (status!=0) {
		fprintf (stderr, "Error %d interpolating uu\n", status);
		return status;
	      }
	    }
	  }
	}
      }
    }


    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<output->atm.Nzcld; ks++)
	if (output->atm.threed[ks])
	  for (is=0; is<output->atm.Nxcld; is++)
	    for (js=0; js<output->atm.Nycld; js++) {
	      
	      status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_h, output->abs3d[ks][is][js], 
			       output->wl.nlambda_h, output->wl.lambda_s, output->abs3d[ks][is][js], 
			       linear, 0);
	      
	      if (status!=0) {
		fprintf (stderr, "Error %d interpolating abs3d\n", status);
		return status;
	      }
	    }
  }
  

  free(output->wl.lambda_h);
  
  output->wl.nlambda_h = output->wl.nlambda_s;

  output->wl.lambda_h = (float *) calloc (output->wl.nlambda_s, sizeof(float));

  for (iv=0; iv<output->wl.nlambda_h; iv++)
    output->wl.lambda_h[iv] = (float) output->wl.lambda_s[iv];

  return status;
  
}




/**************************************************************/
/* Interpolate the transmittance from the internal grid to    */
/* high resolution grid.                                      */
/**************************************************************/

int interpolate_transmittance (input_struct input, output_struct *output)
{
  int lu=0, is=0, js=0, ks=0, iv=0, ivh=0, iu=0, j=0;
  int linear=1, status=0;

  /* If only one wavelength is required, or in correlated-k mode, */
  /* arrays are not interpolated but copied                       */
  /* ??? need to add the condition that the extraterrestrial  ??? */
  /* ??? spectrum was not read from file but set to 1         ??? */

  if (output->wl.nlambda_t == 1 || (input.ck_scheme!=CK_NONE && input.ck_scheme!=CK_LOWTRAN)) {
    
    for (ivh=0; ivh<output->wl.nlambda_h; ivh++) {
      
      if (output->wl.nlambda_t == 1)
	iv = ivh;
      else 
	iv = output->wl.map_e2h[ivh];
      
      for (lu=0; lu<input.atm.nzout; lu++) {
	if (input.rte.solver == POLRADTRAN) {

	  for (is=0;is<input.rte.pol_nstokes;is++) {
	    output->up_flux   [lu][is][ivh] = output->up_flux_t   [lu][is][iv];
	    output->down_flux [lu][is][ivh] = output->down_flux_t [lu][is][iv];
	  }

	  for (j=0; j<input.rte.nphi; j++)
	    for (iu=0; iu<input.rte.nstr/2; iu++)
	      for (is=0; is<input.rte.pol_nstokes; is++) {
		output->down_rad [lu][j][iu][is][ivh] = output->down_rad_t [lu][j][iu][is][iv];
		output->up_rad   [lu][j][iu][is][ivh] = output->up_rad_t   [lu][j][iu][is][iv];
	      }
          if (input.heating != NO_HEAT) {
            output->heat   [lu][ivh] = output->heat_t   [lu][iv];
            output->w_zout [lu][ivh] = output->w_zout_t [lu][iv];
	  }
	}
	else {
	  output->flup   [lu][ivh] = output->flup_t   [lu][iv];
	  output->rfldir [lu][ivh] = output->rfldir_t [lu][iv];
	  output->rfldn  [lu][ivh] = output->rfldn_t  [lu][iv];
	  output->uavg   [lu][ivh] = output->uavg_t   [lu][iv];
	  output->uavgdn [lu][ivh] = output->uavgdn_t [lu][iv];
	  output->uavgso [lu][ivh] = output->uavgso_t [lu][iv];
	  output->uavgup [lu][ivh] = output->uavgup_t [lu][iv];
          if (input.heating != NO_HEAT) {
            output->heat   [lu][ivh] = output->heat_t   [lu][iv];
            output->w_zout [lu][ivh] = output->w_zout_t [lu][iv];
	  }

	  /* radiances */
	  for (iu=0; iu<input.rte.numu; iu++) {
	    output->u0u[lu][iu][ivh] = output->u0u_t[lu][iu][iv];

	    for (j=0; j<input.rte.nphi; j++)
	      output->uu[lu][j][iu][ivh] = output->uu_t[lu][j][iu][iv];
	  }

	  /* 3D fields */
	  if (input.rte.mc.passback3D)
	    for (is=0; is<input.rte.mc.Nx_sample; is++)
	      for (js=0; js<input.rte.mc.Ny_sample; js++) {
		output->rfldir3d   [lu][is][js][ivh] = output->rfldir3d_t   [lu][is][js][iv];
		output->rfldn3d    [lu][is][js][ivh] = output->rfldn3d_t    [lu][is][js][iv];
		output->flup3d     [lu][is][js][ivh] = output->flup3d_t     [lu][is][js][iv];
		output->uavgso3d   [lu][is][js][ivh] = output->uavgso3d_t   [lu][is][js][iv];
		output->uavgdn3d   [lu][is][js][ivh] = output->uavgdn3d_t   [lu][is][js][iv];
		output->uavgup3d   [lu][is][js][ivh] = output->uavgup3d_t   [lu][is][js][iv];
		output->radiance3d [lu][is][js][ivh] = output->radiance3d_t [lu][is][js][iv];
	      }
	}
      }
      
      /* absorption is defined on the 3D cloud grid, not on the user-defined grid */
      if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
	for (ks=0; ks<output->atm.Nzcld; ks++)
	  if (output->atm.threed[ks])
	    for (is=0; is<output->atm.Nxcld; is++)
	      for (js=0; js<output->atm.Nycld; js++)
		output->abs3d [ks][is][js][ivh] = output->abs3d_t [ks][is][js][iv];
      
      output->sza_h [ivh] = output->atm.sza_t [iv];
    }
  } 
  else {   /* interpolation to different wavelengths required */

    /* interpolate solar zenith angle */
    status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->atm.sza_t, 
		      output->wl.nlambda_h, output->wl.lambda_h, output->sza_h, linear, 0);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
      return status;
    }
    

    for (lu=0; lu<input.atm.nzout; lu++) {

      if (input.rte.solver == POLRADTRAN) {
	for (is=0; is<input.rte.pol_nstokes; is++) {
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->down_flux_t[lu][is], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->down_flux[lu][is], linear, 0);
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->up_flux_t[lu][is], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->up_flux[lu][is], linear, 0);
          /*                               ???                                         */
          /* ??? is it not nessesary to interpolate here also: down_rad and up_rad ??? */
          /*                               ???                                         */
          if (input.heating != NO_HEAT) {
	    status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->heat_t[lu], 
                              output->wl.nlambda_h, output->wl.lambda_h, output->heat[lu], linear, 0);
	    status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->w_zout_t[lu], 
                              output->wl.nlambda_h, output->wl.lambda_h, output->w_zout[lu], linear, 0);
	  }
	}
      }
      else {
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->flup[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavg_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgdn_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgso_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lu], linear, 0);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgup_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lu], linear, 0);
        if (input.heating != NO_HEAT) {
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->heat_t[lu], 
                            output->wl.nlambda_h, output->wl.lambda_h, output->heat[lu], linear, 0);
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->w_zout_t[lu], 
                            output->wl.nlambda_h, output->wl.lambda_h, output->w_zout[lu], linear, 0);
	}
	
	/* 3D fields */
	if (input.rte.mc.passback3D)
	  for (is=0; is<input.rte.mc.Nx_sample; is++)
	    for (js=0; js<input.rte.mc.Ny_sample; js++) {
	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->rfldir3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->rfldn3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->flup3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgso3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->uavgso3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgdn3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgup3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->uavgup3d[lu][is][js], linear, 0);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->radiance3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->radiance3d[lu][is][js], linear, 0);

	    }
      }
      
      if (input.rte.solver == POLRADTRAN) {

	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.nstr/2; iu++) {
	    for (is=0; is<input.rte.pol_nstokes; is++) {
	      
	      status = arb_wvn (output->wl.nlambda_t, output->wl.lambda_t, output->down_rad_t[lu][j][iu][is], 
				output->wl.nlambda_h, output->wl.lambda_h, output->down_rad[lu][j][iu][is], 
				linear, 0);

	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }

	      status = arb_wvn (output->wl.nlambda_t, output->wl.lambda_t, output->up_rad_t[lu][j][iu][is], 
				output->wl.nlambda_h, output->wl.lambda_h, output->up_rad[lu][j][iu][is], 
				linear, 0);

	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }
	    }
	  }
	}
      }
      else {
	for (iu=0; iu<input.rte.numu; iu++) {
	  status = arb_wvn (output->wl.nlambda_t, output->wl.lambda_t, output->u0u_t[lu][iu], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lu][iu], 
			    linear, 0);
	  
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	    return status;
	  }
	}
	
	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.numu; iu++) {
	    
	    status = arb_wvn (output->wl.nlambda_t, output->wl.lambda_t, output->uu_t[lu][j][iu], 
			      output->wl.nlambda_h, output->wl.lambda_h, output->uu[lu][j][iu], 
			      linear, 0);

	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	      return status;
	    }
	  }
	}
      }
    }

    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      for (ks=0; ks<output->atm.Nzcld; ks++)
	if (output->atm.threed[ks])
	  for (is=0; is<output->atm.Nxcld; is++)
	    for (js=0; js<output->atm.Nycld; js++) {
	      
	      status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->abs3d_t[ks][is][js], 
			       output->wl.nlambda_h, output->wl.lambda_h, output->abs3d[ks][is][js], 
			       linear, 0);
	      
	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }
	    }
  }

  return 0;
}





/**************************************************************/
/* Multiply the transmittance with the extraterrestrial       */
/* irradiance.                                                */
/**************************************************************/

int multiply_extraterrestrial (input_struct input, output_struct *output)
{
  int iv=0, status=0;
  double ffactor=0, rfactor=0, hfactor=0; 
  double watt_factor = 1.0; 
  double hfactor2 = 0.0;

  /* Figure out if we have to convert from mW to W. This should       */
  /* cover the input stuff that comes with uvspec. Now, if the user   */
  /* decides to change the input solar flux it may get interesting... */

  /* THIS mW to W CONVERTION IS USED FOR HEATING RATES, SHOULD IT BECOME DEFAULT? */
  switch(input.ck_scheme) {
  case CK_FU:
  case CK_KATO:
  case CK_KATO2:
  case CK_KATO2_96:
  case CK_AVHRR_KRATZ:
    watt_factor = 1.0;
    break;
  case CK_LOWTRAN:
  case CK_GENERIC:
  case CK_NONE:
    switch (input.source) {
    case SRC_SOLAR:
      watt_factor = mW2W;
      break;
    case SRC_THERMAL:
      watt_factor = 1.0;
      break;
    default:
      fprintf (stderr, "Error, unknown source %d\n", input.source);
      return -1;
    }
    break;
  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", input.ck_scheme);
    return -1;
    break;
  }

  hfactor  = watt_factor * s2day / c_p;
  hfactor2 = hfactor;

  /* multiply with extraterrestrial irradiance */
  for (iv=0; iv<output->wl.nlambda_h; iv++) {

    switch (input.source) {
    case SRC_THERMAL:
      ffactor = output->wl.filter[iv];
      rfactor = output->wl.filter[iv];
      break;

    case SRC_SOLAR:
      switch (input.processing) {
      case PROCESS_INT:
      case PROCESS_SUM:
      case PROCESS_RGB:
      case PROCESS_RGBNORM:
	ffactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	rfactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	break;

      case PROCESS_NONE:
	switch (input.calibration) {
	case OUTCAL_ABSOLUTE:
	  ffactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	  rfactor = output->wl.fbeam[iv] * output->wl.filter[iv];
	  break;
	  
	case OUTCAL_TRANSMITTANCE:
	  ffactor = 1.0 * output->wl.filter[iv];
	  rfactor = 1.0 * output->wl.filter[iv];
	  break;
	  
	case  OUTCAL_REFLECTIVITY:
	  ffactor = 1.0 / cos(output->sza_h[iv]*PI/180.0) * output->wl.filter[iv];
	  rfactor = PI  / cos(output->sza_h[iv]*PI/180.0) * output->wl.filter[iv];
	  break;

	default:
	  fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
	  return -1;
	}

	break;

      default:
	fprintf (stderr, "Error, unknown output processing %d\n", input.processing);
	return -1;
      }

      break;

    default:
      fprintf (stderr, "Error, unknown source %d\n", input.source);
      return -1;
    }

    switch (input.source) {
    case SRC_SOLAR:
      hfactor = output->wl.fbeam[iv] * hfactor2;
      break;
    case SRC_THERMAL:
      break;
    default:
      fprintf (stderr, "Error, unknown source %d\n", input.source);
      return -1;
    }

    /*****************************************************************************************/
    /* now scale irradiances with ffactor, radiances with rfactor, heating rate with hfactor */
    /*****************************************************************************************/

    status = scale_output (input,
			   &(output->rfldir), &(output->rfldn),  &(output->flup),
			   &(output->uavgso), &(output->uavgdn), &(output->uavgup),
			   &(output->uavg), &(output->u0u), &(output->uu), &(output->heat), &(output->w_zout),
			   &(output->down_flux), &(output->up_flux), &(output->down_rad), &(output->up_rad),
			   &(output->rfldir3d), &(output->rfldn3d), &(output->flup3d), &(output->uavgso3d),
			   &(output->uavgdn3d), &(output->uavgup3d), &(output->radiance3d), 
			   output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed, 
			   &(output->abs3d),
			   ffactor, rfactor, hfactor, iv);

    if (status!=0) {
      fprintf (stderr, "Error %d returned by scale_output()\n", status);
      return status;
    }

  }

  return 0;
}


/***********************************************************************************/
/* Function: optical_properties                                           @61_30i@ */
/* Description:                                                                    */
/*  Calculates optical depth, single scattering albedo and                         */
/*  phase function from absorption and scattering cross sections                   */
/*  corresponding gas and particulate matter concentrations for                    */
/*  one wavelength.                                                                */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i61_30@ */
/***********************************************************************************/

int optical_properties (input_struct input, output_struct *output, int iv, int iq, int verbose)
{
  static int first=1;
  int k=0, lc=0;
  int status=0;
  double babsa=0, babsd=0, babsi=0, babso=0, babs=0;
  double bscaa=0, bscad=0, bscai=0, bscar=0, bsca=0;
  double babso_md=0, babs_md=0; 
  double bext_md=0; /* The absorption of gases minus the one specified with a 
                        matrix using the dens_file command (for airmass factor calculations). */

  double babsatot=0, babsdtot=0, babsitot=0, babsotot=0;
  double bscaatot=0, bscadtot=0, bscaitot=0, bscartot=0;

  double moma=0, momd=0, momi=0;
  double bext=0, deltaz=0, g1a=0, g2a=0, fa=0, g1d=0, g2d=0, fd=0, g1i=0, g2i=0, fi=0;
  double sumsca=0;
  double p1=0, p2=0, p3=0, p4=0, p5=0, p6=0;
  FILE *f=NULL;

  int nlev = output->atm.nlev_common;
  int nlyr = nlev-1;

  double momag1=0, momag2=0;
  double momdg1=0, momdg2=0;
  double momig1=0, momig2=0;

  double rayleigh_mom2=0;
  double Delta = 0, Deltap = 0;

  /* 2nd moment for Rayleigh scattering, including depolarization */
  rayleigh_mom2 = 0.2 * (1.0-output->rayleigh_depol[iv]) / (2.0+output->rayleigh_depol[iv]);

  /* Deltas in Eq. 2.16, Hansen and Travis, Space Science Rev., 16, 527-610, 1974.*/
  Delta         = (1.0 -     output->rayleigh_depol[iv]) / (1.0 + 0.5*output->rayleigh_depol[iv]);
  Deltap        = (1.0 + 2.0*output->rayleigh_depol[iv]) / (1.0 -     output->rayleigh_depol[iv]);


  if (verbose) {
    fprintf (stderr, "... second moment for Rayleigh scattering: %f\n", rayleigh_mom2);
    fprintf (stderr, "*** optical_properties()\n");
    if (output->cf.nlev==0) {
      fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------------------------\n");
      fprintf (stderr, "   lc |  z[km] |  Rayleigh |      Aerosol              |       Water cloud             |        Ice cloud                                 |   Molecular \n");
      fprintf (stderr, "      |        |    dtau   |  scatter.      abs.  asy. |    scatter.        abs.  asy. |    scatter.        abs.  asy.     f    g1     g2 |  absorption \n");
      fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
    else {
      fprintf (stderr, " -----------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
      fprintf (stderr, "   lc |  z[km] |  Rayleigh |      Aerosol              |       Water cloud             |        Ice cloud                                 |   Molecular | Cloud   \n");
      fprintf (stderr, "      |        |    dtau   |  scatter.      abs.  asy. |    scatter.        abs.  asy. |    scatter.        abs.  asy.     f    g1     g2 |  absorption | fraction\n");
      fprintf (stderr, " -----------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }      
  }

  if (first!=0) {
    first = 0;
    
    /* allocate memory for profiles */

    output->dtauc        = (float *) calloc (nlyr, sizeof(float));
    output->dtauc_md     = (float *) calloc (nlyr, sizeof(float));
    output->ssalb        = (float *) calloc (nlyr, sizeof(float));


    output->mc.z       = (float *) calloc (nlev, sizeof(float));
    output->mc.temper  = (float *) calloc (nlev, sizeof(float));

    output->mc.dtair   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtmol   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtaer   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtcld   = (float *) calloc (nlev, sizeof(float));
    output->mc.dtice   = (float *) calloc (nlev, sizeof(float));
    output->mc.omaer   = (float *) calloc (nlev, sizeof(float));
    output->mc.omcld   = (float *) calloc (nlev, sizeof(float));
    output->mc.omice   = (float *) calloc (nlev, sizeof(float));

    output->mc.g1aer   = (float *) calloc (nlev, sizeof(float));
    output->mc.g2aer   = (float *) calloc (nlev, sizeof(float));
    output->mc.ffaer   = (float *) calloc (nlev, sizeof(float));

    output->mc.g1cld   = (float *) calloc (nlev, sizeof(float));
    output->mc.g2cld   = (float *) calloc (nlev, sizeof(float));
    output->mc.ffcld   = (float *) calloc (nlev, sizeof(float));

    output->mc.g1ice   = (float *) calloc (nlev, sizeof(float));
    output->mc.g2ice   = (float *) calloc (nlev, sizeof(float));
    output->mc.ffice   = (float *) calloc (nlev, sizeof(float));

    output->mc.refind  = (float *) calloc (nlev, sizeof(float));

    output->mc.lwccld  = (float *) calloc (nlev, sizeof(float));
    output->mc.recld   = (float *) calloc (nlev, sizeof(float));

    output->mc.momaer  = (float **) calloc (nlev, sizeof(float *));
    output->mc.nmomaer = (int *)    calloc (nlev, sizeof(int));

    for (lc=0; lc<nlev; lc++) {
      output->mc.ffcld[lc]  = 1.0;
      output->mc.ffice[lc]  = 1.0;
      output->mc.refind[lc] = 1.0;
    }
    
    /* output->atm.nmom+1 stores the maximum number of moments */
    /* for all wavelengths and layers, minus 1                 */

    status = ASCII_calloc_float(&output->pmom, nlyr, output->atm.nmom+1);
    
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for output->pmom\n", status);
      return status;
    }

  }  

  babsatot=0;
  babsdtot=0;
  babsitot=0;
  babsotot=0;

  bscaatot=0;
  bscadtot=0;
  bscaitot=0;
  bscartot=0;

  for (lc=0; lc<nlyr; lc++) {

    switch (input.rte.solver) {
    case SSSI:
      /* special treatment for SSSI: cloud single scattering albedo is */
      /* set to 0 because cloud scattering is treated explicitely      */
      /* through the tabulated cloud reflectivity; cloud then only     */
      /* reduce radiance through Lambert-Beer                          */
      
      output->wc.optprop.ssa [iv][lc] = 0;
      output->ic.optprop.ssa [iv][lc] = 0;

      break;

    case SSS2:
      /* special treatment for SSSI: cloud single scattering albedo is */
      /* set to 0 because cloud scattering is treated explicitely      */
      /* through the tabulated cloud reflectivity; cloud then only     */
      /* reduce radiance through Lambert-Beer                          */
      /* ??? this will be changed ???                                  */
      output->wc.optprop.ssa [iv][lc] = 0;
      output->ic.optprop.ssa [iv][lc] = 0;

      break;

    default:
      break;
    }


    /* The factor 1.e+5 converts z from km to cm.*/
    deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;

    /* only in the Fu and Liou case, the Rayleigh scattering */
    /* cross section depends on the subband                  */
    switch(input.ck_scheme) {
    case CK_FU:
      bscar = output->atm.optprop.tau_rayleigh_t[lc][iv][iq];
      break;

    case CK_KATO:
    case CK_KATO2:
    case CK_KATO2_96:
    case CK_AVHRR_KRATZ:
    case CK_GENERIC:
    case CK_LOWTRAN:
    case CK_NONE:
      bscar = output->atm.optprop.tau_rayleigh_t[lc][iv][0];
      break;
      
    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme %d\n", input.ck_scheme);
      return -1;

      break;
    }

    switch (output->atm.molabs) {
    case MOLABS_CALC:
      babso    = output->atm.optprop.tau_molabs_t[lc][iv][iq];
      babso_md = output->atm.optprop.tau_molabs_md_t[lc][iv][iq];
      break;
      
    case MOLABS_FILE_MONO:
      /* babso = output->atm.optprop.tau_molabs_user[lc]; */
      babso = output->atm.optprop.tau_molabs_t[lc][iv][iq];
      babso_md=0;
      break;

    case MOLABS_FILE_SPEC:
      /* babso = output->atm.optprop.tau_molabs_user_t[iv][lc]; */
      babso = output->atm.optprop.tau_molabs_t[lc][iv][iq];
      babso_md=0;
      break;

    case MOLABS_NONE:
      babso = 0;
      babso_md=0;
      break;
      
    default:
      fprintf (stderr, "Error, unknown molecular absorption option %d\n", 
	       output->atm.molabs);
      return -1;
    }

    /* scattering by aerosols, clouds, and ice clouds */
    bscaa = output->aer.optprop.ssa[iv][lc] * output->aer.optprop.dtau[iv][lc];
    bscad = output->wc.optprop.ssa [iv][lc] * output->wc.optprop.dtau [iv][lc];
    bscai = output->ic.optprop.ssa [iv][lc] * output->ic.optprop.dtau [iv][lc];
	  
    g1a   = output->aer.optprop.g1[iv][lc];
    g2a   = output->aer.optprop.g2[iv][lc];
    fa    = output->aer.optprop.ff[iv][lc];

    g1d   = output->wc.optprop.g1 [iv][lc];
    g2d   = output->wc.optprop.g2 [iv][lc];
    fd    = output->wc.optprop.ff [iv][lc];

    g1i   = output->ic.optprop.g1 [iv][lc];
    g2i   = output->ic.optprop.g2 [iv][lc];
    fi    = output->ic.optprop.ff [iv][lc];

    /*  Phase function */
    sumsca =  bscar + bscad +  bscaa + bscai;
    if (sumsca < FLT_MIN)
      sumsca = FLT_MIN;


    /* moments of the phase functions */

    momag1=bscaa;
    momag2=bscaa;

    momdg1=bscad;
    momdg2=bscad;

    momig1=bscai;
    momig2=bscai;

    output->pmom[lc][0] = 1.0;

    for (k=1; k<=output->atm.nmom; k++) {

      /* aerosol */
      if (output->aer.optprop.nmom[iv][lc] > 0) {
	if (k<output->aer.optprop.nmom[iv][lc])
	  moma = bscaa * output->aer.optprop.moment[iv][lc][k] / 
	    output->aer.optprop.moment[iv][lc][0];
	else 
	  moma=0;
      }
      else {  /* double Henyey-Greenstein */
	momag1 *= g1a;
	momag2 *= g2a;

	moma = momag1*fa + (1.0-fa)*momag2;
      }
      
      /* water clouds */
      if (output->wc.optprop.nmom[iv][lc] > 0) {
	if (k<output->wc.optprop.nmom[iv][lc])
	  momd = bscad * output->wc.optprop.moment[iv][lc][k] / output->wc.optprop.moment[iv][lc][0];
	else 
	  momd=0;
      }
      else {  /* double Henyey-Greenstein */
	momdg1 *= g1d;
	momdg2 *= g2d;

	momd = momdg1*fd + (1.0-fd)*momdg2;
      }

      /* ice clouds */
      if (output->ic.optprop.nmom[iv][lc] > 0) {
	if (k<output->ic.optprop.nmom[iv][lc])
	  momi = bscai * output->ic.optprop.moment[iv][lc][k] / output->ic.optprop.moment[iv][lc][0];
	else 
	  momi=0;
      }
      else {  /* double Henyey-Greenstein */
	momig1 *= g1i;
	momig2 *= g2i;

	momi = momig1*fi + (1.0-fi)*momig2;
      }

      output->pmom[lc][k] = (momd + moma + momi) / sumsca;

      if (k == 2)   /* Rayleigh scattering, including depolarization */
	output->pmom[lc][k] += bscar/sumsca*rayleigh_mom2;
    }

    /* absorption by aerosols, clouds, and ice clouds */
    babsa  = (1.0 - output->aer.optprop.ssa[iv][lc]) * output->aer.optprop.dtau[iv][lc];
    babsd  = (1.0 - output->wc.optprop.ssa [iv][lc]) * output->wc.optprop.dtau [iv][lc];
    babsi  = (1.0 - output->ic.optprop.ssa [iv][lc]) * output->ic.optprop.dtau [iv][lc];
    
    /* absorption */
    if (input.absorption) {
      babs    = babsa + babso    + babsd + babsi;
      babs_md = babsa + babso_md + babsd + babsi;
      if (babs > FLT_MAX) 
	babs = FLT_MAX;
      if (babs_md > FLT_MAX) 
	babs_md = FLT_MAX;
    }
    else {
      babs  = 0;
      babso = 0;
      babsa = 0;
      babsd = 0;
      babsi = 0;
      babs_md  = 0;
      babso_md = 0;
    }

    /* Scattering coefficient */
    if (input.scattering) {
      bsca = bscar + bscad + bscaa + bscai;
      if (bsca > FLT_MAX) 
	bsca = FLT_MAX;
    }
    else {
      bsca  = 0;
      bscar = 0;
      bscaa = 0;
      bscad = 0;
      bscai = 0;
    }

    /* Extinction coefficient */
    bext = babs + bsca;
    if (bext > FLT_MAX) 
      bext = FLT_MAX;
    bext_md = babs_md + bsca;
    if (bext_md > FLT_MAX) 
      bext_md = FLT_MAX;

    if (verbose) {
      fprintf (stderr, "%5d | %6.2f | %9.6f | %9.6f %9.6f %5.3f | %11.6f %11.6f %5.3f | %11.6f %11.6f %5.3f %5.3f %5.3f %6.3f | %11.6f", 
	       lc, output->atm.zd[lc+1]+output->alt.altitude, 
	       bscar,
	       bscaa, babsa, g1a*fa + (1.0-fa)*g2a,
	       bscad, babsd, g1d*fd + (1.0-fd)*g2d,
	       bscai, babsi, g1i*fi + (1.0-fi)*g2i,
	       fi, g1i, g2i,
	       babso);

      
      if (output->cf.nlev>0)
	fprintf (stderr, " | %.3f\n", output->cf.cf[lc]);
      else
	fprintf (stderr, "\n");
    }

    /* total column for verbose output */
    babsatot+=babsa;
    babsdtot+=babsd;
    babsitot+=babsi;
    babsotot+=babso;

    bscaatot+=bscaa;
    bscadtot+=bscad;
    bscaitot+=bscai;
    bscartot+=bscar;

    /* Optical properties for the mc model */
    output->mc.dtair[nlyr-1-lc] = bscar;
    output->mc.dtmol[nlyr-1-lc] = babso;
    output->mc.dtaer[nlyr-1-lc] = babsa + bscaa;
    output->mc.dtcld[nlyr-1-lc] = babsd + bscad;
    output->mc.dtice[nlyr-1-lc] = babsi + bscai;
		  	      
    output->mc.omaer[nlyr-1-lc] = output->aer.optprop.ssa[iv][lc];
    output->mc.omcld[nlyr-1-lc] = output->wc.optprop.ssa [iv][lc];
    output->mc.omice[nlyr-1-lc] = output->ic.optprop.ssa [iv][lc];
    		  	      
    output->mc.g1aer[nlyr-1-lc] = output->aer.optprop.g1[iv][lc];
    output->mc.g2aer[nlyr-1-lc] = output->aer.optprop.g2[iv][lc];
    output->mc.ffaer[nlyr-1-lc] = output->aer.optprop.ff[iv][lc];

    output->mc.g1cld[nlyr-1-lc] = output->wc.optprop.g1 [iv][lc];
    output->mc.g2cld[nlyr-1-lc] = output->wc.optprop.g2 [iv][lc];
    output->mc.ffcld[nlyr-1-lc] = output->wc.optprop.ff [iv][lc];

    output->mc.g1ice[nlyr-1-lc] = output->ic.optprop.g1 [iv][lc];
    output->mc.g2ice[nlyr-1-lc] = output->ic.optprop.g2 [iv][lc];
    output->mc.ffice[nlyr-1-lc] = output->ic.optprop.ff [iv][lc];

    /* ??? consequently, one should apply input.atm.interpol_method_refind; for reasons */
    /* ??? of lazyness we simply average the refractive index at the adjacent levels    */
    /* ??? to get the layer property for MYSTIC                                         */
    output->mc.refind[nlyr-1-lc] = 0.5*(output->atm.microphys.refind [lc]+output->atm.microphys.refind [lc+1]);

    /* effective droplet radius             */
    /* ??? this is a temporary solution ??? */
    output->mc.recld [nlyr-1-lc] = output->wc.microphys.effr_layer[lc];
    output->mc.lwccld[nlyr-1-lc] = output->wc.microphys.lwc_layer [lc];

    /* Set optical depth and single scattering albedo */
    if (bext < FLT_MIN) 
      bext = FLT_MIN;
    if (bext_md < FLT_MIN) 
      bext_md = FLT_MIN;

    output->dtauc[lc]    = bext;
    output->dtauc_md[lc] = bext_md;
    output->ssalb[lc] = bsca/bext;

    if (1.0-output->ssalb[lc] < FLT_MIN) 
      output->ssalb[lc] = 1.0 - FLT_MIN;
    
    if (input.rte.solver == POLRADTRAN) {
      if ((f = fopen(&output->atm.pol_scat_files[lc*64], "w")) == NULL) return 1;
      fprintf(f,"%e\n",bext);
      fprintf(f,"%e\n",bsca);
      fprintf(f,"%e\n",bsca/bext);

      fprintf(f,"%d\n", input.rte.nstr); 

      momag1=bscaa;
      momag2=bscaa;

      momig1=bscai;
      momig2=bscai;

      momdg1=bscad;
      momdg2=bscad;

      for (k=0; k<=output->atm.nmom; k++) { 
	if (output->aer.optprop.nmom[iv] > 0) {
	  if (k<output->aer.optprop.nmom[iv][lc])
	    moma = bscaa * output->aer.optprop.moment[iv][lc][k] / 
	      output->aer.optprop.moment[iv][lc][0];
	  else 
	    moma=0;
	}
	else {
	  if (k>0) {
	    momag1 *= g1a;
	    momag2 *= g2a;
	  }
	  momd = momag1*fa + (1.0-fa)*momag2;
	}

	if (k>0) {
	  momdg1 *= g1d;
	  momdg2 *= g2d;
	}
	momd = momdg1*fd + (1.0-fd)*momdg2;

	if (k>0) {
	  momig1 *= g1i;
	  momig2 *= g2i;
	}
	momi = momig1*fi + (1.0-fi)*momig2;
	
	/* The first numbers in the columns are the Legendre coefficients
	   for the Rayleigh scattering phase matrix. Note that disort wants
	   phase function moments g_l while polradtran wants Legendre 
	   coefficients C_l = (2*l+1)g_l. p1 is the scalar story of everything. */
	if (k==0 && output->atm.rayleigh) {     
	  p1 =  1.0;
	  p2 = -0.5*Delta;
	  p3 =  0.0;
	  p4 =  0.0;
	  p5 =  Delta;
	  p6 =  0.0;
	} 
	else if (k==1 && output->atm.rayleigh) { 
	  p1 =  (0.0 + (2*k+1)*(momd + moma + momi))/sumsca; 
	  p2 =  0.0; 
	  p3 =  1.5*Delta; 
	  p4 =  0.0; 
	  p5 =  0.0; 
	  p6 =  1.5*Delta*Deltap; 
	} 
	else if (k==2 && output->atm.rayleigh) { 
	  p1 =  0.5*Delta*bscar/sumsca + (2*k+1)*(momd + moma + momi)/ sumsca; 
	  p2 =  0.5*Delta; 
	  p3 =  0.0; 
	  p4 =  0.0; 
	  p5 =  0.5*Delta; 
	  p6 =  0.0; 
	}
	else {
	  if (k==0) { p1 = 1.0; } 
	  else {      p1 =  (0.0+ (2*k+1)*(momd + moma + momi)/sumsca);  }
	  p2 =  0.0; 
	  p3 =  0.0; 
	  p4 =  0.0; 
	  p5 =  0.0; 
	  p6 =  0.0; 
	} 
	fprintf(f,"%d %f %f %f %f %f %f\n", k, p1, p2, p3, p4, p5, p6); 
      }
      fclose(f);
    }
  }  /* for (lc=0; lc<nlyr; lc++) */


  output->mc.dtair[nlyr] = 0.0;
  output->mc.dtmol[nlyr] = 0.0;
  output->mc.dtaer[nlyr] = 0.0; 
  output->mc.dtcld[nlyr] = 0.0;
  output->mc.dtice[nlyr] = 0.0;
                     
  output->mc.omaer[nlyr] = 0.0;
  output->mc.omcld[nlyr] = 0.0;
  output->mc.omice[nlyr] = 0.0;

  output->mc.omaer[nlyr] = 0.0;
  output->mc.omcld[nlyr] = 0.0;
  output->mc.omice[nlyr] = 0.0;

  /* loop includes nlyr because altitude and temperature */
  /* are defined per level                               */
  for (lc=0; lc<=nlyr; lc++) {
    output->mc.z     [nlyr-lc] = output->atm.zd[lc];
    output->mc.temper[nlyr-lc] = output->atm.microphys.temper[lc];
  }

  /* loop stops at nlyr-1 because aerosol properties */
  /* are defined per layer                           */
  for (lc=0; lc<nlyr; lc++) {

    output->mc.nmomaer[nlyr-lc-1] = output->aer.optprop.nmom[iv][lc];
    output->mc.momaer [nlyr-lc-1] = calloc (output->mc.nmomaer[nlyr-lc-1], sizeof(float));
    
    /* ??? should free this memory later ??? */

    for (k=0; k<output->mc.nmomaer[nlyr-lc-1]; k++)
      output->mc.momaer[nlyr-lc-1][k] = 
	output->aer.optprop.moment[iv][lc][k] / output->aer.optprop.moment[iv][lc][0];
  }

#if HAVE_SSSI
  if (input.rte.solver==SSSI) {
    /* Total cloud optical thickness and top cloud level */
    /* for the SSSI approximation                        */
    output->sssi.tautot = bscadtot + babsdtot + bscaitot + babsitot;
    for (lc=0; lc<nlyr; lc++) 
      if (output->wc.optprop.dtau [iv][lc]>0 || output->ic.optprop.dtau [iv][lc]>0)
	break;

    output->sssi.lctop = lc; 
    
    /* the uppermost layer determines cloud phase */
    output->sssi.type = (output->wc.optprop.dtau [iv][lc] > output->ic.optprop.dtau [iv][lc] ?
			 ISCCP_WATER : ISCCP_ICE);
  }
#endif

#if HAVE_SSS2
  /* ??? this will be changed ??? */
  if (input.rte.solver==SSS2) {
    /* Total cloud optical thickness and top cloud level */
    /* for the SSS2 approximation                        */
    output->sssi.tautot = bscadtot + babsdtot + bscaitot + babsitot;
    for (lc=0; lc<nlyr; lc++) 
      if (output->wc.optprop.dtau [iv][lc]>0 || output->ic.optprop.dtau [iv][lc]>0)
	break;

    output->sssi.lctop = lc; 
    
    /* the uppermost layer determines cloud phase */
    output->sssi.type = (output->wc.optprop.dtau [iv][lc] > output->ic.optprop.dtau [iv][lc] ?
			 ISCCP_WATER : ISCCP_ICE);
  }
#endif
  /* output vertical (total) sum */
  if (verbose) {
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "%5s | %6.2f | %9.6f | %9.6f %9.6f %5.3f | %11.6f %11.6f %5.3f | %11.6f %11.6f %5.3f %5.3f %5.3f %6.3f | %11.6f\n", 
	     "sum", 0.0/0.0,
	     bscartot,
	     bscaatot, babsatot, 0.0/0.0,
	     bscadtot, babsdtot, 0.0/0.0,
	     bscaitot, babsitot, 0.0/0.0,
	     0.0/0.0, 0.0/0.0, 0.0/0.0,
	     babsotot);
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------------------------\n");
  }
	   
  return 0;
}



/***********************************************************************************/
/* Function: crs_rayleigh_bodhaine                                                 */
/* Description:                                                                    */
/*    Calculates the Rayleigh scattering cross section according to the formula    */
/*    by Bodhaine et al, `On Rayleigh optical depth calculations', J. Atm. Ocean   */
/*    Technol., 16, 1854-1861, 1999.                                               */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float* lambda:         Pointer to ary holding the wavelength in nanometers.    */
/*  int n_lambda:          Number of wavelength points.                            */
/*                                                                                 */
/* Return value:                                                                   */
/*  double pointer to ary holding the Rayleigh scattering cross section as a       */
/*  function of wavelength.                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

int crs_rayleigh_bodhaine (float *lambda, int n_lambda, float mixing_ratio_co2, 
			   float *depol, float ***crs) 
{
  double from_nm_to_cm = 1.0e-7;  
  double from_nm_to_um = 1.0e-3;
  double N_s           = 2.546899e+19; 
  double ray_const     = 24*pow(PI,3)/pow(N_s,2);
  double F_N_2=0, F_O_2=0, F_air=0;
  double co2 =0;
  double ref_ratio = 0, n_300=0, n=0;  /* n = refractive index of air at desired */ 
                                       /*     co2 concentration                  */
  double lambda_cm=0.0, lambda_um=0.0;
  int iv=0;

  co2 = mixing_ratio_co2*1.e-4; /* This is co2 in parts per volume by percent, top of page */
                                /* 1858 of Bodhaine et al paper.                           */

  for (iv=0; iv<n_lambda; iv++) {

    lambda_cm = lambda[iv] * from_nm_to_cm;   /* Funny to use two different units */
    lambda_um = lambda[iv] * from_nm_to_um;   /* for lambda in the same paper.....*/

    n_300     = (8060.51 + (2480990/(132.274-pow(lambda_um,-2)))
		 + (17455.7/(39.32957-pow(lambda_um,-2))))*1e-08 ;  /* Eq. (18) */
    n         = (1+0.54*(mixing_ratio_co2*1.e-6-0.0003))*n_300 + 1;  /* Eq. (19) */
    ref_ratio = pow(n*n-1,2)/pow(n*n+2,2);

    F_N_2 = 1.034 + 3.17e-4/pow(lambda_um,2);                                /* Eq. (5) */ 
    F_O_2 = 1.096 + 1.385e-3/pow(lambda_um,2) + 1.448e-4/pow(lambda_um,4);   /* Eq. (6) */
    F_air = (78.084*F_N_2 + 20.946*F_O_2 + 0.934*1.0 + co2*1.15)/(78.084+20.946+0.934+co2); /* Eq. (23) */
    
    if (depol[iv] >= 0)
      F_air = (6.0+3.0*depol[iv]) / (6.0-7.0*depol[iv]);
    else
      depol[iv] = 6.0 * (F_air-1.0) / (3.0+7.0*F_air);

    crs[0][MOL_AIR][iv] =  (ray_const / pow(lambda_cm,4)) * ref_ratio * F_air;
  }

  return 0;
}


/***********************************************************************************/
/* Function: crs_rayleigh_nicolet                                                  */
/* Description:                                                                    */
/*    Calculates the Rayleigh scattering cross section according to the formula    */
/*    by Nicolet, M., `On the molecular scattering in the terrestrial atmosphere:  */
/*    an empirical formula for its calculation in the homosphere', Planet. Space   */
/*    Sci., 32, 1467-1468, 1984.                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float* lambda:         Pointer to ary holding the wavelength in nanometers.    */
/*  int n_lambda:          Number of wavelength points.                            */
/*                                                                                 */
/* Return value:                                                                   */
/*  double pointer to ary holding the Rayleigh scattering cross section as a       */
/*  function of wavelength.                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

int crs_rayleigh_nicolet (float *lambda, int n_lambda, float ***crs) 
{
  float ray_const  = 4.02e-28;
  float from_nm_um = 1.0e-3;
  float lambda_um=0.0, x=0.0;
  int iv=0;
  
  for (iv=0; iv<n_lambda; iv++) {

    lambda_um = lambda[iv] * from_nm_um;

    if (lambda_um > 0.55)
      x = 0.04;
    else
      x = 0.389*lambda_um + 0.09426/lambda_um - 0.3228;
    
    crs[0][MOL_AIR][iv] = ray_const / pow(lambda_um,4+x);
  }

  return 0;
}



/***********************************************************************************/
/* Sum the 3D results over wavelength and print data to mc.sum                     */ 
/***********************************************************************************/

static int write_spectral3D (input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iv=0;

  char flxfilename[FILENAME_MAX] = "";
  char radfilename[FILENAME_MAX] = "";
  char absfilename[FILENAME_MAX] = "";

  FILE *fflx=NULL, *fabs=NULL, *frad=NULL;
  
  /* generate output file names */
  strcpy (flxfilename, input.rte.mc.filename[FN_MC_BASENAME]);
  strcpy (absfilename, input.rte.mc.filename[FN_MC_BASENAME]);
  strcpy (radfilename, input.rte.mc.filename[FN_MC_BASENAME]);

  strcat (flxfilename, ".flx.spc");
  strcat (radfilename, ".rad.spc");

  /* extension for absorption/heating/actinic etc. file */
  switch (input.rte.mc.absorption) {
  case MCABS_ACTINIC:
    strcat (absfilename, ".act.spc");
    break;

  case MCABS_ABSORPTION:
  case MCABS_EMISSION:
  case MCABS_HEATING:
  case MCABS_NONE:
    strcat (absfilename, ".abs.spc");
    break;

  default:
    fprintf (stderr, "Error, unknown absorption type %d\n", 
	     input.rte.mc.absorption);
    break;
  }



  if ((fflx = fopen (flxfilename, "w"))==NULL) {
    fprintf (stderr, "Error opening %s for writing\n", flxfilename);
    return -1;
  }

  if ((frad = fopen (radfilename, "w"))==NULL) {
    fprintf (stderr, "Error opening %s for writing\n", radfilename);
    return -1;
  }

  for (iv=0; iv<output->wl.nlambda_h_print3D; iv++)
    for (ks=0; ks<input.atm.nzout; ks++)
      for (is=0; is<input.rte.mc.Nx_sample; is++)
	for (js=0; js<input.rte.mc.Ny_sample; js++) {
	  
	  fprintf (fflx, "%9.3f %4d %4d %4d %g %g %g %g %g %g\n", 
		   output->wl.lambda_h[iv],
		   is, js, ks, 
		   output->rfldir3d[ks][is][js][iv],
		   output->rfldn3d [ks][is][js][iv],
		   output->flup3d  [ks][is][js][iv],
		   output->uavgso3d[ks][is][js][iv],
		   output->uavgdn3d[ks][is][js][iv],
		   output->uavgup3d[ks][is][js][iv]);

	  fprintf (frad, "%9.3f %4d %4d %4d %g\n", 
		   output->wl.lambda_h[iv],
		   is, js, ks, 
		   output->radiance3d[ks][is][js][iv]);
	}
  
  fclose(fflx); 
  fclose(frad);


  if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE) {

    if ((fabs = fopen (absfilename, "w"))==NULL) {
      fprintf (stderr, "Error opening %s for writing\n", absfilename);
      return -1;
    }

    for (ks=0; ks<output->atm.Nzcld; ks++)
      if (output->atm.threed[ks])    /* only for 3D layers, BM07122005 */
	for (is=0; is<output->atm.Nxcld; is++)
	  for (js=0; js<output->atm.Nycld; js++)
	    for (iv=0; iv<output->wl.nlambda_h_print3D; iv++)
	      fprintf (fabs, "%9.3f %4d %4d %4d %g\n", 
		       output->wl.lambda_h[iv],
		       is, js, ks, 
		       output->abs3d[ks][is][js][iv]);
    fclose(fabs); 
  }
  
  return 0;
}


/***********************************************************************************/
/* Sum the 1D results over wavelength and write data to the first array element    */ 
/***********************************************************************************/

int sum3D (input_struct input, output_struct *output)
{ 
  int iv=0, is=0, js, ks;
  double ffactor=0, rfactor=0;
  float incident=0, fbeam=0; 
  int status=0;

  /* calculate incident flux (incident) and sum extraterrestrial irradiance (fbeam) */
  for (iv=0; iv<output->wl.nlambda_h; iv++) {
    incident += output->wl.filter[iv] * output->wl.fbeam[iv] * cos(output->sza_h[iv]*PI/180.0);
    fbeam    += output->wl.filter[iv] * output->wl.fbeam[iv];
  }

  /* default scaling factors                   */
  ffactor = 1.0;   /* irradiance multiplicator */
  rfactor = 1.0;   /* radiance multiplicator   */
  
  /* if transmittance, then divide by the extraterrestrial flux */
  if (input.source != SRC_THERMAL) {
    switch (input.calibration) {
    case OUTCAL_ABSOLUTE:
      break;

    case OUTCAL_TRANSMITTANCE:
    case OUTCAL_BRIGHTNESS:
      ffactor = 1.0 / fbeam;   /* irradiance multiplicator */
      rfactor = 1.0 / fbeam;   /* radiance multiplicator   */
      break;

    case OUTCAL_REFLECTIVITY:

      ffactor = 1.0 / incident;
      rfactor = PI  / incident;
      break;
      
    default:
      fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
      return -1;
    }
  }


  /* fluxes and radiances */
  for (ks=0; ks<input.atm.nzout; ks++)
    for (is=0; is<input.rte.mc.Nx_sample; is++)
      for (js=0; js<input.rte.mc.Ny_sample; js++) {
	
	for (iv=1; iv<output->wl.nlambda_h; iv++)  {
	  output->rfldir3d   [ks][is][js][0] += output->rfldir3d   [ks][is][js][iv];
	  output->rfldn3d    [ks][is][js][0] += output->rfldn3d    [ks][is][js][iv];
	  output->flup3d     [ks][is][js][0] += output->flup3d     [ks][is][js][iv];
	  output->uavgso3d   [ks][is][js][0] += output->uavgso3d   [ks][is][js][iv];
	  output->uavgdn3d   [ks][is][js][0] += output->uavgdn3d   [ks][is][js][iv];
	  output->uavgup3d   [ks][is][js][0] += output->uavgup3d   [ks][is][js][iv];
	  
	  output->radiance3d [ks][is][js][0] += output->radiance3d [ks][is][js][iv];
	}
	
	/* calibrate */ 
	if (output->wl.nlambda_h > 0) {
	  output->rfldir3d   [ks][is][js][0] *= ffactor;
	  output->rfldn3d    [ks][is][js][0] *= ffactor;
	  output->flup3d     [ks][is][js][0] *= ffactor;
	  output->uavgso3d   [ks][is][js][0] *= ffactor;
	  output->uavgdn3d   [ks][is][js][0] *= ffactor;
	  output->uavgup3d   [ks][is][js][0] *= ffactor;

	  output->radiance3d [ks][is][js][0] *= rfactor;
	}
      }

  if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
    for (ks=0; ks<output->atm.Nzcld; ks++)
      if (output->atm.threed[ks])    /* only for 3D layers, BM07122005 */
	for (is=0; is<output->atm.Nxcld; is++)
	  for (js=0; js<output->atm.Nycld; js++) {
	    
	    for (iv=1; iv<output->wl.nlambda_h; iv++)
	      output->abs3d[ks][is][js][0] += output->abs3d[ks][is][js][iv];
  
	    output->abs3d[ks][is][js][0] *= ffactor; /* ??? need to check if that is the correct factor ??? */
	  }
  

  /* finally, convert to brightness temperature if requested */
  if (input.source == SRC_THERMAL && input.calibration==OUTCAL_BRIGHTNESS) {
    
    if (!input.quiet)
      fprintf (stderr, " ... converting 3D radiances and fluxes to brightness temperatures\n");
    
    /* convert irradiances / radiances to brightness temperatures */
    status = output2bt3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by output2bt()\n", status);
      return status;
    }
  }

  return status;
}



/**************************************************************************************/
/* Integrate the 3D results over wavelength and write data to the first array element */ 
/**************************************************************************************/

int integrate3D (input_struct input, output_struct *output)
{ 
  int iv=0, is=0, js=0, ks=0;
  int status=0;
  double ffactor=0, rfactor=0;
  float *yint=NULL;
  float incident=0, fbeam=0;

  /* calculate integrated incident flux */
  if ((yint = (float *) calloc (output->wl.nlambda_h, sizeof(double))) == NULL)
    error_calloc ("yint","integrate1D", &(status));

  for (iv=0; iv<output->wl.nlambda_h; iv++)
    yint[iv] = output->wl.filter[iv] * output->wl.fbeam[iv] * cos(output->sza_h[iv]*PI/180.0);

  incident = integrate_float (output->wl.lambda_h, yint, output->wl.nlambda_h);

  fprintf (stderr, "3D %f\n", incident);

  /* calculate integrated extraterrestrial irradiance */
  for (iv=0; iv<output->wl.nlambda_h; iv++)
    yint[iv] = output->wl.filter[iv] * output->wl.fbeam[iv];

  fbeam = integrate_float (output->wl.lambda_h, yint, output->wl.nlambda_h);  


  /* default scaling factors                   */
  ffactor = 1.0;   /* irradiance multiplicator */
  rfactor = 1.0;   /* radiance multiplicator   */
  
  /* if transmittance, then divide by the extraterrestrial flux */
  if (input.source != SRC_THERMAL) {
    switch (input.calibration) {
    case OUTCAL_ABSOLUTE:
      break;

    case OUTCAL_TRANSMITTANCE:
    case OUTCAL_BRIGHTNESS:
      ffactor = 1.0 / fbeam;   /* irradiance multiplicator */
      rfactor = 1.0 / fbeam;   /* radiance multiplicator   */
      break;

    case OUTCAL_REFLECTIVITY:

      ffactor = 1.0 / incident;
      rfactor = PI  / incident;
      break;
      
    default:
      fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
      return -1;
    }
  }


  /* fluxes and radiances */
  for (ks=0; ks<input.atm.nzout; ks++)
    for (is=0; is<input.rte.mc.Nx_sample; is++)
      for (js=0; js<input.rte.mc.Ny_sample; js++) {

	output->rfldir3d   [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->rfldir3d [ks][is][js], 
							      output->wl.nlambda_h); 
	
	output->rfldn3d    [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->rfldn3d [ks][is][js], 
							      output->wl.nlambda_h); 
	
	output->flup3d     [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->flup3d [ks][is][js], 
							      output->wl.nlambda_h); 
	
	output->uavgso3d   [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->uavgso3d [ks][is][js], 
							      output->wl.nlambda_h); 
	
	output->uavgdn3d   [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->uavgdn3d [ks][is][js], 
							      output->wl.nlambda_h); 
	
	output->uavgup3d   [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->uavgup3d [ks][is][js], 
							      output->wl.nlambda_h); 
	
	output->radiance3d [ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->radiance3d [ks][is][js], 
							      output->wl.nlambda_h); 

	
	/* calibrate */ 
	if (output->wl.nlambda_h > 0) {
	  output->rfldir3d   [ks][is][js][0] *= ffactor;
	  output->rfldn3d    [ks][is][js][0] *= ffactor;
	  output->flup3d     [ks][is][js][0] *= ffactor;
	  output->uavgso3d   [ks][is][js][0] *= ffactor;
	  output->uavgdn3d   [ks][is][js][0] *= ffactor;
	  output->uavgup3d   [ks][is][js][0] *= ffactor;

	  output->radiance3d [ks][is][js][0] *= rfactor;
	}
      }

  if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
    for (ks=0; ks<output->atm.Nzcld; ks++)
      if (output->atm.threed[ks])    /* only for 3D layers, BM07122005 */
	for (is=0; is<output->atm.Nxcld; is++)
	  for (js=0; js<output->atm.Nycld; js++) {
	    
	    for (iv=1; iv<output->wl.nlambda_h; iv++)
	      output->abs3d[ks][is][js][0] = integrate_float (output->wl.lambda_h, 
							      output->abs3d[ks][is][js],
							      output->wl.nlambda_h); 

	    if (output->wl.nlambda_h > 0)
	      output->abs3d[ks][is][js][0] *= ffactor; /* ??? need to check if that is the correct factor ??? */
	  }
  

  /* finally, convert to brightness temperature if requested */
  if (input.source == SRC_THERMAL && input.calibration==OUTCAL_BRIGHTNESS) {
    
    if (!input.quiet)
      fprintf (stderr, " ... converting 3D radiances and fluxes to brightness temperatures\n");
    
    /* convert irradiances / radiances to brightness temperatures */
    status = output2bt3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by output2bt()\n", status);
      return status;
    }
  }

  return status;
}







/***************************************/
/* Post - processing of the output     */ 
/* sum or integration over wavelength, */ 
/* calculation of heating rates or     */
/* conversion to RGB                   */
/***************************************/

int processing1D (input_struct input, output_struct *output)
{
  int status=0;

  switch (input.processing) {

  case PROCESS_NONE:
    break;

  case PROCESS_SUM:
    status = sum1D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d summing 1D output over wavelength\n", status);
      return status;
    }
    break;
    
  case PROCESS_INT:
    status = integrate1D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d integrating 1D output over wavelength\n", status);
      return status;
    }
    break;

  case PROCESS_RGB:
  case PROCESS_RGBNORM:
    status = spec2rgb (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d converting spectral output to RGB\n", status);
      return status;
    }
    break;

  default:
    fprintf (stderr, "Error, unknown processing scheme %d\n", input.processing);
    return -1;
  }

  return 0;
}


/***************************************/
/* Post - processing of the output     */ 
/* sum or integration over wavelength, */ 
/* calculation of heating rates or     */
/* conversion to RGB                   */
/***************************************/

int processing3D (input_struct input, output_struct *output)
{
  int status=0;

  output->wl.nlambda_h_print3D = output->wl.nlambda_h;

  switch (input.processing) {

  case PROCESS_NONE:
    break;

  case PROCESS_SUM:
    status = sum3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d summing 1D output over wavelength\n", status);
      return status;
    }
    output->wl.nlambda_h_print3D=1;
    break;
    
  case PROCESS_INT:
    status = integrate3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d integrating 1D output over wavelength\n", status);
      return status;
    }
    output->wl.nlambda_h_print3D=1;
    break;

  case PROCESS_RGB:
  case PROCESS_RGBNORM:
    fprintf (stderr, "Error, conversion to RGB for 3D results not yet implemented!\n"); 
    return -1;
    break;

  default:
    fprintf (stderr, "Error, unknown processing scheme %d\n", input.processing);
    return -1;
  }

  /* write 3D data to files */
  status = write_spectral3D (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d writing spectral 3D output\n", status);
    return status;
  }


  return 0;
}





/***********************************************************************************/
/* Sum the 1D results over wavelength and write data to the first array element    */ 
/***********************************************************************************/

int sum1D (input_struct input, output_struct *output)
{ 
  int lev=0, iv=0, iu=0, j=0, is=0;
  int status=0;

  /* calculate incident flux (output->incident) and sum extraterrestrial irradiance (wl.fbeam[0]) */
  for (iv=0; iv<output->wl.nlambda_h; iv++) {
    output->incident               += output->wl.filter[iv] * output->wl.fbeam[iv] * cos(output->sza_h[iv]*PI/180.0);
    if (iv!=0) 
      output->wl.fbeam[0] += output->wl.filter[iv] * output->wl.fbeam[iv];
    else /* iv == 0 */
      output->wl.fbeam[0]  = output->wl.filter[0]  * output->wl.fbeam[0] ; 
  }

  /* callocate integrated values */
  status = calloc_int_values(input, output);
  if (status!=0)  {
    fprintf (stderr, "error allocating integrated values, status %d\n", status);
    return status;
  }

  /* fluxes and radiances */
  for (lev=0; lev<input.atm.nzout; lev++) {
    for (iv=0; iv<output->wl.nlambda_h; iv++) {

      output->rfldir_int[lev] += (double) output->rfldir[lev][iv];
      output->rfldn_int [lev] += (double) output->rfldn [lev][iv];
      output->flup_int  [lev] += (double) output->flup  [lev][iv];
      output->uavg_int  [lev] += (double) output->uavg  [lev][iv];
      output->uavgso_int[lev] += (double) output->uavgso[lev][iv];
      output->uavgdn_int[lev] += (double) output->uavgdn[lev][iv];
      output->uavgup_int[lev] += (double) output->uavgup[lev][iv];
      if (input.heating != NO_HEAT) {
        output->heat_int  [lev] += (double) output->heat  [lev][iv];
        output->w_zout_int[lev] += (double) output->w_zout[lev][iv]; 
      }

      for (iu=0; iu<input.rte.numu; iu++) {
        output->u0u_int[lev][iu] += output->u0u[lev][iu][iv];

	for (j=0; j<input.rte.nphi; j++)
	  output->uu_int[lev][j][iu] += output->uu[lev][j][iu][iv];
      }
    }
  }

  /* polarized fluxes and radiances */
  if (input.rte.solver == POLRADTRAN) 
    for (iv=0; iv<output->wl.nlambda_h; iv++) {
      for (lev=0; lev<input.atm.nzout; lev++) 
	for (is=0; is<input.rte.pol_nstokes; is++) {
	  output->down_flux_int [lev][is] += output->down_flux [lev][is][iv];
	  output->up_flux_int   [lev][is] += output->up_flux   [lev][is][iv];

	  for (iu=0; iu<input.rte.nstr/2; iu++)
	    for (j=0; j<input.rte.nphi; j++) {
	      output->down_rad_int [lev][j][iu][is] += output->down_rad [lev][j][iu][is][iv]; 
	      output->up_rad_int   [lev][j][iu][is] += output->up_rad   [lev][j][iu][is][iv];
	    }
	}
    }


  /* store integrated values in the first entry of the wavelength index, if any result fields are defined */
  if (output->wl.nlambda_h > 0)
    status += double2float_integrated_values(input, output);

  if (status!=0)  {
    fprintf (stderr, "Error, conversion from double to float not possible");
    fprintf (stderr, "sum1D(), ancillary.c, status %d\n", status);
    return status;
  }

  status += scaling_integrated_values(input, output);
  if (status!=0)  {
    fprintf (stderr, "Error, scaling integrated values.\n");
    fprintf (stderr, "scaling_integrated_values(), in ancillary.c, status %d\n", status);
    return status;
  }

  return status;
}


/**************************************************************************************/
/* Integrate the 1D results over wavelength and write data to the first array element */ 
/**************************************************************************************/

int integrate1D (input_struct input, output_struct *output)
{ 
  int lev=0, iv=0, iu=0, j=0, is=0;
  int status=0;
  double *xint=NULL, *yint=NULL;

  /* callocate integrated values */
  status = calloc_int_values(input, output);
  if (status!=0)  {
    fprintf (stderr, "error allocating integrated values, status %d\n", status);
    return status;
  }

  /* calculate incident flux */
  xint = (double *) calloc (output->wl.nlambda_h, sizeof(double));
  if (xint == NULL) error_calloc("xint","integrate1D", &(status));
  yint = (double *) calloc (output->wl.nlambda_h, sizeof(double));
  if (yint == NULL) error_calloc("yint","integrate1D", &(status));
  for (iv=0; iv<output->wl.nlambda_h; iv++) {
    xint[iv] = (double) output->wl.lambda_h[iv];
    yint[iv] = output->wl.filter[iv] * output->wl.fbeam[iv] 
                     * cos(output->sza_h[iv]*PI/180.0);
    /* unit factor should alway be 1 here == output always in W/(m2 nm) */ 
    /* therefor it is not included here in the code */
  }

  output->incident = integrate(xint, yint, output->wl.nlambda_h);

  fprintf (stderr, "1D %f\n", output->incident);

  /* sum extraterrestrial irradiance */
  for (iv=0; iv<output->wl.nlambda_h; iv++)
    yint[iv] = output->wl.filter[iv] * output->wl.fbeam[iv];
    /* unit factor should alway be 1 here == output always in W/(m2 nm) */ 
    /* therefor it is not included here in the code */

  output->wl.fbeam[0] = (float) integrate(xint, yint, output->wl.nlambda_h);  

  /* fluxes and radiances */
  for (lev=0; lev<input.atm.nzout; lev++) {

    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->rfldir[lev][iv];
    output->rfldir_int[lev] = integrate(xint, yint, output->wl.nlambda_h);
    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->rfldn[lev][iv];
    output->rfldn_int[lev]  = integrate(xint, yint, output->wl.nlambda_h);
    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->flup[lev][iv];
    output->flup_int[lev]   = integrate(xint, yint, output->wl.nlambda_h);
    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->uavg[lev][iv];
    output->uavg_int[lev]   = integrate(xint, yint, output->wl.nlambda_h);
    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->uavgso[lev][iv];
    output->uavgso_int[lev] = integrate(xint, yint, output->wl.nlambda_h);
    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->uavgdn[lev][iv];
    output->uavgdn_int[lev] = integrate(xint, yint, output->wl.nlambda_h);
    for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->uavgup[lev][iv];
    output->uavgup_int[lev] = integrate(xint, yint, output->wl.nlambda_h);
    if (input.heating != NO_HEAT) { 
      for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->heat[lev][iv];
      output->heat_int[lev]   = integrate(xint, yint, output->wl.nlambda_h);
      for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->w_zout[lev][iv];
      output->w_zout_int[lev] = integrate(xint, yint, output->wl.nlambda_h);      
    }

    for (iu=0; iu<input.rte.numu; iu++) {
      for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = (double) output->u0u[lev][iu][iv];
      output->u0u_int[lev][iu] = integrate(xint, yint, output->wl.nlambda_h);
        
      for (j=0; j<input.rte.nphi; j++) {
	for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = (double) output->uu[lev][j][iu][iv];
	output->uu_int[lev][j][iu] = integrate(xint, yint, output->wl.nlambda_h);
      }
    }
  }

  /* polarized fluxes and radiances */
  if (input.rte.solver == POLRADTRAN) {
    
    for (lev=0; lev<input.atm.nzout; lev++) {
      for (is=0; is<input.rte.pol_nstokes; is++) {
	 for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->down_flux[lev][is][iv];
	   output->down_flux_int[lev][is] = (float) integrate(xint, yint, output->wl.nlambda_h);
	 for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->up_flux[lev][is][iv];
	   output->up_flux_int[lev][is] = (float) integrate(xint, yint, output->wl.nlambda_h);

	 for (iu=0; iu<input.rte.nstr/2; iu++) {
	   for (j=0; j<input.rte.nphi; j++) {
	     for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->down_rad[lev][j][iu][is][iv];
	       output->down_rad_int[lev][j][iu][is] = (float) integrate(xint, yint, output->wl.nlambda_h);
	     for (iv=0; iv<output->wl.nlambda_h; iv++) yint[iv] = output->up_rad[lev][j][iu][is][iv];
	       output->up_rad_int[lev][j][iu][is] = (float) integrate(xint, yint, output->wl.nlambda_h);
	   }
	 }
      }
    }
  }

  free(xint);
  free(yint); 

  /* store integrated values in the first entry of the wavelength index */
  if (output->wl.nlambda_h > 0)
    status += double2float_integrated_values(input, output);

  if (status!=0)  {
    fprintf (stderr, "Error, conversion from double to float not possible");
    fprintf (stderr, "integrate1D(), ancillary.c, status %d\n", status);
    return status;
  }

  status += scaling_integrated_values(input, output);
  if (status!=0)  {
    fprintf (stderr, "Error, scaling integrated values.\n");
    fprintf (stderr, "integrate1D(), in ancillary.c, status %d\n", status);
    return status;
  }  

  return status;
}


/*********************************/
/* used by integrate1D and sum1D */
/*********************************/
int calloc_int_values(input_struct input, output_struct *output)
{
  int status=0;
  int lev=0, j=0, iu=0;

  /* callocate wavelength integrated fluxes and radiances */
  output->rfldir_int = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->rfldir_int == NULL) error_calloc("output->rfldir_int","calloc_int_values", &(status));
  output->rfldn_int  = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->rfldn_int == NULL) error_calloc("output->rfldn_int","calloc_int_values", &(status)); 
  output->flup_int   = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->flup_int == NULL) error_calloc("output->flup_int","calloc_int_values", &(status));
  output->uavg_int   = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->uavg_int == NULL) error_calloc("output->uavg_int","calloc_int_values", &(status));
  output->uavgso_int = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->uavgso_int == NULL) error_calloc("output->uavgso_int","calloc_int_values", &(status));
  output->uavgdn_int = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->uavgdn_int == NULL) error_calloc("output->uavgdn_int","calloc_int_values", &(status));
  output->uavgup_int = (double *) calloc (input.atm.nzout, sizeof(double));
  if (output->uavgup_int == NULL) error_calloc("output->uavgup_int","calloc_int_values", &(status));
  if (input.heating != NO_HEAT) {
    output->heat_int = (double *) calloc (input.atm.nzout, sizeof(double));
    if (output->heat_int == NULL) error_calloc("output->heat_int","calloc_int_values", &(status)); 
    output->w_zout_int = (double *) calloc (input.atm.nzout, sizeof(double));
    if (output->w_zout_int == NULL) error_calloc("output->w_zout_int","calloc_int_values", &(status));
  }

  output->u0u_int    = (double **) calloc (input.atm.nzout, sizeof(double *));
  if (output->u0u_int == NULL) error_calloc("output->u0u_int","calloc_int_values", &(status));
  for (lev=0; lev<input.atm.nzout; lev++)
    output->u0u_int[lev] = (double *) calloc (input.rte.numu, sizeof(double));
  
  output->uu_int     = (double ***) calloc (input.atm.nzout, sizeof(double *));
  if (output->uu_int == NULL) error_calloc("output->uu_int","calloc_int_values", &(status));  
  for (lev=0; lev<input.atm.nzout; lev++) {
    output->uu_int[lev] = (double **) calloc (input.rte.nphi, sizeof(double *)); 
    for (j=0; j<input.rte.nphi; j++) 
      output->uu_int[lev][j] = (double *) calloc (input.rte.numu, sizeof(double));
  }

  /* callocate wavelength integrated PolRadtran flux and intensities on the output grid */ 
  if (input.rte.solver == POLRADTRAN) {
    output->down_flux_int = (double **) calloc (input.atm.nzout, sizeof(double *));
    if (output->down_flux_int == NULL) error_calloc("output->down_flux_int","calloc_int_values", &(status)); 
    for (lev=0; lev<input.atm.nzout; lev++)
      output->down_flux_int[lev] = (double *) calloc (input.rte.pol_nstokes,sizeof(double));

    output->up_flux_int = (double **) calloc (input.atm.nzout, sizeof(double *));
    if (output->up_flux_int == NULL) error_calloc("output->up_flux_int","calloc_int_values", &(status)); 
    for (lev=0; lev<input.atm.nzout; lev++)
      output->up_flux_int[lev] = (double *) calloc (input.rte.pol_nstokes,sizeof(double));
    
    output->down_rad_int = (double ****) calloc (input.atm.nzout, sizeof(double *));
    if (output->down_rad_int == NULL) error_calloc("output->down_rad_int","calloc_int_values", &(status));
    for (lev=0; lev<input.atm.nzout; lev++) {
      output->down_rad_int[lev] = (double ***) calloc(input.rte.nphi, sizeof(double *));
      for (j=0; j<input.rte.nphi; j++) {
        output->down_rad_int[lev][j] = (double **) calloc(input.rte.nstr/2, sizeof(double *));
        for (iu=0; iu<input.rte.nstr/2; iu++) {
          output->down_rad_int[lev][j][iu] = (double *) calloc(input.rte.pol_nstokes, sizeof(double));
        }
      }
    }

    output->up_rad_int = (double ****) calloc (input.atm.nzout, sizeof(double *));
    if (output->up_rad_int == NULL) error_calloc("output->up_rad_int","calloc_int_values", &(status));
    for (lev=0; lev<input.atm.nzout; lev++) {
      output->up_rad_int[lev] = (double ***) calloc(input.rte.nphi, sizeof(double *));
      for (j=0; j<input.rte.nphi; j++) {
        output->up_rad_int[lev][j] = (double **) calloc(input.rte.nstr/2, sizeof(double *));
        for (iu=0; iu<input.rte.nstr/2; iu++) {
          output->up_rad_int[lev][j][iu] = (double *) calloc(input.rte.pol_nstokes, sizeof(double));
        }
      }
    }
  }

  return status;
}

void error_calloc (char *variable, char *function, int *status)
{
  fprintf (stderr,"Error allocating memory for (%s) in %s (ancillary.c)\n", variable, function);
  fflush(stderr);
  *status = *status-1;
}


/***************************************************************************/
/* abuse first entry of the wavelength index to store the integrated value */ 
/* and to convert double (nessesary for heating rates) to floats           */
/* entry (n+1) or entry (-1) would be nicer                                */
/***************************************************************************/

int double2float_integrated_values(input_struct input,output_struct *output)
{
  int lev, is=0, j=0, iu=0;
  int status=0;  

  if (input.rte.solver == POLRADTRAN) {  /* polRadtran output */
    for (lev=0; lev<input.atm.nzout; lev++)
      for (is=0; is<input.rte.pol_nstokes; is++) {

	/* irradiances */
	output->down_flux [lev][is][0] = (float) output->down_flux_int[lev][is];
	output->up_flux   [lev][is][0] = (float) output->up_flux_int  [lev][is];
	
	/* radiances   */
	for (iu=0; iu<input.rte.nstr/2; iu++)
	  for (j=0; j<input.rte.nphi; j++) {
	    output->down_rad [lev][j][iu][is][0] = (float) output->down_rad_int [lev][j][iu][is];
	    output->up_rad   [lev][j][iu][is][0] = (float) output->up_rad_int   [lev][j][iu][is];
	  }
      }
  }
  else { /* unpolarized output */
  
    for (lev=0; lev<input.atm.nzout; lev++) { 

      /* irradiance / actinic flux */
      output->rfldir[lev][0] = (float) output->rfldir_int[lev];
      output->rfldn [lev][0] = (float) output->rfldn_int [lev];
      output->flup  [lev][0] = (float) output->flup_int  [lev];
      output->uavg  [lev][0] = (float) output->uavg_int  [lev];
      output->uavgso[lev][0] = (float) output->uavgso_int[lev];
      output->uavgdn[lev][0] = (float) output->uavgdn_int[lev];
      output->uavgup[lev][0] = (float) output->uavgup_int[lev];
      if (input.heating != NO_HEAT) {
        output->heat  [lev][0] = (float) output->heat_int  [lev]; 
        output->w_zout[lev][0] = (float) output->w_zout_int[lev];     
      }

      /* radiances   */
      for (iu=0; iu<input.rte.numu; iu++) {
        output->u0u[lev][iu][0] = (float) output->u0u_int[lev][iu];
      
        for (j=0; j<input.rte.nphi; j++)
          output->uu[lev][j][iu][0] = (float) output->uu_int[lev][j][iu];
      }
    }
  }

  return status;
}


/*********************************/
/* used by integrate1D and sum1D */
/*********************************/

int scaling_integrated_values(input_struct input, output_struct *output)
{
  double ffactor=0, rfactor=0, hfactor=1;
  int status=0;

  /* if transmittance, then divide by the extraterrestrial flux */
  if (input.source != SRC_THERMAL) {
    switch (input.calibration) {
    case OUTCAL_ABSOLUTE:
      ffactor = 1.0;   /* irradiance multiplicator */
      rfactor = 1.0;   /* radiance multiplicator   */
      break;

    case OUTCAL_TRANSMITTANCE:
    case OUTCAL_BRIGHTNESS:
      ffactor = 1.0 / output->wl.fbeam[0];   /* irradiance multiplicator */
      rfactor = 1.0 / output->wl.fbeam[0];   /* radiance multiplicator   */
      break;

    case OUTCAL_REFLECTIVITY:

      ffactor = 1.0 / output->incident;
      rfactor = PI  / output->incident;
      break;
      
    default:
      fprintf (stderr, "Error, unknown output calibration %d\n", input.calibration);
      return -1;
    }

    /**************************************************************/
    /* now scale irradiances with ffactor, radiances with rfactor */
    /**************************************************************/

    if (output->wl.nlambda_h>0) {
      
      /* in iv==0 the integrated values are stored */
      status = scale_output (input,
			     &(output->rfldir), &(output->rfldn),  &(output->flup),
			     &(output->uavgso), &(output->uavgdn), &(output->uavgup),
			     &(output->uavg), &(output->u0u), &(output->uu), &(output->heat), &(output->w_zout),
			     &(output->down_flux), &(output->up_flux), &(output->down_rad), &(output->up_rad),
			     &(output->rfldir3d), &(output->rfldn3d), &(output->flup3d), &(output->uavgso3d),
			     &(output->uavgdn3d), &(output->uavgup3d), &(output->radiance3d), 
			     output->atm.Nxcld, output->atm.Nycld, output->atm.Nzcld, output->atm.threed, 
			     &(output->abs3d),
			     ffactor, rfactor, hfactor, 0);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by scale_output()\n", status);
	return status;
      }
    }
  }
  else {
    if (input.calibration==OUTCAL_BRIGHTNESS) {

      if (!input.quiet)
	fprintf (stderr, " ... converting radiances to brightness temperatures\n");

      /* convert irradiances / radiances to brightness temperatures */
      status = output2bt (input, output);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by output2bt()\n", status);
	return status;
      }
    }
  }
  
  /* abuse first entry to store the integrated value, this is not so nice programmed */
  if (output->wl.nlambda_h>0)
    output->wl.nlambda_h = 1;
  
  return 0;
}


/***********************************************************************************/
/* Convert spectra to RGB and write data to the first three array elements         */ 
/***********************************************************************************/

int spec2rgb (input_struct input, output_struct *output)
{ 
  int lev=0, iu=0, j=0, is=0, status=0, norm=0;

  if (output->wl.nlambda_h<3) {
    fprintf (stderr, "Fatal error, need at least 3 wavelengths to store RGB\n");
    return -1;
  }

  /* normalized or weighted with brightness */
  norm=1;
  if (input.processing==PROCESS_RGB)
    norm=0;

  for (lev=0; lev<input.atm.nzout; lev++) {
    status += spectrum_to_rgb_overwrite (output->rfldir[lev], output->wl.nlambda_h, norm);
    status += spectrum_to_rgb_overwrite (output->rfldn [lev], output->wl.nlambda_h, norm);
    status += spectrum_to_rgb_overwrite (output->flup  [lev], output->wl.nlambda_h, norm);
    status += spectrum_to_rgb_overwrite (output->uavgso[lev], output->wl.nlambda_h, norm);
    status += spectrum_to_rgb_overwrite (output->uavgdn[lev], output->wl.nlambda_h, norm);
    status += spectrum_to_rgb_overwrite (output->uavgup[lev], output->wl.nlambda_h, norm);
    
    for (iu=0; iu<input.rte.numu; iu++) {
      status += spectrum_to_rgb_overwrite (output->u0u[lev][iu], output->wl.nlambda_h, norm);

      for (j=0; j<input.rte.nphi; j++)
	status += spectrum_to_rgb_overwrite (output->uu[lev][j][iu], output->wl.nlambda_h, norm);
    }
  }

  /* polarized fluxes and radiances */
  if (input.rte.solver == POLRADTRAN) 

    for (lev=0; lev<input.atm.nzout; lev++) 
      for (is=0; is<input.rte.pol_nstokes; is++) {
	status += spectrum_to_rgb_overwrite (output->down_flux[lev][is], output->wl.nlambda_h, norm);
	status += spectrum_to_rgb_overwrite (output->up_flux  [lev][is], output->wl.nlambda_h, norm);

	for (iu=0; iu<input.rte.nstr/2; iu++)
	  for (j=0; j<input.rte.nphi; j++) {
	    status += spectrum_to_rgb_overwrite (output->down_rad[lev][j][iu][is], output->wl.nlambda_h, norm);
	    status += spectrum_to_rgb_overwrite (output->up_rad  [lev][j][iu][is], output->wl.nlambda_h, norm);
	  }
      }
    

  if (output->wl.nlambda_h>0)
    output->wl.nlambda_h = 3;

  output->wl.lambda_h[0] = 600;
  output->wl.lambda_h[1] = 500;
  output->wl.lambda_h[2] = 400;

  return 0;
}



/***********************************************************************************/
/* Scale uvspec output fields with a given factor; ffactor is the scaling factor   */ 
/* for irradiances/actinic fluxes and rfactor is the scaling factor for radiances  */
/***********************************************************************************/

int scale_output (input_struct input,
		  float ***p_rfldir, float ***p_rfldn,  float ***p_flup,
		  float ***p_uavgso, float ***p_uavgdn, float ***p_uavgup,
		  float ***p_uavg, float ****p_u0u, float *****p_uu, float ***p_heat, float ***p_w_zout,
		  float ****p_down_flux, float ****p_up_flux, float ******p_down_rad, float ******p_up_rad,
		  float *****p_rfldir3d, float *****p_rfldn3d, float *****p_flup3d, float *****p_uavgso3d,
		  float *****p_uavgdn3d, float *****p_uavgup3d, float *****p_radiance3d, 
		  int Nx, int Ny, int Nz, int *threed, 
		  float *****p_abs3d,       
		  double ffactor, double rfactor, double hfactor, int iv)
     
/* changed output_struc to a bunch of pointers, as this function is also used in solve_rte(), UH, 2006-02 */
{
  int lev=0, iu=0, j=0, is=0, js=0, ks=0;

  if (input.rte.solver == POLRADTRAN) {  /* polRadtran output */
    for (lev=0; lev<input.atm.nzout; lev++) { 
      for (is=0; is<input.rte.pol_nstokes; is++) {

	/* irradiances */
	(*p_down_flux) [lev][is][iv] *= ffactor;
	(*p_up_flux)   [lev][is][iv] *= ffactor;
	
	/* radiances   */
	for (iu=0; iu<input.rte.nstr/2; iu++)
	  for (j=0; j<input.rte.nphi; j++) {
	    (*p_down_rad) [lev][j][iu][is][iv] *= rfactor;
	    (*p_up_rad)   [lev][j][iu][is][iv] *= rfactor;
	  }
      }

      /* heating rate */
      if (input.heating != NO_HEAT) {
        (*p_heat)  [lev][iv] *= hfactor;
        (*p_w_zout)[lev][iv] *= hfactor;
      }
    }
  }
  else { /* unpolarized output */

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* irradiance / actinic flux / heating rate */
      (*p_rfldir) [lev][iv] *= ffactor;
      (*p_rfldn)  [lev][iv] *= ffactor;
      (*p_flup)   [lev][iv] *= ffactor;
      (*p_uavg)   [lev][iv] *= ffactor;
      (*p_uavgso) [lev][iv] *= ffactor;
      (*p_uavgdn) [lev][iv] *= ffactor;
      (*p_uavgup) [lev][iv] *= ffactor;
      if (input.heating != NO_HEAT) {
        (*p_heat)  [lev][iv] *= hfactor;
        (*p_w_zout)[lev][iv] *= hfactor;
      }
      
      /* radiances   */
      for (iu=0; iu<input.rte.numu; iu++) {
	(*p_u0u) [lev][iu][iv] *= rfactor;
	
	for (j=0; j<input.rte.nphi; j++)
	  (*p_uu) [lev][j][iu][iv] *= rfactor;
      }

      /* 3D irradiances and radiances */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {
	    (*p_rfldir3d)   [lev][is][js][iv] *= ffactor;
	    (*p_rfldn3d)    [lev][is][js][iv] *= ffactor;
	    (*p_flup3d)     [lev][is][js][iv] *= ffactor;
	    (*p_uavgso3d)   [lev][is][js][iv] *= ffactor;
	    (*p_uavgdn3d)   [lev][is][js][iv] *= ffactor;
	    (*p_uavgup3d)   [lev][is][js][iv] *= ffactor;
	    (*p_radiance3d) [lev][is][js][iv] *= rfactor;
	  }
    }
  
    /* 3D absorption fields; here the loop goes over all 3D boxes */
    if (input.rte.mc.passback3D && input.rte.mc.absorption!=MCABS_NONE)
      
      for (ks=0; ks<Nz; ks++) 
	if (threed[ks])             /* only for 3D layers, BM07122005 */
	  for (is=0; is<Nx; is++) 
	    for (js=0; js<Ny; js++) 
	      (*p_abs3d) [ks][is][js][iv] *= ffactor;
  }

  return 0;
}


/***********************************************************************************/
/* Convert radiance to brightness temperature for a given filter function.         */ 
/***********************************************************************************/

float radiance2bt (float rad, float *wvnmlo, float *wvnmhi, float *filter, int n)
{
  int iv=0;
  float t1=0, t2=0, plkrad1=0, plkrad2=0, r=0;

  float accur = 1e-5;
  float dt=0.001;
  
  t1=273;
  t2 = t1+dt;

  if (rad<=0)
    return 0;

  plkrad1=0;
  plkrad2=0;
  for (iv=0; iv<n; iv++) {
    F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t1, &r);
    plkrad1 += filter[iv] * r;
    
    F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t2, &r);
    plkrad2 += filter[iv] * r;
  }

  while (fabs((plkrad1 - rad)/rad) > accur) {
    t1 = t1+(rad-plkrad1)/(plkrad2-plkrad1)*dt;
    t2 = t1+dt;

    plkrad1=0;
    plkrad2=0;
    for (iv=0; iv<n; iv++) {
      F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t1, &r);
      plkrad1 += filter[iv] * r;
      
      F77_FCN (cplkavg) (&(wvnmlo[iv]), &(wvnmhi[iv]), &t2, &r);
      plkrad2 += filter[iv] * r;
    }
  }

  return t1;
}


/***********************************************************************************/
/* Convert uvspec output to brightness temperatures; currently it is assumed that  */
/* the first element (iv==0) stores the spectrally integrated quantities;          */
/* only these will be considered.                                                  */
/***********************************************************************************/

static int output2bt (input_struct input, output_struct *output)
{
  int lev=0, iu=0, j=0, is=0;

  if (input.rte.solver == POLRADTRAN) { /* polRadtran output */
    for (lev=0; lev<input.atm.nzout; lev++) 
      for (is=0; is<input.rte.pol_nstokes; is++) {
	
	output->down_flux[lev][is][0] = 
	  radiance2bt (output->down_flux[lev][is][0] / PI,
		       output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	
	output->up_flux[lev][is][0] = 
	  radiance2bt (output->up_flux[lev][is][0] / PI,
		       output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	
	/* radiances */
	for (iu=0; iu<input.rte.nstr/2; iu++)
	  for (j=0; j<input.rte.nphi; j++) {
	    
	    output->down_rad [lev][j][iu][is][0] = 
	      radiance2bt (output->down_rad [lev][j][iu][is][0],
			   output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	    
	    output->up_rad [lev][j][iu][is][0] = 
	      radiance2bt (output->up_rad [lev][j][iu][is][0],
			   output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  }
      }
  }
  else { /* unpolarized output */
    
    for (lev=0; lev<input.atm.nzout; lev++) {
      
      /* irradiance / actinic flux */
      output->rfldir[lev][0] = 
	radiance2bt (output->rfldir[lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      
      output->rfldn [lev][0] = 
	radiance2bt (output->rfldn [lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      
      output->flup [lev][0] = 
	radiance2bt (output->flup [lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      
      output->uavg [lev][0] = 
	radiance2bt (output->uavg [lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      
      output->uavgso[lev][0] = 
	radiance2bt (output->uavgso[lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      
      output->uavgdn[lev][0] = 
	radiance2bt (output->uavgdn[lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      
      output->uavgup[lev][0] = 
	radiance2bt (output->uavgup[lev][0] / PI,
		     output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);      

      /* radiances */
      for (iu=0; iu<input.rte.numu; iu++) {
	output->u0u[lev][iu][0] = 
	  radiance2bt (output->u0u[lev][iu][0],
		       output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	
	for (j=0; j<input.rte.nphi; j++)
	  output->uu[lev][j][iu][0] = 
	    radiance2bt (output->uu[lev][j][iu][0],
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
      }
    }
  }
  return 0;
}


/***********************************************************************************/
/* Convert uvspec output to brightness temperatures; currently it is assumed that  */
/* the first element (iv==0) stores the spectrally integrated quantities;          */
/* only these will be considered.                                                  */
/***********************************************************************************/

static int output2bt3D (input_struct input, output_struct *output)
{
  int lev=0, is=0, js=0;

  for (lev=0; lev<input.atm.nzout; lev++) {
    
    
    /* 3D fields */
    if (input.rte.mc.passback3D)
      for (is=0; is<input.rte.mc.Nx_sample; is++)
	for (js=0; js<input.rte.mc.Ny_sample; js++) {
	  
	  output->rfldir3d[lev][is][js][0] = 
	    radiance2bt (output->rfldir3d[lev][is][js][0] / PI,
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  
	  output->rfldn3d[lev][is][js][0] = 
	    radiance2bt (output->rfldn3d[lev][is][js][0] / PI,
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  
	  output->flup3d[lev][is][js][0] = 
	    radiance2bt (output->flup3d[lev][is][js][0] / PI,
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  
	  output->uavgso3d[lev][is][js][0] = 
	    radiance2bt (output->uavgso3d[lev][is][js][0] / PI,
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  
	  output->uavgdn3d[lev][is][js][0] = 
	    radiance2bt (output->uavgdn3d[lev][is][js][0] / PI,
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  
	  output->uavgup3d[lev][is][js][0] = 
	    radiance2bt (output->uavgup3d[lev][is][js][0] / PI,
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	  
	  output->radiance3d [lev][is][js][0] = 
	    radiance2bt (output->radiance3d [lev][is][js][0],
			 output->wl.wvnmlo_h, output->wl.wvnmhi_h, output->wl.filter, output->wl.nlambda_h);
	}
  }
  
  /* conversion of 3D absorption to BT is probably useless, hence we don't do it */

  return 0;
}



static int read_photon_file (char *filename, float *lambda_t, int nlambda_t, float **fraction)
{
  float *wvl=NULL;
  int iv=0, n=0, status=0;
  
  status = read_2c_file_float (filename, &wvl, fraction, &n);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* compare wavelength grids */
  if (nlambda_t != n) {
    fprintf (stderr, "Error, number of wavelengths in %s differing from\n", filename);
    fprintf (stderr, "internal wavelength grid\n");
    return -1;
  }
  
  for (iv=0; iv<n; iv++)
    if (lambda_t[iv] != wvl[iv]) {
      fprintf (stderr, "Error, wavelengths in %s differ from internal wavelength grid: %f vs. %f\n", 
	       filename, lambda_t[iv], wvl[iv]);
      return -1;
    }
  
  return 0;
}


/******************************************************************************/
/* select the required wavelength range from an array of wavelengths (lambda) */
/******************************************************************************/

static int select_wavelength_indices (float *lambda, int nlambda,
				      float *lambda_lower, float *lambda_upper, 
				      int start_index, int end_index, 
				      int quiet,
				      int *lower, int *upper)
{
  int iv=0, tmp=0;

  if (start_index>0 && end_index>0) {
  /* if wavelength indices are defined */
    
    /* check if start_index and end_index sorted; if not, sort */
    if (start_index > end_index) {
      tmp=start_index;
      start_index = end_index;
      end_index = tmp;
    }

    /* check if we are out of range */
    if (end_index > nlambda) {
      fprintf (stderr, "Error, selected wavelength index %d is out of range \n", end_index);
      return -1;
    }

    /* the selected range is valid */
    *lower = start_index-1;
    *upper = end_index-1;
    
    *lambda_lower = lambda[*lower];
    *lambda_upper = lambda[*upper];
  } 
  else {
  /* if wavelength is specified search for suitable indices */

    while (lambda[iv] <= *lambda_lower)
      if (++iv==nlambda)
	break;
    
    if (iv>0)
      iv-=1;
	    
    *lower = iv;
    
    iv=0;
    while (lambda[iv] < *lambda_upper) {
    if (++iv==nlambda)
      break;
    }    
    
    if (iv==nlambda)
      iv-=1;
    
    *upper = iv;
  }

  return 0;
}


/**************************************************************************/
/*  average return the average density per layer in the array **y_average */
/*  **y_average will be automatically callocated                          */
/**************************************************************************/

int average_dens(float *dens, float *dens_air, float *zd, int nlev, int interpol_method, float **dens_avg, int allocate)
{
  int status=0;
  int n_layer;
  float *mix;    /* mixing ratio*/
  int lc;

  n_layer = nlev-1;

  if (allocate)
    if (((*dens_avg) = (float  *) calloc (n_layer, sizeof(float))) == NULL) {
      fprintf (stderr, "Error allocating memory for (*dens_avg) in average (ancillary.c)\n");
      return -1;
    }


  switch (interpol_method) {
  case INTERP_METHOD_SPLINE: 
    status = spline_average(zd, dens, nlev, dens_avg);
    break;
  case INTERP_METHOD_LINEAR:
    for (lc=0; lc<n_layer; lc++) (*dens_avg)[lc] = 0.5 * (dens[lc] + dens[lc+1]);   
    break;
  case INTERP_METHOD_LOG:
    for (lc=0; lc<n_layer; lc++) (*dens_avg)[lc] = log_average(dens[lc],dens[lc+1]);
    break;
  case INTERP_METHOD_LINMIX:                       /* linear mixing ratio integration                */
  case INTERP_METHOD_LOG_SPLINE:                   /* no log_spline intergration implemented jet !!! */ 
    mix = (float  *) calloc (nlev, sizeof(float));
    if (mix == NULL) {
      fprintf (stderr, "Error allocating memory for mixing_ratio in average (ancillary.c)\n");
      return -1;
    }
    for (lc=0; lc<nlev; ++lc) mix[lc] = dens[lc]/dens_air[lc]; /* calculating to mixing ratio */
    for (lc=0; lc<n_layer; lc++)
       (*dens_avg)[lc] = linmix_average(mix[lc],mix[lc+1],dens_air[lc],dens_air[lc+1]);
    free(mix);
    break;
  default:
    fprintf (stderr, "Error, unknown interpolation method input.");
    return -1;  
  }

  if (status!=0) {
    fprintf (stderr, "Error %d calculating average concentration (average, in ancillary.c)\n", status);
    return status;
  }

  return status;
}

/****************************************************************/
/*  log_average returns the logarithmic average density between */
/*  two layers with d[ens]1 und d[ens]2                         */
/*  assuming logarithmic variation with hight                   */
/****************************************************************/

float log_average(float d1,float d2)
/* small function to calculate the average density */
/*        */
{
  float avg = 0;
  float test1 = -666., test2 = -666.0;

  test1 = (d1 < d2 ? d1 : d2);
  test2 = fabs(d2 - d1);

  if (test1<=0 || test2 <= 0.001 * d1)    /* in this case numerically unstable => */
    avg = 0.5 * (d1 + d2);                /* use linear interpolation instead     */
  else 
    avg = (d2-d1) / log(d2/d1);
 
  return avg; 
}

/*******************************************************************/
/*  linmix_average returns the linear mixing ratio average density */
/*  between two layers with m[ixing ratio]1 and m2 and             */
/*  air number density n1 and n2                                   */
/*  assuming linear variation of the mix ratio with hight          */
/*  and logarithmic variation of the air number dens with hight    */
/*******************************************************************/

float linmix_average(float m1,float m2,float n1,float n2)
/* small function to calculate the average density                   */
/* assuming lineary varying mixing ratio and logarithmic air density */
{
  float avg = 0;
  float test1 = -666., test2 = -666.0;
  float log_n = 0;      

  test1 = (n1 < n2 ? n1 : n2);
  test2 = fabs (n1 - n2);


  if (test1 <= 0 || test2 <= 0.001*n1)      /* in this case numerically unstable => */
    avg =  0.5 * (m1*n1 + m2*n2);           /* use linear interpolation instead     */
  else {
    log_n = log(n2/n1);
    avg =  1/(log_n*log_n) * ((n2*m2-n1*m1)*log_n - (m2-m1)*(n2-n1));
  }

  return avg;  
}

/************************************************************/
/* small function to calculate the average density          */
/* assuming cubic spline variation                          */
/* returns an array instead of number as previous functions */
/************************************************************/

int spline_average(float *x, float *y, int n, float **y_average)

{
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  int m;
  int i;
  double dx, dx2, dx3;
  int status;
  int descend=0;
  double *x_sort=NULL, *y_sort=NULL;
  double *y_average_d;

  x_sort = (double *) calloc (n, sizeof(double));
  y_sort = (double *) calloc (n, sizeof(double));

  if (x[1] <= x[0]) 
    descend = 1;

  /* number of layers = number of level - 1 */
  m = n-1;

  if (!descend) {
    for (i=0; i<n; i++) {
      x_sort[i] = (double) x[i];
      y_sort[i] = (double) y[i];
    }
  }
  else {
    for (i=0; i<n; i++) {
      x_sort[i] = (double) x[n-1-i];
      y_sort[i] = (double) y[n-1-i];
    }
  }

  status = spline_coeffc (x_sort, y_sort, n, &a0, &a1, &a2, &a3);
  if (status!=0)  {
    fprintf (stderr, "spline_average: sorry cannot do spline interpolation\n");
    fprintf (stderr, "spline_coeffc() returned status %d\n", status);
    return status;
  }

  y_average_d  = (double *) calloc (m, sizeof(double));

  for (i=0; i<m; i++) {
    dx  = (double) (x_sort[i+1] - x_sort[i]);
    dx2 = (double) (dx * dx);
    dx3 = (double) (dx * dx2);
    y_average_d[i] =   ((double)  0.25) * a3[i]*dx3 
                     + ((double) 1./3.) * a2[i]*dx2 
                     + ((double)  0.5 ) * a1[i]*dx 
                     +                    a0[i];
  }

  if (!descend)
    for (i=0; i<m; i++)
      (*y_average)[i] = (float) y_average_d[i];
  else
    for (i=0; i<m; i++)
      (*y_average)[i] = (float) y_average_d[m-1-i];

  return 0;
}
