/*--------------------------------------------------------------------
 * $Id: atmosphere.c,v 1.85 2006/12/17 23:45:52 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>

#include "uvspec.h"
#include "ckdfu.h"
#include "cloud.h"
#include "ascii.h"

#include "fortran_and_c.h"
#include "f77-uscore.h"

#define EPSILON 1E-6

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif

float F77_FCN (dewpt) (float *wvpress);

float F77_FCN (ept) (float *t, float *td, float *p);


/************************************/
/* prototypes of internal functions */
/************************************/

static int get_surface_elevation_from_map (float latitude, float longitude, alt_out_struct *alt, 
                                           char *filename, int verbose, int quiet);

static int read_atmosphere (char *filename, atm_out_struct *out, 
                            int interpol_method_press, int interpol_method_temper, int *interp_method_gas,
                            int cdf, int quiet, int verbose);

/* static int read_density (char *filename, */
/* 			 atm_out_struct *atm_out, int dens_id, int *interpol_method, */
/* 			 int quiet, int verbose);  */

static int read_radiosonde_file (input_struct input, output_struct *output, double boltzmann);

static int read_and_combine_density (char *filename, atm_out_struct *out,
				     int dens_id, int *interpol_method_gas, int interpol_method_temper, 
                                     int unit_profile, float *mol_mass,
				     float *dens_air_all, float *temper_all, float *z_all, int nlev_all,
                                     float *dens_air_atm, float *z_atm, int nlev_atm,
				     int solver, int quiet, int verbose);

static int scale_density_profile (float **dens, float **dens_avg, int alloc_avg, int i,
                          float input_column, float *output_column, 
                          float *scale_factor, int interpol_method_gas, 
                          input_struct input, output_struct *output);


static int convert_profile_units    (float **ptr_dens, float *z_dens, int n, int dens_id,   
                                     float *dens_air_atm, float *temper_atm, float *z_atm, int n_atm,
                                     int interpol_method_air, int interpol_method_temper,
                                     int unit_profile, float *mol_mass, int quiet);

static int calculate_or_read_refind (char *filename,
				     float **refind, float *z_all, int nlev_all,
				     int interpol_method_refind, 
                                     float *t, float *p,
                                     int quiet);

static int combine_profiles (float **dens, float *z, int nlev, int sza,
                             float *dens_all, float *dens_air_all, float *z_all, int nlev_all, 
                             int interpol_method_dens, int interpol_method_air,
                             float ***dens_comb, float **z_comb, int *nlev_comb, int quiet);

static int read_molecular_absorption (char *filename, 
				      float **zd, int *nlyr,
				      int quiet,
				      double **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic);

static int read_molecular_absorption_z (char* filename, int quiet,
					float **zd, int *nlyr);

static char* interpol_number2string(int method_number);


/**************************************************************/
/* Setup model atmosphere.                                    */
/**************************************************************/

int setup_altitude (input_struct input, output_struct *output)
{
  int   status=0;
  int   cdf=0;

  char  filename[FILENAME_MAX];
  double  boltzmann = 1.38065e-23;

  /* where to get the atmosphere from */
  if (input.ck_scheme == CK_GENERIC) {
    strcpy (filename, input.filename[FN_CK_GENERIC]);
    cdf=1;

    if (!input.quiet)
      if (strlen(input.filename[FN_ATMOSPHERE]) > 0)
	fprintf (stderr, " ... ignoring %s, reading atmosphere from %s\n",
		 input.filename[FN_ATMOSPHERE], input.filename[FN_CK_GENERIC]);
  }
  else {
    strcpy (filename, input.filename[FN_ATMOSPHERE]);
    cdf=0;
  }

  /* read atmosphere file */
  status = read_atmosphere (filename, &(output->atm), 
                            input.atm.interpol_method_press, 
                            input.atm.interpol_method_temper, 
                            input.atm.interpol_method_gas,
                            cdf, input.quiet, input.verbose);

  if (status!=0) {
    fprintf (stderr, "Error %d reading atmosphere from %s\n",
	     status, filename);
    return status;
  }

  /* read radiosonde */
  if (strlen(input.atm.rs_filename) > 0)
    status = read_radiosonde_file(input, output, boltzmann);

  if (status!=0) {
    fprintf (stderr, "Error %d reading radiosonde data from %s\n",
	     status, input.atm.rs_filename);
    return status;
  }


  /* determine altitude */
  switch (input.alt.source) {
  case ALT_NOT_DEFINED:
    output->alt.altitude = output->atm.zd[output->atm.nlev-1];
    output->alt.source = ALT_FROM_ATM;
    break;
  case ALT_DIRECT_INPUT:
    output->alt.altitude = input.alt.altitude;
    output->alt.source   = input.alt.source;
    break;
  case ALT_FROM_MAP:

    /* read surface elevation from map */
    status = get_surface_elevation_from_map (input.latitude, input.longitude, &(output->alt), 
                                             input.filename[FN_ALTITUDE_MAP], input.verbose, input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d reading surface elevation map in setup_altitude (atmosphere.c) \n", status);
      return status;
    }

    output->alt.source = input.alt.source;
    break;
  default:
    fprintf (stderr, "\nError, determining surface altitude! \n");
    return -1;
  }

  if (output->alt.altitude < output->atm.zd[output->atm.nlev-1] || output->alt.altitude > output->atm.zd[0]) {
    fprintf (stderr, "Error -5, altitude = %7.3f km is outside \n", output->alt.altitude);
    fprintf (stderr, "          of the range of the atmosphere file = [%8.3f km,%8.3f km] \n", 
                      output->atm.zd[output->atm.nlev-1], output->atm.zd[0]);
    return -5;
  }

  return status;
}


static int get_surface_elevation_from_map (float lat, float lon, alt_out_struct *alt, char *filename, int verbose, int quiet)
{

  int status = 0;

#if HAVE_LIBNETCDF

  int ncid=0;
  int    idd_nlat, idd_nlon=0, id_data=0;
  size_t     nlat=0,   nlon=0;

  float i, j;
  size_t start[2] = {0,0};
  int altitude=0;

  if (verbose) {
    fprintf (stderr, " ... search surface elevation in map: \n");
    fprintf (stderr, "     %s \n", filename);
    fprintf (stderr, "     map size = [%8.3f,%8.3f] x [%8.3f,%8.3f]\n", 
                          alt->map_lat_min/3600, alt->map_lat_max/3600, 
                          alt->map_lon_min/3600, alt->map_lon_max/3600);
  }

  /* map reference system:                      */
  /* southpole = 0 -> northpole = 180,          */
  /*   shift of all lat not nessesary, as we need differences 
       (that would be [-90,90]-> lat+90-> [0,180]) */
  /* Greenwich = 0->(eastward)->360 Greenwich */
  /*   it is taken account for this in uvspec.c (altitude checks) */

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
    return status;
  }

  /* get dimension id for "nlat" */
  status = nc_inq_dimid (ncid, "nlat", &idd_nlat);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nlat" */
  status = nc_inq_dimlen (ncid, idd_nlat, &nlat);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* get dimension id for "nlon" */
  status = nc_inq_dimid (ncid, "nlon", &idd_nlon);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  /* get dimension length for "nlon" */
  status = nc_inq_dimlen (ncid, idd_nlon, &nlon);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  /* get variable id for "elev" (elevation) */
  status = nc_inq_varid (ncid, "elev", &id_data);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d, while getting ncid from %s\n", status, filename);
    return status;
  }

  if (verbose)
    fprintf (stderr, "     found %d x %d data points\n", nlat, nlon);

  i = (lat - alt->map_lat_min) / (alt->map_lat_max - alt->map_lat_min) * nlat;
  j = (lon - alt->map_lon_min) / (alt->map_lon_max - alt->map_lon_min) * nlon;

  /* rounding (float -> int) */
  /* latitude, index in y-direction, from south to north */ 
  start[0] =  i;
  /* correction integer rounding at the northern boundary of the map */
  if (start[0]==nlat) start[0] = nlat-1;

  /* locations on the boundaries are treated like the northward pixel */
  /* index in x-direction, from Greenwich to east */
  start[1] =  j;
  /* locations on the boundaries are treated like the eastward pixel */
  /* correction integer rounding at the eastern boundary of the map */
  if (start[1]==nlon) start[1] = nlon-1;  

  if (verbose) {
    fprintf (stderr, "     look up surface elevation at: lat = %f, lon = %f, ", lat/3600, lon/3600);
    fprintf (stderr, " element: %4d x %4d \n", start[0]+1,start[1]+1);
  }

  status = nc_get_var1_int (ncid, id_data, start, &altitude);

  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  alt->altitude = altitude/1000.0;  /* m -> km*/
  if (!quiet)
    fprintf (stderr, " ... surface elevation is: %f km\n", alt->altitude);

#else
  fprintf (stderr, " ******************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot       *\n");
  fprintf (stderr, " * use the altitude_map option. Please get netcdf and rebuild.    *\n");
  fprintf (stderr, " ******************************************************************\n");
  return -1;
#endif

  return status;
}

/**************************************************************/
/* Setup model atmosphere.                                    */
/**************************************************************/

int setup_atmosphere (input_struct input, output_struct *output)
{
  int i=0, status=0, lc=0, iv=0, iz=0, ialt=0, ipa=0, iscale=0, is=0, i_gas=0;
  float   scale_factor=0;
  float  *zuser     = NULL;
  float  *zd_user   = NULL;
  int     nlyr_user = -999;
  int     nuser     = -999;

  char   *method    = NULL;

  float  *z_all     = NULL;
  int     nlev_all  = -999;

  float  *tmp_lambda=NULL, *tmp_dtauf=NULL, *tmp_dtauf_t=NULL;
  float **tmp_dtau=NULL;
  int     tmp_nlambda=0;

  float   tmpdens=0, idealdiff=0;
  int     idealgas=0, firstnonideal=0;
  
  double  boltzmann = 1.38065e-23;
  float   kappa = 0.0;

  int     monochromatic=0;

  int     nlev_atm=-999;
  float  *dens_air_atm=NULL;
  float  *z_atm=NULL;
  char   *gas;
  float   column_h2o_in_clouds=0.0;
  int    *water_cloud_indices=NULL;
  int    *ice_cloud_indices=NULL;
  float   press_h2o = -999;
  float   tmp_temper = -999;
  float   tmp_temper_d = -999;
  
  /* allocate memory for layer average densities */
  output->atm.microphys.dens_avg = calloc(MOL_NN, sizeof(float *));
  if (output->atm.microphys.dens_avg == NULL) {
    fprintf (stderr, "Error allocating memory for dens_avg\n");
    return -1;
  }


  if (input.verbose) {
    fprintf (stderr, "*** Interpolation methods \n");
    /* write the interpolate into the verbose file */ 
    method = interpol_number2string(input.atm.interpol_method_press);
    fprintf(stderr, " ... z_interpolate pressure          %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_temper);
    fprintf(stderr, " ... z_interpolate temperature       %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_refind);
    fprintf(stderr, " ... z_interpolate refraction index  %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[0]);
    fprintf(stderr, " ... z_interpolate air  density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[1]);
    fprintf(stderr, " ... z_interpolate O3   density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[2]);
    fprintf(stderr, " ... z_interpolate O2   density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[3]);
    fprintf(stderr, " ... z_interpolate H2O  density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[4]);
    fprintf(stderr, " ... z_interpolate CO2  density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[5]);
    fprintf(stderr, " ... z_interpolate NO2  density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[6]);
    fprintf(stderr, " ... z_interpolate BRO  density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[7]);
    fprintf(stderr, " ... z_interpolate OCLO density      %s interpolation \n", method);
    method = interpol_number2string(input.atm.interpol_method_gas[8]);
    fprintf(stderr, " ... z_interpolate HCHO density      %s interpolation \n", method);
  }

  free(method);


  /*******************************************/
  /* read additional atmospheric input files */
  /*******************************************/

  /* just for short writing */
  nlev_atm = output->atm.nlev;

  /* copy the z-grid of the atmosphere file */
  z_atm = (float *) calloc (nlev_atm, sizeof(float));
  for (lc=0; lc<output->atm.nlev; lc++)
    z_atm[lc] = output->atm.zd[lc];

  /* copy the air number density of the atmosphere file for linmix interpolations*/
  dens_air_atm =    calloc (nlev_atm, sizeof(float));
  for (lc=0; lc<output->atm.nlev; lc++)
    dens_air_atm[lc] = output->atm.microphys.dens[MOL_AIR][lc];
   
  /* initialisation of the common z-grid: z_all */
  nlev_all = 0; 
  set_common_z_grid (output->atm.zd, nlev_atm, 
		     &(z_all), &(nlev_all));


  /********************************************************************************/
  /* combine all levels from other additional input files to a common grid: z_all */ 
  /********************************************************************************/

  /* dens                  */
  /* refind                */
  /* rayscat               */
  /* molabs                */
  /* zout                  */
  /* altitude 2nd argument */
  /* altitude 1st argument */

  /* densities (including rh-file) */
  for (i=0; i<MOL_NN; i++)
    if (strlen(input.atm.filename[i]) > 0)
      status += add_file2grid (&(z_all), &(nlev_all),
			       input.atm.filename[i]);


  /* refractive index */
  if (strlen(input.filename[FN_REFIND]) > 0)
    status += add_file2grid (&(z_all), &(nlev_all), 
			     input.filename[FN_REFIND]);


  /* rayleigh scattering cross section profile, user-defined */
  if (strlen(input.filename[FN_RAYLEIGH]) > 0) {
    status = read_molecular_absorption_z (input.filename[FN_RAYLEIGH], input.quiet,
					  &zd_user, &nlyr_user);
    if (status!=0) {
      fprintf (stderr, "Error reading z levels from user-defined Rayleigh scattering file %s\n",
	       input.filename[FN_RAYLEIGH]);
      fprintf (stderr, "Please check format and retry!\n");
      return -1;
    }

    /* add new levels to the common atmospheric profile */
    set_common_z_grid (zd_user, nlyr_user+1, &z_all, &nlev_all);

    /* free memory */
    free (zd_user);
  }


  /* molecular absorption cross section profile, user-defined  */
  if (strlen(input.filename[FN_MOLABS]) > 0) {
    status = read_molecular_absorption_z (input.filename[FN_MOLABS], input.quiet,
					  &zd_user, &nlyr_user);
    if (status!=0) {
      fprintf (stderr, "Error reading z levels from user-defined molecular absorption file %s\n", 
	       input.filename[FN_MOLABS]);
      fprintf (stderr, "Please check format and retry!\n");
      return -1;
    }

    /* add new levels to the common atmospheric profile */
    set_common_z_grid (zd_user, nlyr_user+1, &z_all, &nlev_all);

    /* free memory */
    free (zd_user);
  }

  /* POLRADTRAN requires zout levels as atmosphere levels */
  if (input.rte.solver == POLRADTRAN && input.atm.zout_interpolate == NO_ZOUT_INTERPOLATE) {
    input.atm.zout_interpolate = ZOUT_INTERPOLATE;
    if (!input.quiet)
      fprintf (stderr, " ... zout_interpolate is turned on by default for rte_solver POLRADTRAN\n");
  }


  /* add user-defined (ground) altitude   */
  /* as level to the common z-grid        */
  zuser = (float *) calloc (1, sizeof(float));
  zuser[0] = output->alt.altitude;
  set_common_z_grid (zuser, 1, &(z_all), &nlev_all);
  free (zuser);


  /* user-defined equidistant altitude grid */
  if (input.atm.zout_interpolate == ZOUT_INTERPOLATE) {
    if (input.alt.altitude_dz > 0) {
      nuser = (int) ((z_all[0] - z_all[nlev_all-1]) / 
	               input.alt.altitude_dz          + 0.5) + 1;
    
      zuser = (float *) calloc (nuser, sizeof(float));

      for (i=0; i<nuser; i++) 
        zuser[i] = z_all[nlev_all-1] + input.alt.altitude_dz * (float) i;
    
      while (zuser[nuser-1] > z_all[0])
        nuser--;

      /* add user-defined grid to atmospheric grid */
      set_common_z_grid (zuser, nuser, &z_all, &nlev_all);

      free (zuser);
    }
  }
  /* else if (input.atm.zout_interpolate == NO_ZOUT_INTERPOLATE) */
  /* altitude second argument heights are added to the z-grid    */
  /* before setup_redistribution                                 */


  /* allocating and calculating output-levels */
  /* relative to sea level and relative to surface level */
  output->atm.nzout    = input.atm.nzout;
  output->atm.zout_sur = calloc (output->atm.nzout, sizeof (float));
  output->atm.zout_sea = calloc (output->atm.nzout, sizeof (float));
  

  /* initialisation of zout_surface und zout_sealevel */
  for (iz=0;iz<output->atm.nzout;iz++) {

    if (input.atm.zout_source == ZOUT_ABOVE_SUR) {
      output->atm.zout_sur[iz] = input.atm.zout_sur[iz];     
      if      (input.atm.zout_sur[iz] == ZOUT_MYSTIC)    /* marker for 2D-SURFACE IN MYSTIC */
        output->atm.zout_sea[iz] = ZOUT_MYSTIC;
      else if (input.atm.zout_sur[iz] == ZOUT_SURFACE) { /* marker for 2D-SURFACE IN MYSTIC */
        output->atm.zout_sea[iz] = output->alt.altitude;
        output->atm.zout_sur[iz] = 0;
      }
      else if (input.atm.zout_sur[iz] == ZOUT_TOA) {   /* marker for TOP_OF_ATMOSPHERE */
        output->atm.zout_sea[iz] = output->atm.zd[0];
        output->atm.zout_sur[iz] = output->atm.zd[0] - output->alt.altitude;
      }
      else { /* normal case */
        output->atm.zout_sea[iz] = input.atm.zout_sur[iz] + output->alt.altitude;
      }
    }
    else if (input.atm.zout_source == ZOUT_ABOVE_SEA) {
      output->atm.zout_sea[iz] = input.atm.zout_sea[iz];     
      if      (input.atm.zout_sea[iz] == ZOUT_MYSTIC)    /* marker for MYSTIC */
        output->atm.zout_sur[iz] = ZOUT_MYSTIC;
      else if (input.atm.zout_sea[iz] == ZOUT_SURFACE) { /* marker for SURFACE_ELEVATION */
        output->atm.zout_sea[iz] = output->alt.altitude;
        output->atm.zout_sur[iz] = 0;
      }
      else if (input.atm.zout_sea[iz] == ZOUT_TOA) {     /* marker for TOP_OF_ATMOSPHERE */
        output->atm.zout_sea[iz] = output->atm.zd[0];
        output->atm.zout_sur[iz] = output->atm.zd[0] - output->alt.altitude;
      }
      else { /* normal case */
        output->atm.zout_sur[iz] = input.atm.zout_sea[iz] - output->alt.altitude;
      }
    }
    else {
      fprintf (stderr, "Error determening output levels zout.\n");
      return -1;
    }
    /* check if values are above bottom */
    if (output->atm.zout_sea[iz] < output->alt.altitude && output->atm.zout_sea[iz] != ZOUT_MYSTIC) {
      fprintf (stderr, "Error zout[%d] = %f km is below the ground = %f km\n",
                        iz, output->atm.zout_sea[iz],output->alt.altitude);
      return -2;
    }
    /* check if values are above atmosphere */
    if (output->atm.zout_sea[iz] > output->atm.zd[0]) {
      fprintf (stderr, "Error zout[%d] = %f km is above top of atmosphere = %f km\n", 
                        iz, output->atm.zout_sea[iz],output->atm.zd[0]);
      return -3;
    }
  } /* for (iz=0; iz<output->atm.nzout; iz++) { ... */


  /* add new levels to the common atmospheric profile */
  /* before interpolation is done */
  if (input.atm.zout_interpolate == ZOUT_INTERPOLATE) {
    for (iz=0; iz<output->atm.nzout; iz++) {
      if (output->atm.zout_sur[iz] != ZOUT_MYSTIC) {
        set_common_z_grid (&(output->atm.zout_sea[iz]), 1, &(z_all), &(nlev_all));
      }
    }
  }
  /* else if (input.atm.zout_interpolate == NO_ZOUT_INTERPOLATE) */
  /* zout level are not added to any z-grid */
  /* when input.atm.zout_interpolate == NO_ZOUT_INTERPOLATE */
  /* solver get the zout levels separat and evaluate there by themselfes*/

  if (status!=0) {
    fprintf (stderr, "Error creating common atmospheric grid\n");
    return status;
  }

/*   for (lc=0; lc<nlev_all; lc++) */
/*    fprintf (stderr, " z[%d] = %f\n", lc, z_all[lc]); */

  /************************************************************************/
  /* READING ADDITIONAL INPUT FILES                                       */ 
  /* INTERPOLATION TO THE COMMON Z-GRID: z_all                            */
  /* if input only covers a part of the vertical range of atmosphere_file */
  /* information of the atmosphere_file is assumed in the regions,        */
  /* which are not covered by the additional input file                   */
  /************************************************************************/

  /* cloud saturation */
  /* pressure    */
  /* temperature */
  /* dens (dens_files might also be dens_tab files) */
  /* water vapour saturation inside clouds */
  /* refind      */

  /* Pressure */ 

  status += interpolate_profile (z_atm, &(output->atm.microphys.press), output->atm.nlev, 
                                 z_all, nlev_all, input.atm.interpol_method_press, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating pressure (setup_atmosphere, in atmosphere.c)\n", status);
    return status;
  }

  /* Temperature */
  status += interpolate_profile (z_atm, &(output->atm.microphys.temper), output->atm.nlev, 
                                 z_all, nlev_all, input.atm.interpol_method_temper, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating temperature (setup_atmosphere, in atmosphere.c)\n", status);
    return status;
  }

  /*   /\* this is not 100% consistent with interpolation of p and T *\/  */
  /* Air */
  status += interpolate_profile (z_atm, &(output->atm.microphys.dens[MOL_AIR]), output->atm.nlev, 
                                 z_all, nlev_all, input.atm.interpol_method_gas[MOL_AIR], input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating dens_air (setup_atmosphere, in atmosphere.c)\n", status);
    return status;
  }

  /*   /\* determine air number density from pressure and temperature *\/ */
  /*   free(output->atm.microphys.dens[MOL_AIR]); */
  /*   output->atm.microphys.dens[MOL_AIR] = calloc (nlev_all, sizeof(float)); */
  /*   for (lc=0; lc<nlev_all; lc++) */
  /*     output->atm.microphys.dens[MOL_AIR][lc] =  output->atm.microphys.press[lc] /  */
  /*                                                (boltzmann * output->atm.microphys.temper[lc]) * 100.0 / 1e6; */

  /* calculate layer average temperature  */ 
  /* this is here, as average_dens() needs a suitable dens_air */
  status = average_dens(output->atm.microphys.temper, output->atm.microphys.dens[MOL_AIR], 
                        z_all, nlev_all,
                        input.atm.interpol_method_temper, 
                        &(output->atm.microphys.temper_avg), YES);

  if (status!=0) {
    fprintf (stderr, "Error %d calculating average temperature (setup_atmosphere, in atmosphere.c)\n", status);
    return status;
  }


  /*****************************************/
  /* density profiles (including rh-file)  */
  /*****************************************/

  for (i=0; i<MOL_NN; i++) {
      
    if (i != MOL_AIR) {
      /* call it in any case as densities are also interpolated onto the z_all grid */ 
      /* it also calculates the otherwise empty rh profile */
      /* even if it doesn't read rh from the file          */
      status = read_and_combine_density (input.atm.filename[i], &output->atm, 
                                         i, input.atm.interpol_method_gas, input.atm.interpol_method_temper, 
                                         input.atm.unit_profile[i], input.atm.mol_mass,
					 output->atm.microphys.dens[MOL_AIR], output->atm.microphys.temper, z_all, nlev_all,
                                         dens_air_atm, z_atm, nlev_atm,
					 input.rte.solver, input.quiet, input.verbose);
      if (status !=0) {
	fprintf (stderr, "Error %d reading density profile\n", status);
        return status;
      }
    }
  }

  /* allocate memory for scale factors */
  output->scale_factor = calloc(MOL_NN, sizeof(float));
  if (output->scale_factor == NULL) {
    fprintf (stderr, "Error allocating memory for dens_avg\n");
    return -1;
  }
 
  /* allocate memory for vertical columns */  
  output->column = calloc(MOL_NN, sizeof(float));
  if (output->column == NULL) {
    fprintf (stderr, "Error allocating memory for dens_avg\n");
    return -1;
  }

  if (output->atm.microphys.nsza_denstab > 0) {
    /* allocate memory for scale factors */

    status  = ASCII_calloc_float (&output->scale_factor_denstab, MOL_NN, output->atm.microphys.nsza_denstab);
    if (status != 0) {
      fprintf (stderr, "Error allocating memory for scale_factor_denstab in setup_atmosphere (atmosphere.c) \n");
      return -1;
    }
 
    /* allocate memory for vertical columns */  
    status  = ASCII_calloc_float (&output->column_denstab, MOL_NN, output->atm.microphys.nsza_denstab);
    if (status != 0) {
      fprintf (stderr, "Error allocating memory for column_denstab in setup_atmosphere (atmosphere.c) \n");
      return -1;
    }
  }


  /*****************************************/
  /* water vapour saturation inside clouds */
  /*****************************************/

  /* calculate h2o column */
  if ((output->atm.microphys.dens_avg[MOL_H2O] = calloc(nlev_all, sizeof(float)))==NULL) {
    fprintf (stderr, "Error: Allocation of dens_avg[MOL_H2O] in setup_atmosphere() \n");
    return -1;
  }
  status = column (output->atm.microphys.dens[MOL_H2O], output->atm.zd, output->atm.nlev, output->alt.altitude,
                   input.atm.interpol_method_gas[MOL_H2O], output->atm.microphys.dens[MOL_AIR],
                   &(output->atm.microphys.dens_avg[MOL_H2O]), NO, &(output->column[MOL_H2O]));

  /* inside ice clouds */
  if (input.ic.wv_saturation == SATURATION_CLOUD)
    for (ipa=0; ipa<output->nipa; ipa++)
      saturate_wv_inside_iceclouds(input, output, ipa, 0); /* (in ancillary.c) */

  /* inside water clouds */
  if (input.wc.wv_saturation == SATURATION_CLOUD)
    for (ipa=0; ipa<output->nipa; ipa++)
      saturate_wv_inside_waterclouds(input, output, ipa, 0); /* (in ancillary.c) */

  free(output->atm.microphys.dens_avg[MOL_H2O]);
  /* this is alloacted again, when dens columns are scaled */


  /*********************/
  /* refraction index  */
  /*********************/

  /* calculate_or_read_refind() needs to be called in any case */
  /* because it calculates the otherwise empty refind profile  */
  /* even if it doesn't read refind from the file              */ 

  status = calculate_or_read_refind (input.filename[FN_REFIND], &(output->atm.microphys.refind),
                                     z_all, nlev_all,
                                     input.atm.interpol_method_refind, 
                                     output->atm.microphys.temper,
	                             output->atm.microphys.press,
                                     input.quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d returned by calculate_or_read_refind() in atmosphere.c\n", status);
    return status;
  }


  /***********************************************/
  /*   copy altitude grid to final destination   */
  /***********************************************/

  free (output->atm.zd);
  output->atm.zd = calloc (nlev_all, sizeof(float));

  for (lc=0; lc<nlev_all; lc++)
    output->atm.zd[lc] = z_all[lc];

  output->atm.nlev = nlev_all;
  output->atm.nlyr = nlev_all-1; 

  free (z_all); 

  /*****************************************************************************************/
  /*                               Scaling profiles                                        */
  /*****************************************************************************************/ 

  if (input.verbose)
    fprintf (stderr, "*** Scaling profiles \n");

  /* Test if surface altitude is really in the zd-grid */
  for (ialt=0; ialt<output->atm.nlev; ialt++) 
    if (output->alt.altitude == output->atm.zd[ialt]) 
      break;
    
  if (ialt == output->atm.nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid.\n", 
	     output->alt.altitude);
    return -1;
  }

  /* allocate memory for layer average densities */
  for (i=0; i<MOL_NN; i++) {
    output->atm.microphys.dens_avg[i] = calloc(output->atm.nlyr, sizeof(float));
    if (output->atm.microphys.dens_avg[i] == NULL) {
      fprintf (stderr, "Error allocating memory for dens_avg[i]\n");
      return -1;
    }
  }

  /*****************************************************************************************/
  /*                      Scaling profiles with COLUMN VALUEs                              */
  /*                and calculating average densities in the layers                        */
  /*****************************************************************************************/ 

  /* BrO, HCHO, OCLO are not included in normal calculations */
  /* but might be added by dens- or denstab files            */
  
  for (i=0; i<MOL_NN; i++) {
  
    switch (i) {
    /************************************/
    /* Scale pressure (and air density) */
    /************************************/
    case(MOL_AIR):
      if (input.pressure >= 0.0) {
      
        scale_factor = input.pressure / output->atm.microphys.press[ialt];
        for (lc=0; lc<output->atm.nlev; lc++) {
          output->atm.microphys.press[lc]         *= scale_factor;
          output->atm.microphys.dens[MOL_AIR][lc] *= scale_factor;
          output->atm.microphys.dens[MOL_O2][lc]  *= scale_factor;
          output->atm.microphys.dens[MOL_CO2][lc] *= scale_factor;
        } 
    
        if (input.verbose) {
          fprintf (stderr, " ... scaling pressure    from     %7.2f hPa   to     %7.2f hPa   at %7.3f km\n", 
	          output->atm.microphys.press[ialt]/scale_factor, output->atm.microphys.press[ialt],output->atm.zd[ialt]);
          fprintf (stderr, " ... scaling air density from %11.5g cm-3  to %11.5g cm-3  at %7.3f km\n", 
	           output->atm.microphys.dens[MOL_AIR][ialt]/scale_factor, output->atm.microphys.dens[MOL_AIR][ialt],
                   output->atm.zd[ialt]);
        }
      }
      status = column (output->atm.microphys.dens[i], output->atm.zd, output->atm.nlev, output->alt.altitude, 
                       input.atm.interpol_method_gas[i], output->atm.microphys.dens[MOL_AIR],
                       &(output->atm.microphys.dens_avg[i]), NO, &(output->column[i]));
      break;

    /********************/
    /* Scale gas column */
    /********************/
    case(MOL_O3):
    case(MOL_O2):
    /* H20 is a special case, see below */
    case(MOL_CO2):
    case(MOL_NO2):
    case(MOL_BRO):
    case(MOL_OCLO):
    case(MOL_HCHO):
    
      status = scale_density_profile (&(output->atm.microphys.dens[i]), 
                      &(output->atm.microphys.dens_avg[i]), YES, i,
                      input.atm.column[i], &(output->column[i]),
                      &(output->scale_factor[i]), input.atm.interpol_method_gas[i],
                      input, output);

      if (status!=0) {
        fprintf (stderr, "Error %d returned by scale_density_profile()\n", status);
        fprintf (stderr, "      while scaling dens in setup_atmosphere() (atmosphere.c)\n");
        return status;
      }

      /* air mass factor profiles */
      if (i == output->atm.microphys.denstab_id) {
          
        for (is=0; is<output->atm.microphys.nsza_denstab; is++) {

          status = scale_density_profile (&(output->atm.microphys.denstab[i][is]), 
                          &(output->atm.microphys.denstab_avg[i][is]), NO, i,
                          input.atm.column[i], &(output->column_denstab[i][is]),
                          &(output->scale_factor_denstab[i][is]), input.atm.interpol_method_gas[i],
                          input, output);

          if (status!=0) {
            fprintf (stderr, "Error %d returned by scale_density_profile()\n", status);
            fprintf (stderr, "      while scaling denstab in setup_atmosphere() (atmosphere.c)\n");
            return status;
          }

          for (lc=0; lc<output->atm.nlev; lc++)
               output->atm.microphys.denstab_amf[is][lc] *= output->scale_factor_denstab[i][is];

        } /* for (is=0; is<output->atm.microphys.nsza_denstab; is++) { */
      }

      break;

    case MOL_H2O:
      /****************************/
      /* Scale precipitable water */
      /****************************/
      gas = gas_number2string(i);

      status = column (output->atm.microphys.dens[i], output->atm.zd, output->atm.nlev, output->alt.altitude,
                       input.atm.interpol_method_gas[i], output->atm.microphys.dens[MOL_AIR],
                       &(output->atm.microphys.dens_avg[i]),YES,&(output->column[i]));

      if (status!=0) {
        fprintf (stderr, "\nError %d calculating %s column!\n", status, gas);
        return status;
      }

      output->precip_water = output->column[i] * DU2particle * (M_H2O / N_MOL_CKDFU * 10.);
                             /*     DU         *  DU->cm-2   *       cm-2->kg/m2         */

      /* alternative input option dens_column, convert to kg/m2 */
      if (input.precip_water < 0.0 && input.atm.column[MOL_H2O] >= 0.0) {
        switch (input.atm.unit_column[MOL_H2O]) {
        case DU:
          input.precip_water = input.atm.column[MOL_H2O] * DU2particle * (M_H2O / N_MOL_CKDFU * 10.);
                               /*         DU             *  DU->cm-2   *       cm-2->kg/m2         */
          /* not changed outside this function as input is not a pointer, UH */
          break;
        case CM_2:
          input.precip_water = input.atm.column[MOL_H2O] * (M_H2O / N_MOL_CKDFU * 10.);
                               /*        cm-2            *       cm-2->kg/m2         */ 
          /* not changed outside this function as input is not a pointer, UH */         
          break;
        default:
          fprintf (stderr, "\nError, unknown unit for %s column\n", gas);
          return -1;
        }
      }

      if (input.precip_water >= 0.0) {

        /* vorbose output */
        if (!input.quiet)
          fprintf (stderr, " ... scaling precipitable water from %10.6f kg/m2 to ", 
	           output->precip_water);

	/* determine scale factor and scale profile */
        if ((input.wc.wv_saturation == SATURATION_IPA && output->wc.occurrence == TRUE) || 
            (input.ic.wv_saturation == SATURATION_IPA && output->ic.occurrence == TRUE)) {
          fprintf (stderr, "\nError: Combination of cloud-saturation for IPA and dens_column h2o not implemented!\n");
          fprintf (stderr, "Please use only one of those options or rh_file or dens_file h2o.\n");
          return -1; 
	}
        else {
          /* scaling profile for right column value */
          /* problem: changing level properties, but you want sum of layer properties */
          /* and a variable integration/interpolation method has to be considered */

          /* allocate cloud_indices */
          if ((water_cloud_indices = calloc(output->atm.nlev, sizeof(int)))==NULL) {
             fprintf (stderr, "Error: Allocation of cloud_indices in setup_atmosphere (atmosphere.c)\n");
             return -1;
          }

          if ((ice_cloud_indices = calloc(output->atm.nlev, sizeof(int)))==NULL) {
             fprintf (stderr, "Error: Allocation of cloud_indices in setup_atmosphere (atmosphere.c)\n");
             return -1;
          }

          /* doing this with an iteration, until column value is right, 20 iterations should be more than enough */
          for (iscale=0;iscale<20;iscale++) { 

            /* sum up water column (in particle/cm-2) inside the clouds */
            column_h2o_in_clouds = 0.0;

            /* find water clouds levels */
            find_cloud_indices (output->wc.optprop.dtau [iv], output->wc.zd, output->wc.nlev, 
                                output->atm.zd, output->atm.nlev,
                                &(water_cloud_indices));

            /* add ice clouds levels */
            find_cloud_indices (output->ic.optprop.dtau [iv], output->ic.zd, output->ic.nlev, 
                                output->atm.zd, output->atm.nlev,
                                &(ice_cloud_indices));                        

            for (lc=0;lc<output->atm.nlev;lc++) {
              if (output->wc.occurrence == TRUE && input.wc.wv_saturation == SATURATION_CLOUD &&
		  (water_cloud_indices[lc]   == 1 ||   /* we are dealing here with layer properties */
		   water_cloud_indices[lc+1] == 1)) {  /* we are dealing here with layer properties */
		
                if (water_cloud_indices[lc]   == 1 && 
                    water_cloud_indices[lc+1] != 1)  { /* adjacent layer to the cloud */
                  column_h2o_in_clouds += output->atm.microphys.dens_avg[MOL_H2O][lc] *
		                          (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1.0E5 / 2.0; /* dz, 1E5 = km to cm */
	        }
                else if (water_cloud_indices[lc]   == 1 &&
		         water_cloud_indices[lc+1] == 1) {  /* full cloud layer */ 
                  /* fprintf(stderr,"\n water-cloud from %5.1f km to %5.1f km, dens = %.5e",
		       output->atm.zd[lc],output->atm.zd[lc+1],output->atm.microphys.dens[MOL_H2O][lc]); */
                  column_h2o_in_clouds += output->atm.microphys.dens_avg[MOL_H2O][lc] *
		                          (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1.0E5; /* dz, 1E5 = km to cm */
	        }
	        else if (water_cloud_indices[lc]   != 1 &&
                         water_cloud_indices[lc+1] == 1) { /* adjacent layer to the cloud */
                  column_h2o_in_clouds += output->atm.microphys.dens_avg[MOL_H2O][lc] *
		                          (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1.0E5 / 2.0; /* dz, 1E5 = km to cm */
	        }
	      }
              /* do not sum up a layer twice, if water- AND ice cloud are defined both in this layer */
              else if (output->ic.occurrence == TRUE && input.ic.wv_saturation == SATURATION_CLOUD &&
		       (ice_cloud_indices[lc]   == 1 ||   /* we are dealing here with layer properties */
			ice_cloud_indices[lc+1] == 1)) {  /* we are dealing here with layer properties */
		
                if (ice_cloud_indices[lc]   == 1 && 
                    ice_cloud_indices[lc+1] != 1) { /* adjacent layer to the cloud */
                  column_h2o_in_clouds += output->atm.microphys.dens_avg[MOL_H2O][lc] *
		                          (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1.0E5 / 2.0; /* dz, 1E5 = km to cm */
	        }
                else if (ice_cloud_indices[lc]   == 1 &&
		         ice_cloud_indices[lc+1] == 1) { /* full cloud layer */ 
                  /* fprintf(stderr,"\n ice-cloud from %5.1f km to %5.1f km, dens = %.5e",
		         output->atm.zd[lc],output->atm.zd[lc+1],output->atm.microphys.dens[MOL_H2O][lc]); */
                  column_h2o_in_clouds += output->atm.microphys.dens_avg[MOL_H2O][lc] *
		                          (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1.0E5; /* dz, 1E5 = km to cm */
	        }
	        else if (ice_cloud_indices[lc]   != 1 &&
                         ice_cloud_indices[lc+1] == 1) { /* adjacent layer to the cloud */
                  column_h2o_in_clouds += output->atm.microphys.dens_avg[MOL_H2O][lc] *
		                          (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1.0E5 / 2.0; /* dz, 1E5 = km to cm */
	        }
	      }
	    } /* end of: for (lc=0;lc<output->atm.nlev;lc++) {... */

            column_h2o_in_clouds *= (M_H2O / N_MOL_CKDFU * 10.); /*  cm-2 -> kg/m2   */
         
            if (input.verbose) {
              /* fprintf (stderr,"\n   water column inside clouds = %7.4f kg/m2\n",column_h2o_in_clouds); */
              /* fprintf (stderr,  "   lc   z[km]    old_h2o      new_h2o      factor\n"); */
	    }

            /* check, if there is enough water to saturate clouds */
            if (input.precip_water - column_h2o_in_clouds < 0.0) {
              fprintf (stderr, "\nError: precip_water %e kg/m2 is smaller than the\n", input.precip_water);
              fprintf (stderr,   "water column required to saturate clouds: %e\n",column_h2o_in_clouds);
              return -1;
	    }


            /* determine scale factor */
            if (input.precip_water == 0) {
              output->scale_factor[i] = 0;
	    }
            else if (output->precip_water == 0) {
              fprintf (stderr, "\nError, can't scale a H2O profile which is == 0, column = 0 kg/m2");
              return -1;
	    }
            else if (output->precip_water - column_h2o_in_clouds == 0) {
              output->scale_factor[i] = 0;
	    }
            else
              /* (also valid for no clouds at all) */
              output->scale_factor[i] = (input.precip_water   - column_h2o_in_clouds) / 
                                        (output->precip_water - column_h2o_in_clouds);


            /* scale profile */
            for (lc = 0; lc<output->atm.nlev; ++lc) {

              /* /\* for additional verbose output *\/ */
/*               old_h2o = output->atm.microphys.dens[MOL_H2O][lc]; */

              if ((input.wc.wv_saturation == SATURATION_CLOUD && output->wc.occurrence == TRUE &&
		   water_cloud_indices[lc] == 1) ||
                  (input.ic.wv_saturation == SATURATION_CLOUD && output->ic.occurrence == TRUE &&
		   ice_cloud_indices[lc] == 1)) {
                /* fprintf(stderr,"cloud in %5.1f km",output->atm.zd[lc]); */
	      }
              else
                output->atm.microphys.dens[i][lc] *= output->scale_factor[i];

              /* additional verbose output */
              /* fprintf (stderr,"%5d  %6.2f  %.5e ", lc, output->atm.zd[lc], old_h2o); */
              /* fprintf (stderr," %.5e  %8.4f\n", output->atm.microphys.dens[MOL_H2O][lc],  */
              /*                                   output->atm.microphys.dens[i][lc]/old_h2o); */
	    } /* end of: for (lc = 0; lc<output->atm.nlev; ++lc) { ... */

            /* recalculate the h2o column AND dens_avg */
            status = column (output->atm.microphys.dens[i], output->atm.zd, output->atm.nlev, output->alt.altitude, 
                             input.atm.interpol_method_gas[i], output->atm.microphys.dens[MOL_AIR],
                             &(output->atm.microphys.dens_avg[i]),NO,&(output->column[i]));

            if (status!=0) {
              fprintf (stderr, "\nError %d calculating %s column!\n", status, gas);
              return status;
            }
            output->precip_water = output->column[i] * DU2particle * (M_H2O / N_MOL_CKDFU * 10.);   
                                   /*     DU         *  DU->cm-2   *       cm-2->kg/m2         */

            if (fabs(output->precip_water-input.precip_water) < 0.00001)
              break;
	  } /* end of: for (iscale=0;iscale<10;iscale++) { ... */
	  
          free(water_cloud_indices);
          free(ice_cloud_indices);

        } /* end of: else { ... (if ((input.wc.wv_saturation == SATURATION_IPA) */

        /* recalculate the h2o column */
        status = column (output->atm.microphys.dens[i], output->atm.zd, output->atm.nlev, output->alt.altitude, 
                         input.atm.interpol_method_gas[i], output->atm.microphys.dens[MOL_AIR],
                         &(output->atm.microphys.dens_avg[i]),NO,&(output->column[i]));

        if (status!=0) {
          fprintf (stderr, "\nError %d calculating %s column!\n", status, gas);
          return status;
        }
        output->precip_water = output->column[i] * DU2particle * (M_H2O / N_MOL_CKDFU * 10.);   
                                 /*     DU         *  DU->cm-2   *       cm-2->kg/m2         */

        if (fabs(output->precip_water-input.precip_water) >= 0.0001) {
          fprintf (stderr, "\nError: scaling water column to %10.6f kg/m2 in setup_atmosphere().\n",input.precip_water);
          fprintf (stderr,   "Reached end of iteration: %10.6f kg/m2 \n",output->precip_water);
          return -1;
	}

        /* verbose output */
        if (!input.quiet)
          fprintf (stderr, "%10.6f kg/m2\n", output->precip_water);

      } /* end of: if (input.precip_water >= 0.0) {... */
      break;
    default:
      fprintf (stderr, "Error, unknown gas_number %d when scaling gas profile (atmosphere.c)\n", i);
      return -1;
    }
  }


  /*****************************************************************************************/
  /*                      Scaling profiles with MIXING RATIOs                              */
  /*                         at the user-defined altitude                                  */
  /*****************************************************************************************/ 

  for (ialt=0; ialt<output->atm.nlev; ialt++)
    if (output->alt.altitude == output->atm.zd[ialt])
      break;
    
  if (ialt == output->atm.nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid.\n", 
	     output->alt.altitude);
    return -1;
  }

  output->mixing_ratio = calloc (MX_NN, sizeof(float));

  /* O2 */
  if (input.mixing_ratio[MX_O2]<0) {
    /* If not specified, use the mixing ratio at the user-defined altitude */
    output->mixing_ratio[MX_O2] = output->atm.microphys.dens[MOL_O2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_O2] = input.mixing_ratio[MX_O2];
    
    scale_factor = (input.mixing_ratio[MX_O2] / 1e6) / 
      (output->atm.microphys.dens[MOL_O2] [ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling O2-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_O2][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	       input.mixing_ratio[MX_O2]); 

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_O2][lc] *= scale_factor;
  }
  

  /* CO2 */
  if (input.mixing_ratio[MX_CO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_CO2] = output->atm.microphys.dens[MOL_CO2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_CO2] = input.mixing_ratio[MX_CO2];
    
    scale_factor = (input.mixing_ratio[MX_CO2] / 1e6) / 
      (output->atm.microphys.dens[MOL_CO2][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling CO2-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_CO2][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	      input.mixing_ratio[MX_CO2]); 

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_CO2][lc] *= scale_factor;
  }
  

  /* H2O */
  if (input.mixing_ratio[MX_H2O]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_H2O] = output->atm.microphys.dens[MOL_H2O][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_H2O] = input.mixing_ratio[MX_H2O];
    
    scale_factor = (input.mixing_ratio[MX_H2O] / 1e6) / 
      (output->atm.microphys.dens[MOL_H2O][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling H2O-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_H2O][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	      input.mixing_ratio[MX_H2O]); 

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_H2O][lc] *= scale_factor;
  }
  
  /* NO2 */
  if (input.mixing_ratio[MX_NO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_NO2] = output->atm.microphys.dens[MOL_NO2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_NO2] = input.mixing_ratio[MX_NO2];
    
    scale_factor = (input.mixing_ratio[MX_NO2] / 1e6) / 
      (output->atm.microphys.dens[MOL_NO2][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling NO2-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_NO2][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	      input.mixing_ratio[MX_NO2]);

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_NO2][lc] *= scale_factor;
  }
  
  /* default CH4, N2O, F11, F12, and F22 are 'hardwired'; */
  /* these gases are assumed to be well-mixed throughout  */
  /* the whole atmosphere                                 */

  if (input.mixing_ratio[MX_CH4] < 0)
    output->mixing_ratio[MX_CH4] = 1.6; 
  else {
    output->mixing_ratio[MX_CH4] = input.mixing_ratio[MX_CH4];
    if (!input.quiet)
      fprintf(stderr," ... scaling CH4-mixing ratio from 1.6 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_CH4]);
  }

  if (input.mixing_ratio[MX_N2O] < 0)
    output->mixing_ratio[MX_N2O] = 0.28; 
  else {
    output->mixing_ratio[MX_N2O] = input.mixing_ratio[MX_N2O];
    if (!input.quiet)
      fprintf(stderr," ... scaling N2O-mixing ratio from 0.28 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_N2O]);
  }
  
  if (input.mixing_ratio[MX_F11] < 0)
    output->mixing_ratio[MX_F11] = 0.268E-3; 
  else {
    output->mixing_ratio[MX_F11] = input.mixing_ratio[MX_F11];
    if (!input.quiet)
      fprintf(stderr," ... scaling F11-mixing ratio from 0.268E-3 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_F11]);
  }
  
  if (input.mixing_ratio[MX_F12] < 0)
    output->mixing_ratio[MX_F12] = 0.503E-3; 
  else {
    output->mixing_ratio[MX_F12] = input.mixing_ratio[MX_F12];
    if (!input.quiet)
      fprintf(stderr," ... scaling F12-mixing ratio from 0.503E-3 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_F12]);   
  }

  if (input.mixing_ratio[MX_F22] < 0)
    output->mixing_ratio[MX_F22] = 0.105E-3; 
  else {
    output->mixing_ratio[MX_F22] = input.mixing_ratio[MX_F22];
    if (!input.quiet)
      fprintf(stderr," ... scaling F22-mixing ratio from 0.105E-3 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_F22]);
  }

  /* Read Rayleigh cross section if specified */
  if (strlen(input.filename[FN_RAYLEIGH]) > 0) {
    status = read_molecular_absorption (input.filename[FN_RAYLEIGH],
					&zd_user, &nlyr_user,
					input.quiet,
					&(output->atm.optprop.tau_rayleigh_user), 
					&tmp_dtau, &tmp_lambda, &tmp_nlambda,
					&monochromatic);

    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_molecular_absorption (\"%s\")\n", 
	       status, input.filename[FN_RAYLEIGH]);
      return status;
    }

    if (monochromatic) {
      output->atm.rayleigh = RAYLEIGH_FILE_MONO;

      if (!input.quiet)
	fprintf (stderr, " ... read monochromatic molecular absorption from %s\n", 
		 input.filename[FN_RAYLEIGH]);
    }
    else {
      output->atm.rayleigh = RAYLEIGH_FILE_SPEC;
      
      if (!input.quiet)
	fprintf (stderr, " ... read Rayleigh scattering cross section from %s\n", 
		 input.filename[FN_RAYLEIGH]);
      
      /***************************************************************/
      /* interpolate optical thickness to transmittance grid, linear */
      /***************************************************************/

      /* allocate memory for optical thickness on internal transmittance grid */
      output->atm.optprop.tau_rayleigh_user_t = calloc (output->wl.nlambda_t, sizeof(double *));
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	output->atm.optprop.tau_rayleigh_user_t[iv] = calloc (nlyr_user, sizeof(double));
      
      switch (output->wl.type) {
      case WLGRID_CK:
      case WLGRID_USER:
      case WLGRID_BANDS:
      case WLGRID_UVSPEC:
      case WLGRID_MOLABS:
	
	/* memory for temporary arrays */
	tmp_dtauf   = calloc (tmp_nlambda,          sizeof (float));
	tmp_dtauf_t = calloc (output->wl.nlambda_t, sizeof (float));
	
	/* loop over layers */
	for (lc=0; lc<nlyr_user; lc++) {
	  
	  /* interpolate tmp_dtauf to internal wavelength grid */

	  /* copy double array to temporary float array */
	  /* ??? should improve that by providing a ??? */
	  /* ??? an arb_wvn for arrays of double    ??? */
	  for (iv=0; iv<tmp_nlambda; iv++)
	    tmp_dtauf[iv] = tmp_dtau[lc][iv];
	  
	  status = arb_wvn (tmp_nlambda, tmp_lambda, tmp_dtauf, 
			    output->wl.nlambda_t, output->wl.lambda_t, 
			    tmp_dtauf_t,
			    1, 0);
	  
	  /* copy temporary float array to final destination */
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->atm.optprop.tau_rayleigh_user_t[iv][lc] = tmp_dtauf_t[iv];
	  
	  
	  if (status != 0) {
	    fprintf (stderr, "Error %d interpolating molecular absorption grid to internal\n", status);
	    fprintf (stderr, "wavelength grid\n");
	    return status;
	  }
	}

	free(tmp_dtauf); free(tmp_dtauf_t);
	break;
	  
      default:
	fprintf (stderr, "Error, unknown wavelength grid type %d\n", 
		 output->wl.type);
	return -1;
      }

      free(tmp_lambda); 
      for (lc=0; lc<nlyr_user; lc++)
	free (tmp_dtau[lc]);
      free (tmp_dtau);
    }


    /* redistribute tau_rayleigh_user and tau_rayleigh_user_t to common atmospheric grid */
    status = 0;
    if (monochromatic) {
      status += redistribute_1D ((void **) (&(output->atm.optprop.tau_rayleigh_user)),
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1, 
				 1, REDISTRIBUTE_DOUBLE);
    }
    else {
      status += redistribute_2D ((void ***) (&(output->atm.optprop.tau_rayleigh_user_t)), 
				 output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1,
				 1, REDISTRIBUTE_DOUBLE);
    }

    if (status!=0) {
      fprintf (stderr, "Error interpolating user-defined molecular absorption coefficients\n");
      fprintf (stderr, "to common atmospheric vertical grid.\n");
      return status;
    }
  }			


  /* Read molecular absorption cross section if specified */
  if (strlen(input.filename[FN_MOLABS]) > 0) {
    status = read_molecular_absorption (input.filename[FN_MOLABS], 
					&zd_user, &nlyr_user,
					input.quiet,
					&(output->atm.optprop.tau_molabs_user), 
					&tmp_dtau, &tmp_lambda, &tmp_nlambda,
					&monochromatic);

    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_molecular_absorption (\"%s\")\n", 
	       status, input.filename[FN_MOLABS]);
      return status;
    }

    if (monochromatic) {
      output->atm.molabs = MOLABS_FILE_MONO;

      if (!input.quiet)
	fprintf (stderr, " ... read monochromatic molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
    }
    else {
      output->atm.molabs = MOLABS_FILE_SPEC;
      
      if (!input.quiet)
	fprintf (stderr, " ... read spectral molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
      
      /***************************************************************/
      /* interpolate optical thickness to transmittance grid, linear */
      /***************************************************************/

      /* allocate memory for optical thickness on internal transmittance grid */
      output->atm.optprop.tau_molabs_user_t = calloc (output->wl.nlambda_t, sizeof(double *));
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	output->atm.optprop.tau_molabs_user_t[iv] = calloc (nlyr_user, sizeof(double));
      
      switch (output->wl.type) {
      case WLGRID_CK:
      case WLGRID_USER:
      case WLGRID_BANDS:
      case WLGRID_UVSPEC:
	
	/* memory for temporary arrays */
	tmp_dtauf   = calloc (tmp_nlambda,          sizeof (float));
	tmp_dtauf_t = calloc (output->wl.nlambda_t, sizeof (float));
	
	/* loop over layers */
	for (lc=0; lc<nlyr_user; lc++) {
	  
	  /* interpolate tmp_dtauf to internal wavelength grid */

	  /* copy double array to temporary float array */
	  /* ??? should improve that by providing a ??? */
	  /* ??? an arb_wvn for arrays of double    ??? */
	  for (iv=0; iv<tmp_nlambda; iv++)
	    tmp_dtauf[iv] = tmp_dtau[lc][iv];
	  
	  status = arb_wvn (tmp_nlambda, tmp_lambda, tmp_dtauf, 
			    output->wl.nlambda_t, output->wl.lambda_t, 
			    tmp_dtauf_t,
			    1, 0);
	  
	  /* copy temporary float array to final destination */
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->atm.optprop.tau_molabs_user_t[iv][lc] = tmp_dtauf_t[iv];
	  
	  
	  if (status != 0) {
	    fprintf (stderr, "Error %d interpolating molecular absorption grid to internal\n", status);
	    fprintf (stderr, "wavelength grid\n");
	    return status;
	  }
	}

	free(tmp_dtauf); free(tmp_dtauf_t);
	break;
	  
      case WLGRID_MOLABS:
	/* no need to interpolate, just copy */

	/* copy temporary float array to final destination */
	for (lc=0; lc<nlyr_user; lc++)
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->atm.optprop.tau_molabs_user_t[iv][lc] = tmp_dtau[lc][iv];
	
	break;
	
      default:
	fprintf (stderr, "Error, unknown wavelength grid type %d\n", 
		 output->wl.type);
	return -1;
      }

      free(tmp_lambda); 
      for (lc=0; lc<nlyr_user; lc++)
	free (tmp_dtau[lc]);
      free (tmp_dtau);
    }


    /* redistribute tau_molabs_user and tau_molabs_user_t to common atmospheric grid */
    status = 0;
    if (monochromatic) {
      status += redistribute_1D ((void **) (&(output->atm.optprop.tau_molabs_user)),
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1, 
				 1, REDISTRIBUTE_DOUBLE);
    }
    else {
      status += redistribute_2D ((void ***) (&(output->atm.optprop.tau_molabs_user_t)), 
				 output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1,
				 1, REDISTRIBUTE_DOUBLE);
    }

    if (status!=0) {
      fprintf (stderr, "Error interpolating user-defined molecular absorption coefficients\n");
      fprintf (stderr, "to common atmospheric vertical grid.\n");
      return status;
    }
  }			

  
  if (input.verbose) {
 
    fprintf (stderr, "--------------------------------------------------------------------------------------------------------------------------");
    if (output->column[MOL_BRO]  !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_OCLO] !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_HCHO] !=0)
      fprintf (stderr, "--------------");
    fprintf (stderr, "\n");
    fprintf (stderr, "  lc |  z[km]  |  Pressure  | Temp.  |    Air      |   Ozone     |     O2      | Water vap.  |    CO2      |    NO2      | ");
    if (output->column[MOL_BRO]  !=0)
      fprintf (stderr, "    BRO      |");
    if (output->column[MOL_OCLO] !=0)
      fprintf (stderr, "    OCLO     |");
    if (output->column[MOL_HCHO] !=0)
      fprintf (stderr, "    HCHO     |");
    fprintf (stderr, "\n");
    fprintf (stderr, "     |         |   [hPa]    |  [K]   |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    | ");
    if (output->column[MOL_BRO]  !=0)
      fprintf (stderr, "   [cm-3]    |");
    if (output->column[MOL_OCLO] !=0)
      fprintf (stderr, "   [cm-3]    |");
    if (output->column[MOL_HCHO] !=0)
      fprintf (stderr, "   [cm-3]    |");
    fprintf (stderr, "\n");
    fprintf (stderr, "--------------------------------------------------------------------------------------------------------------------------");
    if (output->column[MOL_BRO]  !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_OCLO] !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_HCHO] !=0)
      fprintf (stderr, "--------------");
    fprintf (stderr, "\n");


    for (lc=0; lc<output->atm.nlev; lc++) {
      if (output->atm.zd[lc] >= output->alt.altitude) {
        fprintf (stderr, "%5d  %7.3f   %10.5f   %6.2f   %.5e   %.5e   %.5e   %.5e   %.5e   %.5e",
	         lc, output->atm.zd[lc],
	         output->atm.microphys.press[lc], output->atm.microphys.temper[lc],output->atm.microphys.dens[MOL_AIR][lc],
	         output->atm.microphys.dens[MOL_O3][lc],output->atm.microphys.dens[MOL_O2][lc], output->atm.microphys.dens[MOL_H2O][lc],
                 output->atm.microphys.dens[MOL_CO2][lc],output->atm.microphys.dens[MOL_NO2][lc]);
        if (output->column[MOL_BRO]  !=0)
          fprintf (stderr, "   %.5e",output->atm.microphys.dens[MOL_BRO][lc]);
        if (output->column[MOL_OCLO] !=0)
          fprintf (stderr, "   %.5e",output->atm.microphys.dens[MOL_OCLO][lc]);
        if (output->column[MOL_HCHO] !=0)
          fprintf (stderr, "   %.5e",output->atm.microphys.dens[MOL_HCHO][lc]);
        fprintf (stderr, "\n");
      }
    }

    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------");
    if (output->column[MOL_BRO]  !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_OCLO] !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_HCHO] !=0)
      fprintf (stderr, "--------------");
    fprintf (stderr, "\n");
    fprintf (stderr, "%5s | %6.2f | %10.5f | %6.2f | %7.1ecm-2 |  %7.3f DU | %7.1ecm-2 | %6.3f kg/m2 | %7.1ecm-2 | %7.5f DU |",
	             "sum", 0.0/0.0, 0.0/0.0, 0.0/0.0, output->column[MOL_AIR]*DU2particle, output->column[MOL_O3], 
                      output->column[MOL_O2]*DU2particle, output->precip_water, output->column[MOL_CO2]*DU2particle, output->column[MOL_NO2]);
      if (output->column[MOL_BRO]  !=0)
        fprintf (stderr, " %7.5f DU |",output->column[MOL_BRO]);
      if (output->column[MOL_OCLO] !=0)
        fprintf (stderr, " %7.5f DU |",output->column[MOL_OCLO]);
      if (output->column[MOL_HCHO] !=0)
        fprintf (stderr, " %7.5f DU |",output->column[MOL_HCHO]);
      fprintf (stderr, "\n");
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------");
    if (output->column[MOL_BRO]  !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_OCLO] !=0)
      fprintf (stderr, "--------------");
    if (output->column[MOL_HCHO] !=0)
      fprintf (stderr, "--------------");
    fprintf (stderr, "\n");

  }


  /* SPECIAL VERTICAL FIELD AT "ZOUT-LEVELS"          */
  /* for heating rate calculations and for output user */

  /* allocation */

  /* gas densities */
  if ((output->atm.microphys.dens_zout = calloc (MOL_NN, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for dens_zout in setup_atmosphere() (in atmosphere.c)\n");
    return -30;
  }
  for (i_gas=0;i_gas<MOL_NN;i_gas++) {
    /* gas density */
    if ((output->atm.microphys.dens_zout[i_gas] = calloc (output->atm.nlev, sizeof (float))) == NULL) {
      fprintf (stderr,"Error, allocating memory for dens_zout[%d] in setup_atmosphere() (in atmosphere.c)\n",i_gas);
      return -31;
    }
  }

  /* pressure */
  if ((output->atm.microphys.press_zout       = calloc (output->atm.nlev, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for press_zout in setup_atmosphere() (in atmosphere.c)\n");
    return -32;
  }
  /* temperature */
  if ((output->atm.microphys.temper_zout      = calloc (output->atm.nlev, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for temper_zout in setup_atmosphere() (in atmosphere.c)\n");
    return -33;
  }
  /* potential temperature theta */
  if ((output->atm.microphys.theta_zout       = calloc (output->atm.nlev, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for theta_zout in setup_atmosphere() (in atmosphere.c)\n");
    return -34;
  }
  /* dewpoint temperature */
  if ((output->atm.microphys.temper_d_zout    = calloc (output->atm.nzout, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for theta_zout in setup_atmosphere() (in atmosphere.c)\n");
    return -35;
  }
  /* equivalent potential temperature theta_e */
  if ((output->atm.microphys.theta_e_zout = calloc (output->atm.nzout, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for theta_equiv_zout in setup_atmosphere() (in atmosphere.c)\n");
    return -36;
  }

  /* calculation or dummy value */
  if (output->atm.zout_sur[0] != ZOUT_MYSTIC) { /* all cases except MYSTIC 2D surface */

    /* AIR DENSITY (must be before other gases, because result */
    /*  is needed for linmix interpolation of the other gases) */

    /* copy original values */
    for (iz=0; iz<output->atm.nlev; iz++)
      output->atm.microphys.dens_zout[MOL_AIR][iz] = output->atm.microphys.dens[MOL_AIR][iz];

    /* interpolation */
    status = interpolate_profile (output->atm.zd, &(output->atm.microphys.dens_zout[MOL_AIR]), output->atm.nlev,
                                  output->atm.zout_sea, output->atm.nzout, 
                                  input.atm.interpol_method_gas[MOL_AIR],input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error interpolating dens_zout[MOL_AIR]\n");
      return status;
    }

    /* GAS DENISTY */
    for (i_gas=0;i_gas<MOL_NN;i_gas++) {

      /* other than air */
      if (i_gas != MOL_AIR) {

        /* copy atmosphere values */
        for (iz=0; iz<output->atm.nlev; iz++)
          output->atm.microphys.dens_zout[i_gas][iz] = output->atm.microphys.dens[i_gas][iz];

        /* interpolation */
        status += interpolate_density (output->atm.zd, &(output->atm.microphys.dens_zout[i_gas]), output->atm.nlev, 
                                       output->atm.zout_sea, output->atm.nzout, input.atm.interpol_method_gas[i_gas],
                                       output->atm.microphys.dens[MOL_AIR], output->atm.microphys.dens_zout[MOL_AIR], input.quiet);
        if (status!=0) {
          fprintf (stderr, "Error interpolating dens_zout[%d]\n",i_gas);
          return status;
        }
      }
    }

    /* PRESSURE */
    /* copy original values, size is adjusted automagically in interpolate_profile */
    for (iz=0; iz<output->atm.nlev; iz++)
      output->atm.microphys.press_zout[iz] = output->atm.microphys.press[iz];
        
    status = interpolate_profile (output->atm.zd, &output->atm.microphys.press_zout, output->atm.nlev,
                                  output->atm.zout_sea, output->atm.nzout, 
                                  input.atm.interpol_method_press,input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error interpolating press_zout\n");
      return status;
    }

    /* TEMPERATURE */
    /* copy original values, size is adjusted automagically in interpolate_profile */
    for (iz=0; iz<output->atm.nlev; iz++)
      output->atm.microphys.temper_zout[iz] = output->atm.microphys.temper[iz];
        
    status = interpolate_profile (output->atm.zd, &output->atm.microphys.temper_zout, output->atm.nlev,
                                  output->atm.zout_sea, output->atm.nzout, 
                                  input.atm.interpol_method_temper,input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error interpolating temper_zout\n");
      return status;
    }

    /* potential temperature THETA */
    /* calculate theta from p and T */
    kappa = (c_p-c_v)/c_p;  /* == R / cp */
    for (iz=0;iz<output->atm.nzout;iz++) 
      output->atm.microphys.theta_zout[iz] = output->atm.microphys.temper_zout[iz] * 
                                             pow((1000./output->atm.microphys.press_zout[iz]),kappa);
    
    /* calculate DEWPOINT TEMPERATURE from dens_h2o */
    for (lc=0;lc<output->atm.nzout;lc++) {
      press_h2o = output->atm.microphys.dens_zout[MOL_H2O][lc] * boltzmann * output->atm.microphys.temper[lc] * 1E+04; 
                  /* 10E4 == J/cm-3 -> hPa */
      if (press_h2o!=0.0) {
        output->atm.microphys.temper_d_zout[lc] = F77_FCN(dewpt)(&(press_h2o));
        output->atm.microphys.temper_d_zout[lc] += 273.15;
      }
      else {
        output->atm.microphys.temper_d_zout[lc] = 0.0;    /* 0 Kelvin */
      }
    }

    /* calculate EQUIVALENT POTENTIAL TEMPERATURE from  T, T_D, and p */

    for (lc=0;lc<output->atm.nzout;lc++) {

      tmp_temper   = output->atm.microphys.temper_zout[lc]   - 273.15; /* K -> degree C */
      tmp_temper_d = output->atm.microphys.temper_d_zout[lc] - 273.15; /* K -> degree C */
 
      output->atm.microphys.theta_e_zout[lc] = F77_FCN (ept) (&(tmp_temper), 
                                                              &(tmp_temper_d), 
							      &(output->atm.microphys.press_zout[lc])); 

      output->atm.microphys.theta_e_zout[lc] += 273.15; /* degree C -> K */
    }
    /* approximately theta_e = (T + L_v R/c_p h2o_mix) (p_0/p(z))^kappa) */
    /* but L_v is temperature dependent */

  }
  else { /* for MYSTIC 2D surface set dummy value */
    for (i_gas=0;i_gas<MOL_NN;i_gas++)
      output->atm.microphys.dens_zout[i_gas][0] = -999.0;
    output->atm.microphys.press_zout[0]         = -999.0;
    output->atm.microphys.temper_zout[0]        = -999.0;
    output->atm.microphys.theta_zout[0]         = -999.0;
  }

  /* -------------------------------------------------------------------------------------------*/
  /*      every z-level added to zd_common will be taken care of in setup_redistribute()        */
  /* -------------------------------------------------------------------------------------------*/

  /*************************************************************/
  /*  Initialise atm.zd_common grid, all properties will be at */
  /*  this grid latest after setup_redistribute().             */
  /*  output->atm.microphys.xxx is defined on output->atm.zd   */ 
  /*  until setup_redistribute()                               */ 
  /*************************************************************/

  /* Set common vertical resolution */
  output->atm.nlev_common = 0;
  set_common_z_grid (output->atm.zd, output->atm.nlev,
		     &output->atm.zd_common, &output->atm.nlev_common);

  /* if zout_interpolate is not specified, data for                  */
  /* altitude (2nd argument) levels will be added in redistribute.c  */
  /* and zout-Levels will not be added anywhere to the z-grid        */
  /* solvers do that for us                                          */

  /* add watercloud levels, if no cloud water vapour saturation is specified */
  /* (no interpolation, but redistribution on these levels */
  if (input.wc.wv_saturation == SATURATION_NONE) {
    /* add the water cloud levels to the common profile */
    if (input.wcipa) {
      for (ipa=0; ipa<output->nipa; ipa++)
        set_common_z_grid (output->wcipa[ipa].zd, output->wcipa[ipa].nlev, 
                           &output->atm.zd_common, &output->atm.nlev_common);
    }
    else {
      set_common_z_grid (output->wc.zd, output->wc.nlev, 
                         &output->atm.zd_common, &output->atm.nlev_common);
    }
  }

  /* add icecloud levels, if no cloud water vapour saturation is specified */
  /* (no interpolation, but redistribution on these levels */
  if (input.ic.wv_saturation == SATURATION_NONE) {
    /* add the ice cloud levels to the common profile */
    if (input.icipa) {
      for (ipa=0; ipa<output->nipa; ipa++) {
        set_common_z_grid (output->icipa[ipa].zd, output->icipa[ipa].nlev, 
                           &output->atm.zd_common, &output->atm.nlev_common);
      }
    } 
    else {
      set_common_z_grid (output->ic.zd, output->ic.nlev, 
                         &output->atm.zd_common, &output->atm.nlev_common);
    }
  }

  /* this is just a check for stupid programming errors, in usual case this is unnessesary */

  /* Finally, check if profiles of pressure, temperature and air density are consistent */
  /* with ideal gas law over 6 orders of magnitude in pressure; check only first six    */
  /* orders of magnitude because for very small pressures roundoff errors due to too    */
  /* few digits cause problems                                                          */
  firstnonideal=1;
  for (lc=0; lc<output->atm.nlev; lc++) {

    idealgas=1;

    if (output->atm.microphys.press[lc]/output->atm.microphys.press[output->atm.nlev-1]>1e-6) {
      tmpdens = output->atm.microphys.press[lc]*1E-04 / (boltzmann*output->atm.microphys.temper[lc]);
      
      if ((tmpdens==0 && output->atm.microphys.dens[MOL_AIR][lc]!=0) || 
	  (tmpdens!=0 && output->atm.microphys.dens[MOL_AIR][lc]==0))
	idealgas=0;
      
      if (output->atm.microphys.dens[MOL_AIR][lc]>0)
	if ((idealdiff = fabs ((tmpdens - output->atm.microphys.dens[MOL_AIR][lc]) / output->atm.microphys.dens[MOL_AIR][lc])) > 0.01)
	  idealgas=0;
      
      if (!idealgas && firstnonideal) {
	firstnonideal=0;
	fprintf (stderr, "\n");
	fprintf (stderr, "*** ATTENTION: pressure, temperature, and air density not\n");
	fprintf (stderr, "*** consistent with ideal gas law at the following levels\n");
	fprintf (stderr, "*** (assuming a Boltzmann constant %.6e):\n", boltzmann);
	fprintf (stderr, "***    z [km]  density [cm-3]  p/kT [cm-3]  difference\n");
      }
      
      if (!idealgas) {
	fprintf (stderr, "***   %7.2f  %.4e      %.4e    %6.1f%%\n", 
		 output->atm.zd[lc], output->atm.microphys.dens[MOL_AIR][lc], tmpdens, idealdiff*100.0);
      }
    }
  }
  
  if (!firstnonideal) {
    fprintf (stderr, "*** Please check pressure, temperature, and air density in\n");
    fprintf (stderr, "*** %s for consistency!\n", input.filename[FN_ATMOSPHERE]);
    fprintf (stderr, "\n");
    return -1;
  } 


  /* some cloud stuff had to be moved here for sake of changing the order */

  /* initialize wcloud levels, if there is no wcloud defined */
  /* some functions might need that */
  if (strlen(input.filename[FN_WCLOUD]) > 0) {
  }
  else if (strlen(input.filename[FN_WCLOUD_IPA]) > 0) {
  }
  else if (input.wcipa)  {

    /******************************/
    /* compare 3D and 1D profiles */
    /******************************/
    
    /* ??? does the following still make sense after the redistribution ???  */
    /* make sure that 1D and 3D properties are defined on the same altitude grid */
    status = compare_levels_ipa_df (output->wc.z3D, output->wc.Nz3D, output->atm.zd, output->atm.nlyr);
    
    if (status!=0) {
      fprintf (stderr, "\nFATAL error: z-grids of 1D atmosphere and 3D clouds are different\n");
      return status;
    }

    if (!input.quiet)
      fprintf (stderr, " ... 3D cloud definition seems to be o.k.\n");
  }

  free(output->wc.z3D);

  return 0;
}




/**************************************************************/
/* Setup optical properties of trace gases.                   */
/**************************************************************/

int setup_gases (input_struct input, output_struct *output)
{
  static int first=1;
  int status=0;

  double bscar=0;
  double babso=0, babso_o3=0, babso_co2=0, babso_o2=0;
  double babso_no2=0, babso_bro=0, babso_oclo=0, babso_hcho=0, babso_md=0;
  double babso_amf=0, babso_amftot=0, vertical_column=0, crs_amf=0;
  double bscartot=0;
  double babsotot=0, babsotot_o3=0, babsotot_co2=0, babsotot_o2=0; 
  double babsotot_no2=0, babsotot_bro=0, babsotot_oclo=0, babsotot_hcho=0;
  double deltaz=0;
  int iq=0, is=0, isthis=0, iv=0, lc=0, ipa=0;
  int nq_rayleigh=0;
  int nlev = output->atm.nlev;
  int nlyr = output->atm.nlyr;

  int iq1=0, iq2=0, iq3=0, iq4=0;
  int nq=0, nq1=0, nq2=0, nq3=0, nq4=0;

  double factor=0;
  double weight[KATO_COMPONENTS];

  float *fu_rayleigh = calloc (nlyr, sizeof(float));
  float density=-999.0;
  int gas_nr;

  float *dens_h2o_org=NULL;
  /* double ***tau_molabs_t_org=NULL; */ 

  /* number of quadrature points for Rayleigh scattering */
  switch (input.ck_scheme) {
  case CK_FU:
    nq_rayleigh=FU_MAXINT;
    break;
    
  case CK_KATO:
  case CK_KATO2:
  case CK_KATO2_96:
  case CK_AVHRR_KRATZ:
  case CK_GENERIC:
  case CK_LOWTRAN:
  case CK_NONE:
    nq_rayleigh=1;
    break;

  default:
    fprintf (stderr, "Error: unimplemented correlated-k scheme %d\n", input.ck_scheme);
    return -1;
    break;
  }

  if (first!=0) {
    first = 0;
   
    /* allocate memory for profiles */
    output->atm.cdento   = (float *) calloc (nlyr, sizeof(float));
  
    output->atm.wght_t   = calloc (output->wl.nlambda_t, sizeof (double *));
    output->atm.nq_t     = calloc (output->wl.nlambda_t, sizeof (int));

    output->mc_photons_t = calloc (output->wl.nlambda_t, sizeof (double));


    /* copy ck weights */
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      switch (input.ck_scheme) {
      case CK_FU:
        output->atm.nq_t[iv] = output->ck.wght[FU_WGHT][iv+1];
        break;
        
      case CK_KATO:
      case CK_KATO2:
      case CK_KATO2_96:
        output->atm.nq_t[iv] = output->ck.wght[KATO_CO2][iv+1] *
                               output->ck.wght[KATO_O3] [iv+1] *
                               output->ck.wght[KATO_O2] [iv+1] *
                               output->ck.wght[KATO_H2O][iv+1];
	break;

      case CK_AVHRR_KRATZ:
        output->atm.nq_t[iv] = output->ck.wght[AVHRR_KRATZ_WGHT][iv+1];
        break;

      case CK_GENERIC:
        output->atm.nq_t[iv] = output->ck.wght[GENERIC_WGHT][iv+1];
        break;

      case CK_LOWTRAN:
        output->atm.nq_t[iv] = LOWTRAN_MAXINT;
        /* this is the maximum number, but can be less */
        /* this is optimised inside the solve_rte() ipa-loop, UH 2006-07-19 */
        break;

      case CK_NONE:
	output->atm.nq_t[iv] = 1;
	break;
	
      default:
	fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
	return -1;
	break;
      }

      output->atm.wght_t [iv] = calloc (output->atm.nq_t[iv], sizeof (double));

      /* allocate memory for photon fractions */
      output->mc_photons_t [iv] = calloc (output->atm.nq_t[iv], sizeof (double));


      /* copy photon fractions */
      switch(input.ck_scheme) {
      case CK_NONE:
        
        iq=0; /* no loop over quadrature points required */
        output->mc_photons_t [iv][iq] = output->wl.pfraction[iv];
	
	break;
	
      case CK_KATO:
      case CK_KATO2:
      case CK_KATO2_96:
        nq1 = output->crs_ck.profile[KATO_CO2-1][iv].ngauss;
        nq2 = output->crs_ck.profile[KATO_O3-1 ][iv].ngauss;
        nq3 = output->crs_ck.profile[KATO_O2-1 ][iv].ngauss;
        nq4 = output->crs_ck.profile[KATO_H2O-1][iv].ngauss;
      
        iq=0;  /* reset quadrature counter */
        /* loop over quadrature points */
        for (iq1=0; iq1<nq1; iq1++)  /* CO2 */
          for (iq2=0; iq2<nq2; iq2++)  /* O3 */
            for (iq3=0; iq3<nq3; iq3++)  /* O2 */
              for (iq4=0; iq4<nq4; iq4++)  /* H2O */
                switch (input.source) {
                case SRC_SOLAR: /* solar source */
                  output->mc_photons_t [iv][iq++] = ((double *****) output->ck.x_solar)  [iv+1][iq1+1][iq2+1][iq3+1][iq4+1];
                  break;
                case SRC_THERMAL: /* thermal source */
                  output->mc_photons_t [iv][iq++] = ((double *****) output->ck.x_thermal)[iv+1][iq1+1][iq2+1][iq3+1][iq4+1];
                  break;
                default:
                  fprintf (stderr, "Error, unknown source %d\n", input.source);
                  return -1;
                }
        break;
      
      case CK_FU:
      case CK_AVHRR_KRATZ:
      case CK_GENERIC:
      case CK_LOWTRAN:
        
        if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
            (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {
          nq = output->crs_ck.ipa_profile[ipa][0][iv].ngauss;
	}
        else
          nq = output->crs_ck.profile[0][iv].ngauss;
        
        for (iq=0; iq<nq; iq++)
          switch (input.source) {
          case SRC_SOLAR: /* solar source */
            output->mc_photons_t [iv][iq] = ((double **) output->ck.x_solar)  [iv+1][iq+1];
            break;
          case SRC_THERMAL: /* thermal source */
            output->mc_photons_t [iv][iq] = ((double **) output->ck.x_thermal)[iv+1][iq+1];
            break;
          default:
            fprintf (stderr, "Error, unknown source %d\n", input.source);
            return -1;
          }
        
        
        break;
        
      default:
        fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
        return -1;
        break;
      } /* end switch */
    } /* end wavelength loop */


    /* Rayleigh scattering optical thickness */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_rayleigh_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;

    /* Molecular absorption optical thickness */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_molabs_t, 
        					 nlyr, output->wl.nlambda_t, 
        					 output->atm.nq_t)) != 0)
      return status;
    
    if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
        (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) { 
      dens_h2o_org = (float *) calloc (nlev, sizeof(float));
      if (dens_h2o_org == NULL) {
        fprintf (stderr, "Error allocating memory for dens_h2o_org in function setup_gases()\n");
        return -1;
      }

/*       if ((status = ASCII_calloc_double_3D_arylen (&(tau_molabs_t_org),  */
/*                                                    nlyr, output->wl.nlambda_t,  */
/*                                                    output->atm.nq_t)) != 0) */
/*       return status; */

      output->atm.ipa_wght_t   = calloc (output->nipa, sizeof (double **));
      for (ipa=0; ipa<output->nipa; ipa++) {
        output->atm.ipa_wght_t[ipa]   = calloc (output->wl.nlambda_t, sizeof (double *));
        for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++)
          output->atm.ipa_wght_t[ipa][iv] = calloc (output->atm.nq_t[iv], sizeof (double));
      }

      if ((status = ASCII_calloc_double_4D_arylen (&output->atm.optprop.tau_molabs_t_ipa, 
                                                   output->nipa, nlyr, output->wl.nlambda_t,
                                                   output->atm.nq_t)) != 0)
        return status;

    }

    /* Molecular absorption optical thickness for AMF calculations */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_molabs_md_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;
    
    /* Calculate column density of air (total density of the atmosphere). */
    /* Required to calculate Rayleigh optical depth.                      */

    if (output->atm.rayleigh==RAYLEIGH_CALC) {
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1e5 converts z from km to cm.*/
	deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
      
	/* average air density over a layer assuming  dens = exp (-c * z) */
        output->atm.cdento[lc] = log_average(output->atm.microphys.dens[MOL_AIR][lc],
                                             output->atm.microphys.dens[MOL_AIR][lc+1]) * deltaz;
      }
    }
  } /* End if first block */


  if (output->atm.microphys.denstab_id) {
    /* Find matrix profile corresponding to the sza this rte calculation is done for */
    /* This is airmass factor stuff */
    iv=0;  /* No solar zenith variations for a single amf calculation */
    for (is=0; is<output->atm.microphys.nsza_denstab; is++)
      if (fabs (output->atm.microphys.sza_denstab[is]-output->atm.sza_t[iv]) < 1.e-04)
	isthis = is;

    for (lc=0; lc<nlyr; lc++) { 
      output->atm.microphys.dens[output->atm.microphys.denstab_id][lc] = 
	output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc];
    }
  }


  /***********************/
  /* Rayleigh scattering */
  /***********************/

  /* loop over wavelength */
  for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {

    for (iq=0; iq<output->atm.nq_t[iv]; iq++) {

      /* FU special: The Rayleigh optical depth depends on the subband */
      if (input.ck_scheme == CK_FU) {
#if HAVE_FULIOU
	status = ckdfuray (iv+1, iq+1, cos(output->atm.sza_t[iv]*PI/180.0), nlev,
			   fu_rayleigh);
	
	if (status!=0) {
	  fprintf (stderr, "Error, status %d returned by ckdfuray()\n", status);
	  return status;
	}    
      }
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif  
	
	
      /* loop over layers */
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1.e+5 converts z from km to cm */
	deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
	
	switch (output->atm.rayleigh) {  /* if Rayleigh scattering is not switched off */
	case RAYLEIGH_CALC:
	  if (input.ck_scheme != CK_FU) {
	    bscar = output->atm.cdento[lc] * output->crs.crs[0][MOL_AIR][iv];
	  }
	  else 
	    bscar = fu_rayleigh[lc];
	  break;
	  
	case RAYLEIGH_FILE_MONO:
	  bscar = output->atm.optprop.tau_rayleigh_user[lc];
	  break;
	  
	case RAYLEIGH_FILE_SPEC:
	  bscar = output->atm.optprop.tau_rayleigh_user_t[iv][lc];
	  break;
	  
	case RAYLEIGH_NONE:
	  bscar = 0;
	  break;
	  
	default:
	  fprintf (stderr, "Error, unknown Rayleigh option %d\n", output->atm.rayleigh);
	  return -1;
	}
	
	output->atm.optprop.tau_rayleigh_t[lc][iv][iq] = bscar;
      }
    }
  } /* End wavelength loop */


  /***********************/
  /* Absorption by gases */
  /***********************/

  /* save the original water vapour profile */
  if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
      (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) { 
    for (lc=0; lc<nlev; lc++)
      dens_h2o_org[lc] = output->atm.microphys.dens[MOL_H2O][lc];
  }

  /* loop for indipendent columns */
  for (ipa=0; ipa<output->nipa; ipa++) {
    
    /* saturate water vapour inside clouds (last argument is iv=0) */
    if (input.wcipa && input.wc.wv_saturation == SATURATION_IPA) saturate_wv_inside_waterclouds(input, output, ipa, 0);
    if (input.icipa && input.ic.wv_saturation == SATURATION_IPA) saturate_wv_inside_iceclouds  (input, output, ipa, 0);
    /* replace the changed wv profile with the original one at the end of the ic loop! */
     
    /* loop over wavelength */
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      switch(input.ck_scheme) {
      case CK_NONE:
      
        iq=0; /* no loop over quadrature points required */
        
        bscartot      = 0;
      
        babsotot      = 0; 
        babsotot_o3   = 0; 
        babsotot_co2  = 0; 
        babsotot_o2   = 0; 
        babsotot_no2  = 0; 
        babsotot_bro  = 0; 
        babsotot_oclo = 0; 
        babsotot_hcho = 0; 
        babso_amftot  = 0;
        
        vertical_column=0;
     
        /* loop over layers */
        for (lc=0; lc<nlyr; lc++) {
           
          /* The factor 1.e+5 converts z from km to cm.*/
          deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
        
          babso      = 0;
          babso_o3   = 0;
          babso_co2  = 0;
          babso_o2   = 0;
          babso_no2  = 0;
          babso_bro  = 0;
          babso_oclo = 0;
          babso_hcho = 0;
          babso_amf  = 0;
          babso_md   = 0;
           
          switch (output->atm.molabs) {
          case MOLABS_CALC: 
          
            for (gas_nr = 1; gas_nr < MOL_NN; gas_nr++) {
         
              density = output->atm.microphys.dens_avg[gas_nr][lc];
              
              switch (gas_nr) {
                case MOL_O3: 
                  babso_o3   += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_o3;
                  if (output->atm.microphys.denstab_id != MOL_O3)   babso_md += babso_o3;
                  break;
                case MOL_O2:
                  babso_o2   += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_o2;
                  if (output->atm.microphys.denstab_id != MOL_O2)   babso_md += babso_o2;
                  break;
                case MOL_CO2: 
                  babso_co2  += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_co2;
                  if (output->atm.microphys.denstab_id != MOL_CO2)  babso_md += babso_co2;
                  break;
                case MOL_NO2:
                  babso_no2  += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_no2;
                  if (output->atm.microphys.denstab_id != MOL_NO2)  babso_md += babso_no2;
                  break;
                case MOL_BRO: 
                  babso_bro  += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_bro;
                  if (output->atm.microphys.denstab_id != MOL_BRO)  babso_md += babso_bro; 
                  break;
                case MOL_OCLO:               
                  babso_oclo += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_oclo;
                  if (output->atm.microphys.denstab_id != MOL_OCLO) babso_md += babso_oclo;
                  break;
                case MOL_HCHO:
                  babso_hcho += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] + output->crs.crs[lc+1][gas_nr][iv]) * density;
                  babso += babso_hcho;
                  if (output->atm.microphys.denstab_id != MOL_HCHO) babso_md += babso_hcho;
                  break;
              }
           
              if (output->atm.microphys.denstab_id == gas_nr) {
                /* This is airmass factor stuff */
                /* Use column for matrix profile corresponding to the SZA this RTE calculation is done for */
                switch (input.atm.interpol_method_gas[gas_nr]) {
                case INTERP_METHOD_LINEAR:
                  density =         0.5 * (output->atm.microphys.denstab[gas_nr][isthis][lc] + 
                                           output->atm.microphys.denstab[gas_nr][isthis][lc+1]);
                  break;
                case INTERP_METHOD_LOG:
                  density = log_average   (output->atm.microphys.denstab[gas_nr][isthis][lc],
                                           output->atm.microphys.denstab[gas_nr][isthis][lc+1]);
                  break;
                case INTERP_METHOD_LINMIX:
                case INTERP_METHOD_SPLINE:
                case INTERP_METHOD_LOG_SPLINE: 
                  density = linmix_average(output->atm.microphys.denstab[gas_nr][isthis][lc],
                                           output->atm.microphys.denstab[gas_nr][isthis][lc+1],
                                           output->atm.microphys.dens[MOL_AIR][lc],
                                           output->atm.microphys.dens[MOL_AIR][lc+1]);
                  break;
                default:
                  fprintf (stderr, "Error: unkown z_interpol_gas option for denstab calculation \n");
                  return -1;
                  break;
                }
              
                babso_amf += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] +
                                             output->crs.crs[lc+1][gas_nr][iv]) * density;
                vertical_column += deltaz * density;
                crs_amf = output->crs.crs[lc+1][output->atm.microphys.denstab_id][iv];
              }
            }
            
            output->atm.optprop.tau_molabs_t [lc][iv][iq] = babso;
            if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                (input.icipa && input.ic.wv_saturation == SATURATION_IPA))
              output->atm.optprop.tau_molabs_t_ipa[ipa][lc][iv][iq] = 
                 output->atm.optprop.tau_molabs_t [lc][iv][iq]; 
/*             if (input.wc.wv_saturation && input.wcipa &&  */
/*                 output->atm.microphys.dens[MOL_H2O][lc]   == dens_h2o_org[lc]  &&  */
/*                 output->atm.microphys.dens[MOL_H2O][lc+1] == dens_h2o_org[lc+1]) { */
/*               tau_molabs_t_org[lc][iv][iq] = output->atm.optprop.tau_molabs_t[lc][iv][iq]; */
/* 	    } */
            output->atm.optprop.tau_molabs_md_t[lc][iv][iq] = babso_md;
            break;
            
          case MOLABS_FILE_MONO:
            babso = output->atm.optprop.tau_molabs_user[lc];
            output->atm.optprop.tau_molabs_t [lc][iv][iq] = babso;
            if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                (input.icipa && input.ic.wv_saturation == SATURATION_IPA))
              output->atm.optprop.tau_molabs_t_ipa[ipa][lc][iv][iq] = 
		 output->atm.optprop.tau_molabs_t [lc][iv][iq];
            break;
            
          case MOLABS_FILE_SPEC:
            babso = output->atm.optprop.tau_molabs_user_t[iv][lc];
            output->atm.optprop.tau_molabs_t [lc][iv][iq] = babso;
            if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                (input.icipa && input.ic.wv_saturation == SATURATION_IPA))
              output->atm.optprop.tau_molabs_t_ipa[ipa][lc][iv][iq] = 
                 output->atm.optprop.tau_molabs_t [lc][iv][iq];
            break;
            
          case MOLABS_NONE:
            babso = 0;
            output->atm.optprop.tau_molabs_t [lc][iv][iq]   = babso;
            if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                (input.icipa && input.ic.wv_saturation == SATURATION_IPA))
              output->atm.optprop.tau_molabs_t_ipa[ipa][lc][iv][iq] = 
                 output->atm.optprop.tau_molabs_t [lc][iv][iq];
            break;
             
          default:
            fprintf (stderr, "Error, unknown molecular absorption option %d\n", output->atm.molabs);
            return -1;
          }
          
          output->atm.wght_t[iv][iq] = 1;  /* default weight */
          
      
          if (input.verbose) {
            bscar=output->atm.optprop.tau_rayleigh_t[lc][iv][iq];
            bscartot+=bscar;
          
            babsotot      += babso;
            babsotot_o3   += babso_o3;
            babsotot_co2  += babso_co2;
            babsotot_o2   += babso_o2;
            babsotot_no2  += babso_no2;
            babsotot_bro  += babso_bro;
            babsotot_oclo += babso_oclo;
            babsotot_hcho += babso_hcho;
            babso_amftot  += babso_amf;
          
            if (lc == 0) {
              fprintf (stderr, "\n*** wavelength: iv = %d, %f nm\n", iv, output->wl.lambda_t[iv]);
              if (output->wl.lambda_t[iv] > 500.0)
                fprintf (stderr, "*** WARNING: water vapour and ozone absorption are inexact for wavelength above 500nm!\n");
              fprintf(stderr,"*** Rayleigh scattering cross section: %e\n", output->crs.crs[0][MOL_AIR][iv]);
              fprintf(stderr,"*** setup_gases(), layer properties, layer range = from z-level written to the z-level above \n");
              fprintf(stderr," ----------------------------------------------------------------------------------------------------------------------------------------------\n");
              fprintf(stderr,"   lc |    z[km] |  Rayleigh   |   Molecular \n");
              fprintf(stderr,"      |          |    dtau     |   absorption        o3             o2          co2           no2            bro           oclo          hcho\n");
              fprintf(stderr," ----------------------------------------------------------------------------------------------------------------------------------------------\n");
            }
            if (output->atm.zd[lc+1] >= output->alt.altitude) {
              fprintf (stderr, "%5d | %8.2f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f\n", 
                       lc, output->atm.zd[lc+1],
                       bscar, babso, babso_o3, babso_o2, babso_co2, babso_no2, babso_bro, babso_oclo, babso_hcho);
	    }
          }
        } /* End layer loop */

        if (input.verbose) {
          fprintf (stderr, " ----------------------------------------------------------------------------------------------------------------------------------------------\n");
          fprintf (stderr, "%5s | %8.2f | %11.6f | %11.6f | %12.6e| %12.6e| %12.6e| %12.6e| %12.6e| %12.6e| %12.6e\n", 
                   "sum", 0.0/0.0,
                   bscartot,
                   babsotot, babsotot_o3, babsotot_o2, babsotot_co2, babsotot_no2, babsotot_bro, babsotot_oclo, babsotot_hcho);
          fprintf (stderr, " ----------------------------------------------------------------------------------------------------------------------------------------------\n");
        }

        break;
        
      case CK_KATO:
      case CK_KATO2:
      case CK_KATO2_96:
        nq1 = output->crs_ck.profile[KATO_CO2-1][iv].ngauss;
        nq2 = output->crs_ck.profile[KATO_O3-1 ][iv].ngauss;
        nq3 = output->crs_ck.profile[KATO_O2-1 ][iv].ngauss;
        nq4 = output->crs_ck.profile[KATO_H2O-1][iv].ngauss;
      
        if (input.verbose) {
          fprintf (stderr, "*** subbands:\n");
          fprintf (stderr, "***  CO2: %2d\n", nq1);
          fprintf (stderr, "***  O3:  %2d\n", nq2);
          fprintf (stderr, "***  O2:  %2d\n", nq3);
          fprintf (stderr, "***  H2O: %2d\n", nq4);
        }
      
        for (lc=0; lc<output->atm.nlyr; lc++) {
        
          /* The factor 1e5 converts z from km to cm.*/
          deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
        
          iq=0;  /* reset quadrature counter */
          /* loop over components (4D correlated-k) */
          for (iq1=0; iq1<nq1; iq1++) {  /* CO2 */
            
            /* set CO2 cross section and weights */
            if (!(input.wcipa && input.wc.wv_saturation == SATURATION_IPA) &&
                !(input.icipa && input.ic.wv_saturation == SATURATION_IPA))
              output->crs.crs[lc][MOL_CO2][iv] = 
                output->crs_ck.profile[KATO_CO2-1][iv].crs[lc][iq1]; 
	    else 
              output->crs.crs[lc][MOL_CO2][iv] = 
                output->crs_ck.ipa_profile[ipa][KATO_CO2-1][iv].crs[lc][iq1];
          
            weight[KATO_CO2-1] = output->crs_ck.profile[KATO_CO2-1][iv].weight[iq1];
          
            for (iq2=0; iq2<nq2; iq2++) { /* O3 */
            
              /* set O3 cross section and weights */
              if (!(input.wcipa && input.wc.wv_saturation == SATURATION_IPA) &&
                  !(input.icipa && input.ic.wv_saturation == SATURATION_IPA))
                output->crs.crs[lc][MOL_O3][iv]  = 
                  output->crs_ck.profile[KATO_O3-1][iv].crs[lc][iq2];
	      else 
                output->crs.crs[lc][MOL_O3][iv]  = 
                  output->crs_ck.ipa_profile[ipa][KATO_O3-1][iv].crs[lc][iq2];
            
              weight[KATO_O3-1] = output->crs_ck.profile[KATO_O3-1][iv].weight[iq2];
            
              for (iq3=0; iq3<nq3; iq3++) { /* O2 */
              
                /* set O2 cross section and weights */
                if (!(input.wcipa && input.wc.wv_saturation == SATURATION_IPA) &&
                    !(input.icipa && input.ic.wv_saturation == SATURATION_IPA))
                  output->crs.crs[lc][MOL_O2][iv]  =
                    output->crs_ck.profile[KATO_O2-1][iv].crs[lc][iq3];
		else 
                  output->crs.crs[lc][MOL_O2][iv]  =
                    output->crs_ck.ipa_profile[ipa][KATO_O2-1][iv].crs[lc][iq3];
              
                weight[KATO_O2-1] = output->crs_ck.profile[MOL_O2][iv].weight[iq3];
              
                for (iq4=0; iq4<nq4; iq4++) { /* H2O */
        	
                  /* set H2O cross section and weights */
                  if (!(input.wcipa && input.wc.wv_saturation == SATURATION_IPA) &&
                      !(input.icipa && input.ic.wv_saturation == SATURATION_IPA))
                    output->crs.crs[lc][MOL_H2O][iv]  = 
                      output->crs_ck.profile[KATO_H2O-1][iv].crs[lc][iq4];
		  else 
                    output->crs.crs[lc][MOL_H2O][iv]  = 
                      output->crs_ck.ipa_profile[ipa][KATO_H2O-1][iv].crs[lc][iq4];
        	
                  weight[KATO_H2O-1] = output->crs_ck.profile[MOL_H2O][iv].weight[iq4];
        	
                  /* quadrature weight */
                  factor = weight[KATO_CO2-1]*weight[KATO_O3-1]*weight[KATO_O2-1]*weight[KATO_H2O-1];
                
                   
                  /* sum absorption */
                  babso = 0;
                
                  /* O3 absorption */
                  babso += deltaz * output->crs.crs[lc][MOL_O3][iv]  * output->atm.microphys.dens_avg[MOL_O3][lc];
                
                  /* O2 absorption */
                  babso += deltaz * output->crs.crs[lc][MOL_O2][iv]  * output->atm.microphys.dens_avg[MOL_O2][lc];
                 
                  /* H2O absorption */
                  babso += deltaz * output->crs.crs[lc][MOL_H2O][iv] * output->atm.microphys.dens_avg[MOL_H2O][lc];
                  
                  /* CO2 absorption */
                  babso += deltaz * output->crs.crs[lc][MOL_CO2][iv] * output->atm.microphys.dens_avg[MOL_CO2][lc];
                  
                  /* copy absorption cross section and weight */
                  /* to final destination                     */
                  output->atm.wght_t[iv][iq] = factor;
                
                  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
                  if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                      (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) 
                    output->atm.optprop.tau_molabs_t_ipa[ipa][lc][iv][iq] = 
                       output->atm.optprop.tau_molabs_t[lc][iv][iq];
                  iq++;
                
                }  /* end of correlated-k (iq4) loop */
              }  /* end of correlated-k (iq3) loop */
            }  /* end of correlated-k (iq2) loop */
          }  /* end of correlated-k (iq1) loop */
        } /* End layer loop */
  
        break;
        
      case CK_FU:
      case CK_AVHRR_KRATZ:
      case CK_LOWTRAN:
      case CK_GENERIC:
      
        if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
            (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {
          nq = output->crs_ck.ipa_profile[ipa][0][iv].ngauss;
	}
        else
          nq = output->crs_ck.profile[0][iv].ngauss;

        /* loop over quadrature points */
      
        for (iq=0; iq<nq; iq++) {
        
          /* set absorption coefficient */
          for (lc=0; lc<output->atm.nlyr; lc++) {
            
            if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                (input.icipa && input.ic.wv_saturation == SATURATION_IPA))
              output->crs_ck.profile[0][iv].crs[lc][iq] = output->crs_ck.ipa_profile[ipa][0][iv].crs[lc][iq];
   
            /* The factor 1e5 converts z from km to cm.*/
            deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
          
            output->atm.optprop.tau_molabs_t[lc][iv][iq] =
              deltaz/1e5 * output->crs_ck.profile[0][iv].crs[lc][iq];
                   
            if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
                (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {
              output->atm.optprop.tau_molabs_t_ipa[ipa][lc][iv][iq] = output->atm.optprop.tau_molabs_t[lc][iv][iq];
	    }
          }
          
          /* quadrature weight */
          if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
              (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {
            factor = output->crs_ck.ipa_profile[ipa][0][iv].weight[iq];
            output->atm.ipa_wght_t[ipa][iv][iq] = factor;
	  }
          else {
            factor = output->crs_ck.profile[0][iv].weight[iq];
            output->atm.wght_t[iv][iq] = factor;
	  }
        }
      
        break;
      
      default:
        fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
        return -1;
        break;
      }
    } /* End wavelength loop */

    /* replace the wv-dens and rel. hum. with the original profiles again */
    /* this is important for the next call of saturate_wv_inside_clouds() */
    if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
        (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) { 
      for (lc=0; lc<nlev; lc++)
        output->atm.microphys.dens[MOL_H2O][lc] = dens_h2o_org[lc];
    }

  } /* End of indipendent pixel (ipa) loop */


  free(fu_rayleigh);
  if ((input.wcipa && input.wc.wv_saturation == SATURATION_IPA) ||
      (input.icipa && input.ic.wv_saturation == SATURATION_IPA)) {
    free(dens_h2o_org);
    /* free(tau_molabs_t_org); */
  }

  return 0;  /* ok */
}



/**************************************************************/
/* Setup temperature at output levels.                        */
/**************************************************************/

int setup_temperature (input_struct input, output_struct *output)
{
  int status=0, lev=0, iz=0;
  float *zout=NULL;

  /* interpolate output temperature to zout */
 
  output->atm.temper_out  = (float *) calloc (output->atm.nlev, sizeof(float));
  
  /* copy temperature profile */
  for (lev=0; lev<output->atm.nlev; lev++) {
    output->atm.temper_out[lev] = output->atm.microphys.temper [lev];
  }
 
  /* Now we need to define a temporary zout grid;        */
  /* for rte_solver montecarlo it is possible to         */
  /* define an output altitude of -999 which stands      */
  /* for the surface altitude which may be a function    */
  /* of (x,y). This value is replaced by 0 which means   */
  /* that the surface temperature at the surface output  */
  /* level is set to temperature(z=0) which is not quite */
  /* correct if a 2D altitude grid has been defined.     */
  /* However, the only place where this information is   */
  /* currently used is the flexstor header which is      */
  /* required for photolysis calculations.               */

  zout = calloc (output->atm.nzout, sizeof(float));
  for (iz=0; iz<output->atm.nzout; iz++) {
    if (output->atm.zout_sea[iz] != -999)
      zout[iz] = output->atm.zout_sea[iz];
    else
      zout[iz] = 0;
  }
  
  /* interpolate zout_temperature to zout-grid with variable interpolation method */
  status += interpolate_profile (output->atm.zd , &(output->atm.temper_out), output->atm.nlev,
                                 zout, input.atm.nzout, input.atm.interpol_method_temper, 
                                 input.quiet);

  if (status != 0) {
    fprintf (stderr, "Error interpolating temperature to output grid\n");
    return status;
  }

  free(zout);


  /* Finally, copy the temperature of the lowest level to */
  /* output->temperature if the latter is not already set */
  /* ATTENTION: moved this code to redistribute.c because */
  /* we want the surface temperature to equal the         */
  /* z=altitude rather than z=0                           */

  /*
  if (output->surface_temperature < 0)
    output->surface_temperature = output->atm.microphys.temper[output->atm.nlev-1];
  */

  return 0;
}




/**************************************************************/
/* Read atmosphere file and write to atm_out_struct           */
/**************************************************************/

static int read_atmosphere (char *filename, atm_out_struct *out, 
                            int interpol_method_press, int interpol_method_temper, int *interpol_method_gas,
                            int cdf, int quiet, int verbose) 

{
  int rows=0, min_columns=0, max_columns=0;
  int lev=0, status = 0;
  float **atm_data=NULL;
  double boltzmann = 1.38065e-23;
  
  int nlev=0, nlyr=0;

  float tmpdens=0, idealdiff=0;
  int idealgas=0, firstnonideal=0;
  float *z_atm=NULL;

#if HAVE_LIBNETCDF
  int i=0;
  int ncid=0;
  int idd_nlev=0;
  int id_z=0, id_press=0, id_temper=0;
  int id_air=0, id_o3=0, id_o2=0, id_h2o=0, id_co2=0;
  
  size_t start[1] = {0};
  size_t count[1] = {0};
  size_t dimlen=0;
#endif

  if (cdf==0) {   /* classic ASCII atmospheric model */
    status = ASCII_file2float (filename, 
			       &rows, &max_columns, &min_columns, 
			       &atm_data);

    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    if (max_columns!=min_columns) {
      fprintf (stderr, "*** Warning: Inconsistent number of columns in %s\n", filename);
      fprintf (stderr, "*** min = %d, max =%d\n", min_columns, max_columns);
      return -1;
    }
    
    if (min_columns<3) {
      fprintf (stderr, "Error, too few columns in %s! Need at least %d.\n", filename, 3);
      return -1;
    }
        
    nlev = rows;
    nlyr = rows-1;
  
    /* allocate memory for density array */
    out->microphys.dens = calloc(MOL_NN, sizeof(float *));
    if (out->microphys.dens == NULL) {
      fprintf (stderr, "Error allocating memory for dens\n");
      return -1;
    }
    
    out->microphys.press  = ASCII_column_float (atm_data, nlev, ATMFILE_PRESS);
    out->microphys.temper = ASCII_column_float (atm_data, nlev, ATMFILE_TEMPER);
    z_atm                 = ASCII_column_float (atm_data, nlev, ATMFILE_ZD);

    /* add atmosphere z-grid to the common grid out->... */
    set_common_z_grid (z_atm, nlev, &(out->zd), &(out->nlev)); /* in redistribute.c */


    /**************************************/
    /* the following columns are optional */
    /**************************************/

    if (min_columns>3) { /* air */
      out->microphys.dens[MOL_AIR] = ASCII_column_float(atm_data, nlev, ATMFILE_AIR);
    }
    else {
      out->microphys.dens[MOL_AIR] = (float *) calloc (nlev, sizeof(float));
      
      /* Set air density */
      for (lev=0; lev<nlev; lev++)
	out->microphys.dens[MOL_AIR][lev] = out->microphys.press[lev]*1E-04/(boltzmann*out->microphys.temper[lev]);
    }
    
    if (min_columns>4) /* o3   */
      out->microphys.dens[MOL_O3]  = ASCII_column_float(atm_data, nlev, ATMFILE_O3);
    else
      out->microphys.dens[MOL_O3]  = (float *) calloc (nlev, sizeof(float));
    
    if (min_columns>5)  /* oxygen           */
      out->microphys.dens[MOL_O2]  = ASCII_column_float(atm_data, nlev, ATMFILE_O2);
    else 
      out->microphys.dens[MOL_O2]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>6)  /* water vapour     */
      out->microphys.dens[MOL_H2O] = ASCII_column_float(atm_data, nlev, ATMFILE_H2O);
    else 
      out->microphys.dens[MOL_H2O]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>7)  /* carbon dioxide   */
      out->microphys.dens[MOL_CO2] = ASCII_column_float(atm_data, nlev, ATMFILE_CO2);
    else 
      out->microphys.dens[MOL_CO2]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>8)  /* nitrogen dioxide */
      out->microphys.dens[MOL_NO2] = ASCII_column_float(atm_data, nlev, ATMFILE_NO2);
    else 
      out->microphys.dens[MOL_NO2]  = (float *) calloc (nlev, sizeof(float));

    /* Allocate memory for species not in atmosphere file */
    out->microphys.dens[MOL_BRO]  = (float *) calloc (nlev, sizeof(float));
    out->microphys.dens[MOL_OCLO] = (float *) calloc (nlev, sizeof(float));
    out->microphys.dens[MOL_HCHO] = (float *) calloc (nlev, sizeof(float));

    ASCII_free_float(atm_data, rows);
  }
  else {

#if HAVE_LIBNETCDF

    /* open netcdf file */
    status = nc_open (filename, NC_NOWRITE, &ncid);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
      return status;
    }

    /* get dimension id for "nlev" */
    status = nc_inq_dimid (ncid, "nlev", &idd_nlev);
  
    /* get dimension length for "nlev" */
    status = nc_inq_dimlen (ncid, idd_nlev, &dimlen);
    nlev = dimlen;
    nlyr = dimlen-1;


    start[0] = 0;
    count[0] = nlev;

    /* allocate memory for profiles */
    out->microphys.press  = calloc (nlev, sizeof (float));
    out->microphys.temper = calloc (nlev, sizeof (float));
    z_atm                 = calloc (nlev, sizeof (float));

    /* get variable id for "z" */
    status = nc_inq_varid (ncid, "z", &id_z);
    
    /* read altitude "z" */
    status = nc_get_vara_float (ncid, id_z, start, count, z_atm);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading z from %s\n", status, filename);
      return status;
    }

    /* add atmosphere z-grid to the common grid out->... */
    set_common_z_grid (z_atm, nlev, &(out->zd), &(out->nlev));

    /* get variable id for "pressure" */
    status = nc_inq_varid (ncid, "pressure", &id_press);
    
    /* read "pressure" */
    status = nc_get_vara_float (ncid, id_press, start, count, out->microphys.press);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading pressure from %s\n", status, filename);
      return status;
    }


    /* get variable id for "temperature" */
    status = nc_inq_varid (ncid, "temperature", &id_temper);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading temperature from %s\n", status, filename);
      return status;
    }
    
    /* read "temperature" */
    status = nc_get_vara_float (ncid, id_temper, start, count, out->microphys.temper);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading temperature from %s\n", status, filename);
      return status;
    }


    /* allocate memory for density array */
    out->microphys.dens = calloc(MOL_NN, sizeof(float *));
    if (out->microphys.dens == NULL) {
      fprintf (stderr, "Error allocating memory for dens\n");
      return -1;
    }

    for (i=0; i<MOL_NN; i++) {
      out->microphys.dens[i] = calloc(nlev, sizeof(float));
      if (out->microphys.dens[i] == NULL) {
	fprintf (stderr, "Error allocating memory for dens[%d]\n", i);
	return -1;
      }
    }


    /* get variable id for "density" */
    status = nc_inq_varid (ncid, "density", &id_air);
    
    /* read "density" */
    status = nc_get_vara_float (ncid, id_air, start, count, out->microphys.dens[MOL_AIR]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading density from %s\n", status, filename);
      return status;
    }


    /* get variable id for "o3" */
    status = nc_inq_varid (ncid, "o3", &id_o3);
    
    /* read "o3" */
    status = nc_get_vara_float (ncid, id_o3, start, count, out->microphys.dens[MOL_O3]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading o3 from %s\n", status, filename);
      return status;
    }


    /* get variable id for "o2" */
    status = nc_inq_varid (ncid, "o2", &id_o2);
    
    /* read "o2" */
    status = nc_get_vara_float (ncid, id_o2, start, count, out->microphys.dens[MOL_O2]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading o2 from %s\n", status, filename);
      return status;
    }


    /* get variable id for "h2o" */
    status = nc_inq_varid (ncid, "h2o", &id_h2o);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading h2o from %s\n", status, filename);
      return status;
    }
    
    /* read "h2o" */
    status = nc_get_vara_float (ncid, id_h2o, start, count, out->microphys.dens[MOL_H2O]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading h2o from %s\n", status, filename);
      return status;
    }


    /* get variable id for "co2" */
    status = nc_inq_varid (ncid, "co2", &id_co2);
    
    /* read "co2" */
    status = nc_get_vara_float (ncid, id_co2, start, count, out->microphys.dens[MOL_CO2]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading co2 from %s\n", status, filename);
      return status;
    }


    /* we don't have no2 in the netcdf file, hence we keep it 0;        */
    /* also we don't read n2o, co, and ch4 because we never used        */
    /* those before (should change that by introducing all relevant     */
    /* gases into dens; should also get rid of ndens  */
    /* and replace that by MOL_NN)                                      */

    nc_close (ncid);
    
#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif

  }

  
  /* interpolate the atmospheric profiles to the grid */
  status += interpolate_atmosphere (z_atm,
                                    &(out->microphys.press),
                                    &(out->microphys.temper),
                                    &(out->microphys.dens),
                                    nlev,
                                    out->zd, out->nlev,
                                    interpol_method_press,
                                    interpol_method_temper,
                                    interpol_method_gas,
                                    quiet);
  if (status!=0) {
    fprintf (stderr, "Error during interpolate_atmosphere in read_atmosphere (atmosphere.c)\n");
    return status;
  }

  /* Finally, check if profiles of pressure, temperature and air density are consistent */
  /* with ideal gas law over 6 orders of magnitude in pressure; check only first six    */
  /* orders of magnitude because for very small pressures roundoff errors due to too    */
  /* few digits cause problems                                                          */
  firstnonideal=1;
  for (lev=0; lev<out->nlev; lev++) {

    idealgas=1;

    if (out->microphys.press[lev]/out->microphys.press[nlev-1]>1e-6) {
      tmpdens = out->microphys.press[lev]*1E-04 / (boltzmann*out->microphys.temper[lev]);
      
      if ((tmpdens==0 && out->microphys.dens[MOL_AIR][lev]!=0) || 
	  (tmpdens!=0 && out->microphys.dens[MOL_AIR][lev]==0))
	idealgas=0;
      
      if (out->microphys.dens[MOL_AIR][lev]>0)
	if ((idealdiff = fabs ((tmpdens - out->microphys.dens[MOL_AIR][lev]) / out->microphys.dens[MOL_AIR][lev])) > 0.01)
	  idealgas=0;
      
      if (!idealgas && firstnonideal) {
	firstnonideal=0;
	fprintf (stderr, "\n");
	fprintf (stderr, "*** ATTENTION: pressure, temperature, and air density not\n");
	fprintf (stderr, "*** consistent with ideal gas law at the following levels\n");
	fprintf (stderr, "*** (assuming a Boltzmann constant %.6e):\n", boltzmann);
	fprintf (stderr, "***    z [km]  density [cm-3]  p/kT [cm-3]  difference\n");
      }
      
      if (!idealgas) {
	fprintf (stderr, "***   %7.2f  %.4e      %.4e    %6.1f%%\n", 
		 out->zd[lev], out->microphys.dens[MOL_AIR][lev], tmpdens, idealdiff*100.0);
      }
    }
  }
  
  if (!firstnonideal) {
    fprintf (stderr, "*** Please check pressure, temperature, and air density in\n");
    fprintf (stderr, "*** %s for consistency!\n", filename);
    fprintf (stderr, "\n");
    return -1;
  } 

  return 0;
}


/******************************************************************/
/* Read radiosonde file                                           */
/* and convert p and T data to a z-grid                           */
/* additional option to manipulate atmosphere profiles UH 2006-06 */
/******************************************************************/

static int read_radiosonde_file (input_struct input, output_struct *output, double boltzmann)

{
  int rows=0, max_columns=0, min_columns=0;
  float **atm_data = NULL;
  int     n_rs     = 0;
  float  *p_rs     = NULL;
  float  *T_rs     = NULL;  
  float  *z_rs     = NULL;
  float  *ln_p_rs  = NULL;
  float  *ln_p_atm = NULL;
  float **dens_rs  = NULL;
  int     gas_nr   = 0;
  int     lc       = 0;
  int     status   = 0;
  float  *gas_rs   = NULL;
  double  avogadro = 6.022142e23;
  double  R        = 0;
  float   g0       = 9.8;
/*   float   g0       = 9.80665; */
  float   g        = 0;
  float   r_earth  = 6371000.0; /* radius of the earth in m */
  int     i        = 0;
  char   *gas      = NULL;
  int     n_comb   = 0;
  int     n_above  = 0;
  float  *z_comb   = NULL;
  float  *p_comb   = NULL;
  float  *T_comb   = NULL;  
  float **dens_comb= NULL;
  int     n_outside= 0;
  float   dz       = 0;
  float   tmpdens  = 0.0;
  float   scale_factor = 0.0;

  if (input.verbose)
    fprintf (stderr," ... reading radiosonde data file %s\n", input.atm.rs_filename);

  /* read radiosonde */
  status = ASCII_file2float (input.atm.rs_filename, 
                             &rows, &max_columns, &min_columns, 
                             &atm_data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, input.atm.rs_filename);
    return status;
  }

  /* reading data from file */
  if (max_columns!=(2+input.atm.n_rs_gas) || min_columns!=(2+input.atm.n_rs_gas)) {
    fprintf (stderr, "Error: Inconsistent number of columns in %s\n", input.atm.rs_filename);
    fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    fprintf (stderr, "     Radiosonde data is expected to have 2 + %d columns\n", input.atm.n_rs_gas);
    fprintf (stderr, "     First pressure [hPa], second temperature [K], and specified gas densities\n");
    return -1;
  }

  n_rs  = rows;

  /* copy radiosonde p-data to p_rs */
  p_rs = ASCII_column_float (atm_data, n_rs, 0);

  /* copy radiosonde T-data to T_rs */
  T_rs = ASCII_column_float (atm_data, n_rs, 1);

  /* Allocate z_rs */
  if ((z_rs = (float *) calloc (n_rs, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for z_rs in read_radiosonde_file (in atmosphere.c)\n");
    return -2;
  }

  R = boltzmann * avogadro * 1000. / input.atm.mol_mass[MOL_AIR];
  if (input.verbose) {
    fprintf (stderr," ...  converting p and T to a z-grid\n");
    fprintf (stderr," ...  assuming: R = %12.6f, g(z=0km) = %12.6f\n",R,g0); 
  }

  /* initialise lowest level */
  if (output->alt.altitude != 0) { 
    z_rs[n_rs-1] = 0.0;
    g = g0;
  }
  else {
    z_rs[n_rs-1] = output->alt.altitude;
    g = g0 * (r_earth * r_earth) / ((r_earth+z_rs[n_rs-1]*1000.)*(r_earth+z_rs[n_rs-1]*1000.));
  }

  /* calculate radiosonde z-grid (hydrostatic equation) */
  for (lc=n_rs-1;lc>=1;lc--) {

    /* equation for no variation of g inside one layer */
    /*  z_rs[lc-1]=z_rs[lc] + 0.001 * R/g*0.5*(T_rs[lc]+T_rs[lc-1])*log(p_rs[lc]/p_rs[lc-1]);     */
    /*                        /\* 0.001 <==> m -> km *\/                                          */
    /*  g = g0 * (r_earth * r_earth) / ((r_earth+1000.0*z_rs[lc-1])*(r_earth+1000.0*z_rs[lc-1])); */

    /* equation taking care of variation of g inside one layer */
    z_rs[lc-1]= - r_earth + 1/
		  (1/(r_earth+1000.0*z_rs[lc])
                    - R/(g0*r_earth*r_earth)*0.5*(T_rs[lc]+T_rs[lc-1])*log(p_rs[lc]/p_rs[lc-1]));
    z_rs[lc-1]*=0.001;  /* m -> km */

   /* addiotional verbose output */
   /*  fprintf (stderr,"dz1 = %8.4f, dz2 = %8.4f\n", */
   /*                   0.001 * R/g*0.5*(T_rs[lc]+T_rs[lc-1])*log(p_rs[lc]/p_rs[lc-1]), */
   /*                   z_rs[lc-1]-z_rs[lc]); */
  }


  /* calculate new densities on the vertical grid of the radiosonde data */

  /* vertical ln(p/p0)-grid for background atmosphere */
  if ((ln_p_atm = (float *) calloc (output->atm.nlev, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for ln_p_atm in read_radiosonde_file (in atmosphere.c)\n");
    return -3;
  }
  for (lc=0; lc<output->atm.nlev;lc++)
    ln_p_atm[lc]=-log(output->atm.microphys.press[lc]/output->atm.microphys.press[output->atm.nlev-1]);

  /* vertical ln(p/p0)-grid for radiosonde */
  if ((ln_p_rs = (float *) calloc (n_rs, sizeof (float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for ln_p_rs in read_radiosonde_file (in atmosphere.c)\n");
    return -4;
  }
  for (lc=n_rs-1; lc>=0;lc--) {
    if (p_rs[lc]>=output->atm.microphys.press[0])
      ln_p_rs[lc]=-log(p_rs[lc]/p_rs[n_rs-1]);
    else
      n_outside = lc+1;
  }


  /* Allocate dens_rs */
  if ((dens_rs = (float **) calloc (MOL_NN, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for dens_rs in read_radiosonde_file (in atmosphere.c)\n");
    return -5;
  }

  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++) {
    if ((dens_rs[gas_nr] = (float *) calloc (output->atm.nlev, sizeof (float))) == NULL) {
      fprintf (stderr,"Error, allocating memory for dens_rs[%d] in read_radiosonde_file (in atmosphere.c)\n",gas_nr);
      return -6;
    }
  }

  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
    for (lc=0; lc<output->atm.nlev;lc++)
      dens_rs[gas_nr][lc] = output->atm.microphys.dens[gas_nr][lc]; 

  /* Interpolate dens to the vertical grid of the radiosonde */
  /* AIR (this is not 100% consistent with interpolation of p and T)*/
  status += interpolate_profile (ln_p_atm, &(dens_rs[MOL_AIR]), output->atm.nlev, 
                                 ln_p_rs+n_outside, n_rs-n_outside, input.atm.interpol_method_gas[MOL_AIR], input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating dens_rs[air] in read_radiosonde_file (in atmosphere.c)\n", status);
    return status;
  }
    

  /* Ozone, O2, water vapour, CO2, NO2, BRO, OClO, HCHO */
  /* considering mixing ratios, they have to be interpolate with dens[MOL_AIR] from atmosphere file */
  /* as they are also densities from the atmosphere file */
  for (gas_nr=0; gas_nr<MOL_NN; gas_nr++) {
    if (gas_nr != MOL_AIR)
      status += interpolate_density (ln_p_atm, &(dens_rs[gas_nr]), output->atm.nlev, 
                                     ln_p_rs+n_outside, n_rs-n_outside, input.atm.interpol_method_gas[gas_nr],
                                     output->atm.microphys.dens[MOL_AIR], dens_rs[MOL_AIR], input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating dens_rs[%d] in read_radiosonde_file (in atmosphere.c)\n", status, gas_nr);
      return status;
    }
  }

  /* not nice, but working, shift the first (n_rs-n_outside) points and forget about the rest */
  if (n_outside > 0) {
    for (lc=0; lc<n_rs-n_outside;lc++) {
      z_rs[lc] = z_rs[lc+n_outside];
      p_rs[lc] = p_rs[lc+n_outside];
      T_rs[lc] = T_rs[lc+n_outside];
    }
  }

  /* p and T is from the radiosonde data, but dens_air is from the atmosphere */
  /* this will violate the ideal gas law */
  /* therefor all atmospheric densities will be scaled accordingly */

  for (lc=0; lc<n_rs-n_outside; lc++) {
    tmpdens = p_rs[lc] * 1E-04 / (boltzmann * T_rs[lc]);
    scale_factor = tmpdens/dens_rs[MOL_AIR][lc];
    for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
      dens_rs[gas_nr][lc] *= scale_factor;
  }


  /* if there are any gas profiles provided by the radiosonde file  */
  /* take them to replace the standard profile                      */
  /* if they are given as mixing ratios, they have to be multiplied */
  /* with a dens_air derived from the radiosonde data (p, T)        */
  for (i=0; i < input.atm.n_rs_gas; i++) {
    gas_nr = input.atm.rs_gas[i];
    gas = gas_number2string(gas_nr);
    if (input.verbose)
      fprintf (stderr, " ... Reading %s profile from radiosonde data, unit=%d (1=cm-3, 2=m-3, 3=MMR, 4=VMR, 5=rel.hum.)\n",
                        gas, input.atm.rs_unit[gas_nr]);

    gas_rs = ASCII_column_float (atm_data, n_rs, i + 2);

    for (lc=0; lc<n_rs-n_outside;lc++) 
      dens_rs[gas_nr][lc] = gas_rs[lc+n_outside];

    /* convert units of the dens_file to cm-3 */
    status =  convert_profile_units (&(dens_rs[gas_nr]), z_rs, (n_rs-n_outside), gas_nr,
                                     dens_rs[MOL_AIR], T_rs, z_rs, (n_rs-n_outside),
                                     input.atm.interpol_method_gas[MOL_AIR], input.atm.interpol_method_temper,
                                     input.atm.rs_unit[gas_nr], input.atm.mol_mass, input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d during convert_profile_unit in read_radiosonde (atmosphere.c)\n", status);
      return status;
    }

/*     for (lc=0; lc<n_rs-n_outside;lc++) { */
/*       fprintf (stderr, "z=%6.3f, p=%12.6f, T=%8.3f, gas=%12.6e \n",z_rs[lc],p_rs[lc],T_rs[lc],dens_rs[gas_nr][lc]); */
/*     } */

  }

  n_rs = n_rs-n_outside;
  free(atm_data);


  /* now we have a complete new set of variables on the rs grid */

  /* fprintf(stderr,"radiosonde data completed with data from atmosphere file\n"); */
  /* fprintf (stderr," lc    z           p       T         air           O3            O2            ");*/
  /* fprintf (stderr,"H2O            CO2           NO2           BRO          OCLO         HCHO\n"); */
  /* for (lc=0; lc<n_rs;lc++) */
  /*   fprintf (stderr,"%3d %7.3f %11.5f %7.2f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e\n", */
  /*                     lc, z_rs[lc], p_rs[lc], T_rs[lc], */
  /*                     dens_rs[0][lc],dens_rs[1][lc],dens_rs[2][lc],dens_rs[3][lc],dens_rs[4][lc], */
  /*                     dens_rs[5][lc],dens_rs[6][lc],dens_rs[7][lc],dens_rs[8][lc]); */


  /* combine radiosonde profiles and background profile above */
  for (lc=0;lc<output->atm.nlev;lc++)
    if (output->atm.microphys.press[lc]>=p_rs[0])
      break;

  n_above = lc;
  n_comb  = n_above + n_rs;

  /* Allocate memory for combined profiles */
  if ((z_comb = (float *) calloc (n_comb, sizeof(float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for z_comb in read_radiosonde_file (in atmosphere.c)\n");
    return -7;
  }
  if ((p_comb = (float *) calloc (n_comb, sizeof(float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for z_comb in read_radiosonde_file (in atmosphere.c)\n");
    return -8;
  }
  if ((T_comb = (float *) calloc (n_comb, sizeof(float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for z_comb in read_radiosonde_file (in atmosphere.c)\n");
    return -9;
  }

  if ((dens_comb = (float **) calloc (MOL_NN, sizeof (float *))) == NULL) {
    fprintf (stderr,"Error, allocating memory for dens_comb in read_radiosonde_file (in atmosphere.c)\n");
    return -10;
  }

  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++) {
    if ((dens_comb[gas_nr] = (float *) calloc (n_comb, sizeof (float))) == NULL) {
      fprintf (stderr,"Error, allocating memory for dens_comb[%d] in read_radiosonde_file (in atmosphere.c)\n",gas_nr);
      return -11;
    }
  }


/*   fprintf (stderr,"combine profiles, n_comb = %3d \n", n_comb); */
/*   fprintf (stderr," lc    z           p       T         air           O3            O2            H2O            CO2           NO2           BRO          OCLO         HCHO\n"); */

  /* points inside the radiosonde profile, starting from the ground (lc = n_comb-1) */
  for (lc=n_comb-1;lc>(n_comb-1-n_rs);lc--) {
    p_comb[lc] = p_rs[lc-(n_above)];
    T_comb[lc] = T_rs[lc-(n_above)];
    z_comb[lc] = z_rs[lc-(n_above)];
    for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
      dens_comb[gas_nr][lc] = dens_rs[gas_nr][lc-(n_above)];
/*     fprintf (stderr,"%3d %7.3f %11.5f %7.2f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e\n", */
/*                       lc, z_comb[lc], p_comb[lc], T_comb[lc], */
/*                       dens_comb[0][lc],dens_comb[1][lc],dens_comb[2][lc],dens_comb[3][lc],dens_comb[4][lc], */
/*                       dens_comb[5][lc],dens_comb[6][lc],dens_comb[7][lc],dens_comb[8][lc]); */
  }

  if (n_comb-1-n_rs >= 0) {

    /* transition layer to upper part of the atmosphere, given by background atmosphere */
/*     fprintf (stderr,"tansition layer %3d\n", n_comb-1-n_rs); */
    lc=n_comb-1-n_rs;
    p_comb[lc] = output->atm.microphys.press [n_above-1];
    T_comb[lc] = output->atm.microphys.temper[n_above-1];

    z_comb[lc] = z_comb[lc+1]+ 0.001 * 0.5 * (T_rs[0]+output->atm.microphys.temper[n_above-1])
                                 * R/g * log(p_rs[0]/output->atm.microphys.press[n_above-1]);

/*     fprintf (stderr,"between:lc=%3d, p_rs=%12.6f, p_atm=%12.6f, T_rs=%7.2f, T_atm=%7.2f, z_below=%7.3f, z_new=%7.3f\n", */
/*               lc,p_rs[0],output->atm.microphys.press[n_above-1],T_rs[0],output->atm.microphys.temper[n_above-1],z_comb[lc+1],z_comb[lc]);  */
        
    /* adjustment of the new z-grid */
    dz = output->atm.zd[n_above-1]-z_comb[lc];
    /* fprintf (stderr,"z_atm=%7.3f, z_rs=%7.3f, dz=%7.3f\n",output->atm.zd[n_above-1],z_comb[lc],dz); */

    for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
      dens_comb[gas_nr][lc] = output->atm.microphys.dens[gas_nr][n_above-1];
    
/*     fprintf (stderr,"%3d %7.3f %11.5f %7.2f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e\n", */
/*                      lc, z_comb[lc], p_comb[lc], T_comb[lc], */
/*                      dens_comb[0][lc],dens_comb[1][lc],dens_comb[2][lc],dens_comb[3][lc],dens_comb[4][lc], */
/*                      dens_comb[5][lc],dens_comb[6][lc],dens_comb[7][lc],dens_comb[8][lc]); */

    /* points above the radiosonde data are copied from the background atmosphere */
    /* but height is shifted by dz */
    for (lc=n_above-2;lc>=0;lc--) {
      if (output->atm.zd[lc]-dz < output->atm.zd[0]) {
        p_comb[lc] = output->atm.microphys.press[lc];
        T_comb[lc] = output->atm.microphys.temper[lc];
        z_comb[lc] = output->atm.zd[lc]-dz;
        for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
          dens_comb[gas_nr][lc] = output->atm.microphys.dens[gas_nr][lc];
/*         fprintf (stderr,"%3d %7.3f %11.5f %7.2f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e\n", */
/*                         lc, z_comb[lc], p_comb[lc], T_comb[lc], */
/*                         dens_comb[0][lc],dens_comb[1][lc],dens_comb[2][lc],dens_comb[3][lc],dens_comb[4][lc], */
/*                         dens_comb[5][lc],dens_comb[6][lc],dens_comb[7][lc],dens_comb[8][lc]); */
      } 
      else {
        n_outside = lc+1;
        if (input.verbose)
          fprintf (stderr,"exceeded upper edge of the atmosphere: %6.2f km, %3d points left \n",output->atm.zd[0], n_outside);
      }
    }
  }

  /* radiosonde arrays have done their job */
  free(p_rs);
  free(T_rs);
  free(z_rs);
  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
    free(dens_rs[gas_nr]);
  free(dens_rs);


  /* copy data to final destination */

  /* reallocate output->atm.zd */
  free(output->atm.zd);
  free(output->atm.microphys.press);
  free(output->atm.microphys.temper);
  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
    free(output->atm.microphys.dens[gas_nr]);

  if ((output->atm.zd               = (float *) calloc ((n_comb-n_outside), sizeof(float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for output->atm.zd in read_radiosonde_file (in atmosphere.c)\n");
    return -12;
  }

  if ((output->atm.microphys.press  = (float *) calloc ((n_comb-n_outside), sizeof(float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for output->atm.microphys.press in read_radiosonde_file (in atmosphere.c)\n");
    return -13;
  }

  if ((output->atm.microphys.temper = (float *) calloc ((n_comb-n_outside), sizeof(float))) == NULL) {
    fprintf (stderr,"Error, allocating memory for output->atm.microphys.temper in read_radiosonde_file (in atmosphere.c)\n");
    return -14;
  }

  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
    if ((output->atm.microphys.dens[gas_nr] = (float *) calloc ((n_comb-n_outside), sizeof(float))) == NULL) {
      fprintf (stderr,"Error, allocating memory for output->atm.zd in read_radiosonde_file (in atmosphere.c)\n");
      return -15;
    }

  /* copy all to final destination */

  output->atm.nlev = (n_comb-n_outside);

  for (lc=n_outside;lc<n_comb;lc++) {
    output->atm.zd                      [lc-n_outside] = z_comb[lc];
    output->atm.microphys.press         [lc-n_outside] = p_comb[lc];
    output->atm.microphys.temper        [lc-n_outside] = T_comb[lc];
    for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
      output->atm.microphys.dens[gas_nr][lc-n_outside] = dens_comb[gas_nr][lc];
  }

  /* combined arrays have done their job */
  free(z_comb);
  free(p_comb);
  free(T_comb);
  for (gas_nr = 0; gas_nr < MOL_NN; gas_nr++)
    free(dens_comb[gas_nr]);
  free(dens_comb);

  /* if read_radionsonde() is call after setup_clouds and saturate_clouds is activated, */
  /* add cloud levels here and interpolate_atmoshere() */

  return status;
}


/**************************************************************/
/* Read density file                                          */
/* Combine with background atmosphere and                     */
/* Interpolate on the common grid: z_all                      */
/**************************************************************/

static int read_and_combine_density (char *filename, atm_out_struct *out,
				     int dens_id, int *interpol_method_gas, int interpol_method_temper, 
                                     int unit_profile, float *mol_mass, 
				     float *dens_air_all, float *temper_all, float *z_all, int nlev_all,
                                     float *dens_air_atm, float *z_atm, int nlev_atm,
                                     int solver, int quiet, int verbose)

{
  int rows=0, min_columns=0, max_columns=0, nlev=0;
  int status = 0;
  float **atm_data = NULL;
  int lc           = -666;
  int i            = -666; 
  float *zd        = NULL;
  float **dens     = NULL;
  static int first_dens = 1;
  float *z_comb         = NULL;
  float **dens_comb     = NULL;
  int nlev_comb         = -666;
  float *dens_air_comb  = NULL;
  int iv=0, is=0, isthis=0;

  if ((strlen(filename) == 0) && (out->nlev == nlev_all)) {
    /* nothing to do, everything is fine */
  }
  else { /* something to do */
    
    /* densfile specified in the input file */
    if (strlen(filename) > 0) {

      if (verbose)
        fprintf (stderr," ... reading density file %s\n", filename);

      /* read densfile */
      status = ASCII_file2float (filename, 
                                 &rows, &max_columns, &min_columns, 
                                 &atm_data);
  
      if (status!=0) {
        fprintf (stderr, "Error %d reading %s\n", status, filename);
        return status;
      }

      if (max_columns==2) {   /* 2-columns: 1 altitude, 2 density */
    
        /* reading data from file */
        if (max_columns!=min_columns) {
          fprintf (stderr, "Error: Inconsistent number of columns in %s\n", filename);
          fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
          fprintf (stderr, "     Should have been %d. First altitude, second density.\n", 2);
          return -1;
        }
 
        nlev  = rows;

        /* copy dens-file z-grid to zd */
        zd   = ASCII_column_float (atm_data, nlev, 0);

        /* allocate dens-array */
        status  = ASCII_calloc_float(&dens, 1, nlev);
        if (status!=0) {
	  fprintf (stderr, "Error %d allocating memory for dens\n", status); 
          fprintf (stderr, "     in read_and_combine_density (atmosphere.c)\n");
	  return status; 
        }

        /* lc is the 2nd index */
        for (lc=0; lc<rows; lc++) dens[0][lc] = atm_data[lc][1];
        
        /* convert units of the dens_file to cm-3 */
        status =  convert_profile_units (&(dens[0]), zd, rows, dens_id,
					 dens_air_all, temper_all, z_all, nlev_all,
                                         interpol_method_gas[MOL_AIR], interpol_method_temper, 
                                         unit_profile, mol_mass, quiet);
	
        if (status!=0) {
          fprintf (stderr, "Error %d during convert_profile_unit \n", status);
          fprintf (stderr, "     in read_and_combine_density (atmosphere.c)\n");
          return status;
        }

        /* combine file data and background */
        status = combine_profiles (dens, zd, nlev, 1,
                                   out->microphys.dens[dens_id], dens_air_all, out->zd, out->nlev,
                                   interpol_method_gas[dens_id], interpol_method_gas[MOL_AIR],
                                   &(dens_comb), &(z_comb), &(nlev_comb), quiet);

        if (status!=0) {
          fprintf (stderr, "Error %d during combine_profiles \n", status);
          fprintf (stderr, "     in read_and_combine_density (atmosphere.c)\n");
          return status;
        }

        /* check, if dens = 0 in the uppermost level, for log-interpolation types */
        if (interpol_method_gas[dens_id] == INTERP_METHOD_LOG || 
	    interpol_method_gas[dens_id] == INTERP_METHOD_LOG_SPLINE) {
          for (i=0; i<1; i++)
            if (dens_comb[i][0] == 0.0) {
              fprintf (stderr, "Error: dens(%5.1f km) = 0.0 cannot be interpolated with log assumption\n",z_comb[0]);
              fprintf (stderr, "  ***  By removing this line from the file %s, values for\n", filename);
              fprintf (stderr, "  ***  heights above the last entry in densfile will be \n");
              fprintf (stderr, "  ***  filled with the values given by the background atmosphere.\n");
              fprintf (stderr, "       in read_and_combine_density (atmosphere.c)\n");
              return -1;
            }
        }
        
        /* first determine air_dens profile on z_comb for linmix interpolation */
        if (interpol_method_gas[dens_id] == INTERP_METHOD_LINMIX) {
          dens_air_comb = (float *) calloc (nlev_comb, sizeof(float));
          status += arb_wvn (nlev_all, z_all,  dens_air_all, 
                             nlev_comb,  z_comb, dens_air_comb, 
                             interpol_method_gas[MOL_AIR], 1);
	}

        /* interpolate combined data on the z_all grid */
        status += interpolate_density (z_comb, &(dens_comb[0]), nlev_comb, 
                                       z_all, nlev_all, interpol_method_gas[dens_id],
                                       dens_air_comb, dens_air_all, quiet);

        if (status!=0) {
          fprintf (stderr, "Error %d interpolating density to common atmospheric grid\n", status);
          fprintf (stderr, "     in read_and_combine_density (atmosphere.c)\n");
          return status;
        }
        free(dens_air_comb);

        /* put result on final array */
        free(out->microphys.dens[dens_id]);
        out->microphys.dens[dens_id] = (float *) calloc (nlev_all, sizeof (float));
    
        for (lc=0; lc<nlev_all; lc++)
          out->microphys.dens[dens_id][lc] = dens_comb[0][lc];
      }

      else if (max_columns > 2) {   /* density matrix for airmass calculations */

        if (max_columns!=min_columns) {
          fprintf (stderr, "Error: Inconsistent number of columns in %s\n", filename);
          fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
          fprintf (stderr, "     in read_and_combine_density (atmosphere.c)\n");
          return -1;
        }

        if (solver != SDISORT) {
          fprintf (stderr, "Error: Dens-file %s has the format of a matrix.\n", filename);
          fprintf (stderr, "            This input type is only possible for solver SDISORT.\n");
          fprintf (stderr, "            in read_and_combine_density (atmosphere.c)\n");
          return -1;
        }

    
        /* first row/column contains sza/altitude */
        out->microphys.nsza_denstab = max_columns-1;
        nlev = rows-1;

        zd  = (float *) calloc (nlev, sizeof (float));

        if (out->microphys.denstab_id > 0) {
          fprintf (stderr, "Error: dens_file with matrix instead of profile\n");
          fprintf (stderr, "       already specified. The matrix may be used\n");
          fprintf (stderr, "       only for one specie. Profiles for as many as you like\n");
          fprintf (stderr, "       in read_and_combine_density (atmosphere.c)\n");
          return -1;
        }

        if (first_dens) {
          out->microphys.denstab_id  = dens_id;
          out->microphys.sza_denstab = (float *) calloc (out->microphys.nsza_denstab, sizeof (float));

          status  = ASCII_calloc_float(&out->microphys.denstab_amf,
                                       out->microphys.nsza_denstab, nlev);
          if (status!=0) {
            fprintf (stderr, "Error %d allocating memory for out->denstab_amf\n", status);
            fprintf (stderr, "       in read_and_combine_density (atmosphere.c)\n");
            return status;
          }
        }

        /* copy first row to solar zenith angle array  */
        for (i=1; i<=out->microphys.nsza_denstab; i++)
          out->microphys.sza_denstab[i-1] = atm_data[0][i];
    
        /* copy first column to altitude array zd */
        for (lc=1; lc<=nlev; lc++)
          zd[lc-1] = atm_data[lc][0];


        /* index of lc has to be at the last position */
        for (i=1; i<=out->microphys.nsza_denstab; i++)
          for (lc=1; lc<=nlev; lc++)
            out->microphys.denstab_amf [i-1][lc-1] = atm_data[lc][i];


        /* convert units of the dens_file to cm-3 */
        for (i=0; i<out->microphys.nsza_denstab; i++) {
          status =  convert_profile_units (&(out->microphys.denstab_amf[i]), zd, nlev, dens_id, 
                                           dens_air_all, temper_all, z_all, nlev_all,
                                           interpol_method_gas[MOL_AIR], interpol_method_temper, 
                                           unit_profile, mol_mass, quiet);
          if (status!=0) {
            fprintf (stderr, "Error %d during convert_profile_unit in read_and_combine_density\n", status);
            fprintf (stderr, "       in read_and_combine_density (atmosphere.c)\n");
            return status;
          }
        }

        /* combine file data and background */
        status = combine_profiles (out->microphys.denstab_amf, zd, nlev, out->microphys.nsza_denstab,
                                   out->microphys.dens[dens_id], dens_air_all, out->zd, out->nlev,
                                   interpol_method_gas[dens_id], interpol_method_gas[MOL_AIR],
                                   &(dens_comb), &(z_comb), &(nlev_comb), quiet);

        if (status!=0) {
          fprintf (stderr, "Error %d during combine_profiles (2) \n", status);
          fprintf (stderr, "       in read_and_combine_density (atmosphere.c)\n");
          return status;
        }

/* 	for (lc=1; lc<=nlev_all; lc++) */
/*           fprintf (stderr, "lc = %d, z_comb = %5.2f, dens_comb %11.7e\n", lc, z_comb[lc], dens_comb[0][lc]); */

        /* check, if dens = 0 in the uppermost level, for log-interpolation types */
        if (interpol_method_gas[dens_id] == INTERP_METHOD_LOG || 
	    interpol_method_gas[dens_id] == INTERP_METHOD_LOG_SPLINE) {
          for (i=0; i<out->microphys.nsza_denstab; i++)
            if (dens_comb[i][0] == 0.0) {
              fprintf (stderr, "Error: dens(%5.1f km) = 0.0 cannot be interpolated with log assumption\n",z_comb[0]);
              fprintf (stderr, "  ***  By removing this line from the file %s, values for\n", filename);
              fprintf (stderr, "  ***  hights above the last entry in densfile will be \n");
              fprintf (stderr, "  ***  filled with the values given by the background atmosphere.\n");
              fprintf (stderr, "       in read_and_combine_density (atmosphere.c)\n");
              return -1;
            }
        }

        /* first determine air_dens profile on z_comb for linmix interpolation */
        if (interpol_method_gas[dens_id] == INTERP_METHOD_LINMIX) {
          dens_air_comb = (float *) calloc (nlev_comb, sizeof(float));
          status += arb_wvn (nlev_all, z_all,  dens_air_all, 
                             nlev_comb,  z_comb, dens_air_comb, 
                             interpol_method_gas[MOL_AIR], 1);
	}

        if (status!=0) {
          fprintf (stderr, "Error %d interpolating air_dens to z_comb grid\n", status); 
          fprintf (stderr, "      in read_and_combine_density (atmosphere.c)\n");
          return status;
        }

        /* interpolate combined data on the z_all grid */
        for (i=0; i<out->microphys.nsza_denstab; i++) {
          status += interpolate_density (z_comb, &(dens_comb[i]), nlev_comb, 
                                         z_all, nlev_all, interpol_method_gas[dens_id],
                                         dens_air_comb, dens_air_all, quiet);

          if (status!=0) {
            fprintf (stderr, "Error %d interpolating denstab (dens_comb) to z_all grid\n", status); 
            fprintf (stderr, "      in read_and_combine_density (atmosphere.c)\n");
            return status;
          }
        }

        free(dens_air_comb);

        /* reallocate denstab_amf*/
        ASCII_free_float(out->microphys.denstab_amf, out->microphys.nsza_denstab);
        status  = ASCII_calloc_float(&out->microphys.denstab_amf,
                                      out->microphys.nsza_denstab, nlev_all);
        if (status!=0) {
          fprintf (stderr, "Error %d allocating memory for out->denstab_amf\n", status);
          fprintf (stderr, "      in read_and_combine_density (atmosphere.c)\n");
          return status;
        }

        /* copy combined data on final place  */
        for (i=0; i<out->microphys.nsza_denstab; i++)
          for (lc=0;lc<nlev_all;lc++)
            out->microphys.denstab_amf[i][lc]=dens_comb[i][lc];
           
        /* callocate denstab array */
        if (first_dens) {
          status  = ASCII_calloc_float_3D (&out->microphys.denstab, MOL_NN, out->microphys.nsza_denstab, nlev_all);
          if (status!=0) {
            fprintf (stderr, "Error %d allocating memory for out->denstab\n", status);
            fprintf (stderr, "      in read_and_combine_density (atmosphere.c)\n");
            return status;
          }
          status  = ASCII_calloc_float_3D (&out->microphys.denstab_avg, MOL_NN, out->microphys.nsza_denstab, nlev_all);
          if (status!=0) {
            fprintf (stderr, "Error %d allocating memory for out->denstab_avg\n", status);
            fprintf (stderr, "      in read_and_combine_density (atmosphere.c)\n");
            return status;
          }
        }

        /* copy the combined 2-dim double array to denstab */
        for (i=0; i<out->microphys.nsza_denstab; i++)
          for (lc=0; lc<nlev_all; lc++)
            out->microphys.denstab [dens_id][i][lc] = out->microphys.denstab_amf[i][lc];

        isthis=-666;
        /* replace the profile in the atmosphere-data according to sza */
        iv=0;  /* No solar zenith variations for a single amf calculation */
        for (is=0; is<out->microphys.nsza_denstab; is++)
          if (fabs (out->microphys.sza_denstab[is]-out->sza_t[iv]) < 1.e-04)
	    isthis = is;

        if (isthis < 0) {
          /* sza not found in the denstab-file */
          if (!quiet) {
            fprintf (stderr, "*** Warning: requested sza is not specified in the denstab-file.\n");
            fprintf (stderr, "*** Using the trace gas profile from the atmosphere_file.\n");
          }

          /* interpolate data of the background-atmosphere */
          /* from atm-file grid onto the z_all-grid  */
          status += interpolate_density (z_atm, &(out->microphys.dens[dens_id]), nlev_atm, 
                                         z_all, nlev_all, interpol_method_gas[dens_id],
                                         dens_air_atm, dens_air_all, quiet);
          if (status!=0) {
            fprintf (stderr, "Error %d interpolating background denstab to z_all grid\n", status);
            fprintf (stderr, "in read_and_combine_density (atmosphere.c)\n"); 
            return status;
          }
        }
        else {
          /* normal case, sza found in the denstab file */
          if (verbose)
            fprintf (stderr, " ... found denstab profile for sza = %6.2f\n", out->sza_t[0]);

          free(out->microphys.dens[dens_id]);
          out->microphys.dens[dens_id] = (float *) calloc (nlev_all, sizeof (float));

          for (lc=0; lc<nlev_all; lc++) {
            out->microphys.dens[dens_id][lc] = out->microphys.denstab[dens_id][isthis][lc];
          }
        }

        first_dens = 0;

      } /* if (max_columns==2) */

      free(zd);
      free(dens);
      free(z_comb);
      free(dens_comb);
  
      ASCII_free_float(atm_data, rows);

    }       /* if (strlen(filename) > 0) */
    else {  /* (there is no densfile, but z-grids are different)    */

      /* interpolation from atm-file grid onto the z_all-grid  */
      status += interpolate_density (z_atm, &(out->microphys.dens[dens_id]), nlev_atm, 
                                     z_all, nlev_all, interpol_method_gas[dens_id],
                                     dens_air_atm, dens_air_all, quiet);
      if (status!=0) {
         fprintf (stderr, "Error %d interpolating density to z_all grid\n", status); 
         fprintf (stderr, "      in read_and_combine_density (atmosphere.c)\n");
         return status;
      }     

    } /* if (strlen(filename) > 0) */

  } /* else { ( something to do ) */ 

  return status;
}

/**************************************************************/
/* Scale a density profile from it own column to input.column */
/**************************************************************/

static int scale_density_profile (float **dens, float **dens_avg, int alloc_avg, int i,
                          float input_column, float *output_column, 
                          float *scale_factor, int interpol_method_gas, 
                          input_struct input, output_struct *output)
{
  char   *gas;
  int lc=0;
  int status=0;

  gas = gas_number2string(i);
  if (strcmp(" -1 ",gas)==0) {
    fprintf (stderr, "Error wrong gas identifier in atmophere.c\n");
    return -1;
  }

  if (i == MOL_AIR) {
    fprintf (stderr, "Error you can't use the function scale_density_profile() for \n");
    fprintf (stderr, "      AIR, please have a look into setup_atmosphere() (atmosphere.c) \n");
    fprintf (stderr, "     ----- \n");
    return -2;
  }

  if (i == MOL_H2O) {
    fprintf (stderr, "Error when using function scale_density_profile for \n");
    fprintf (stderr, "      WATER VAPOUR, saturate water vapour inside cloud is not considered.\n");
    fprintf (stderr, "     -------------- That's why, it is forbitten to use it for water. \n");
    fprintf (stderr, "                    Please have a look into setup_atmosphere() (atmosphere.c) \n");
    return -3;
  }

  status = column ((*dens), output->atm.zd, output->atm.nlev, output->alt.altitude, 
                   interpol_method_gas, output->atm.microphys.dens[MOL_AIR],
                   &(*dens_avg), alloc_avg, &(*output_column));

  if (status!=0) {
    fprintf (stderr, "Error %d calculating %s column\n", status, gas);
    return status;
  }

  if (input_column >= 0.0) {
    switch (input.atm.unit_column[i]) {
    case DU:
      /* Everything ok, do nothing */
      if (!input.quiet)
        fprintf (stderr, " ... scaling %s column from %11.5g DU    to ", 
	         gas, (*output_column));
      break;
    case CM_2:
      if (!input.quiet)
        fprintf (stderr, " ... scaling %s column from %11.5g cm^-2 to ", 
	         gas, (*output_column)*DU2particle);
      input_column /= DU2particle; /* not changed outside this function as input is not a pointer, UH */
      break;
    default:
      fprintf (stderr, "Error, unknown unit for %s column\n", gas);
      return -4;
    }
    
    /* determine scale factor */
    if ((*output_column) > 0.0)
      (*scale_factor) = input_column / (*output_column);
    else
      (*scale_factor) = 0;

    /* scale profile */
    for (lc = 0; lc<output->atm.nlev; ++lc)
      (*dens)[lc] *= (*scale_factor);

    status = column ((*dens), output->atm.zd, output->atm.nlev, output->alt.altitude, 
                     interpol_method_gas, output->atm.microphys.dens[MOL_AIR],
                     &(*dens_avg), NO, output_column);

    if (status!=0) {
      fprintf (stderr, "\nError %d calculating %s column!\n", status, gas);
      return status;
    }

    switch (input.atm.unit_column[i]) {
    case DU:
      if (!input.quiet) fprintf (stderr, "%11.5g DU\n", (*output_column));
      break;
    case CM_2:
      if (!input.quiet) fprintf (stderr, "%11.5g cm^-2\n", (*output_column)*DU2particle);
      break;
    default:
      fprintf (stderr, "\nError, unknown unit for %s column\n", gas);
      return -4;
    }
  }

  return status;

}


/******************************************/
/* Converts the unit of a profile to cm-3 */
/******************************************/

static int convert_profile_units    (float **ptr_dens, float *z_dens, int nlev, int dens_id,   
                                     float *dens_air_atm, float *temper_atm, float *z_atm, int nlev_atm,
                                     int interpol_method_air, int interpol_method_temper,
                                     int unit_profile, float *mol_mass, int quiet)

{
  int    lc=0;
  float  mol_mass_ratio=0.0; 
  float *dens_air_filegrid = NULL; 
  float *temper_filegrid   = NULL;
  int    status=0;

  switch (unit_profile) {
    case CM_3: 
      /* do nothing, particles unit is already fine, particles per cubic centimeter */
      break;
    case M_3: /* particles per cubic meter */
      for (lc=0; lc<nlev; lc++)
        if ((*ptr_dens)[lc] != -1.0)
          (*ptr_dens)[lc] *= 1.E-6; /* m-3 -> cm-3 */
      break;
    case MMR: /* mass mixing ratio   */
    case VMR: /* volumn mixing ratio */

      /* multiply mass mixing ratio with dens_air (and if nessesary molecular mass ratio) */

      /* interpolate n_air to z-grid of the dens file */
      dens_air_filegrid = (float *) calloc (nlev_atm, sizeof (float));
      if (dens_air_filegrid == NULL) {
        fprintf (stderr, "Error allocating memory for dens_air_filegrid in function convert_profile_units\n");
        return -1;
      }
      for (lc=0; lc<nlev_atm; lc++) dens_air_filegrid[lc] = dens_air_atm[lc];
    
      status += interpolate_profile (z_atm, &(dens_air_filegrid), nlev_atm, 
                           z_dens, nlev, interpol_method_air, quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d interpolating density of air to z-grid of the dens-file\n", status);
        return status;
      }

      /* determine molecular mass ratio in order to convert mass mixing ratios */
      if (unit_profile == MMR)
        mol_mass_ratio = mol_mass[MOL_AIR]/mol_mass[dens_id];
      else if (unit_profile == VMR) 
        mol_mass_ratio = 1.0;
      else { 
        fprintf (stderr, "Error. Unknown unit for dens_file = %d!\n", unit_profile);
        return -1;
      }

      for (lc=0; lc<nlev; lc++)
        if ((*ptr_dens)[lc] != -1.0)
          (*ptr_dens)[lc] *= (dens_air_filegrid[lc] * mol_mass_ratio);

      free(dens_air_filegrid);

      break;
    case RH: /* relative humidity, only for water vapour */
      if (dens_id != MOL_H2O) {
        fprintf (stderr, "Error, during convert_profile_unit (atmosphere.c)\n");
        fprintf (stderr, "Unit RH is only possible for water vapour!\n");
        return -2;
      } 

      /* interpolate Temperature to z-grid of the dens file */
      temper_filegrid = (float *) calloc (nlev_atm, sizeof (float));
      if (temper_filegrid == NULL) {
        fprintf (stderr, "Error allocating memory for temper_filegrid in function convert_profile_units\n");
        return -1;
      }
      for (lc=0; lc<nlev_atm; lc++) temper_filegrid[lc] = temper_atm[lc];
      
      status += interpolate_profile (z_atm, &(temper_filegrid), nlev_atm, 
                           z_dens, nlev, interpol_method_temper, quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d interpolating temperature to z-grid of the dens-file\n", status);
        return status;
      }

      for (lc=0; lc<nlev; lc++)
        if ((*ptr_dens)[lc] != -1.0)
          (*ptr_dens)[lc] *=  vapor_pressure(temper_filegrid[lc]) / 100.0;

      free(temper_filegrid);

      break;
    default:
      fprintf (stderr, "Error. Unknown unit for dens_file = %d!\n", unit_profile);
      fprintf (stderr, "This is a program bug, please contact the programmers!\n");
      fprintf (stderr, "Please include the used input-file!\n");
      return -3;
      break;
  }
  return 0;
}


/**************************************/
/* Combine two profiles               */
/**************************************/

static int combine_profiles (float **dens, float *z, int nlev, int nsza, 
                             float *dens_atm, float *dens_air_atm, float *z_atm, int nlev_atm, 
                             int interpol_method_dens, int interpol_method_air,
                             float ***dens_comb, float **z_comb, int *nlev_comb, int quiet)

{  /* _atm means z-grid of the atmosphere files */
   /* comb means combined                */
   /* output are dens_comb and z_comb    */

  int lc=0,i=0;
  int n_up=-666, n_dn=-666;
  float *dens_atm_filegrid=NULL;
  float *dens_air_filegrid=NULL;
  int    status = 0;

  /* z-range of dens-file is larger than those of the background atmosphere file */
  /* reduce the data to the range of the background atmosphere file */
  if (z[0] > z_atm[0] || z[nlev-1] < z_atm[nlev_atm-1]) {
    fprintf (stderr, "Error in function combine_profiles (atmosphere.c) \n");  
    fprintf (stderr, "     Range of the densfile z = [%f,%f] is larger than \n",z[nlev-1],z[0]);
    fprintf (stderr, "     those of the atmosphere file z_atm = [%f,%f] \n",z_atm[nlev_atm-1],z_atm[0]);
    fprintf (stderr, "     Can't do interpolation, as information of dens_air on filegrid is missing \n");
    return -1;
  }


  /* determine AIR density on background atmosphere grid and filegrid */
  /* copy air density of the background atmosphere */
  dens_air_filegrid = calloc (nlev_atm, sizeof (float));
  for (lc=0; lc<nlev_atm; lc++)
    dens_air_filegrid[lc] = dens_air_atm[lc];

  /* interpolate air density to the z-grid of the dens_file */  
  status += interpolate_profile (z_atm, &(dens_air_filegrid), nlev_atm, 
                                 z, nlev, interpol_method_air, quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d interpolating air density of air to z-grid of the dens-file\n", status);
    fprintf (stderr, "   in combine_profiles (atmosphere.c)\n"); 
    return status;
  }

  /* copy density of the background atmosphere */
  dens_atm_filegrid = calloc (nlev_atm, sizeof (float));
  for (lc=0; lc<nlev_atm; lc++)
    dens_atm_filegrid[lc] = dens_atm[lc];

  /* interpolate density to the z-grid of the dens-file*/
  status += interpolate_density (z_atm, &(dens_atm_filegrid), nlev_atm, 
                                 z, nlev, interpol_method_dens,
                                 dens_air_atm, dens_air_filegrid, quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d interpolating dens of backgroud atmosphere to z-grid of the dens-file\n", status);
    fprintf (stderr, "   in combine_profiles (atmosphere.c)\n"); 
    return status;
  }

  if (z[0] == z_atm[0] &&  z[nlev-1] == z_atm[nlev_atm-1]) {
    /* upper and lower boundary are identical, no combination nessesary */
    *nlev_comb = nlev;

    *z_comb    = calloc (*nlev_comb, sizeof (float));
    *dens_comb = calloc (nsza,      sizeof (float *));
    for (i=0; i<nsza; i++)
      (*dens_comb)[i] = calloc (*nlev_comb, sizeof (float));

    for (lc=0; lc<*nlev_comb; lc++)
      (*z_comb)[lc]    = z[lc];

    for (i=0; i<nsza; i++) 
      for (lc=0; lc<*nlev_comb; lc++)
        if (dens[i][lc] != -1.0)
          (*dens_comb)[i][lc] = dens[i][lc];
	else
          (*dens_comb)[i][lc] = dens_atm_filegrid[lc];
  }
  else {

    /* get z_atm index above and below the dens-data */
    n_up = 0;
    for  (lc=0; lc<nlev_atm; lc++) {
      if(z_atm[lc] <= z[0]) {        
        n_up = lc-1;     /* last index of z_atm above the z-grid-range */
        break;
      }
    }
    if (n_up < -1) {
      fprintf (stderr, "Error in function combine_profiles!\n");  
      fprintf (stderr, "upper index not found, n_up = %d \n", n_up);   
      return -2;
    }
    
    n_dn = nlev_atm-1;   
    for  (lc=(nlev_atm-1); lc>=0; lc--)
      if(z_atm[lc] >= z[nlev-1]) {
        n_dn = lc;       /* last index of z_atm within the z-grid-range */
	break;
      }

    if (n_dn < 0) {
      fprintf (stderr, "Error in function combine_profiles!\n");  
      fprintf (stderr, "lower index not found, n_down = %d \n", n_dn);   
      return -3;
    }

    (*nlev_comb) = (n_up + 1) + nlev + ((nlev_atm - 1) - n_dn);
    
    /* combine the two profiles, dens (on z) has priority */
    /* Allocation */
    *z_comb    = calloc (*nlev_comb, sizeof (float  ));
    *dens_comb = calloc (nsza,       sizeof (float *));
    for (i=0; i<nsza; i++)
      (*dens_comb)[i] = calloc (*nlev_comb, sizeof (float));

    /* combine z-grid */
    for (lc=0; lc<=n_up; lc++)
         (*z_comb)[lc]          =    z_atm[lc];
    for (lc=0; lc<nlev; lc++)
         (*z_comb)[lc+(n_up+1)] =    z[lc];
    for (lc=1; lc<(nlev_atm-n_dn);lc++)
         (*z_comb)[lc+(n_up+nlev)] =    z_atm[lc+n_dn];

    /* combine data */
    for (i=0; i<nsza; i++) {
      for (lc=0; lc<=n_up; lc++)
        (*dens_comb)[i][lc] = dens_atm[lc];
      for (lc=0; lc<nlev; lc++)
        if (dens[i][lc] != -1.0)
          (*dens_comb)[i][lc+(n_up+1)] = dens[i][lc];
	else
          (*dens_comb)[i][lc+(n_up+1)] = dens_atm_filegrid[lc];
      for (lc=1; lc<(nlev_atm-n_dn);lc++)
        (*dens_comb)[i][lc+(n_up+nlev)] = dens_atm[lc+n_dn];
    }
  }
  
/*   for (lc=0; lc<*nlev_comb;lc++) */
/*     fprintf (stderr,"z_comb[%3d] = %7.3f, dens_comb[%3d] = %11.7e\n", lc, (*z_comb)[lc], lc, (*dens_comb)[0][lc]); */

  return 0;
}



/**************************************************************/
/* Read refraction index file                                 */
/* Combine with background atmosphere and                     */
/* Interpolate on the common grid: z_all                      */
/**************************************************************/

static int calculate_or_read_refind  (char *filename,
				      float **refind, float *z_all, int nlev_all,
				      int interpol_method_refind, 
                                      float *t, float *p,
                                      int quiet)

{
  int status=0, nlev=0, lc=0;

  float *z=NULL;
  float *refind_all=NULL;
  float **refind_comb=NULL;
  float *z_comb=NULL;
  int nlev_comb=-666;
  float n_0 = 1.000293; /* refraction under normal conditions */   
  float t_0 = 15.0;     /* normal temperature in celsius      */
  float p_0 = 1013.25;  /* normal pressure in mbar            */
  float ar  = 0.00366;  /* parameter for calculating refind   */
  float *dens_air_dummy = NULL;


  /* read refractive index from file */
  if (strlen(filename)>0) {

    fprintf (stderr, " ... reading refractive index from %s\n", filename);

    status = read_2c_file_float (filename, &z, refind, &nlev);   
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    if (z[0] == z_all[0] && z[nlev-1] == z_all[nlev_all-1]) {
      /* interpolate refind profile to common atmospheric grid */
      status += interpolate_profile (z, refind, nlev,
                                     z_all, nlev_all,
                                     interpol_method_refind, quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d interpolating relative humidity to common atmospheric grid\n", status);
        return status;
      }

    }
    else {
      /* combine refind_file and atmosphere-file data*/

      /* calculate refind of the atmosphere-file data */
      refind_all = (float *) calloc (nlev_all, sizeof(float));      
      if (refind_all == NULL) {
        fprintf (stderr, "Error allocating memory for refind_all in function calculate_or_read_refind \n");
        return -1;
      }

      for (lc = 0; lc < nlev_all; lc++)
        refind_all[lc] = 1.0 + (n_0-1.0) * (1.0 + ar*t_0) / (1.0 + ar * (t[lc]-273.15)) * p[lc] / p_0; 

      /* air dens dummy */
      dens_air_dummy = (float *) calloc (nlev_all, sizeof(float)); 

      /* combine refind-file data and atmosphere-file refind-data  */
      status = combine_profiles (refind, z, nlev, 1,
				 refind_all, dens_air_dummy, z_all, nlev_all, /* dens_air_dummy is dummy argument*/
				 interpol_method_refind, 1,                   /* 1              is dummy argument*/
				 &(refind_comb), &(z_comb), &(nlev_comb), quiet);
      
      if (status!=0) {
        fprintf (stderr, "Error %d during combine_profiles while interpolating refind (from refind-file to z_comb\n", status);
        fprintf (stderr, "   in calculate_or_read_refind (atmosphere.c)"); 
        return status;
      }

      free(dens_air_dummy);
      free(*refind);
      (*refind) = (float *) calloc (nlev_all, sizeof(float));

      status += arb_wvn (nlev_comb, z_comb, refind_comb[0],
                         nlev_all,  z_all, *refind,
                         interpol_method_refind, 1);

      free(z_comb);
      free(refind_comb[0]);
      free(refind_comb);
      free(refind_all);
    }
    free(z);
  }
  else {  /* if (strlen(filename)>0) => no refind-input-file */

    /* allocate memory */
    /* free(*refind); */     
    *refind = (float *) calloc (nlev_all, sizeof(float));
    if (*refind == NULL) {
      fprintf (stderr, "Error allocating memory for refind in function calculate_or_read_refind \n");
      return -1;
    }
    for (lc = 0; lc < nlev_all; lc++)
      (*refind)[lc] = 1.0 + (n_0-1.0) * (1.0 + ar*t_0) / (1.0 + ar * (t[lc]-273.15)) * p[lc] / p_0;
  }

  return 0;   /* o.k. */
}



/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z]. Memory for dtau[] is allocated automatically.              */ 
/***********************************************************************************/

static int read_molecular_absorption (char* filename, 
				      float **zd, int *nlyr,
				      int quiet,
				      double **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int lu=0, iv=0;

  int fulfilled=0;

  float **data=NULL;

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nlyr=0, idd_nwvl=0;
  int id_wvl=0, id_z=0, id_tau=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t taustart[2] = {0,0};
  size_t taucount[2] = {0,0};

  double *tmp_zd=NULL, *tmp_lambda=NULL;

  size_t dimlen=0;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    if (!quiet) {
      fprintf (stderr, "*** %s is not a netcdf file, trying to open as ASCII instead.\n", filename);
      fprintf (stderr, "*** It is recommended to use the netCDF format because access is much faster.\n");
    }
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... reading profiles from netcdf file %s\n", filename);

    /* get dimension id for "nlyr" */
    status = nc_inq_dimid (ncid, "nlyr", &idd_nlyr);
    
    /* get dimension length for "nlyr" */
    status = nc_inq_dimlen (ncid, idd_nlyr, &dimlen);
    *nlyr = dimlen;

    /* get dimension id for "nwvl" */
    status = nc_inq_dimid (ncid, "nwvl", &idd_nwvl);
    
    /* get dimension length for "nwvl" */
    status = nc_inq_dimlen (ncid, idd_nwvl, &dimlen);
    *nlambda = dimlen;

    /* read levels */
    tmp_zd = calloc (*nlyr+1, sizeof(double));
    *zd    = calloc (*nlyr+1, sizeof(float));

    /* get variable id for "z" */
    status = nc_inq_varid (ncid, "z", &id_z);

    count[0] = *nlyr+1;

    /* read "z" */
    status = nc_get_vara_double (ncid, id_z, start, count, tmp_zd);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    
    /* read wavelengths */
    tmp_lambda = calloc (*nlambda, sizeof(double));
    *lambda    = calloc (*nlambda, sizeof(float));

    /* get variable id for "wvl" */
    status = nc_inq_varid (ncid, "wvl", &id_wvl);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    
    /* read "wvl" */
    count[0] = *nlambda;

    status = nc_get_vara_double (ncid, id_wvl, start, count, tmp_lambda);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }


    /* get variable id for "tau" */
    status = nc_inq_varid (ncid, "tau", &id_tau);

    /* read optical thickness */
    taustart[1] = 0;  /* start with first optical thickness */
    taucount[0] = 1;  /* read one layer at a time           */
    
    *dtau_spec = calloc (*nlyr, sizeof(float *));
    for (lu=0; lu<*nlyr; lu++) {
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));

      taustart[0] = lu;
      taucount[1] = *nlambda;

      status = nc_get_vara_float (ncid, id_tau, taustart, taucount, (*dtau_spec)[lu]);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
	return status;
      }
    }

    /* copy data to final destination and free temporary memory */
    for (lu=0; lu<=*nlyr; lu++)
      (*zd)[lu] = tmp_zd[lu];

    for (iv=0; iv<*nlambda; iv++)
      (*lambda)[iv] = tmp_lambda[iv];

    free (tmp_zd); free (tmp_lambda);

    /* this is obviously a wavelength-dependent cross section */
    *monochromatic=0;

    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading profiles from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* a rough format check */
  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  if (min_columns!=max_columns) {
    fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	     filename, min_columns, max_columns);
    return -1;
  }


  /* check if first column sorted in descending order;    */
  /* if yes, we assume old format: two columns z and dtau */
  fulfilled=1;
  for (lu=0; lu<rows-1; lu++)
    if (data[lu][0] < data[lu+1][0])
      fulfilled = 0;
  
  if (fulfilled) {

    /* here we assume that the file comes in the old format,  */
    /* two columns z and dtau; in this case, the first column */
    /* must equal the atmospheric profile                     */

    *monochromatic=1;

    *nlyr = rows-1;
    *zd        = calloc (*nlyr+1, sizeof (float));
    *dtau_mono = calloc (*nlyr,   sizeof (double));

    for (lu=0; lu<*nlyr+1; lu++)
      (*zd)[lu] = data[lu][0];

    for (lu=0; lu<*nlyr; lu++)
      (*dtau_mono)[lu] = data[lu+1][1];
  }
  else {
    /* now we need to assume that we deal with a new format */
    /* where the first row must be the altitude levels      */
    
    /* check if first column sorted in ascending order; */
    /* if yes, we assume wavelength-dependent format    */
    fulfilled=1;
    for (iv=1; iv<rows-1; iv++)
      if (data[iv][0] > data[iv+1][0])
	fulfilled = 0;

    if (!fulfilled) {
      fprintf (stderr, "Error reading %s\n", filename);
      return -1;
    }
    
    /* now we know that the format is correct, and the only thing   */
    /* we need to do is to copy the data to their final destination */
    
    *monochromatic=0;

    *nlyr    = min_columns-1;
    *nlambda = rows-1;

    /* allocate memory */
    *zd = calloc (*nlyr+1, sizeof (float));
    *lambda    = calloc (*nlambda, sizeof (float));
    *dtau_spec = calloc (*nlyr,    sizeof (float *));
    for (lu=0; lu<*nlyr; lu++)
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));


    for (lu=0; lu<=*nlyr; lu++)
      (*zd)[lu] = data[0][lu];

    for (iv=0; iv<*nlambda; iv++) {
      (*lambda)[iv] = data[iv+1][0];
      
      for (lu=0; lu<*nlyr; lu++)
	(*dtau_spec)[lu][iv] = data[iv+1][lu+1];
    }
  }

  /* free memory */
  ASCII_free_float (data, rows);

  return 0;
}


/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z].                                                            */
/***********************************************************************************/

static int read_molecular_absorption_z (char* filename, int quiet,
					float **zd, int *nlyr) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int lu=0, iv=0;

  int fulfilled=0;

  
  float **data=NULL;
  

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nlyr=0;
  int id_z=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t dimlen=0;

  double *z=NULL;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {
    /* get dimension id for "nlyr" */
    status = nc_inq_dimid (ncid, "nlyr", &idd_nlyr);
    
    /* get dimension length for "nlyr" */
    status = nc_inq_dimlen (ncid, idd_nlyr, &dimlen);
    *nlyr = dimlen;

    /* read levels */
    z   = calloc (*nlyr+1, sizeof(double));
    *zd = calloc (*nlyr+1, sizeof(float));

    /* get variable id for "z" */
    status = nc_inq_varid (ncid, "z", &id_z);

    count[0] = *nlyr+1;

    /* read "z" */
    status = nc_get_vara_double (ncid, id_z, start, count, z);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* copy profile to final destination and free temporary memory */
    for (lu=0; lu<=*nlyr; lu++)
      (*zd)[lu] = z[lu];
    free (z);

    nc_close (ncid);
    
    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading altitudes from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* a rough format check */
  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  if (min_columns!=max_columns) {
    fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	     filename, min_columns, max_columns);
    return -1;
  }


  /* check if first column sorted in descending order;    */
  /* if yes, we assume old format: two columns z and dtau */
  fulfilled=1;
  for (lu=0; lu<rows-1; lu++)
    if (data[lu][0] < data[lu+1][0])
      fulfilled = 0;
  
  if (fulfilled) {
    *nlyr = rows-1;
    *zd = calloc (rows, sizeof (float));

    for (lu=0; lu<rows; lu++)
      (*zd)[lu] = data[lu][0];
  }
  else {
    /* now we need to assume that we deal with a new format */
    /* where the first row must be the altitude levels      */
    
    /* check if first column sorted in ascending order; */
    /* if yes, we assume wavelength-dependent format    */
    fulfilled=1;
    for (iv=1; iv<rows-1; iv++)
      if (data[iv][0] > data[iv+1][0])
	fulfilled = 0;

    if (!fulfilled) {
      fprintf (stderr, "Error reading %s\n", filename);
      return -1;
    }
    
    *nlyr = min_columns-1;
    *zd = calloc (min_columns, sizeof (float));
    for (lu=0; lu<min_columns; lu++)
      (*zd)[lu] = data[0][lu];
  }

  /* free memory */
  ASCII_free_float (data, rows);

  return 0;
}




/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z]. Check if the levels equal those defined in zd[]. Memory    */
/*  for dtau[] is allocated automatically.                                         */ 
/***********************************************************************************/

int read_molecular_absorption_lambda (char* filename, int quiet,
				      float **lambda, int *nlambda, int *monochromatic) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int iv=0;

  int fulfilled=0;

  float **data=NULL;
  
#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nwvl=0;
  int id_wvl=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t dimlen=0;

  double *wvl=NULL;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {
    /* get dimension id for "nwvl" */
    status = nc_inq_dimid (ncid, "nwvl", &idd_nwvl);
    
    /* get dimension length for "nwvl" */
    status = nc_inq_dimlen (ncid, idd_nwvl, &dimlen);
    *nlambda = dimlen;

    /* read levels */
    wvl     = calloc (*nlambda, sizeof(double));
    *lambda = calloc (*nlambda, sizeof(float));

    /* get variable id for "wvl" */
    status = nc_inq_varid (ncid, "wvl", &id_wvl);

    count[0] = *nlambda;

    /* read "wvl" */
    status = nc_get_vara_double (ncid, id_wvl, start, count, wvl);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* copy profile to final destination and free temporary memory */
    for (iv=0; iv<*nlambda; iv++)
      (*lambda)[iv] = wvl[iv];
    free (wvl);

    *monochromatic=0;

    nc_close (ncid);

    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading wavelengths from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* a rough format check */
  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  if (min_columns!=max_columns) {
    fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	     filename, min_columns, max_columns);
    return -1;
  }


  /* check if first column sorted in descending order;    */
  /* if yes, we assume old format: two columns z and dtau */
  fulfilled=1;
  for (iv=0; iv<rows-1; iv++)
    if (data[iv][0] < data[iv+1][0])
      fulfilled = 0;
  
  if (fulfilled)
    *monochromatic=1;
  else {
    /* now we need to assume that we deal with a new format */
    /* where the first column must be the wavelengths       */
    
    *monochromatic=0;

    fprintf (stderr, " ... now testing as wavelength-dependent absorption file\n");
    fflush (stderr);
    
    /* check if first column sorted in ascending order; */
    /* if yes, we assume wavelength-dependent format    */
    fulfilled=1;
    for (iv=1; iv<rows-1; iv++)
      if (data[iv][0] > data[iv+1][0])
	fulfilled = 0;

    if (!fulfilled) {
      fprintf (stderr, "Error reading %s\n", filename);
      return -1;
    }
    
    *nlambda = rows-1;
    *lambda = calloc (rows-1, sizeof (float));
    for (iv=0; iv<*nlambda; iv++)
      (*lambda)[iv] = data[iv+1][0];
  }

  /* free memory */
  ASCII_free_float (data, rows);

  return 0;
}



/**************************************************************/
/* Calculate atmospheric column in DU.                        */
/**************************************************************/

int column (float *dens, float *zd, int nlev, float altitude, int interpol_method, float *dens_air, 
            float **dens_avg, int allocate, float *column)
{

  float deltaz=0;
  int lc=0, ialt=0;
  int status=0;
  
  
  for (ialt=0; ialt<nlev; ialt++)
    /*    if (altitude == zd[ialt])  which one is correct ??? */
    if (altitude >= zd[ialt])
      break;
    
  if (ialt == nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid;\n", altitude);
    fprintf (stderr, "exiting column (in atmosphere.c)!\n");
    return -1;
  }

  status = average_dens(dens, dens_air, zd, nlev, interpol_method, dens_avg, allocate);

  if (status!=0) {
    fprintf (stderr, "Error %d calculating average concentration (column, in atmosphere.c)\n", status);
    return status;
  }

  *column=0;
  for (lc=0; lc<ialt; ++lc) {
    deltaz = (zd[lc] - zd[lc+1]) * 1.0E5;  /* km to cm */
    *column += (*dens_avg)[lc] * deltaz;   /* cm-3 * cm = cm-2 */
  }

  *column /= DU2particle;   /* column [particles per cm^2] to DU */
  
  return 0;
}



/**************************************************************/
/* Convert relative humidity [%] to concentration [1/cm3]     */
/**************************************************************/

double vapor_pressure (double t)
{
  /*
  double c1=-6096.9385;
  double c2=16.635794;
  double c3=-2.711193e-2;
  double c4=1.673952e-5;
  double c5=2.433502;
  double boltzmann = 1.38065e-23;
  */

  double avogadro  = 6.022142e23;
  
  double tzero = 273.15;
  double a = tzero/t;
  double mh2o = M_H2O/avogadro;


  /* Sonntag, Meteorol. Zeitschr, n.f. 3, 51-66, 1994                                 */
  /* return 100.0 / boltzmann / t * exp (c1/t +c2 +c3*t +c4*t*t + c5*log(t)) / 1.0e6; */
  
  /* SBDART, -> CRC handbook */
  return a * exp (18.916758 - a * (14.845878 + a*2.4918766)) / mh2o / 1.0e6;
}



/* converts number of interpolation method to human readable string*/
static char* interpol_number2string(int method_number) {

  char *method=NULL;

  /* all method names MUST have the same numer of characters */
  method = (char *) calloc (strlen("spline             ")+1, sizeof (char));

  switch (method_number) {
  case INTERP_METHOD_SPLINE:
    strcpy (method, "spline             ");
    break;
  case INTERP_METHOD_LINEAR:
    strcpy (method, "linear             ");
    break;
  case INTERP_METHOD_LOG:
    strcpy (method, "logarithmic        ");
    break;
  case INTERP_METHOD_LINMIX:
    strcpy (method, "linear mixing ratio");
    break;
  case INTERP_METHOD_LOG_SPLINE:
    strcpy (method, "logarithmic spline ");
    break;
  default:
    fprintf (stderr, "Error. Unknown interpolation method z_interpolate = %d!\n", method_number);
    fprintf (stderr, "This is a program bug, please contact the programmers!\n");
    fprintf (stderr, "Please include the used input-file!\n");
    return "       -1          ";
    break;
  }

  return method;
}

/* converts number of MOL_** to human readable string*/
char* gas_number2string(int gas_number) {

  char *gas=NULL;
  /* all gas names should have 4 Characters */
  gas = (char *) calloc (strlen("Air ")+1, sizeof (char));

  switch (gas_number) {
  case MOL_AIR:
    strcpy (gas, "Air ");
    break;
  case MOL_O3:
    strcpy (gas, "O3  ");
    break;
  case MOL_O2:
    strcpy (gas, "O2  ");
    break;
  case MOL_H2O:
    strcpy (gas, "H2O ");
    break;
  case MOL_CO2:
    strcpy (gas, "CO2 ");
    break;
  case MOL_NO2:
    strcpy (gas, "NO2 ");
    break;
  case MOL_BRO:
    strcpy (gas, "BRO ");
    break;
  case MOL_OCLO:
    strcpy (gas, "OCLO");
    break;
  case MOL_HCHO:
    strcpy (gas, "HCHO");
    break;
  default:
    fprintf (stderr, "Error. Unknown indentifier MOL_** = %d !\n", gas_number);
    fprintf (stderr, "This is a program bug, please contact the programmers!\n");
    fprintf (stderr, "Please include the used input-file!\n");
    return " -1 ";
    break;
  }

  return gas;
}
